// Product automaton -*- c++ -*-

#include "Product.h"
#include "Property.h"
#include "GlobalMarking.h"
#include "Net.h"
#include "Graph.h"

Product::Product (class Graph& graph, const class Property& prop) :
  myGraph (graph), myProp (prop), myHash (), myStates (0)
{
}

Product::~Product ()
{
  for (unsigned i = myHash.size (); i--; )
    delete myStates[i];
  delete[] myStates;
}

void
Product::insert (class ProductState* ps)
{
  assert (ps && myHash.find (ps) == myHash.end ());

  // allocate capacity in powers of 2
  size_t newsize = myHash.size (), oldsize = newsize++;
  if (!(oldsize & newsize)) {
    size_t capacity = newsize * 2 - 1;
    class ProductState** states = new class ProductState*[capacity];
    memcpy (states, myStates, oldsize * sizeof *myStates);
    memset (states + oldsize, 0, capacity - oldsize);
    delete[] myStates;
    myStates = states;
  }

  myHash.insert (myStates[oldsize] = ps);
}

/** Flag: has the analysis been interrupted? */
extern volatile bool interrupted;

bool
Product::analyze (Product::Stack& pstack)
{
  /** Proposition values at different states */
  class TritVector result (myProp.getNumExprs (), TritVector::Other);
  /** Search list for the reachability analysis */
  class SearchList states;

  while (!interrupted && !pstack.empty ()) {
    class ProductState& ps = *pstack.top ();
    const class PropertyState& prop = ps.getProp ();
    pstack.pop ();
    states.push (myGraph.find (ps.getState ().getNumber ()));
    myGraph.getNet ().analyze (myProp.getPrinter (), myGraph, states,
			       false, true);
    while (!states.empty ()) {
      const class State& state2 = states.pop (false).getState ();
      {
	class GlobalMarking* m = myGraph.fetchState (state2.getNumber ());
	if (!myProp.eval (*m, result)) {
	  delete m;
	  return false;
	}
	delete m;
      }
      for (unsigned i = prop.getNumSuccessors (); i--; ) {
	const class PropertyState& prop2 = prop[i];
	if (prop2.getProps () == result) {
	  const class ProductState* succ = find (state2, prop2);
	  if (!succ) {
	    if (prop2.isFinal ())
	      return false;
	    class ProductState* ps2 = new class ProductState (state2, prop2);
	    insert (ps2);
	    pstack.push (ps2);
	    succ = ps2;
	  }
	  ps.addSuccessor (*succ);
	}
      }
    }
  }

  return true;
}

bool
Product::analyze (const class GlobalMarking& marking)
{
  /** Proposition values at different states */
  class TritVector result (myProp.getNumExprs (), TritVector::Other);

  const class State& state = myGraph.add (marking).state;

  if (myProp.eval (marking, result) && myProp[0].getProps () == result) {
    /** Stack of product states */
    Stack pstack;
    // graph.getProductStack (pstack);
    class ProductState* ps = new class ProductState (state, myProp[0]);
    insert (ps);
    pstack.push (ps);
    bool result = analyze (pstack);
    assert (!result || pstack.empty () || interrupted);
    // graph.setProductStack (pstack);
    return result;
  }
  else
    return false;
}
