// Product automaton -*- c++ -*-

#ifndef PRODUCT_H_
# define PRODUCT_H_

# include "ProductState.h"

# ifdef HAS_HASH_SET
#  include <hash_set>
# else
#  include <set>
# endif // HAS_HASH_SET
# include <stack>
# include <list>
# include <cassert>

/** Product automaton */
class Product {
  /** Hash and comparison functions */
  struct hc {
    /** Calculate the hash value of a product automaton state
     * @param s		state to be hashed
     * @return		the hash value
     */
    size_t operator() (const class ProductState* s) const {
      return (*s) ();
    }
    /** Equality comparison for product automaton states
     * @param s1	state to be compared
     * @param s2	state to be compared
     * @return		whether the two states are equal
     */
    bool operator () (const class ProductState* s1,
		      const class ProductState* s2) const {
      return *s1 == *s2;
    }
  };
  // Typedefs

# ifdef HAS_HASH_SET
  typedef std::hash_set<class ProductState*,struct hc,struct hc> StateHash;
# else
  typedef std::set<class ProductState*,struct hc> StateHash;
# endif // HAS_HASH_SET
  typedef std::stack<class ProductState*, std::list<class ProductState*> >
  Stack;

public:
  /** Constructor
   * @param graph	the reachability graph
   * @param property	the property automaton
   */
  Product (class Graph& graph, const class Property& prop);
private:
  /** Copy constructor */
  Product (const class Product& old);
  /** Assignment operator */
  class Product& operator= (const class Product& old);
public:
  /** Destructor */
  ~Product ();

  /** Find a state in the product automaton
   * @param state	state in the reachability graph
   * @param props	state of the property automaton
   * @return		the corresponding product state, or NULL
   */
  const class ProductState* find (const class State& state,
				  const class PropertyState& props) const {
    class ProductState ps (state, props);
    StateHash::const_iterator i = myHash.find (&ps);
    return i == myHash.end () ? NULL : *i;
  }

  /** Insert a state to the product automaton
   * @param ps		state to be inserted
   */
  void insert (class ProductState* ps);

  /** Check whether the property expressed by the property automaton holds,
   * calculating the product of the reachability graph and the automaton,
   * performing on-the-fly analysis
   * @param pstack	stack of product automaton states
   * @return		whether the property holds
   */
  bool analyze (Stack& pstack);

  /** Check whether the property expressed by the property automaton holds,
   * calculating the product of the reachability graph and the automaton,
   * performing on-the-fly analysis
   * @param marking	the global marking from which to start the analysis
   * @return		whether the property holds
   */
  bool analyze (const class GlobalMarking& marking);

private:
  /** The reachability graph */
  class Graph& myGraph;
  /** The property automaton */
  const class Property& myProp;
  /** Hash table of the product states */
  StateHash myHash;
  /** The product states */
  class ProductState** myStates;
};

#endif // PRODUCT_H_
