// State of a product automaton -*- c++ -*-

#ifndef PRODUCTSTATE_H_
# define PRODUCTSTATE_H_

# include "State.h"
# include "PropertyState.h"

/** State of a product automaton */
class ProductState {
public:
  /** Default constructor */
  ProductState () : myState (0), myPropertyState (0),
    myNumSuccessors (0), mySuccessors (0) {}
  /** Constructor
   * @param state	the reachability graph state
   * @param props	the property automaton state
   */
  ProductState (const class State& state, const class PropertyState& props) :
    myState (&state), myPropertyState (&props),
    myNumSuccessors (0), mySuccessors (0) {}
  /** Copy constructor */
  ProductState (const class ProductState& old) :
    myState (old.myState), myPropertyState (old.myPropertyState),
    myNumSuccessors (old.myNumSuccessors), mySuccessors (old.mySuccessors) {}
private:
  /** Assignment operator */
  class ProductState& operator= (const class ProductState& old);
public:
  /** Destructor */
  ~ProductState () {
    delete[] mySuccessors;
  }

  /** Equality comparison
   * @param other	state to be compared to
   * @return		whether the states are equal
   */
  bool operator== (const class ProductState& other) const {
    return myState == other.myState &&
      myPropertyState == other.myPropertyState;
  }
  /** Less-than comparison
   * @param other	state to be compared with
   * @return		whether this is smaller than other
   */
  bool operator< (const class ProductState& other) const {
    return myState < other.myState ||
      (myState == other.myState && myPropertyState < other.myPropertyState);
  }

  /** Integer hash function by Robert Jenkins (bob_jenkins@compuserve.com)
   * <URL:http://ourworld.compuserve.com/homepages/bob_jenkins/blockcip.htm>
   * @param key		key to be hashed
   * @return		hash value of key
   */
  static size_t hasher (card_t key) {
    key += key << 12;
    key ^= key >> 22;
    key += key << 4;
    key ^= key >> 9;
    key += key << 10;
    key ^= key >> 2;
    key += key << 7;
    key ^= key >> 12;
    return size_t (key);
  }

  /** Get the hash value of the product state */
  size_t operator() () const {
    return
      hasher (myState->getNumber ()) ^ hasher (myPropertyState->getNumber ());
  }

  /** Determine the reachability graph state */
  const class State& getState () const { return *myState; }
  /** Determine the property automaton state */
  const class PropertyState& getProp () const { return *myPropertyState; }

  /** Determine whether the state is final */
  bool isFinal () const { return myPropertyState->isFinal (); }

  /** Add a successor state
   * @param succ	state to be added
   */
  void addSuccessor (const class ProductState& succ) {
    // allocate capacity in powers of 2
    unsigned oldNum = myNumSuccessors++;
    if (!(oldNum & myNumSuccessors)) {
      unsigned capacity = myNumSuccessors * 2 - 1;
      const class ProductState** succs =
	new const class ProductState*[capacity];
      memcpy (succs, mySuccessors, oldNum * sizeof *mySuccessors);
      memset (succs + oldNum, 0, capacity - oldNum);
      delete[] mySuccessors;
      mySuccessors = succs;
    }

    mySuccessors[oldNum] = &succ;
  }

  /** Get the number of successors */
  unsigned getNumSuccessors () const { return myNumSuccessors; }
  /** Access a successor by index
   * @param index	the index
   * @return		the successor
   */
  const class ProductState& operator[] (unsigned index) const {
    assert (index < getNumSuccessors ());
    return *mySuccessors[index];
  }

private:
  /** The reachability graph state */
  const class State* const myState;
  /** The property automaton state */
  const class PropertyState* const myPropertyState;

  /** Number of successor states */
  unsigned myNumSuccessors;
  /** The successor states */
  const class ProductState** mySuccessors;
};

#endif // PRODUCTSTATE_H_
