// Property automaton -*- c++ -*-

#include "Property.h"
#include "PropertyState.h"
#include "TemporalUnop.h"
#include "Printer.h"
#include "Expression.h"
#include "BoolValue.h"
#include "Printer.h"
#include "GlobalMarking.h"

class Property*
Property::create (const class Printer& printer,
		  const class Expression& formula)
{
  if (formula.getKind () == Expression::eTemporalUnop) {
    const class TemporalUnop& unop =
      static_cast<const class TemporalUnop&>(formula);
    if (unop.getOp () == TemporalUnop::Globally &&
	!unop.getQual () && !unop.getExpr ()->isTemporal ()) {
      const class Expression** exprs = new const class Expression*[1];
      *exprs = unop.getExpr ();
      class Property* prop = new class Property (printer, 1, exprs);
      class PropertyState& init = *prop += TritVector (1, TritVector::Other);
      class PropertyState& final = *prop += TritVector (1, TritVector::False);
      final.flagFinal ();
      init += init += final;
      return prop;
    }
  }

  printer.printRaw ("unsupported temporal logic formula");
  printer.linebreak ();
  return NULL;
}

Property::Property (const class Printer& printer,
		    unsigned numExprs, const class Expression* const* exprs) :
  myPrinter (printer),
  myNumExprs (numExprs), myExprs (exprs),
  myNumStates (0), myStates (0)
{
}

class PropertyState&
Property::operator+= (const class TritVector& props)
{
  unsigned index = myNumStates++;
  if (!(index & myNumStates)) {
    unsigned capacity = myNumStates * 2 - 1;
    class PropertyState** states = new class PropertyState*[capacity];
    memcpy (states, myStates, index * sizeof *myStates);
    memset (states + index, 0, (capacity - index) * sizeof *myStates);
    delete[] myStates;
    myStates = states;
  }
  assert (!myStates[index]);
  return *(myStates[index] = new class PropertyState (index, props));
}

bool
Property::eval (const class GlobalMarking& marking,
		class TritVector& result) const
{
  assert (result.getSize () == myNumExprs);
  class Valuation valuation;
  valuation.setGlobalMarking (&marking);

  for (unsigned i = myNumExprs; i--; ) {
#ifdef EXPR_CACHE
    const class Value* v = myExprs[i]->const_eval (&valuation);
#else // EXPR_CACHE
    class Value* v = myExprs[i]->eval (&valuation);
#endif // EXPR_CACHE
    if (v) {
      assert (valuation.isOK ());
      assert (v->getKind () == Value::vBool);
      if (static_cast<const class BoolValue*>(v)->getValue ())
	result.setTrue (i);
      else
	result.setFalse (i);
#ifndef EXPR_CACHE
      delete v;
#endif // !EXPR_CACHE
    }
    else {
      assert (!valuation.isOK ());
      myPrinter.printRaw ("cannot evaluate property ");
      myPrinter.print (i);
      myPrinter.delimiter (':');
      myPrinter.linebreak ();
      valuation.display (myPrinter);
      marking.display (myPrinter);
      myPrinter.linebreak ();
      return false;
    }
  }

  return true;
}
