// Property automaton -*- c++ -*-

#ifndef PROPERTY_H_
# define PROPERTY_H_

# include <cassert>

/** Property automaton */
class Property {
public:
  /** Create a property automaton from a formula
   * @param printer	Printer object for diagnostics
   * @param formula	the formula to be verified
   * @return		the property automaton, or NULL in case of error
   */
  static class Property* create (const class Printer& printer,
				 const class Expression& formula);
private:
  /** Constructor
   * @param printer	Printer object for diagnostics
   * @param numExprs	size of the alphabet
   * @param exprs	the atomic propositions forming the alphabet
   */
  Property (const class Printer& printer,
	    unsigned numExprs, const class Expression* const* exprs);
  /** Copy constructor */
  explicit Property (const class Property& old);
  /** Assignment operator */
  class Property operator= (const class Property& old);
public:
  /** Destructor */
  ~Property ();

  /** Get the printer object */
  const class Printer& getPrinter () const { return myPrinter; }
  /** Determine the number of expressions in the property automaton */
  unsigned getNumExprs () const { return myNumExprs; }

  /** Add a state to the property automaton
   * @param props	desired truth values for the atomic propositions
   * @return		the added state
   */
  class PropertyState& operator+= (const class TritVector& props);

  /** Access a state of the automaton
   * @param index	number of the state (0 is the initial state)
   * @return		the state
   */
  const class PropertyState& operator[] (unsigned index) const {
    assert (index < myNumStates);
    return *myStates[index];
  }

  /** Evaluate the expressions in a global marking
   * @param marking	the global marking
   * @param result	placeholder for the result (truth values)
   * @return		whether the operation succeeded
   */
  bool eval (const class GlobalMarking& marking,
	     class TritVector& result) const;

private:
  /** The Printer object for diagnostic output */
  const class Printer& myPrinter;
  /** Number of atomic propositions */
  const unsigned myNumExprs;
  /** The atomic propositions (boolean expressions, the alphabet) */
  const class Expression* const* myExprs;
  /** Number of states */
  unsigned myNumStates;
  /** The states */
  class PropertyState** myStates;
};

#endif // PROPERTY_H_
