// A state in the property automaton -*- c++ -*-

#ifndef PROPERTYSTATE_H_
# define PROPERTYSTATE_H_

# include "TritVector.h"
# include "typedefs.h"
# include <cassert>
# include <cstring>

/** final state flag embedded in the state number */
# define PROPERTY_FINAL (card_t (1) << (CARD_T_BIT - 1))
/** visit flag embedded in the state number */
# define PROPERTY_VISIT (card_t (1) << (CARD_T_BIT - 2))
/** mask for getting rid of the flags */
# define PROPERTY_MASK (~(PROPERTY_FINAL | PROPERTY_VISIT))

/** A state in the property automaton */
class PropertyState {
public:
  /** Constructor
   * @param number	number of the state
   * @param props	truth values for the propositions
   */
  PropertyState (unsigned number, const class TritVector& props) :
    myNumber (number), myProps (props),
    myNumSuccessors (0), mySuccessors (0) {
    assert (!(myNumber & ~PROPERTY_MASK));
  }
private:
  /** Copy constructor */
  explicit PropertyState (const class PropertyState& old);
  /** Assignment operator */
  class PropertyState& operator= (const class PropertyState& old);
public:
  /** Destructor */
  ~PropertyState () { delete[] mySuccessors; }

  /** Determine the number of the state */
  card_t getNumber () const { return myNumber & PROPERTY_MASK; }
  /** Determine the truth values for the propositions */
  const class TritVector& getProps () const { return myProps; }
  /** Determine the number of successor states */
  card_t getNumSuccessors () const { return myNumSuccessors; }

  /** Access a successor
   * @param index	zero-based index number to the successor
   * @return		the successor
   */
  const class PropertyState& operator[] (card_t index) const {
    assert (index < getNumSuccessors ());
    return *mySuccessors[index];
  }

  /** Access a successor
   * @param index	zero-based index number to the successor
   * @return		the successor
   */
  class PropertyState& operator[] (card_t index) {
    assert (index < getNumSuccessors ());
    return *mySuccessors[index];
  }

  /** Add a successor state
   * @param other	state to be added
   * @return		*this
   */
  class PropertyState& operator+= (class PropertyState& other) {
    card_t newindex = getNumSuccessors (), oldindex = newindex++;
    if (!(newindex & oldindex)) {
      card_t capacity = newindex * 2 - 1;
      class PropertyState** successors = new class PropertyState*[capacity];
      memcpy (successors, mySuccessors, oldindex * sizeof *successors);
      delete[] mySuccessors;
      mySuccessors = successors;
    }
    mySuccessors[oldindex] = &other;
    myNumSuccessors++;
    return *this;
  }

  /** Determine whether the state is final */
  bool isFinal () const { return bool (myNumber & PROPERTY_FINAL); }
  /** Flag the state final */
  void flagFinal () { myNumber |= PROPERTY_FINAL; }
  /** Determine whether the state has been visited */
  bool isVisited () const { return bool (myNumber & PROPERTY_VISIT); }
  /** Flag the state visited */
  void flagVisited () { myNumber |= PROPERTY_VISIT; }
  /** Clear the visited flag */
  void clearVisited () { myNumber &= ~PROPERTY_VISIT; }

private:
  /** Number of the state */
  card_t myNumber;
  /** Truth values for the atomic propositions */
  const class TritVector myProps;
  /** Number of successor states */
  card_t myNumSuccessors;
  /** The successor states */
  class PropertyState** mySuccessors;
};

#endif // PROPERTYSTATE_H_
