// Ternary digit (trit) vector -*- c++ -*-

#ifndef TRITVECTOR_H_
# define TRITVECTOR_H_

# include <limits.h>
# include <cstring>
# include <cassert>

/* Determine the number of trits per char */
# if CHAR_BIT == 8
#  define TRITS 5
# else
#  define TRITS (CHAR_BIT / 2)
# endif

/** Ternary digit (trit) vector */
class TritVector {
public:
  /** Ternary digits */
  enum Trit { False, True, Other };

  // Constructors and destructors

  /** Constructor
   * @param size	number of elements in the vector
   * @param init	initial value for the elements
   */
  TritVector (unsigned size, enum Trit init);
  /** Copy constructor */
  explicit TritVector (const class TritVector& old) :
    mySize (old.mySize), myTrits (0) {
    memcpy (myTrits = new unsigned char[tritWords (mySize)],
	    old.myTrits, tritWords (mySize));
  }
private:
  /** Assignment operator */
  class TritVector& operator= (const class TritVector& old);
public:
  /** Destructor */
  ~TritVector () { delete[] myTrits; }

  // Accessors

  /** Determine the size of the vector */
  unsigned getSize () const { return mySize; }

  /** Read a ternary digit
   * @param index	zero-based index of the element
   * @return		value of the ternary digit
   */
  enum Trit operator[] (unsigned index) const;

  /** Set a ternary digit to true value
   * @param index	zero-based index of the element
   */
  void setTrue (unsigned index);
  /** Set a ternary digit to false value
   * @param index	zero-based index of the element
   */
  void setFalse (unsigned index);
  /** Set a ternary digit to other value
   * @param index	zero-based index of the element
   */
  void setOther (unsigned index);

  // Other methods

  /** Determine the number of words required for storing a number of trits
   * @param trits	amount of trits
   * @return		number of words required for holding the trits
   */
  static unsigned tritWords (unsigned trits) {
    return (trits + TRITS - 1) / TRITS;
  }

  /** Equality comparison
   * @param other	vector to be compared with (containing no Other bits)
   * @return		whether the vectors are equal, ignoring Other bits
   */
  bool operator== (const class TritVector& other) const {
    assert (mySize == other.mySize);
    if (!memcmp (myTrits, other.myTrits, tritWords (mySize)))
      return true;
    for (unsigned i = mySize; i--; ) {
      enum Trit t = (*this)[i];
      if (t != Other && t != other[i])
	return false;
    }
    return true;
  }

private:
  /** Number of elements in the vector */
  unsigned mySize;
  /** The vector */
  unsigned char* myTrits;
};

#endif // TRITVECTOR_H_
