// binary operator class -*- c++ -*-

#include "BinopExpression.h"
#include "IntType.h"
#include "IntValue.h"
#include "Net.h"
#include "Valuation.h"
#include "Constant.h"

BinopExpression::BinopExpression (enum Op op,
				  class Expression* left,
				  class Expression* right) :
  Expression (),
  myOp (op), myLeft (left), myRight (right)
{
  assert (myLeft->getType ()->getKind () == Type::tInt);
  assert (myRight->getType ()->getKind () == Type::tInt);
  assert (myLeft->isBasic () && myRight->isBasic ());
  setType (Net::getIntType ());
}

BinopExpression::~BinopExpression ()
{
  myLeft->destroy ();
  myRight->destroy ();
}

#ifdef EXPR_CACHE
bool
BinopExpression::invalidate (const class VariableDefinition* variable) const
{
  bool invalidatedLeft = myLeft->invalidate (variable);

  if (myRight->invalidate (variable) || invalidatedLeft) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
BinopExpression::do_clearValuation (const class Valuation* valuation) const
{
  myLeft->clearValuation (valuation);
  myRight->clearValuation (valuation);
}

class Value*
BinopExpression::do_eval (const class Valuation* valuation) const
{
  const class Value* left = myLeft->const_eval (valuation);
  if (!left)
    return NULL;
  const class Value* right = myRight->const_eval (valuation);
  if (!right)
    return NULL;

  assert (left->getType ()->getKind () == Type::tInt);
  assert (left->getKind () == Value::vInt);
  assert (right->getType ()->getKind () == Type::tInt);
  assert (right->getKind () == Value::vInt);

  const class IntValue* ileft = static_cast<const class IntValue*>(left);
  const class IntValue* iright = static_cast<const class IntValue*>(right);

  switch (getOp ()) {
  case bPlus:
    if ((ileft->getValue () > 0 && iright->getValue () > 0 &&
	 INT_T_MAX - ileft->getValue () < iright->getValue ()) ||
	(ileft->getValue () < 0 && iright->getValue () < 0 &&
	 INT_T_MIN - ileft->getValue () > iright->getValue ())) {
      if (valuation)
	valuation->flagOverflow ();
      return NULL;
    }

    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () + iright->getValue ()));
  case bMinus:
    if ((ileft->getValue () > 0 && iright->getValue () < 0 &&
	 INT_T_MAX + iright->getValue () < ileft->getValue ()) ||
	(ileft->getValue () < 0 && iright->getValue () > 0 &&
	 INT_T_MAX + ileft->getValue () < iright->getValue ())) {
      if (valuation)
	valuation->flagOverflow ();
      return NULL;
    }

    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () - iright->getValue ()));
  case bDiv:
    if (iright->getValue ())
      return constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () / iright->getValue ()));
    else {
      if (valuation)
	valuation->flagDivisionByZero ();
      return NULL;
    }
  case bMul:
    if (ileft->getValue () > 0) {
      if (iright->getValue () > 0) {
	if (INT_T_MAX / ileft->getValue () < iright->getValue ()) {
	  if (valuation)
	    valuation->flagOverflow ();
	  return NULL;
	}
      }
      else if (INT_T_MIN / ileft->getValue () > iright->getValue ()) {
	if (valuation)
	  valuation->flagOverflow ();
	return NULL;
      }
    }
    else if (iright->getValue () > 0) {
      if (INT_T_MIN / ileft->getValue () < iright->getValue ()) {
	if (valuation)
	  valuation->flagOverflow ();
	return NULL;
      }
    }
    else if (INT_T_MAX / ileft->getValue () > iright->getValue ()) {
      if (valuation)
	valuation->flagOverflow ();
      return NULL;
    }

    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () * iright->getValue ()));
  case bMod:
    if (iright->getValue ())
      return constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () % iright->getValue ()));
    else {
      if (valuation)
	valuation->flagDivisionByZero ();
      return NULL;
    }
  case bAnd:
    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () & iright->getValue ()));
  case bOr:
    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () | iright->getValue ()));
  case bXor:
    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () ^ iright->getValue ()));
  case bRol:
    if (iright->getValue () < 0 ||
	iright->getValue () > static_cast<int_t>(INT_T_BIT)) {
    shiftError:
      if (valuation)
	valuation->flagShiftError ();
      return NULL;
    }
    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () << iright->getValue ()));
  case bRor:
    if (iright->getValue () < 0 ||
	iright->getValue () > static_cast<int_t>(INT_T_BIT))
      goto shiftError;

    return constrain (valuation,
		      new class IntValue
		      (getType (),
		       ileft->getValue () >> iright->getValue ()));
  }

  assert (false);
  return NULL;
}
#else // EXPR_CACHE
class Value*
BinopExpression::do_eval (const class Valuation* valuation) const
{
  class Value* left = myLeft->eval (valuation);
  if (!left)
    return NULL;
  class Value* right = myRight->eval (valuation);
  if (!right) {
    delete left;
    return NULL;
  }

  assert (left->getType ()->getKind () == Type::tInt);
  assert (left->getKind () == Value::vInt);
  assert (right->getType ()->getKind () == Type::tInt);
  assert (right->getKind () == Value::vInt);

  class IntValue* ileft = static_cast<class IntValue*>(left);
  class IntValue* iright = static_cast<class IntValue*>(right);
  class Value* result;

  switch (getOp ()) {
  case bPlus:
    if ((ileft->getValue () > 0 && iright->getValue () > 0 &&
	 INT_T_MAX - ileft->getValue () < iright->getValue ()) ||
	(ileft->getValue () < 0 && iright->getValue () < 0 &&
	 INT_T_MIN - ileft->getValue () > iright->getValue ())) {
      if (valuation)
	valuation->flagOverflow ();
      delete ileft, delete iright;
      return NULL;
    }

    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () + iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bMinus:
    if ((ileft->getValue () > 0 && iright->getValue () < 0 &&
	 INT_T_MAX + iright->getValue () < ileft->getValue ()) ||
	(ileft->getValue () < 0 && iright->getValue () > 0 &&
	 INT_T_MAX + ileft->getValue () < iright->getValue ())) {
      if (valuation)
	valuation->flagOverflow ();
      delete ileft, delete iright;
      return NULL;
    }

    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () - iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bDiv:
    if (iright->getValue ()) {
      result = constrain (valuation,
			  new class IntValue
			  (getType (),
			   ileft->getValue () / iright->getValue ()));
      delete ileft, delete iright;
      return result;
    }
    else {
      if (valuation)
	valuation->flagDivisionByZero ();
      delete ileft, delete iright;
      return NULL;
    }
  case bMul:
    if (ileft->getValue () > 0) {
      if (iright->getValue () > 0) {
	if (INT_T_MAX / ileft->getValue () < iright->getValue ()) {
	  if (valuation)
	    valuation->flagOverflow ();
	  delete ileft, delete iright;
	  return NULL;
	}
      }
      else if (INT_T_MIN / ileft->getValue () > iright->getValue ()) {
	if (valuation)
	  valuation->flagOverflow ();
	delete ileft, delete iright;
	return NULL;
      }
    }
    else if (iright->getValue () > 0) {
      if (INT_T_MIN / ileft->getValue () < iright->getValue ()) {
	if (valuation)
	  valuation->flagOverflow ();
	delete ileft, delete iright;
	return NULL;
      }
    }
    else if (INT_T_MAX / ileft->getValue () > iright->getValue ()) {
      if (valuation)
	valuation->flagOverflow ();
      delete ileft, delete iright;
      return NULL;
    }

    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () * iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bMod:
    if (iright->getValue ()) {
      result = constrain (valuation,
			  new class IntValue
			  (getType (),
			   ileft->getValue () % iright->getValue ()));
      delete ileft, delete iright;
      return result;
    }
    else {
      if (valuation)
	valuation->flagDivisionByZero ();
      delete ileft, delete iright;
      return NULL;
    }
  case bAnd:
    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () & iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bOr:
    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () | iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bXor:
    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () ^ iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bRol:
    if (iright->getValue () < 0 ||
	iright->getValue () > static_cast<int_t>(INT_T_BIT)) {
    shiftError:
      if (valuation)
	valuation->flagShiftError ();
      delete ileft, delete iright;
      return NULL;
    }
    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () << iright->getValue ()));
    delete ileft, delete iright;
    return result;
  case bRor:
    if (iright->getValue () < 0 ||
	iright->getValue () > static_cast<int_t>(INT_T_BIT))
      goto shiftError;

    result = constrain (valuation,
			new class IntValue
			(getType (),
			 ileft->getValue () >> iright->getValue ()));
    delete ileft, delete iright;
    return result;
  }

  assert (false);
  return NULL;
}
#endif // EXPR_CACHE

class Expression*
BinopExpression::ground (const class Valuation& valuation)
{
  class Expression* left = myLeft->ground (valuation);
  class Expression* right = myRight->ground (valuation);

  if (!left || !right) {
    left->destroy ();
    right->destroy ();
    return NULL;
  }

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class BinopExpression (myOp, left, right);
    expr->setType (getType ());
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
BinopExpression::substitute (class Substitution& substitution)
{
  class Expression* left = myLeft->substitute (substitution);
  class Expression* right = myRight->substitute (substitution);

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class BinopExpression (myOp, left, right);
    expr->setType (getType ());
    return expr;
  }
}

#ifdef EXPR_CACHE
void
BinopExpression::associateVariables (const class Expression& expr,
				     bool assoc) const
{
  myLeft->associateVariables (expr, assoc);
  myRight->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
BinopExpression::forVariables (bool (*operation)
			       (const class VariableDefinition&)) const
{
  return myLeft->forVariables (operation) && myRight->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

inline static const char*
opstring (enum BinopExpression::Op op)
{
  switch (op) {
  case BinopExpression::bPlus:
    return "+";
  case BinopExpression::bMinus:
    return "-";
  case BinopExpression::bDiv:
    return "/";
  case BinopExpression::bMul:
    return "*";
  case BinopExpression::bMod:
    return "%";
  case BinopExpression::bAnd:
    return "&";
  case BinopExpression::bOr:
    return "|";
  case BinopExpression::bXor:
    return "^";
  case BinopExpression::bRol:
    return "<<";
  case BinopExpression::bRor:
    return ">>";
  }

  return "???";
}

void
BinopExpression::prettyPrint (unsigned indent) const
{
  print (indent, "(\n");
  myLeft->prettyPrint (indent + 1);
  print (indent, ") %s (\n", opstring (getOp ()));
  myRight->prettyPrint (indent + 1);
  print (indent, ")\n");
}

#endif // DEBUG
