// Maria boolean binary operator expression class -*- c++ -*-

#include "BooleanBinop.h"
#include "BoolType.h"
#include "BoolValue.h"
#include "Net.h"
#include "Constant.h"

BooleanBinop::BooleanBinop (enum Op op,
			    class Expression* left,
			    class Expression* right) :
  Expression (),
  myOp (op), myLeft (left), myRight (right)
{
  assert (myLeft->getKind () == Expression::eUndefined ||
	  myLeft->getType ()->getKind () == Type::tBool);
  assert (myRight->getKind () == Expression::eUndefined ||
	  myRight->getType ()->getKind () == Type::tBool);
  setType (Net::getBoolType ());
}

BooleanBinop::~BooleanBinop ()
{
  myLeft->destroy ();
  myRight->destroy ();
}

bool
BooleanBinop::isBasic () const
{
  return myLeft->isBasic () && myRight->isBasic ();
}

bool
BooleanBinop::isTemporal () const
{
  return myLeft->isTemporal () || myRight->isTemporal ();
}

const char*
BooleanBinop::getOpString () const
{
  switch (getOp ()) {
  case bAnd:
    return "&&";
  case bOr:
    return "||";
  case bImpl:
    return "=>";
  case bXor:
    return "^^";
  case bEquiv:
    return "<=>";
  }

  return "???";
}

#ifdef EXPR_CACHE
bool
BooleanBinop::invalidate (const class VariableDefinition* variable) const
{
  bool invalidatedLeft = myLeft->invalidate (variable);

  if (myRight->invalidate (variable) || invalidatedLeft) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
BooleanBinop::do_clearValuation (const class Valuation* valuation) const
{
  myLeft->clearValuation (valuation);
  myRight->clearValuation (valuation);
}

class Value*
BooleanBinop::do_eval (const class Valuation* valuation) const
{
  const class Value* left = myLeft->const_eval (valuation);
  if (!left)
    return NULL;
  assert (left->getType ()->getKind () == Type::tBool);
  assert (left->getKind () == Value::vBool);
  const class BoolValue* bleft = static_cast<const class BoolValue*>(left);

  switch (getOp ()) {
  case bAnd:
    if (!bleft->getValue ())
      return constrain (valuation, bleft->copy ());
    break;
  case bOr:
    if (bleft->getValue ())
      return constrain (valuation, bleft->copy ());
    break;
  case bImpl:
    if (!bleft->getValue ())
      return constrain (valuation, new class BoolValue (getType (), true));
    break;
  default:
    break;
  }

  const class Value* right = myRight->const_eval (valuation);
  if (!right) {
    return NULL;
  }
  assert (right->getType ()->getKind () == Type::tBool);
  assert (right->getKind () == Value::vBool);
  const class BoolValue* bright = static_cast<const class BoolValue*>(right);

  switch (getOp ()) {
  case bAnd:
  case bOr:
  case bImpl:
    return constrain (valuation, bright->copy ());
  case bXor:
    return constrain (valuation,
		      new class BoolValue (getType (), !(*bleft == *bright)));
  case bEquiv:
    return constrain (valuation,
		      new class BoolValue (getType (), *bleft == *bright));
  }

  assert (false);
  return NULL;
}
#else // EXPR_CACHE
class Value*
BooleanBinop::do_eval (const class Valuation* valuation) const
{
  class Value* left = myLeft->eval (valuation);
  if (!left)
    return NULL;
  assert (left->getType ()->getKind () == Type::tBool);
  assert (left->getKind () == Value::vBool);
  class BoolValue* bleft = static_cast<class BoolValue*>(left);

  switch (getOp ()) {
  case bAnd:
    if (!bleft->getValue ())
      return constrain (valuation, bleft);
    break;
  case bOr:
    if (bleft->getValue ())
      return constrain (valuation, bleft);
    break;
  case bImpl:
    if (!bleft->getValue ()) {
      bleft->setValue (true);
      return constrain (valuation, bleft);
    }
    break;
  default:
    break;
  }

  class Value* right = myRight->eval (valuation);
  if (!right) {
    delete bleft;
    return NULL;
  }
  assert (right->getType ()->getKind () == Type::tBool);
  assert (right->getKind () == Value::vBool);
  class BoolValue* bright = static_cast<class BoolValue*>(right);

  switch (getOp ()) {
  case bAnd:
  case bOr:
  case bImpl:
    delete bleft;
    return constrain (valuation, bright);
  case bXor:
    bright->setValue (!(*bleft == *bright));
    delete bleft;
    return constrain (valuation, bright);
  case bEquiv:
    bright->setValue (*bleft == *bright);
    delete bleft;
    return constrain (valuation, bright);
  }

  assert (false);
  return NULL;
}
#endif // EXPR_CACHE

class Expression*
BooleanBinop::ground (const class Valuation& valuation)
{
  class Expression* left = myLeft->ground (valuation);
  class Expression* right = myRight->ground (valuation);

  if (!left || !right) {
    left->destroy ();
    right->destroy ();
    return NULL;
  }

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class BooleanBinop (myOp, left, right);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
BooleanBinop::substitute (class Substitution& substitution)
{
  class Expression* left = myLeft->substitute (substitution);
  class Expression* right = myRight->substitute (substitution);

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class BooleanBinop (myOp, left, right);
    expr->setType (getType ());
    return expr;
  }
}

#ifdef EXPR_CACHE
void
BooleanBinop::associateVariables (const class Expression& expr,
				  bool assoc) const
{
  myLeft->associateVariables (expr, assoc);
  myRight->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
BooleanBinop::forVariables (bool (*operation)
			       (const class VariableDefinition&)) const
{
  return myLeft->forVariables (operation) && myRight->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

void
BooleanBinop::prettyPrint (unsigned indent) const
{
  print (indent, "(\n");
  myLeft->prettyPrint (indent + 1);
  print (indent, ") %s (\n", getOpString ());
  myRight->prettyPrint (indent + 1);
  print (indent, ")\n");
}

#endif // DEBUG
