// Maria boolean binary operator expression class -*- c++ -*-

#ifndef BOOLEANBINOP_H_
# define BOOLEANBINOP_H_

# include "Expression.h"

/** Boolean binary operator expression */
class BooleanBinop : public Expression {
public:
  enum Op { bAnd, bOr, bImpl, bXor, bEquiv };

  // Constructors and destructors

  /** Constructs a new BooleanExpression
   * @param op		operator
   * @param left	left-hand-side expression
   * @param right	right-hand-side expression
   */
  BooleanBinop (enum Op op, class Expression* left, class Expression* right);

private:
  /** Copy constructor */
  BooleanBinop (const class BooleanBinop& old);
  /** Assignment operator */
  class BooleanBinop& operator= (const class BooleanBinop& old);
protected:
  /** Destructs BooleanBinop */
  ~BooleanBinop ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBooleanBinop; }

  /** Determine whether this is a basic expression */
  bool isBasic () const;
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const;

  /** Determine the operator */
  enum Op getOp () const { return myOp; }
  /** Determine the string representation of the operator */
  const char* getOpString () const;
  /** Determine the left-hand expression */
  const class Expression* getLeft () const { return myLeft; }
  /** Determine the right-hand expression */
  const class Expression* getRight () const { return myRight; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class BooleanBinop& other) const {
    return myOp == other.myOp &&
      *myLeft == *other.myLeft &&
      *myRight == *other.myRight;
  }

  /** Ordering comparison operator */
  bool operator< (const class BooleanBinop& other) const {
    if (myOp < other.myOp) return true;
    if (other.myOp < myOp) return false;
    if (*myLeft < *other.myLeft) return true;
    if (*other.myLeft < *myLeft) return false;
    return *myRight < *other.myRight;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;
# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The operator */
  enum Op myOp;
  /** Left-hand expression */
  class Expression* myLeft;
  /** Right-hand expression */
  class Expression* myRight;
};

#endif // BOOLEANBINOP_H_
