// Maria buffer expression class -*- c++ -*-

#include "BufferExpression.h"
#include "BufferValue.h"
#include "Constant.h"
#include <cstring>

BufferExpression::BufferExpression (const class BufferType* type) :
  Expression (),
  myComponents (NULL)
{
  assert (type && type->getSize () > 0);
  myComponents = new class Expression*[type->getSize ()];
  memset (myComponents, 0, type->getSize () * sizeof *myComponents);
  setType (type);
}

BufferExpression::~BufferExpression ()
{
  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++)
    myComponents[i]->destroy ();
  delete[] myComponents;
}

void
BufferExpression::setType (const class Type* type)
{
  assert (type && type->getKind () == Type::tBuffer);
  const class Type* itemType =
    static_cast<const class BufferType*>(type)->getItemType ();

  for (card_t size = static_cast<const class BufferType*>(type)->getSize (),
	 i = 0; i < size && myComponents[i]; i++)
    myComponents[i]->setType (itemType);

  Expression::setType (type);
}

#ifdef EXPR_CACHE
bool
BufferExpression::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = false;

  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++)
    if (myComponents[i]->invalidate (variable))
      invalidated = true;

  if (invalidated)
    Expression::invalidate ();

  return invalidated;
}

void
BufferExpression::do_clearValuation (const class Valuation* valuation) const
{
  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++)
    myComponents[i]->clearValuation (valuation);
}
#endif // EXPR_CACHE

class Value*
BufferExpression::do_eval (const class Valuation* valuation) const
{
  class BufferValue* buffer = new class BufferValue (getType ());

  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++) {
    if (class Value* v = myComponents[i]->eval (valuation)) {
      assert (v->getType () ==
	      static_cast<const class BufferType*>
	      (getType ())->getItemType ());
      (*buffer)[i] = v;
    }
    else {
      delete buffer;
      return NULL;
    }
  }

  return constrain (valuation, buffer);
}

class Expression*
BufferExpression::ground (const class Valuation& valuation)
{
  bool same = true;
  const class BufferType* const type =
    static_cast<const class BufferType*>(getType ());
  class BufferExpression* expr = new class BufferExpression (type);

  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++) {
    class Expression* e = myComponents[i]->ground (valuation);
    if (!e) {
      expr->destroy ();
      return NULL;
    }

    if (e != myComponents[i])
      same = false;
    expr->myComponents[i] = e;
  }

  if (same) {
    expr->destroy ();
    return copy ();
  }
  else if (valuation.isOKExcept (Valuation::undefinedValue)) {
    if (class Value* v = expr->eval (&valuation)) {
      class Expression* c = new class Constant (v);
      expr->destroy ();
      return c;
    }
    else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
      expr->destroy ();
      return NULL;
    }
  }

  return expr;
}

class Expression*
BufferExpression::substitute (class Substitution& substitution)
{
  bool same = true;
  const class BufferType* const type =
    static_cast<const class BufferType*>(getType ());
  class BufferExpression* expr = new class BufferExpression (type);

  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++) {
    class Expression* e = myComponents[i]->substitute (substitution);
    if (e != myComponents[i])
      same = false;
    expr->myComponents[i] = e;
  }
  if (same) {
    expr->destroy ();
    return copy ();
  }
  else
    return expr;
}

#ifdef EXPR_CACHE
void
BufferExpression::associateVariables (const class Expression& expr,
				      bool assoc) const
{
  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++)
    myComponents[i]->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
BufferExpression::forVariables (bool (*operation)
				(const class VariableDefinition&)) const
{
  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++)
    if (!myComponents[i]->forVariables (operation))
      return false;
  return true;
}

bool
BufferExpression::isCompatible (const class Value& value,
				const class Valuation& valuation) const
{
  assert (getType () && value.getType ());

  if (value.getKind () != Value::vBuffer)
    return false;

  assert (value.getType ()->getKind () == Type::tBuffer);

  const class BufferValue& bv = static_cast<const class BufferValue&>(value);
  const class BufferType* bt =
    static_cast<const class BufferType*>(getType ());

  assert (bv.getSize () ==
	  static_cast<const class BufferType*>(bv.getType ())->getSize ());

  if (!bv.getType ()->isAssignable (getType ()))
    return false;

  for (card_t size = bt->getSize (), i = 0; i < size; i++) {
    if (myComponents[i]) {
#ifdef EXPR_CACHE
      const class Value* comp = myComponents[i]->const_eval (&valuation);
      if (comp && (!bv[i] || !(*comp == *bv[i])))
	return false;
#else // EXPR_CACHE
      class Value* comp = myComponents[i]->eval (&valuation);
      if (comp && (!bv[i] || !(*comp == *bv[i]))) {
	delete comp;
	return false;
      }
      delete comp;
#endif // EXPR_CACHE
    }
    else
      return !bv[i];
  }

  return true;
}

const class VariableDefinition*
BufferExpression::getLvalue (const class Value*& value,
			     class Valuation& valuation) const
{
  assert (value->getKind () == Value::vBuffer);
  const class BufferValue* bv = static_cast<const class BufferValue*>(value);

  card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
  for (card_t i = 0; i < size && myComponents[i]; i++) {
    const class Value* val = (*bv)[i];
    const class VariableDefinition* v =
      myComponents[i]->getLvalue (val, valuation);
    if (v) {
      value = val;
      return v;
    }
  }

  return NULL;
}

bool
BufferExpression::isLvalue (const class VariableDefinition* variable) const
{
  for (card_t i = static_cast<const class BufferType*>(getType ())->getSize ();
       i--; )
    if (myComponents[i] && myComponents[i]->isLvalue (variable))
      return true;
  return false;
}

#ifdef DEBUG
# include "util.h"

void
BufferExpression::prettyPrint (unsigned indent) const
{
  const class BufferType* type =
    static_cast<const class BufferType*>(getType ());
  print (indent, "{\n");
  for (card_t i = 0; myComponents[i]; ) {
    myComponents[i]->prettyPrint (indent + 1);
    if (++i == type->getSize () || !myComponents[i])
      break;
    print (indent + 1, ",\n");
  }
  print (indent, "}\n");
}

#endif // DEBUG
