// Maria buffer expression class -*- c++ -*-

#ifndef BUFFEREXPRESSION_H_
# define BUFFEREXPRESSION_H_

# include "Expression.h"
# include "BufferType.h"

# include <cassert>

/** Buffer initialization expression */
class BufferExpression : public Expression {
public:
  // Constructors and destructors

  /** Constructs BufferExpression
   * @param type	Type of the expression
   */
  BufferExpression (const class BufferType* type);
private:
  /** Copy constructor */
  BufferExpression (const class BufferExpression& old);
  /** Assignment operator */
  class BufferExpression& operator= (const class BufferExpression& old);
protected:
  /** Destructs BufferExpression */
  ~BufferExpression ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBuffer; }

  /** Set the type of the expression
   * @param type	Type of the expression
   */
  void setType (const class Type* type);

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Append a component
   * @param component	The component expression to be appended
   */
  void append (class Expression* component) {
    assert (component && hasCapacity ());
    assert (component->isBasic ());
    myComponents[getSize ()] = component;
  }

  /** Determine whether there is capacity left */
  bool hasCapacity () const {
    card_t size = static_cast<const class BufferType*>(getType ())->getSize ();
    return size && !myComponents[size - 1];
  }

  /** Determine the current size of the buffer expression */
  card_t getSize () const {
    for (card_t i =
	   static_cast<const class BufferType*>(getType ())->getSize ();
	 i--; ) {
      if (myComponents[i])
	return i + 1;
    }
    return 0;
  }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class BufferExpression& other) const {
    const class BufferType* v =
      static_cast<const class BufferType*>(getType ());
    const class BufferType* u =
      static_cast<const class BufferType*>(other.getType ());
    if (v->isStack () != u->isStack () ||
	v->getSize () != u->getSize ())
      return false;
    for (card_t i = 0; i < v->getSize (); i++)
      if (!myComponents[i]) {
	if (other.myComponents[i])
	  return false;
	else
	  continue;
      }
      else if (!other.myComponents[i] ||
	       !(*(myComponents[i]) == *(other.myComponents[i])))
	return false;

    return true;
  }

  /** Ordering comparison operator */
  bool operator< (const class BufferExpression& other) const {
    const class BufferType* v =
      static_cast<const class BufferType*>(getType ());
    const class BufferType* u =
      static_cast<const class BufferType*>(other.getType ());
    if (v->getSize () < u->getSize ())
      return true;
    if (u->getSize () < v->getSize ())
      return false;
    if (!v->isStack () && u->isStack ())
      return true;
    if (!u->isStack () && v->isStack ())
      return false;
    for (card_t i = v->getSize (); i--; )
      if (!myComponents[i]) {
	if (other.myComponents[i])
	  return true;
	else
	  continue;
      }
      else if (!other.myComponents[i])
	return false;
      else if (*(myComponents[i]) < *(other.myComponents[i]))
	return true;
      else if (*(other.myComponents[i]) < *(myComponents[i]))
	return false;

    return false;
  }

  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Determine whether the expression is compatible with the specified value,
   * neglecting subexpressions that cannot be evaluated
   * @param value	value the expression will be compared to
   * @param valuation	variable substitutions
   * @return		true if the expression is compatible with the value
   */
  bool isCompatible (const class Value& value,
		     const class Valuation& valuation) const;

  /** Get an unbound lvalue and a corresponding value
   * @param value	value the expression should evaluate to
   * @param valuation	variable substitutions
   * @return		the lvalue; set value to the suggested value
   */
  const class VariableDefinition* getLvalue
  (const class Value*& value, class Valuation& valuation) const;

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** The expressions for the buffer components */
  class Expression** myComponents;
};

#endif // BUFFEREXPRESSION_H_
