// Buffer indexing expression (used only during parsing) -*- c++ -*-

#ifndef BUFFERINDEX_H_
# define BUFFERINDEX_H_

# include "Expression.h"

/** Buffer indexing expression (used only during parsing) */
class BufferIndex : public Expression {
public:
  // Constructors and destructors

  /** Constructor
   * @param buffer	The buffer
   * @param index	Indexing expression
   */
  BufferIndex (class Expression* buffer, class Expression* index);
private:
  /** Copy constructor */
  BufferIndex (const class BufferIndex& old);
  /** Assignment operator */
  class BufferIndex& operator= (const class BufferIndex& old);
protected:
  /** Destructor */
  ~BufferIndex ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBufferIndex; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Get the buffer expression */
  class Expression* getBuffer () { return myBuffer->copy (); }
  /** Get the indexing expression */
  class Expression* getIndex () { return myIndex->copy (); }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class BufferIndex& other) const {
    return *myBuffer == *other.myBuffer && *myIndex == *other.myIndex;
  }

  /** Ordering comparison operator */
  bool operator< (const class BufferIndex& other) const {
    if (*myIndex < *other.myIndex) return true;
    if (*other.myIndex < *myIndex) return false;
    return *myBuffer < *other.myBuffer;
  }

  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The buffer expression */
  class Expression* myBuffer;
  /** The indexing expression */
  class Expression* myIndex;
};

#endif // BUFFERINDEX_H_
