// Buffer unary operator class -*- c++ -*-

#include "BufferUnop.h"
#include "BufferType.h"
#include "BufferValue.h"
#include "Net.h"
#include "IntType.h"
#include "IntValue.h"
#include "Constant.h"

BufferUnop::BufferUnop (enum BufferUnop::Op op,
			class Expression* buffer, class Expression* index) :
  myOp (op), myBuffer (buffer), myIndex (index)
{
  assert (myBuffer && myBuffer->getType ()->getKind () == Type::tBuffer);
  assert (!myIndex || myIndex->getType ()->getKind () == Type::tInt);
  assert (myBuffer->isBasic ());
  assert (!myIndex || myIndex->isBasic ());

  switch (myOp) {
  case bPeek:
    setType (static_cast<const class BufferType*>
	     (myBuffer->getType ())->getItemType ());
    break;
  case bRemove:
    setType (myBuffer->getType ());
    break;
  case bFree:
  case bUsed:
    assert (!myIndex);
    setType (Net::getIntType ());
    break;
  default:
    assert (false);
  }
}

BufferUnop::~BufferUnop ()
{
  myBuffer->destroy ();
  myIndex->destroy ();
}

#ifdef EXPR_CACHE
bool
BufferUnop::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = myBuffer->invalidate (variable);
  if (myIndex && myIndex->invalidate (variable))
    invalidated = true;
  if (invalidated)
    Expression::invalidate ();
  return invalidated;
}

void
BufferUnop::do_clearValuation (const class Valuation* valuation) const
{
  myBuffer->clearValuation (valuation);
}

class Value*
BufferUnop::do_eval (const class Valuation* valuation) const
{
  const class Value* buffer = myBuffer->const_eval (valuation);
  if (!buffer)
    return NULL;
  assert (buffer->getType ()->getKind () == Type::tBuffer);

  const class BufferValue* bv = static_cast<const class BufferValue*>(buffer);
  card_t index = 0, capacity = bv->getCapacity ();
  if (myIndex) {
    if (const class Value* iv = myIndex->const_eval (valuation)) {
      assert (iv->getType ()->getKind () == Type::tInt);
      int_t i = static_cast<const class IntValue*>(iv)->getValue ();
      if (i < 0 || card_t (i) >= capacity) {
	if (valuation)
	  valuation->flagBufferViolation ();
	return NULL;
      }
      index = card_t (i);
    }
    else
      return NULL;
  }

  switch (myOp) {
  case bPeek:
    if (!capacity) {
      if (valuation)
	valuation->flagBufferViolation ();
      return NULL;
    }
    return (*bv)[index]->copy ();

  case bRemove:
    if (!capacity) {
      if (valuation)
	valuation->flagBufferViolation ();
      return NULL;
    }
    else {
      class BufferValue* buf = new class BufferValue (getType ());
      card_t i;
      for (i = capacity; --i > index; )
	(*buf)[i - 1] = (*bv)[i]->copy ();
      while (i--)
	(*buf)[i] = (*bv)[i]->copy ();
      return buf;
    }

  case bFree:
    return constrain (valuation,
		      new class IntValue
		      (getType (), bv->getSize () - capacity));
  case bUsed:
    return constrain (valuation,
		      new class IntValue (getType (), capacity));
  }

  assert (false);
  return NULL;
}
#else // EXPR_CACHE
class Value*
BufferUnop::do_eval (const class Valuation* valuation) const
{
  class Value* buffer = myBuffer->eval (valuation);
  if (!buffer)
    return NULL;
  assert (buffer->getType ()->getKind () == Type::tBuffer);

  class BufferValue* bv = static_cast<class BufferValue*>(buffer);
  card_t index = 0, capacity = bv->getCapacity ();
  if (myIndex) {
    if (class Value* iv = myIndex->eval (valuation)) {
      assert (iv->getType ()->getKind () == Type::tInt);
      int_t i = static_cast<const class IntValue*>(iv)->getValue ();
      delete iv;
      if (i < 0 || card_t (i) >= capacity) {
	if (valuation)
	  valuation->flagBufferViolation ();
	delete bv;
	return NULL;
      }
      index = card_t (i);
    }
    else {
      delete bv;
      return NULL;
    }
  }

  switch (myOp) {
  case bPeek:
    if (!capacity) {
      if (valuation)
	valuation->flagBufferViolation ();
      delete bv;
      return NULL;
    }
    else {
      class Value* item = (*bv)[index]->copy ();
      delete bv;
      return item;
    }

  case bRemove:
    if (!capacity) {
      if (valuation)
	valuation->flagBufferViolation ();
      delete bv;
      return NULL;
    }
    else {
      delete (*bv)[index];
      for (card_t i = index; ++i < capacity; )
	(*bv)[i - 1] = (*bv)[i];
      (*bv)[capacity - 1] = NULL;
      return bv;
    }

  case bFree:
    {
      class Value* result =
	constrain (valuation,
		   new class IntValue (getType (), bv->getSize () - capacity));
      delete bv;
      return result;
    }
  case bUsed:
    delete bv;
    return constrain (valuation,
		      new class IntValue (getType (), capacity));
  }

  assert (false);
  return NULL;
}
#endif // EXPR_CACHE

class Expression*
BufferUnop::ground (const class Valuation& valuation)
{
  class Expression* buffer = myBuffer->ground (valuation);
  class Expression* index = myIndex ? myIndex->ground (valuation) : NULL;

  if (!buffer || (myIndex && !index)) {
    buffer->destroy ();
    index->destroy ();
    return NULL;
  }

  if (buffer == myBuffer && index == myIndex) {
    buffer->destroy ();
    index->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class BufferUnop (myOp, buffer, index);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
BufferUnop::substitute (class Substitution& substitution)
{
  class Expression* buffer = myBuffer->substitute (substitution);
  class Expression* index =
    myIndex ? myIndex->substitute (substitution) : NULL;

  if (buffer == myBuffer && index == myIndex) {
    buffer->destroy ();
    index->destroy ();
    return copy ();
  }

  return new class BufferUnop (myOp, buffer, index);
}

#ifdef EXPR_CACHE
void
BufferUnop::associateVariables (const class Expression& expr,
				bool assoc) const
{
  myBuffer->associateVariables (expr, assoc);
  if (myIndex)
    myIndex->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
BufferUnop::forVariables (bool (*operation)
			  (const class VariableDefinition&)) const
{
  return
    myBuffer->forVariables (operation) &&
    myIndex->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

void
BufferUnop::prettyPrint (unsigned indent) const
{
  switch (myOp) {
  case bPeek:
    print (indent, "*\n");
    break;
  case bRemove:
    print (indent, "-\n");
    break;
  case bFree:
    print (indent, "%%\n");
    break;
  case bUsed:
    print (indent, "/\n");
    break;
  }

  if (myIndex) {
    print (indent, "(\n");
    myBuffer->prettyPrint (indent + 1);
    print (indent + 1, "[\n");
    myIndex->prettyPrint (indent + 2);
    print (indent + 1, "]\n");
    print (indent, ")\n");
  }
  else
    myBuffer->prettyPrint (indent);
}

#endif // DEBUG
