// Buffer unary operator class -*- c++ -*-

#ifndef BUFFERUNOP_H_
# define BUFFERUNOP_H_

# include "Expression.h"

/** Buffer Unary operator expression */
class BufferUnop : public Expression {
public:
  /** Unary operators */
  enum Op { bPeek, bRemove, bFree, bUsed };

  // Constructors and destructors

  /** Constructs a new BufferUnop
   * @param op		Operator
   * @param buffer	Buffer the operator is applied to
   * @param index	Indexing expression
   */
  BufferUnop (enum Op op, class Expression* buffer, class Expression* index);

private:
  /** Copy constructor */
  BufferUnop (const class BufferUnop& old);
  /** Assignment operator */
  class BufferUnop& operator= (const class BufferUnop& old);
protected:
  /** Destructor */
  ~BufferUnop ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBufferUnop; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class BufferUnop& other) const {
    return myOp == other.myOp &&
      (myIndex || !other.myIndex) && (other.myIndex || !myIndex) &&
      *myBuffer == *other.myBuffer && (!myIndex || *myIndex == *other.myIndex);
  }

  /** Ordering comparison operator */
  bool operator< (const class BufferUnop& other) const {
    if (myOp < other.myOp) return true;
    if (other.myOp < myOp) return false;
    if (!myIndex && other.myIndex)
      return true;
    if (!other.myIndex && myIndex)
      return false;
    if (myIndex) {
      if (*myIndex < *other.myIndex) return true;
      if (*other.myIndex < *myIndex) return false;
    }
    return *myBuffer < *other.myBuffer;
  }

  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The operator */
  enum Op myOp;
  /** The buffer expression */
  class Expression* myBuffer;
  /** The optional index expression */
  class Expression* myIndex;
};

#endif // BUFFERUNOP_H_
