// Buffer write expression -*- c++ -*-

#include "BufferWrite.h"
#include "BufferType.h"
#include "BufferValue.h"
#include "IntValue.h"
#include "Constant.h"

BufferWrite::BufferWrite (class Expression* buffer, class Expression* item,
			  class Expression* index) :
  myBuffer (buffer), myItem (item), myIndex (index)
{
  assert (myBuffer && myItem);
  assert (myBuffer->getType ()->getKind () == Type::tBuffer);
  assert (myItem->isAssignable (static_cast<const class BufferType*>
				(myBuffer->getType ())->getItemType ()));
  assert (!myIndex || myIndex->getType ()->getKind () == Type::tInt);
  assert (myBuffer->isBasic () && myItem->isBasic ());
  assert (!myIndex || myIndex->isBasic ());

  setType (myBuffer->getType ());
}

BufferWrite::~BufferWrite ()
{
  myBuffer->destroy ();
  myItem->destroy ();
  myIndex->destroy ();
}

#ifdef EXPR_CACHE
bool
BufferWrite::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = myBuffer->invalidate (variable);
  if (myItem->invalidate (variable))
    invalidated = true;
  if (myIndex && myIndex->invalidate (variable))
    invalidated = true;
  if (invalidated)
    Expression::invalidate ();
  return invalidated;
}

void
BufferWrite::do_clearValuation (const class Valuation* valuation) const
{
  myBuffer->clearValuation (valuation);
  myItem->clearValuation (valuation);
}

class Value*
BufferWrite::do_eval (const class Valuation* valuation) const
{
  const class Value* buffer = myBuffer->const_eval (valuation);
  if (!buffer)
    return NULL;
  const class Value* item = myItem->const_eval (valuation);
  if (!item)
    return NULL;

  assert (buffer->getType ()->getKind () == Type::tBuffer);
  assert (item->getType ()
	  ->isAssignable (static_cast<const class BufferType*>
			  (buffer->getType ())->getItemType ()));

  const class BufferValue* bv = static_cast<const class BufferValue*>(buffer);
  card_t index = 0, capacity = bv->getCapacity ();
  if (myIndex) {
    if (const class Value* iv = myIndex->const_eval (valuation)) {
      assert (iv->getType ()->getKind () == Type::tInt);
      int_t i = static_cast<const class IntValue*>(iv)->getValue ();
      if (i < 0 || card_t (i) >= capacity) {
	if (valuation)
	  valuation->flagBufferViolation ();
	return NULL;
      }
      index = card_t (i);
    }
    else
      return NULL;
  }

  if (capacity == bv->getSize ()) {
    if (valuation)
      valuation->flagBufferViolation ();
    return NULL;
  }

  class BufferValue* buf = new class BufferValue (getType ());

  if (static_cast<const class BufferType*>(getType ())->isStack ()) {
    card_t i;
    for (i = capacity; i > index; i--)
      (*buf)[i] = (*bv)[i - 1]->copy ();
    (*buf)[i] = item->copy ();
    while (i--)
      (*buf)[i] = (*bv)[i]->copy ();
  }
  else {
    card_t i;
    for (i = capacity; i > capacity - index; i--)
      (*buf)[i] = (*bv)[i - 1]->copy ();
    (*buf)[i] = item->copy ();
    while (i--)
      (*buf)[i] = (*bv)[i]->copy ();
  }

  return constrain (valuation, buf);
}
#else // EXPR_CACHE
class Value*
BufferWrite::do_eval (const class Valuation* valuation) const
{
  class Value* buffer = myBuffer->eval (valuation);
  if (!buffer)
    return NULL;
  class Value* item = myItem->eval (valuation);
  if (!item) {
    delete buffer;
    return NULL;
  }

  assert (buffer->getType ()->getKind () == Type::tBuffer);
  assert (item->getType ()
	  ->isAssignable (static_cast<const class BufferType*>
			  (buffer->getType ())->getItemType ()));

  class BufferValue* bv = static_cast<class BufferValue*>(buffer);
  card_t index = 0, capacity = bv->getCapacity ();
  if (myIndex) {
    if (class Value* iv = myIndex->eval (valuation)) {
      assert (iv->getType ()->getKind () == Type::tInt);
      int_t i = static_cast<const class IntValue*>(iv)->getValue ();
      delete iv;
      if (i < 0 || card_t (i) >= capacity) {
	if (valuation)
	  valuation->flagBufferViolation ();
	delete buffer, delete item;
	return NULL;
      }
      index = card_t (i);
    }
    else {
      delete buffer, delete item;
      return NULL;
    }
  }

  if (capacity == bv->getSize ()) {
    if (valuation)
      valuation->flagBufferViolation ();
    delete buffer, delete item;
    return NULL;
  }

  card_t i;
  if (static_cast<const class BufferType*>(getType ())->isStack ())
    for (i = capacity; i > index; i--)
      (*bv)[i] = (*bv)[i - 1];
  else
    for (i = capacity; i > capacity - index; i--)
      (*bv)[i] = (*bv)[i - 1];
  (*bv)[i] = item;

  return bv;
}
#endif // EXPR_CACHE

class Expression*
BufferWrite::ground (const class Valuation& valuation)
{
  class Expression* buffer = myBuffer->ground (valuation);
  class Expression* item = myItem->ground (valuation);
  class Expression* index = myIndex ? myIndex->ground (valuation) : NULL;

  if (!buffer || !item || (myIndex && !index)) {
    buffer->destroy ();
    item->destroy ();
    index->destroy ();
    return NULL;
  }

  if (buffer == myBuffer && item == myItem && index == myIndex) {
    buffer->destroy ();
    item->destroy ();
    index->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class BufferWrite (buffer, item, index);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
BufferWrite::substitute (class Substitution& substitution)
{
  class Expression* buffer = myBuffer->substitute (substitution);
  class Expression* item = myItem->substitute (substitution);
  class Expression* index =
    myIndex ? myIndex->substitute (substitution) : NULL;

  if (buffer == myBuffer && item == myItem && index == myIndex) {
    buffer->destroy ();
    item->destroy ();
    index->destroy ();
    return copy ();
  }

  return new class BufferWrite (buffer, item, index);
}

#ifdef EXPR_CACHE
void
BufferWrite::associateVariables (const class Expression& expr,
				 bool assoc) const
{
  myBuffer->associateVariables (expr, assoc);
  myItem->associateVariables (expr, assoc);
  if (myIndex)
    myIndex->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
BufferWrite::forVariables (bool (*operation)
			   (const class VariableDefinition&)) const
{
  return
    myBuffer->forVariables (operation) &&
    myItem->forVariables (operation) &&
    (!myIndex || myIndex->forVariables (operation));
}

#ifdef DEBUG
# include "util.h"

void
BufferWrite::prettyPrint (unsigned indent) const
{
  if (myIndex) {
    print (indent, "(\n");
    myBuffer->prettyPrint (indent + 1);
    print (indent + 1, "[\n");
    myIndex->prettyPrint (indent + 2);
    print (indent + 1, "]\n");
    print (indent, ") +\n");
  }
  else {
    myBuffer->prettyPrint (indent);
    print (indent, "+\n");
  }
  myItem->prettyPrint (indent);
}

#endif // DEBUG
