// Buffer write expression -*- c++ -*-

#ifndef BUFFERWRITE_H_
# define BUFFERWRITE_H_

# include "Expression.h"

/** Buffer write expression */
class BufferWrite : public Expression {
public:
  // Constructors and destructors

  /** Constructor
   * @param buffer	The buffer
   * @param item	The item to be read or written
   * @param index	Indexing expression
   */
  BufferWrite (class Expression* buffer, class Expression* item,
	       class Expression* index);
private:
  /** Copy constructor */
  BufferWrite (const class BufferWrite& old);
  /** Assignment operator */
  class BufferWrite& operator= (const class BufferWrite& old);
protected:
  /** Destructor */
  ~BufferWrite ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eBufferWrite; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class BufferWrite& other) const {
    return (myIndex || !other.myIndex) && (other.myIndex || !myIndex) &&
      *myBuffer == *other.myBuffer &&
      *myItem == *other.myItem &&
      (!myIndex || *myIndex == *other.myIndex);
  }

  /** Ordering comparison operator */
  bool operator< (const class BufferWrite& other) const {
    if (!myIndex && other.myIndex)
      return true;
    if (!other.myIndex && myIndex)
      return false;
    if (myIndex) {
      if (*myIndex < *other.myIndex) return true;
      if (*other.myIndex < *myIndex) return false;
    }
    if (*myBuffer < *other.myBuffer) return true;
    if (*other.myBuffer < *myBuffer) return false;
    return *myItem < *other.myItem;
  }

  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The buffer expression */
  class Expression* myBuffer;
  /** The item expression */
  class Expression* myItem;
  /** The optional index expression */
  class Expression* myIndex;
};

#endif // BUFFERWRITE_H_
