// Maria expression class -*- c++ -*-

#include "Expression.h"
#include "allExpressions.h"

#include <cassert>

Expression::Expression () :
  myReferences (1), myType (0),
#ifdef EXPR_CACHE
  myValue (0), myValuation (0),
#endif // EXPR_CACHE
  myIsAtomic (false)
{
}

Expression::~Expression ()
{
  assert (!myReferences);
#ifdef EXPR_CACHE
  delete myValue;
#endif // EXPR_CACHE
}

void
Expression::setType (const class Type* type)
{
  assert (!myType || type->isAssignable (myType));
  if (type == myType)
    return;
#ifdef EXPR_CACHE
  assert (!myValue ||
	  (getKind () == eConstant && myValue &&
	   type->isConstrained (myValue)));
  myType = type;
  if (myValue)
    myValue->setType (type);
#else // EXPR_CACHE
  myType = type;
#endif // EXPR_CACHE
}

bool
Expression::operator== (const class Expression& other) const
{
  if (this == &other)
    return true;
  if (getKind () != other.getKind () || getType () != other.getType ())
    return false;

  switch (getKind ()) {
  case eVariable:
    return
      static_cast<const class Variable&>(*this) ==
      static_cast<const class Variable&>(other);
  case eConstant:
    return
      static_cast<const class Constant&>(*this) ==
      static_cast<const class Constant&>(other);
  case eUndefined:
    return
      static_cast<const class Undefined&>(*this) ==
      static_cast<const class Undefined&>(other);
  case eStruct:
    return
      static_cast<const class StructExpression&>(*this) ==
      static_cast<const class StructExpression&>(other);
  case eStructComponent:
    return
      static_cast<const class StructComponent&>(*this) ==
      static_cast<const class StructComponent&>(other);
  case eUnion:
    return
      static_cast<const class UnionExpression&>(*this) ==
      static_cast<const class UnionExpression&>(other);
  case eUnionComponent:
    return
      static_cast<const class UnionComponent&>(*this) ==
      static_cast<const class UnionComponent&>(other);
  case eUnionType:
    return
      static_cast<const class UnionTypeExpression&>(*this) ==
      static_cast<const class UnionTypeExpression&>(other);
  case eVector:
    return
      static_cast<const class VectorExpression&>(*this) ==
      static_cast<const class VectorExpression&>(other);
  case eVectorIndex:
    return
      static_cast<const class VectorIndex&>(*this) ==
      static_cast<const class VectorIndex&>(other);
  case eVectorComponent:
    return
      static_cast<const class VectorComponent&>(*this) ==
      static_cast<const class VectorComponent&>(other);
  case eVectorDefined:
    return
      static_cast<const class VectorDefined&>(*this) ==
      static_cast<const class VectorDefined&>(other);
  case eUnop:
    return
      static_cast<const class UnopExpression&>(*this) ==
      static_cast<const class UnopExpression&>(other);
  case eBinop:
    return
      static_cast<const class BinopExpression&>(*this) ==
      static_cast<const class BinopExpression&>(other);
  case eBooleanBinop:
    return
      static_cast<const class BooleanBinop&>(*this) ==
      static_cast<const class BooleanBinop&>(other);
  case eNot:
    return
      static_cast<const class NotExpression&>(*this) ==
      static_cast<const class NotExpression&>(other);
  case eRelop:
    return
      static_cast<const class RelopExpression&>(*this) ==
      static_cast<const class RelopExpression&>(other);
  case eBuffer:
    return
      static_cast<const class BufferExpression&>(*this) ==
      static_cast<const class BufferExpression&>(other);
  case eBufferUnop:
    return
      static_cast<const class BufferUnop&>(*this) ==
      static_cast<const class BufferUnop&>(other);
  case eBufferWrite:
    return
      static_cast<const class BufferWrite&>(*this) ==
      static_cast<const class BufferWrite&>(other);
  case eBufferIndex:
    return
      static_cast<const class BufferIndex&>(*this) ==
      static_cast<const class BufferIndex&>(other);
  case eSet:
    return
      static_cast<const class SetExpression&>(*this) ==
      static_cast<const class SetExpression&>(other);
  case eIfThenElse:
    return
      static_cast<const class IfThenElse&>(*this) ==
      static_cast<const class IfThenElse&>(other);
  case eTemporalBinop:
    return
      static_cast<const class TemporalBinop&>(*this) ==
      static_cast<const class TemporalBinop&>(other);
  case eTemporalUnop:
    return
      static_cast<const class TemporalUnop&>(*this) ==
      static_cast<const class TemporalUnop&>(other);
  case eTemporalQuantifier:
    return
      static_cast<const class TemporalQuantifier&>(*this) ==
      static_cast<const class TemporalQuantifier&>(other);
  case eTypecast:
    return
      static_cast<const class Typecast&>(*this) ==
      static_cast<const class Typecast&>(other);
  case eCardinality:
    return
      static_cast<const class CardinalityExpression&>(*this) ==
      static_cast<const class CardinalityExpression&>(other);
  case eMarking:
    return
      static_cast<const class Marking&>(*this) ==
      static_cast<const class Marking&>(other);
  case eTransitionQualifier:
    return
      static_cast<const class TransitionQualifier&>(*this) ==
      static_cast<const class TransitionQualifier&>(other);
  case ePlaceContents:
    return
      static_cast<const class PlaceContents&>(*this) ==
      static_cast<const class PlaceContents&>(other);
  case eSubmarking:
    return
      static_cast<const class Submarking&>(*this) ==
      static_cast<const class Submarking&>(other);
  case eEmptySet:
    return
      static_cast<const class EmptySet&>(*this) ==
      static_cast<const class EmptySet&>(other);
  }

  assert (false);
  return false;
}

bool
Expression::operator< (const class Expression& other) const
{
  if (this == &other)
    return false;
  if (getKind () < other.getKind ())
    return true;
  if (getKind () > other.getKind ())
    return false;
  if (getType () < other.getType ())
    return true;
  if (getType () > other.getType ())
    return false;

  switch (getKind ()) {
  case eVariable:
    return
      static_cast<const class Variable&>(*this) <
      static_cast<const class Variable&>(other);
  case eConstant:
    return
      static_cast<const class Constant&>(*this) <
      static_cast<const class Constant&>(other);
  case eUndefined:
    return
      static_cast<const class Undefined&>(*this) <
      static_cast<const class Undefined&>(other);
  case eStruct:
    return
      static_cast<const class StructExpression&>(*this) <
      static_cast<const class StructExpression&>(other);
  case eStructComponent:
    return
      static_cast<const class StructComponent&>(*this) <
      static_cast<const class StructComponent&>(other);
  case eUnion:
    return
      static_cast<const class UnionExpression&>(*this) <
      static_cast<const class UnionExpression&>(other);
  case eUnionComponent:
    return
      static_cast<const class UnionComponent&>(*this) <
      static_cast<const class UnionComponent&>(other);
  case eUnionType:
    return
      static_cast<const class UnionTypeExpression&>(*this) <
      static_cast<const class UnionTypeExpression&>(other);
  case eVector:
    return
      static_cast<const class VectorExpression&>(*this) <
      static_cast<const class VectorExpression&>(other);
  case eVectorIndex:
    return
      static_cast<const class VectorIndex&>(*this) <
      static_cast<const class VectorIndex&>(other);
  case eVectorComponent:
    return
      static_cast<const class VectorComponent&>(*this) <
      static_cast<const class VectorComponent&>(other);
  case eVectorDefined:
    return
      static_cast<const class VectorDefined&>(*this) <
      static_cast<const class VectorDefined&>(other);
  case eUnop:
    return
      static_cast<const class UnopExpression&>(*this) <
      static_cast<const class UnopExpression&>(other);
  case eBinop:
    return
      static_cast<const class BinopExpression&>(*this) <
      static_cast<const class BinopExpression&>(other);
  case eBooleanBinop:
    return
      static_cast<const class BooleanBinop&>(*this) <
      static_cast<const class BooleanBinop&>(other);
  case eNot:
    return
      static_cast<const class NotExpression&>(*this) <
      static_cast<const class NotExpression&>(other);
  case eRelop:
    return
      static_cast<const class RelopExpression&>(*this) <
      static_cast<const class RelopExpression&>(other);
  case eBuffer:
    return
      static_cast<const class BufferExpression&>(*this) <
      static_cast<const class BufferExpression&>(other);
  case eBufferUnop:
    return
      static_cast<const class BufferUnop&>(*this) <
      static_cast<const class BufferUnop&>(other);
  case eBufferWrite:
    return
      static_cast<const class BufferWrite&>(*this) <
      static_cast<const class BufferWrite&>(other);
  case eBufferIndex:
    return
      static_cast<const class BufferIndex&>(*this) <
      static_cast<const class BufferIndex&>(other);
  case eSet:
    return
      static_cast<const class SetExpression&>(*this) <
      static_cast<const class SetExpression&>(other);
  case eIfThenElse:
    return
      static_cast<const class IfThenElse&>(*this) <
      static_cast<const class IfThenElse&>(other);
  case eTemporalBinop:
    return
      static_cast<const class TemporalBinop&>(*this) <
      static_cast<const class TemporalBinop&>(other);
  case eTemporalUnop:
    return
      static_cast<const class TemporalUnop&>(*this) <
      static_cast<const class TemporalUnop&>(other);
  case eTemporalQuantifier:
    return
      static_cast<const class TemporalQuantifier&>(*this) <
      static_cast<const class TemporalQuantifier&>(other);
  case eTypecast:
    return
      static_cast<const class Typecast&>(*this) <
      static_cast<const class Typecast&>(other);
  case eCardinality:
    return
      static_cast<const class CardinalityExpression&>(*this) <
      static_cast<const class CardinalityExpression&>(other);
  case eMarking:
    return
      static_cast<const class Marking&>(*this) <
      static_cast<const class Marking&>(other);
  case eTransitionQualifier:
    return
      static_cast<const class TransitionQualifier&>(*this) <
      static_cast<const class TransitionQualifier&>(other);
  case ePlaceContents:
    return
      static_cast<const class PlaceContents&>(*this) <
      static_cast<const class PlaceContents&>(other);
  case eSubmarking:
    return
      static_cast<const class Submarking&>(*this) <
      static_cast<const class Submarking&>(other);
  case eEmptySet:
    return
      static_cast<const class EmptySet&>(*this) <
      static_cast<const class EmptySet&>(other);
  }

  assert (false);
  return false;
}

#ifdef EXPR_CACHE
bool
Expression::invalidate (const class VariableDefinition* variable) const
{
  assert (getKind () == eConstant || getKind () == eUndefined);
  return false;
}

void
Expression::do_clearValuation (const class Valuation* valuation) const
{
  assert (!myValuation);
}
#endif // EXPR_CACHE

class Value*
Expression::do_eval (const class Valuation* valuation) const
{
  assert (false);
  return 0;
}

class PlaceMarking*
Expression::meval (const class Valuation& valuation) const
{
  assert (false);
  return 0;
}

#ifdef EXPR_CACHE
void
Expression::invalidate () const
{
  assert (getKind () != eConstant && getKind () != eMarking);
  delete myValue;
  myValue = 0;
}
#endif // EXPR_CACHE

class Expression*
Expression::ground (const class Valuation& valuation)
{
  return copy ();
}

class Expression*
Expression::substitute (class Substitution& substitution)
{
  return copy ();
}

#ifdef EXPR_CACHE
void
Expression::associateVariables (const class Expression& expr, bool assoc) const
{
}
#endif // EXPR_CACHE

bool
Expression::forVariables (bool (*operation)
			  (const class VariableDefinition&)) const
{
  return true;
}

bool
Expression::isTypeCompatible (const class Value& value) const
{
  assert (value.getType () != NULL);
  return isAssignable (value.getType ());
}

bool
Expression::isCompatible (const class Value& value,
			  const class Valuation& valuation) const
{
  assert (isTypeCompatible (value));

#ifdef EXPR_CACHE
  if (const class Value* v = const_eval (&valuation))
    return value == *v;
#else // EXPR_CACHE
  if (class Value* v = eval (&valuation)) {
    bool compatible = value == *v;
    delete v;
    return compatible;
  }
#endif

  return true;
}

const class VariableDefinition*
Expression::getLvalue (const class Value*& value,
		       class Valuation& valuation) const
{
  return 0;
}

bool
Expression::isLvalue (const class VariableDefinition* variable) const
{
  return false;
}
