// Maria expression class -*- c++ -*-

#ifndef EXPRESSION_H_
# define EXPRESSION_H_

# include "Type.h"
# include "Constraint.h"
# include "Value.h"
# include "Valuation.h"

/** Abstract base class for expressions */
class Expression {
  friend class Constant;
public:

  /** Expression kinds (@see getKind) */
  enum Kind {
    eVariable, eConstant, eUndefined,
    eStruct, eStructComponent, eUnion, eUnionComponent, eUnionType,
    eVector, eVectorIndex, eVectorComponent, eVectorDefined,
    eUnop, eBinop, eBooleanBinop, eNot, eRelop,
    eBuffer, eBufferUnop, eBufferWrite, eBufferIndex,
    eSet, eIfThenElse, eTemporalBinop, eTemporalUnop, eTemporalQuantifier,
    eTypecast, eCardinality, eMarking, eTransitionQualifier,
    ePlaceContents, eSubmarking, eEmptySet
  };

  // Constructors and destructors

  /** Constructs a new Expression */
  Expression ();

private:
  /** Copy constructor */
  Expression (const class Expression& old);
  /** Assignment operator */
  class Expression& operator= (const class Expression& old);
protected:
  /** Destructs Expression */
  virtual ~Expression ();
public:
  /** Get a shared copy of the expression */
  class Expression* copy () { myReferences++; return this; }
  /** Destructs Expression, provided that there aren't any references to it */
  void destroy () { if (this && !--myReferences) delete this; }
  /** Determine whether this is the last copy of an expression */
  bool isLastCopy () const { return myReferences == 1; }

  // Accessors

  /** Determine the type of the expression */
  virtual enum Expression::Kind getKind () const = 0;

  /** Set the atomicity flag
   * @param atomic	specifies whether this should be treated
   *			as an atomic (non-temporal) expression
   */
  void setAtomic (bool atomic) { myIsAtomic = atomic; }

  /** Read the atomicity flag */
  bool isAtomic () const { return myIsAtomic; }

  /** Set the type of the expression
   * @param type	Type of the expression
   */
  virtual void setType (const class Type* type);

  /** Get the type associated with the expression */
  const class Type* getType () const { return myType; }

  /**
   * Determine whether this is a basic expression containing
   * no temporal logic or set operations
   * @return	true if this is a basic expression
   */
  virtual bool isBasic () const = 0;
  /**
   * Determine whether this is a temporal logic expression
   * @return	true if this is a temporal logic expression
   */
  virtual bool isTemporal () const = 0;

# ifdef EXPR_CACHE
  /** Determine whether the value of the expression is in the cache
   * @return	true if the value is cached
   */
  bool isCached () const { return myValue != 0; }
# endif // EXPR_CACHE

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class Expression& other) const;

  /** Ordering comparison operator */
  bool operator< (const class Expression& other) const;

# ifdef EXPR_CACHE
  /** Evaluate the expression or return its cached value
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  const class Value* const_eval (const class Valuation* valuation) const {
    if (!myValue)
      myValue = do_eval (myValuation = valuation);
    else if (getKind () != eConstant && valuation != myValuation) {
      delete myValue;
      myValue = do_eval (myValuation = valuation);
    }
    return myValue;
  }

  /** Evaluate the expression or return a copy of its cached value
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* eval (const class Valuation* valuation) const {
    if (!myValue)
      myValue = do_eval (myValuation = valuation);
    else if (getKind () != eConstant && valuation != myValuation) {
      delete myValue;
      myValue = do_eval (myValuation = valuation);
    }
    return myValue ? myValue->copy () : 0;
  }

  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  virtual bool invalidate (const class VariableDefinition* variable) const;

  /** Clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void clearValuation (const class Valuation* valuation) const {
    if (myValuation != valuation && getKind () != eMarking)
      return;
    myValuation = NULL;
    do_clearValuation (valuation);
  }
  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  virtual void do_clearValuation (const class Valuation* valuation) const;

# else // EXPR_CACHE
  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* eval (const class Valuation* valuation) const {
    return do_eval (valuation);
  }
# endif // EXPR_CACHE

  /** Evaluate the multiset expression
   * @param valuation	variable substitutions and the global marking
   * @return		a marking
   */
  virtual class PlaceMarking* meval (const class Valuation& valuation) const;

private:
  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  virtual class Value* do_eval (const class Valuation* valuation) const;

protected:
  /** Check the constraints of a value
   * @param valuation	Variable substitutions (for error reporting)
   * @param value	the value to be checked
   * @return		value, or NULL in case of error
   */
  class Value* constrain (const class Valuation* valuation,
			  class Value* value) const {
    assert (value != NULL);
    if (const class Constraint* const constraint = myType->getConstraint ()) {
      if (!constraint->isConstrained (value)) {
	if (valuation)
	  valuation->flagConstraintViolation ();
	delete value;
	return NULL;
      }
    }
    return value;
  }
# ifdef EXPR_CACHE
  /** Get the cached value of the expression */
  const class Value* getValue () const { return myValue; }
  /** Unconditionally invalidate the cached value of the expression */
  void invalidate () const;
# endif // EXPR_CACHE

public:
  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  virtual class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  virtual class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  virtual void associateVariables (const class Expression& expr,
				   bool assoc) const;

  /** Associate the variables in the expression
   * @param assoc	false=disassociate
   */
  void associateVariables (bool assoc) const {
    associateVariables (*this, assoc);
  }
# else // EXPR_CACHE
  void associateVariables (bool assoc) const { assoc = assoc; }
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  virtual bool forVariables (bool (*operation)
			     (const class VariableDefinition&)) const;

  /** Determine whether the expression is assignable to the given data type
   * (perform static analysis)
   * @param type	Data type to be checked
   * @return		true if the expression is assignable to the type
   */
  bool isAssignable (const class Type* type) const {
    return myType->isAssignable (type);
  }

  /** Determine whether the expression is type compatible with the specified
   * value (whether it ever could evaluate to the value)
   */
  bool isTypeCompatible (const class Value& value) const;

  /** Determine whether the expression is compatible with the specified value,
   * neglecting subexpressions that cannot be evaluated
   * @param value	value the expression will be compared to
   * @param valuation	variable substitutions
   * @return		true if the expression is compatible with the value
   */
  virtual bool isCompatible (const class Value& value,
			     const class Valuation& valuation) const;

  /** Get an unbound lvalue and a corresponding value
   * @param value	value the expression should evaluate to
   * @param valuation	variable substitutions
   * @return		the lvalue; set value to the suggested value
   */
  virtual const class VariableDefinition* getLvalue
  (const class Value*& value, class Valuation& valuation) const;

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  virtual bool isLvalue (const class VariableDefinition* variable) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  virtual void prettyPrint (unsigned indent) const = 0;
# endif // DEBUG

private:
  /** Number of references to the expression */
  unsigned myReferences;
  /** Type of the expression */
  const class Type* myType;
# ifdef EXPR_CACHE
  /** Cached value of the expression */
  mutable class Value* myValue;
  /** Last valuation used for evaluating the expression */
  mutable const class Valuation* myValuation;
# endif // EXPR_CACHE
  /** Flag: is this an atomic expression */
  bool myIsAtomic;
};

#endif // EXPRESSION_H_
