// Expression list class -*- c++ -*-

#include "ExpressionList.h"
#include "Expression.h"
#include <cstring>

ExpressionList::ExpressionList () :
  myReferences (1), mySize (0), myExprs (0)
{
}

ExpressionList::~ExpressionList ()
{
  assert (!myReferences);
  for (card_t i = 0; i < size (); i++)
    myExprs[i]->destroy ();
  delete[] myExprs;
}

void
ExpressionList::append (class Expression* expression)
{
  assert (expression && (expression->getKind () == Expression::eUndefined ||
			 expression->getType ()));
  if (empty ()) {
    assert (!myExprs);
    myExprs = new class Expression*[1];
  }
  else {
    class Expression** exprs = new class Expression*[mySize + 1];
    memcpy (exprs, myExprs, mySize * sizeof *exprs);
    delete[] myExprs;
    myExprs = exprs;
  }

  myExprs[mySize++] = expression;
}

void
ExpressionList::prepend (class Expression* expression)
{
  assert (expression && (expression->getKind () == Expression::eUndefined ||
			 expression->getType ()));
  if (empty ()) {
    assert (!myExprs);
    myExprs = new class Expression*[1];
  }
  else {
    class Expression** exprs = new class Expression*[mySize + 1];
    memcpy (exprs + 1, myExprs, mySize * sizeof *exprs);
    delete[] myExprs;
    myExprs = exprs;
  }

  *myExprs = expression, mySize++;
}

const class Type*
ExpressionList::getType () const
{
  const class Type* type = NULL;
  for (card_t i = 0; i < mySize; i++)
    if ((type = myExprs[i]->getType ()))
      break;
  return type;
}

bool
ExpressionList::isBasic () const
{
  for (card_t i = 0; i < size (); i++)
    if (!myExprs[i]->isBasic ())
      return false;
  return true;
}

bool
ExpressionList::isTemporal () const
{
  for (card_t i = 0; i < size (); i++)
    if (myExprs[i]->isTemporal ())
      return true;
  return false;
}

bool
ExpressionList::operator== (const class ExpressionList& other) const
{
  if (this == &other)
    return true;
  if (size () != other.size ())
    return false;
  for (card_t i = 0; i < size (); i++)
    if (!(*myExprs[i] == *other.myExprs[i]))
      return false;
  return true;
}

bool
ExpressionList::operator< (const class ExpressionList& other) const
{
  if (this == &other)
    return false;
  if (size () < other.size ())
    return true;
  if (other.size () < size ())
    return false;
  for (card_t i = 0; i < size (); i++) {
    if (*myExprs[i] < *other.myExprs[i])
      return true;
    if (*other.myExprs[i] < *myExprs[i])
      return false;
  }
  return false;
}

#ifdef EXPR_CACHE
bool
ExpressionList::invalidate (const class VariableDefinition* variable)
{
  bool invalidated = false;

  for (card_t i = 0; i < size (); i++)
    if (myExprs[i]->invalidate (variable))
      invalidated = true;

  return invalidated;
}

bool
ExpressionList::invalidate (const class VariableDefinition* variable,
			    card_t index)
{
  bool invalidated = false;

  for (card_t i = 0; i < size (); i++)
    if (myExprs[i]->invalidate (variable) && i == index)
      invalidated = true;

  return invalidated;
}

void
ExpressionList::do_clearValuation (const class Valuation* valuation) const
{
  for (card_t i = 0; i < size (); i++)
    myExprs[i]->clearValuation (valuation);
}
#endif // EXPR_CACHE

class ExpressionList*
ExpressionList::ground (const class Valuation& valuation)
{
  bool same = true;
  class ExpressionList* l = new class ExpressionList;

  for (card_t i = 0; i < size (); i++) {
    class Expression* e = myExprs[i]->ground (valuation);
    if (!e) {
      l->destroy ();
      return NULL;
    }

    if (e != myExprs[i])
      same = false;
    l->append (e);
  }

  if (same) {
    l->destroy ();
    return copy ();
  }

  return l;
}

class ExpressionList*
ExpressionList::substitute (class Substitution& substitution)
{
  bool same = true;
  class ExpressionList* l = new class ExpressionList;

  for (card_t i = 0; i < size (); i++) {
    class Expression* e = myExprs[i]->substitute (substitution);
    if (e != myExprs[i])
      same = false;
    l->append (e);
  }

  if (same) {
    l->destroy ();
    return copy ();
  }

  return l;
}

#ifdef EXPR_CACHE
void
ExpressionList::associateVariables (const class Expression& expr,
				    bool assoc) const
{
  for (card_t i = 0; i < size (); i++)
    myExprs[i]->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
ExpressionList::forVariables (bool (*operation)
			      (const class VariableDefinition&)) const
{
  for (card_t i = 0; i < size (); i++)
    if (!myExprs[i]->forVariables (operation))
      return false;
  return true;
}

bool
ExpressionList::isLvalue (const class VariableDefinition* variable) const
{
  for (card_t i = 0; i < size (); i++)
    if (myExprs[i]->isLvalue (variable))
      return true;
  return false;
}

#ifdef DEBUG
# include "util.h"

void
ExpressionList::prettyPrint (unsigned indent) const
{
  for (card_t i = size (); i--; ) {
    myExprs[i]->prettyPrint (indent);
    if (i)
      print (indent, ":\n");
  }
}

#endif // DEBUG
