// Expression list class -*- c++ -*-

#ifndef EXPRESSIONLIST_H_
# define EXPRESSIONLIST_H_

# include "typedefs.h"
# include <cassert>

/** List of expressions of same type */
class ExpressionList {
public:
  // Constructors and destructors

  /** Constructs a new ExpressionList */
  ExpressionList ();

private:
  /** Copy constructor */
  ExpressionList (const class ExpressionList& old);
  /** Assignment operator */
  class ExpressionList& operator= (const class ExpressionList& old);
protected:
  /** Destructs ExpressionList */
  ~ExpressionList ();
public:
  /** Get a shared copy of the expression list */
  class ExpressionList* copy () { myReferences++; return this; }
  /** Destructs ExpressionList, provided there aren't any references to it */
  void destroy () { if (this && !--myReferences) delete this; }
# ifndef NDEBUG
  /** Determine whether this is the last copy of an expression list */
  bool isLastCopy () const { return myReferences == 1; }
# endif // NDEBUG

  // Accessors

  /** Append an expression to the list */
  void append (class Expression* expression);
  /** Prepend an expression to the list */
  void prepend (class Expression* expression);

  /** Get the type associated with the expression list */
  const class Type* getType () const;

  /**
   * Determine whether this is a basic expression list containing
   * no temporal logic or set operations
   * @return	true if this is a basic expression
   */
  bool isBasic () const;
  /**
   * Determine whether this is a temporal logic expression
   * @return	true if this is a temporal logic expression
   */
  bool isTemporal () const;

  card_t size () const { return mySize; }
  bool empty () const { return !mySize; }

  /** Get an expression from the list
   * @param index	index to the expression
   * @return		the expression
   */
  const class Expression* operator[] (card_t index) const {
    assert (index < size ());
    return myExprs[index];
  }
  /** Get an expression from the list
   * @param index	index to the expression
   * @return		the expression
   */
  class Expression* operator[] (card_t index) {
    assert (index < size ());
    return myExprs[index];
  }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class ExpressionList& other) const;

  /** Ordering comparison operator */
  bool operator< (const class ExpressionList& other) const;

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expressions in the list,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if any expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable);

  /** Conditionally invalidate the cached value of the expressions in the list,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @param index	expression whose invalidation status we want
   * @return		true if the specified expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable, card_t index);

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Partially evaluate the expression list using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression list, or NULL in case of error
   */
  class ExpressionList* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression list with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class ExpressionList* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the expression list with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
    */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Determine whether a variable can ever be bound from the expression list
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** Number of references to the expression list */
  unsigned myReferences;
  /** Number of list members */
  card_t mySize;
  /** The list of expressions */
  class Expression** myExprs;
};

#endif // EXPRESSIONLIST_H_
