// Maria if-then-else expression class -*- c++ -*-

#include "IfThenElse.h"
#include "Type.h"
#include "Constant.h"

IfThenElse::IfThenElse (class Expression* condition,
			class ExpressionList* exprList) :
  Expression (),
  myCondition (condition), myExprList (exprList)
{
  assert (myCondition->getType ()->getNumValues () == myExprList->size ());
  setType (myExprList->getType ());
}

IfThenElse::~IfThenElse ()
{
  myCondition->destroy ();
  myExprList->destroy ();
}

bool
IfThenElse::isBasic () const
{
  return
    myCondition->isBasic () &&
    myExprList->isBasic ();
}

bool
IfThenElse::isTemporal () const
{
  return
    myCondition->isTemporal () ||
    myExprList->isTemporal ();
}

void
IfThenElse::setType (const class Type* type)
{
  Expression::setType (type);
  assert (myExprList->getType () == type);
}

#ifdef EXPR_CACHE
bool
IfThenElse::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = myCondition->invalidate (variable);

  if (myExprList->invalidate (variable))
    invalidated = true;

  if (invalidated)
    Expression::invalidate ();

  return invalidated;
}

void
IfThenElse::do_clearValuation (const class Valuation* valuation) const
{
  myCondition->clearValuation (valuation);
  myExprList->do_clearValuation (valuation);
}

class Value*
IfThenElse::do_eval (const class Valuation* valuation) const
{
  const class Value* cond = myCondition->const_eval (valuation);
  if (!cond)
    return NULL;
  return (*myExprList)[myCondition->getType ()->convert (cond)]
    ->eval (valuation);
}
#else // EXPR_CACHE
class Value*
IfThenElse::do_eval (const class Valuation* valuation) const
{
  if (class Value* cond = myCondition->eval (valuation)) {
    const card_t index = myCondition->getType ()->convert (cond);
    delete cond;
    return (*myExprList)[index]->eval (valuation);
  }
  else
    return NULL;
}
#endif // EXPR_CACHE

class Expression*
IfThenElse::ground (const class Valuation& valuation)
{
  class Expression* expr = NULL;
#ifdef EXPR_CACHE
  if (const class Value* cond = myCondition->const_eval (&valuation))
    if (!(expr = (*myExprList)[myCondition->getType ()->convert (cond)]
	  ->ground (valuation)))
      return NULL;
#else // EXPR_CACHE
  if (class Value* cond = myCondition->eval (&valuation)) {
    const card_t index = myCondition->getType ()->convert (cond);
    delete cond;
    if (!(expr = (*myExprList)[index]->ground (valuation)))
      return NULL;
  }
#endif // EXPR_CACHE
  else {
    class Expression* condition = myCondition->ground (valuation);
    class ExpressionList* exprList = myExprList->ground (valuation);

    if (!condition || !exprList) {
      condition->destroy ();
      exprList->destroy ();
      return NULL;
    }

    if (condition == myCondition && exprList == myExprList) {
      condition->destroy ();
      exprList->destroy ();

      return copy ();
    }
    else
      expr = new class IfThenElse (condition, exprList);
  }

  if (valuation.isOKExcept (Valuation::undefinedValue)) {
    if (class Value* v = expr->eval (&valuation)) {
      class Expression* c = new class Constant (v);
      expr->destroy ();
      return c;
    }
    else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
      expr->destroy ();
      return NULL;
    }
  }

  return expr;
}

class Expression*
IfThenElse::substitute (class Substitution& substitution)
{
  class Expression* condition = myCondition->substitute (substitution);
  class ExpressionList* exprList = myExprList->substitute (substitution);

  if (condition == myCondition && exprList == myExprList) {
    condition->destroy ();
    exprList->destroy ();

    return copy ();
  }
  else
    return new class IfThenElse (condition, exprList);
}

#ifdef EXPR_CACHE
void
IfThenElse::associateVariables (const class Expression& expr, bool assoc) const
{
  myCondition->associateVariables (expr, assoc);
  myExprList->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
IfThenElse::forVariables (bool (*operation)
			  (const class VariableDefinition&)) const
{
  return
    myCondition->forVariables (operation) &&
    myExprList->forVariables (operation);
}

#ifdef EXPR_CACHE
const class VariableDefinition*
IfThenElse::getLvalue (const class Value*& value,
		       class Valuation& valuation) const
{
  if (const class Value* cond = myCondition->const_eval (&valuation)) {
    const card_t index = myCondition->getType ()->convert (cond);
    return (*myExprList)[index]->getLvalue (value, valuation);
  }
  else
    return NULL;
}
#else // EXPR_CACHE
const class VariableDefinition*
IfThenElse::getLvalue (const class Value*& value,
		       class Valuation& valuation) const
{
  if (class Value* cond = myCondition->eval (&valuation)) {
    const card_t index = myCondition->getType ()->convert (cond);
    delete cond;
    return (*myExprList)[index]->getLvalue (value, valuation);
  }
  else
    return NULL;
}
#endif // EXPR_CACHE


#ifdef DEBUG
# include "util.h"

void
IfThenElse::prettyPrint (unsigned indent) const
{
  print (indent, "(\n");
  myCondition->prettyPrint (indent + 1);
  print (indent + 1, "?\n");
  myExprList->prettyPrint (indent + 1);
  print (indent, ")\n");
}

#endif // DEBUG
