// Marking -*- c++ -*-

#include "Marking.h"
#include "Token.h"
#include "Quantifier.h"
#include "IntValue.h"
#include "BoolValue.h"
#include "Valuation.h"
#include "Place.h"
#include "VariableDefinition.h"

/** Calculate the multiplicity of a marking
 * @param valuation	variable substitutions
 * @param expr		the multiplicity expression
 * @param count		multiplicity of the surrounding marking expression
 * @return		multiplicity of the marking expression (count * expr)
 */
inline static card_t
calcMultiplicity (const class Valuation& valuation,
		  const class Expression* expr, card_t count)
{
  assert (expr && count);
#ifdef EXPR_CACHE
  if (const class Value* v = expr->const_eval (&valuation)) {
    assert (v->getKind () == Value::vInt);
    int_t c = static_cast<const class IntValue*>(v)->getValue ();
    if (c < 0) {
      valuation.flagNegativeCardinality ();
      return 0;
    }

    return count * c;
  }
#else // EXPR_CACHE
  if (class Value* v = expr->eval (&valuation)) {
    assert (v->getKind () == Value::vInt);
    int_t c = static_cast<const class IntValue*>(v)->getValue ();
    delete v;
    if (c < 0) {
      valuation.flagNegativeCardinality ();
      return 0;
    }

    return count * c;
  }
#endif // EXPR_CACHE

  assert (!valuation.isOK ());
  return 0;
}

Marking::Marking (const class Place* place, class Marking* child) :
  myChild (child), myNext (0),
  myPlace (place), myMultiplicity (0), myToken (0),
  myProcessed (Marking::unprocessed), myQuantifiers (0),
  myIsQuantified (child && child->myIsQuantified)
{
  if (myPlace)
    Expression::setType (myPlace->getType ());
  else if (myChild)
    Expression::setType (myChild->getType ());
  assert (!myChild ||
	  (myChild->myPlace == myPlace && myChild->getType () == getType ()));
}

Marking::~Marking ()
{
  myChild->destroy ();
  myNext->destroy ();
  myMultiplicity->destroy ();
  myToken->destroy ();
  delete myQuantifiers;
}

void
Marking::setPlace (const class Place* place)
{
  assert (!myPlace);
  myPlace = place;

  if (myChild)
    myChild->setPlace (place);
  if (myNext)
    myNext->setPlace (place);
}

void
Marking::setToken (class Expression* token)
{
  assert (!myToken && !myChild); myToken = token;
  assert (myToken && myToken->getType ());
  assert (myToken->isBasic ());
  Expression::setType (myToken->getType ());
}

void
Marking::setType (const class Type* type)
{
  for (class Marking* m = this; m; m = m->myNext) {
    Expression::setType (type);
    if (myToken) {
      myToken->setType (type);
      assert (!myChild);
    }
    else {
      assert (myChild != NULL);
      myChild->setType (type);
    }
  }
}

#ifdef EXPR_CACHE
/** Conditionally invalidate the cached values of the marking expression,
 * excluding the quantification conditions,
 * provided that they depend on the value of a variable
 * @param variable	variable the value should depend on
 * @return		true if any expression was invalidated
 */
inline static bool
invalidate (const class Marking* m,
	    const class VariableDefinition* variable)
{
  bool invalidated = false;

  for (; m; m = m->getNext ()) {
    if (m->getToken ()) {
      assert (!m->getChild ());
      if (m->getToken ()->invalidate (variable))
	invalidated = true;
    }
    else {
      assert (m->getChild () != NULL);
      if (m->getChild ()->invalidate (variable))
	invalidated = true;
    }

    if (m->getMultiplicity () && m->getMultiplicity ()->invalidate (variable))
      invalidated = true;
  }

  return invalidated;
}
#endif // EXPR_CACHE

/** Evaluate the marking subexpression.
 * Quantification is handled outside this function.
 * @param m		the marking subexpression
 * @param p		the PlaceMarking object that is to receive the tokens
 * @param count		multiplicity of the surrounding marking expression
 * @param valuation	variable substitutions
 * @param rm		flag: remove the marking instead of adding?
 * @return		true on success, false on error (unbound variables)
 */
inline static bool
eval (const class Marking* m,
      class PlaceMarking& p,
      const card_t count,
      const class Valuation& valuation,
      bool rm)
{
  assert (m && count);
  assert (valuation.isOK ());
  goto ok;

  for (; m; m = m->getNext ()) {
    assert (valuation.isOK ());

    if (m->hasQuantifier ())
      return m->eval (p, count, valuation, rm);
  ok:
    card_t tokencount;

    if (m->getMultiplicity ()) {
      tokencount = calcMultiplicity (valuation, m->getMultiplicity (), count);
      if (!valuation.isOK ())
	return false;
    }
    else
      tokencount = count;

    if (m->getToken ()) {
      assert (!m->getChild ());
      if (tokencount) {
	if (rm) {
#ifdef EXPR_CACHE
	  if (const class Value* token =
	      m->getToken ()->const_eval (&valuation))
	    p.remove (token, tokencount);
#else // EXPR_CACHE
	  if (class Value* token = m->getToken ()->eval (&valuation)) {
	    p.remove (token, tokencount);
	    delete token;
	  }
#endif // EXPR_CACHE
	  else
	    return false;
	}
	else if (class Value* token = m->getToken ()->eval (&valuation)) {
	  p.add (token, tokencount);
	}
	else
	  return false;
      }
    }
    else {
      assert (m->getChild () != NULL);
      if (tokencount && !m->getChild ()->eval (p, tokencount, valuation, rm))
	return false;
    }
  }

  return true;
}

bool
Marking::operator== (const class Marking& other) const
{
#define NULLCMP(c) ((c && !other.c) || (!c && other.c))
  if (myPlace != other.myPlace ||
      NULLCMP (myChild) || NULLCMP (myToken) ||
      NULLCMP (myMultiplicity) ||
      NULLCMP (myQuantifiers) ||
      NULLCMP (myNext))
    return false;
#undef NULLCMP

  if (myChild) {
    assert (!myToken);
    if (!(*myChild == *other.myChild))
      return false;
  }
  else {
    assert (myToken != NULL);
    if (!(*myToken == *other.myToken))
      return false;
  }

  if (myMultiplicity && !(*myMultiplicity == *other.myMultiplicity))
    return false;

  if (myQuantifiers && !(*myQuantifiers == *other.myQuantifiers))
    return false;

  return !myNext || *myNext == *other.myNext;
}

bool
Marking::operator< (const class Marking& other) const
{
  if (myPlace < other.myPlace) return true;
  if (other.myPlace < myPlace) return false;
  if (!myChild && other.myChild) return true;
  if (!other.myChild && myChild) return false;

  if (myChild) {
    assert (!myToken);
    if (*myChild < *other.myChild) return true;
    if (*other.myChild < *myChild) return false;
  }

  if (!myToken && other.myToken) return true;
  if (!other.myToken && myToken) return false;

  if (myToken) {
    assert (!myChild);
    if (*myToken < *other.myToken) return true;
    if (*other.myToken < *myToken) return false;
  }
  else
    assert (myChild != NULL);

  if (!myMultiplicity && other.myMultiplicity) return true;
  if (!other.myMultiplicity && myMultiplicity) return false;

  if (myMultiplicity) {
    if (*myMultiplicity < *other.myMultiplicity) return true;
    if (*other.myMultiplicity < *myMultiplicity) return false;
  }

  if (!myQuantifiers && other.myQuantifiers) return true;
  if (!other.myQuantifiers && myQuantifiers) return false;
  if (myQuantifiers) {
    if (*myQuantifiers < *other.myQuantifiers) return true;
    if (*other.myQuantifiers < *myQuantifiers) return false;
  }

  if (!myNext && other.myNext) return true;
  if (!other.myNext && myNext) return false;

  return myNext && *myNext < *other.myNext;
}

#ifdef EXPR_CACHE
bool
Marking::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = myQuantifiers && myQuantifiers->invalidate (variable);

  if (::invalidate (this, variable) || invalidated) {
    if (!variable->isAggregate ())
      setProcessed (unprocessed);
    return true;
  }

  return false;
}

void
Marking::do_clearValuation (const class Valuation* valuation) const
{
  for (const class Marking* m = this; m; m = m->myNext) {
    if (m->myChild) {
      m->myChild->do_clearValuation (valuation);
      assert (!m->myToken);
    }
    else {
      assert (m->myToken != NULL);
      m->myToken->clearValuation (valuation);
    }
  }
}

void
Marking::associate (bool assoc) const
{
  for (const class Marking* m = this; m; m = m->myNext) {
    if (m->myMultiplicity)
      m->myMultiplicity->associateVariables (*m, assoc);
    if (m->myToken) {
      assert (!m->myChild);
      m->myToken->associateVariables (*m, assoc);
    }
    else if (!assoc)
      m->myChild->associate (assoc);
    if (assoc)
      break;
  }
}
#endif // EXPR_CACHE

bool
Marking::eval (class PlaceMarking& p,
	       const card_t count,
	       const class Valuation& valuation,
	       bool rm) const
{
  if (!hasQuantifier ())
    return ::eval (this, p, count, valuation, rm);

  class Valuation v (valuation);
  switch (myQuantifiers->augment (v)) {
  case Quantifier::pass:
    break;
  case Quantifier::fail:
    goto next;
  case Quantifier::undefined:
    valuation.copyErrors (v);
    return false;
  }

  do {
  loop:
    if (!::eval (this, p, count, v, rm)) {
      valuation.copyErrors (v);
      return false;
    }
  next:
    switch (myQuantifiers->advance (v)) {
    case Quantifier::pass:
      goto loop;
    case Quantifier::fail:
      return true;
    case Quantifier::undefined:
      valuation.copyErrors (v);
      return false;
    }
  } while (0);

  return true;
}

class PlaceMarking*
Marking::meval (const class Valuation& valuation) const
{
  class PlaceMarking* const p = new class PlaceMarking;
#ifndef NDEBUG
  p->setType (getType ());
#endif
  if (eval (*p, 1, valuation, false))
    return p;
  delete p;
  return NULL;
}

/** Determine whether the whole marking expression is defined
 * @param m		the marking expression
 * @param valuation	Valuation for evaluating the expression
 * @return		true if everything is defined
 */
inline static bool
isDefined (const class Marking* m, const class Valuation& valuation)
{
  goto ok;

  for (; m; m = m->getNext ()) {
    assert (valuation.isOK ());

    if (m->hasQuantifier ())
      return m->isDefined (valuation);
  ok:
    if (m->getMultiplicity ()) {
      card_t count = calcMultiplicity (valuation, m->getMultiplicity (), 1);
      if (!valuation.isOK ())
	return false;
      if (!count)
	continue;
    }

    if (m->getToken ()) {
      assert (!m->getChild ());
#ifdef EXPR_CACHE
      if (const class Value* v = m->getToken ()->const_eval (&valuation))
	assert (valuation.isOK ());
#else // EXPR_CACHE
      if (class Value* v = m->getToken ()->eval (&valuation)) {
	assert (valuation.isOK ());
	delete v;
      }
#endif // EXPR_CACHE
      else {
	assert ((valuation.isUndefinedValue () &&
		 valuation.isOKExcept (Valuation::undefinedValue)) ||
		(valuation.isUndefinedComponent () &&
		 valuation.isOKExcept (Valuation::undefinedComponent)));
	return false;
      }
    }
    if (m->getChild ()) {
      assert (!m->getToken ());
      return m->getChild ()->isDefined (valuation);
    }
  }

  return true;
}

bool
Marking::isDefined (const class Valuation& valuation) const
{
  if (!hasQuantifier ())
    return ::isDefined (this, valuation);

  class Valuation v (valuation);
  switch (myQuantifiers->augment (v)) {
  case Quantifier::pass:
    break;
  case Quantifier::fail:
    goto next;
  case Quantifier::undefined:
    valuation.copyErrors (v);
    return false;
  }

  do {
  loop:
    v.clearErrors ();
    if (bool defined = ::isDefined (this, v)) {
      assert (v.isOK ());
      assert (valuation.isOK ());
      return true;
    }
    else {
      assert (!v.isOK ());
      valuation.copyErrors (v);
      return false;
    }

  next:
    switch (myQuantifiers->advance (v)) {
    case Quantifier::pass:
      goto loop;
    case Quantifier::fail:
      return true;
    case Quantifier::undefined:
      valuation.copyErrors (v);
      return false;
    }
  } while (0);

  return false;
}

/** Determine whether the whole marking expression has been processed
 * @param m	the marking expression
 * @return	true if the whole expression has been processed
 */
inline static bool
isProcessed (const class Marking* m)
{
  for (; m; m = m->getNext ())
    if (!m->isProcessed ())
      return false;
  return true;
}

/** Get next batch of formal tokens that can be assigned to
 * @param m		the marking expression
 * @param tokens	container for the tokens to be retrieved
 * @param valuation	variable substitutions (including quantification)
 * @param place		PlaceMarking where the tokens will be reserved
 * @param count		token multiplier (for recursive calls)
 * @param orig		original variable substitutions (without quantifiers)
 * @return		an expression that cannot be evaluated,
 *			or NULL if all went as expected
 */
inline static const class Expression*
getNextTokens (const class Marking* m,
	       class TokenList& tokens,
	       class Valuation& valuation,
	       class PlaceMarking& place,
	       card_t count,
	       class Valuation& orig,
	       const class Marking*& proc)
{
  assert (m && proc);
  assert (count);
  assert (valuation.isOK ());
  assert (orig.isOK ());
  if (&valuation != &orig)
    goto ok;
  for (; m; m = m->getNext ()) {
    assert (valuation.isOK ());

    if (m->isProcessed ())
      continue;

    if (m->hasQuantifier ())
      return m->getNextTokens (tokens, valuation, place, count, orig, proc);

  ok:
    card_t tokencount;
    if (m->getMultiplicity ()) {
      if (!(tokencount = calcMultiplicity (valuation,
					   m->getMultiplicity (),
					   count))) {
	if (valuation.isOKExcept (Valuation::undefinedValue)) {
	  if (&valuation == &orig) {
	    m->setProcessed (valuation.isOK ()
			     ? Marking::processed
			     : Marking::skipped);
	    valuation.clearErrors ();
	    proc = NULL;
	    continue;
	  }
	  else if (tokens.empty ()) {
	    valuation.clearErrors ();
	    assert (m == proc);
	    return NULL;
	  }
	}
	return m->getMultiplicity ();
      }
    }
    else
      tokencount = count;

    assert (valuation.isOK ());

    if (!proc)
      proc = m;

    if (m->getToken ()) {
      assert (!m->getChild ());
      if (place.empty ())
	return m;

      if (class Expression* expr = &valuation == &orig
	  ? m->getToken ()->copy ()
	  : m->getToken ()->ground (valuation)) {
	tokens.insert (new class Token (tokencount, expr, orig, place, *m));
	assert (valuation.isOKExcept (Valuation::undefinedValue));
	valuation.clearErrors ();
      }
      else {
	assert (valuation.isConstraintViolation ());
	return m->getToken ();
      }

      if (&valuation == &orig)
	m->setProcessed (Marking::processed);
    }
    else {
      assert (m->getChild () != NULL);
      if (const class Expression* excuse =
	  m->getChild ()->getNextTokens (tokens, valuation, place,
					 tokencount, orig, proc)) {
	assert (!::isProcessed (m->getChild ()));
	return excuse;
      }
      else {
	if (&valuation == &orig)
	  m->setProcessed (::isProcessed (m->getChild ())
			   ? Marking::processed
			   : Marking::unprocessed);
	if (tokens.empty ())
	  continue;
      }
    }

    return NULL;
  }

  return NULL;
}

/** Determine the next marking expression to be processed
 * @param m	initial value for the iteration
 * @param t	if m==t, q cannot be set to m
 * @param q	quantifier marking expression:
 *		if there is an unprocessed marking expression m
 *		followed by the quantifier marking expression q,
 *		return q instead of m
 * @return	the marking expression to be processed, or NULL
 */
inline static const class Marking*
nextProcessed (const class Valuation& valuation,
	       const class Marking* m,
	       const class Marking* t,
	       const class Marking* q)
{
  for (; m; m = m->getNext ()) {
    if (m->getMultiplicity () &&
	!calcMultiplicity (valuation, m->getMultiplicity (), 1) &&
	!valuation.isOK ()) {
      if (m->hasQuantifier () &&
	  valuation.isOKExcept (Valuation::undefinedValue))
	valuation.clearErrors ();
      else {
	valuation.clearErrors ();
	continue;
      }
    }
    if (!m->isProcessed ())
      return q ? q : m;
    if (!q && m != t && m->hasQuantifier ())
      q = m;
    if (m->getChild ())
      if (const class Marking* mm = nextProcessed (valuation,
						   m->getChild (), t, q))
	return mm;
  }

  return NULL;
}

const class Expression*
Marking::getNextTokens (class TokenList& tokens,
			class Valuation& valuation,
			class PlaceMarking& place,
			card_t count,
			class Valuation& orig,
			const class Marking*& proc) const
{
  assert (count);
  assert (valuation.isOK ());

  assert (orig.isOK ());

  if (!hasQuantifier ()) {
    if (isProcessed ())
      if (const class Marking* m =
	  ::nextProcessed (valuation, this, this, NULL))
	return m->getNextTokens (tokens, valuation, place, count, orig, proc);
      else
	return NULL;

    const class Expression* excuse =
      ::getNextTokens (this, tokens, valuation, place, count, orig,
		       proc = this);
    if (&valuation != &orig) {
      assert (!proc || proc->getProcessed () != Marking::processed);
      assert (proc == this || getProcessed () != Marking::unprocessed);
      if (!excuse && proc == this)
	setProcessed (!getChild () || ::isProcessed (getChild ())
		      ? processed : unprocessed);
    }
    else
      proc = NULL;

    if (excuse || !tokens.empty ())
      return excuse;

    const class Marking* m = ::nextProcessed (valuation, this, this, NULL);
    return m && m != this
      ? m->getNextTokens (tokens, valuation, place, count, orig, proc)
      : NULL;
  }

  class Valuation v (valuation);
  switch (myQuantifiers->augment (v)) {
  case Quantifier::pass:
    break;
  case Quantifier::fail:
    switch (myQuantifiers->advance (v)) {
    case Quantifier::pass:
      break;
    case Quantifier::fail:
      return NULL;
    case Quantifier::undefined:
      valuation.copyErrors (v);
#ifdef EXPR_CACHE
      return myQuantifiers->getUndefined ();
#else // EXPR_CACHE
      return myQuantifiers->getUndefined (valuation);
#endif // EXPR_CACHE
    }
    break;
  case Quantifier::undefined:
    valuation.copyErrors (v);
#ifdef EXPR_CACHE
    return myQuantifiers->getUndefined ();
#else // EXPR_CACHE
    return myQuantifiers->getUndefined (valuation);
#endif // EXPR_CACHE
  }

  const class Marking* m = ::nextProcessed (v, this, this, NULL);
  if (!m)
    return NULL;

  do {
  loop:
    assert (valuation.isOK ());
    assert (v.isOK ());
    if (m != this) {
      if (const class Expression* excuse =
	  m->getNextTokens (tokens, v, place, count, orig, proc)) {
	valuation.copyErrors (v);
	assert (m->getProcessed () != Marking::processed);
	return excuse;
      }
    }
    else {
      if (const class Expression* excuse =
	  ::getNextTokens (this, tokens, v, place, count, orig, proc = this)) {
	valuation.copyErrors (v);
	assert (getProcessed () != Marking::processed);
	return excuse;
      }
    }

    assert (v.isOK ());

    switch (myQuantifiers->advance (v)) {
    case Quantifier::pass:
      if (proc)
	proc->setProcessed (unprocessed), proc = NULL;
      goto loop;
    case Quantifier::fail:
      (proc = m)->setProcessed (processed);
      setProcessed (processed);
      return NULL;
    case Quantifier::undefined:
      valuation.copyErrors (v);
#ifdef EXPR_CACHE
      return myQuantifiers->getUndefined ();
#else // EXPR_CACHE
      return myQuantifiers->getUndefined (valuation);
#endif // EXPR_CACHE
    }
  } while (0);

  setProcessed (processed);
  return NULL;
}

class Token*
Marking::getNextToken (class Valuation& valuation,
		       class PlaceMarking& place,
		       const class Expression*& excuse,
		       card_t count) const
{
  assert (count > 0);
  assert (valuation.isOK ());
  assert (!excuse);

  assert (!isQuantified () && !hasQuantifier ());

  for (const class Marking* m = this; m; m = m->getNext ()) {
    if (m->isQuantified ())
      return NULL;
    if (m->isProcessed () &&
	(!m->getChild () || ::isProcessed (m->getChild ())))
      continue;

    card_t tokencount;
    if (m->getMultiplicity ()) {
      if (!(tokencount = calcMultiplicity (valuation,
					   m->getMultiplicity (),
					   count))) {
	if (valuation.isOKExcept (Valuation::undefinedValue)) {
	  m->setProcessed (valuation.isOK ()
			? Marking::processed
			: Marking::skipped);
	  valuation.clearErrors ();
	  continue;
	}
	excuse = m->getMultiplicity ();
	return NULL;
      }
    }
    else
      tokencount = count;

    assert (valuation.isOK ());

    if (m->getToken ()) {
      assert (!m->getChild ());
      if (place.empty ()) {
	excuse = this;
	return NULL;
      }

      m->setProcessed (Marking::processed);
      return new class Token (tokencount, m->getToken ()->copy (),
			      valuation, place, *m);
    }
    else {
      assert (m->getChild () != NULL);
      if (class Token* token = m->getChild ()
	  ->getNextToken (valuation, place, excuse, tokencount)) {
	assert (!excuse);
	m->setProcessed (::isProcessed (m->getChild ())
			 ? Marking::processed
			 : Marking::unprocessed);
	return token;
      }
      else if (excuse) {
	assert (!::isProcessed (m->getChild ()));
	return NULL;
      }
      else {
	assert (valuation.isOK ());
	continue;
      }
    }
  }

  return NULL;
}

bool
Marking::isLvalue (const class VariableDefinition* variable) const
{
  for (const class Marking* m = this; m; m = m->getNext ())
    if (m->getToken ()) {
      assert (!m->getChild ());
      if (m->getToken ()->isLvalue (variable))
	return true;
    }
    else if (m->getChild ()->isLvalue (variable))
      return true;
  return false;
}

void
Marking::addQuantifier (class Quantifier* quantifier)
{
  if (!myQuantifiers)
    myQuantifiers = new class QuantifierList;
  myQuantifiers->prepend (quantifier);
  myIsQuantified = true;
}

#ifdef DEBUG
# include "util.h"

void
Marking::prettyPrint (unsigned indent) const
{
  /** Number of opened parenthesis */
  card_t parens = 0;

  for (const class Marking* m = this;; ) {
    // The variable scoping information is
    // somewhat lost in the process of parsing.
    // This method does not always produce the same result as the input file.

    if (m->myQuantifiers)
      m->myQuantifiers->prettyPrint (indent);

    if (m->hasQuantifier ())
      print (indent++, "(\n"), parens++;

    if (m->myMultiplicity) {
      m->myMultiplicity->prettyPrint (indent);
      print (indent, "#\n");
    }
    if (m->myToken) {
      m->myToken->prettyPrint (indent);
      assert (!m->myChild);
    }
    else {
      assert (m->myChild != NULL);
      print (indent, "(\n");
      m->myChild->prettyPrint (indent + 1);
      print (indent, ")\n");
    }

    if ((m = m->myNext))
      print (indent, ",\n");
    else {
      while (parens)
	print (--indent, ")\n"), parens--;
      break;
    }
  }
}

#endif // DEBUG
