// Marking -*- c++ -*-

#ifndef MARKING_H_
# define MARKING_H_

# include "Expression.h"
# include <cassert>

class Marking : public Expression {
public:
  /** Result of a check */
  enum Result { undefined, pass, fail };
  /** Processing status */
  enum Processed { unprocessed, processed, skipped, skippedPreviously };

  // Constructors and destructors

  /** Constructs a Marking
   * @param place	the place the marking is associated with
   * @param child	child marking
   */
  Marking (const class Place* place, class Marking* child = 0);
private:
  /** Copy constructor */
  Marking (const class Marking& old);
  /** Assignment operator */
  class Marking& operator= (const class Marking& old);
protected:
  /** Destructs Marking */
  ~Marking ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eMarking; }

  /** Get the child marking */
  const class Marking* getChild () const { return myChild; }

  /** Set the place associated with this marking (and its neighbours) */
  void setPlace (const class Place* place);
  /** Get the place associated with this marking */
  const class Place* getPlace () const { return myPlace; }

  /** Set the multiplicity associated with this marking */
  void setMultiplicity (class Expression* multiplicity) {
    assert (!multiplicity || multiplicity->isBasic ());
    myMultiplicity = multiplicity;
  }
  /** Get the multiplicity associated with this marking */
  const class Expression* getMultiplicity () const { return myMultiplicity; }
  /** Get the multiplicity expression associated with this marking */
  class Expression* getMultiplicity () { return myMultiplicity; }

  /** Set the token value expression associated with this marking */
  void setToken (class Expression* token);
  /** Get the token value expression associated with this marking */
  class Expression* getToken () const { return myToken; }

  /** Get the next marking in the list */
  const class Marking* getNext () const { return myNext; }

  /**
   * Determine whether this is a basic expression containing
   * no temporal logic or set operations
   * @return	true if this is a basic expression
   */
  bool isBasic () const { return true; }
  /**
   * Determine whether this is a temporal logic expression
   * @return	true if this is a temporal logic expression
   */
  bool isTemporal () const { return false; }

  /** Set the type of the expression and its subexpressions */
  void setType (const class Type* type);

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class Marking& other) const;

  /** Ordering comparison operator */
  bool operator< (const class Marking& other) const;

#ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the marking expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;

  /** Associate the marking expression with its variables
   * @param assoc	true=associate, false=disassociate
   */
  void associate (bool assoc = true) const;
#else // EXPR_CACHE
  void associate (bool assoc = true) const { assoc = assoc; }
#endif // EXPR_CACHE

  /** Evaluate the marking and add it to or remove it from a PlaceMarking
   * @param target	the PlaceMarking object that to receive the tokens
   * @param count	token multiplier
   * @param valuation	variable substitutions
   * @param rm		flag: remove the marking instead of adding?
   * @return		true if everything evaluated ok; false otherwise
   */
  bool eval (class PlaceMarking& target,
	     const card_t count,
	     const class Valuation& valuation,
	     bool rm) const;
  /** Evaluate the marking and add it to a PlaceMarking
   * @param target	the PlaceMarking object that to receive the tokens
   * @param count	token multiplier
   * @param valuation	variable substitutions
   * @return		true if everything evaluated ok; false otherwise
   */
  bool add (class PlaceMarking& target,
	    const card_t count,
	    const class Valuation& valuation) const {
    return eval (target, count, valuation, false);
  }
  /** Evaluate the marking and remove it from a PlaceMarking
   * @param target	the PlaceMarking object that to receive the tokens
   * @param count	token multiplier
   * @param valuation	variable substitutions
   * @return		true if everything evaluated ok; false otherwise
   */
  bool remove (class PlaceMarking& target,
	       const card_t count,
	       const class Valuation& valuation) const {
    return eval (target, count, valuation, true);
  }

  /** Evaluate the marking expression
   * @param valuation	variable substitutions
   * @return		a PlaceMarking object, or NULL
   */
  class PlaceMarking* meval (const class Valuation& valuation) const;

  /** Determine whether the whole marking expression is defined
   * @param valuation	Valuation for evaluating the expression
   * @return		true if everything is defined
   */
  bool isDefined (const class Valuation& valuation) const;

  /** Set all "skipped" flags to "skippedPreviously" */
  void clearSkipped () const {
    assert (isQuantified ());
    for (const class Marking* m = this; m; m = m->getNext ()) {
      if (m->isProcessed () == skipped)
	m->setProcessed (skippedPreviously);
      if (m->getChild ())
	m->getChild ()->clearSkipped ();
    }
  }

  /** Get next batch of formal tokens that can be assigned to
   * @param tokens	container for the tokens to be retrieved
   * @param valuation	variable substitutions
   * @param place	PlaceMarking where the tokens will be reserved
   * @param count	token multiplier (for recursive calls)
   * @return		an expression that cannot be evaluated,
   *			or NULL if everything went as expected
   * @return		true if the whole marking expression was processed
   */
  const class Expression* getNextTokens (class TokenList& tokens,
					 class Valuation& valuation,
					 class PlaceMarking& place,
					 card_t count = 1) const {
    const class Marking* proc = NULL;
    clearSkipped ();
    return getNextTokens (tokens, valuation, place, count, valuation, proc);
  }

  /** Get next batch of formal tokens that can be assigned to
   * @param tokens	container for the tokens to be retrieved
   * @param valuation	variable substitutions
   * @param place	PlaceMarking where the tokens will be reserved
   * @param count	token multiplier (for recursive calls)
   * @param orig	original variable substitutions (without quantifiers)
   * @param proc	placeholder for the last processed Marking
   * @return		an expression that cannot be evaluated,
   *			or NULL if all went as expected
   */
  const class Expression* getNextTokens (class TokenList& tokens,
					 class Valuation& valuation,
					 class PlaceMarking& place,
					 card_t count,
					 class Valuation& orig,
					 const class Marking*& proc) const;
  /** Get the next formal token that can be assigned to
   * @param valuation	variable substitutions
   * @param place	PlaceMarking where the tokens will be reserved
   * @param excuse	placeholder for returning an expression
   *			that cannot be evaluated (in case of error)
   * @param count	token multiplier (for recursive calls)
   * @return		the Token, or NULL in case of error
   */
  class Token* getNextToken (class Valuation& valuation,
			     class PlaceMarking& place,
			     const class Expression*& excuse,
			     card_t count = 1) const;

  /** Determine whether the expression has been processed */
  bool isProcessed () const {
    return myProcessed == processed || myProcessed == skipped;
  }
  /** Read the flag for getNextTokens () */
  enum Processed getProcessed () const { return myProcessed; }
  /** Set the flag for getNextTokens () */
  void setProcessed (enum Processed proc) const { myProcessed = proc; }

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

  /** Append a marking to the list of markings
   * @param next	Marking to be appended
   */
  void append (class Marking* next) {
    class Marking* m = this;
    while (m->myNext)
      m = m->myNext;
    m->myNext = next;
  }

  /** Prepend a quantifier to the marking
   * @param quantifier	Quantifier to be prepended
   */
  void addQuantifier (class Quantifier* quantifier);

  /** Determine whether the marking expression has a quantifier
   * @return	true if any quantifier has been set
   */
  bool hasQuantifier () const { return myQuantifiers != NULL; }
  /** Determine recursively whether the marking expression is quantified
   * @return	true if there are any quantifiers
   */
  bool isQuantified () const { return myIsQuantified; }

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The child marking */
  class Marking* myChild;
  /** The sibling marking */
  class Marking* myNext;
  /** The place associated with the marking */
  const class Place* myPlace;
  /** The multiplicity of the marking */
  class Expression* myMultiplicity;
  /** The token expression of the marking */
  class Expression* myToken;
  /** Flag: has the marking expression been processed by getNextTokens ()? */
  mutable enum Processed myProcessed;
  /** Quantifiers of the multiplicity expression */
  class QuantifierList* myQuantifiers;
  /** Flag: does the marking expression contain quantifiers? */
  bool myIsQuantified;
};

#endif // MARKING_H_
