// Maria boolean negation operator class -*- c++ -*-

#ifndef NOTEXPRESSION_H_
# define NOTEXPRESSION_H_

# include "Expression.h"

/** Boolean negation operator expression */
class NotExpression : public Expression {
public:
  // Constructors and destructors

  /** Constructs a new NotExpression
   * @param expr	Expression the negation is applied to
   */
  NotExpression (class Expression* expr);

private:
  /** Copy constructor */
  NotExpression (const class NotExpression& old);
  /** Assignment operator */
  class NotExpression& operator= (const class NotExpression& old);
protected:
  /** Destructs NotExpression */
  ~NotExpression ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eNot; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return myExpr->isBasic (); }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return myExpr->isTemporal (); }

  /** Determine the expression */
  const class Expression* getExpr () const { return myExpr; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class NotExpression& other) const {
    return *myExpr == *other.myExpr;
  }

  /** Ordering comparison operator */
  bool operator< (const class NotExpression& other) const {
    return *myExpr < *other.myExpr;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The subexpression */
  class Expression* myExpr;
};

#endif // NOTEXPRESSION_H_
