// Quantifier expression -*- c++ -*-

#include "Quantifier.h"
#include "Expression.h"
#include "VariableDefinition.h"
#include "BoolValue.h"

Quantifier::Quantifier (char* name, const class Type* type,
			class Expression* condition) :
  myVariable (new class VariableDefinition (name, type, true)),
  myCondition (condition)
{
  assert (!condition || condition->getType ()->getKind () == Type::tBool);
}

Quantifier::Quantifier (class VariableDefinition* variable,
			class Expression* condition) :
  myVariable (variable), myCondition (condition)
{
  assert (!condition || condition->getType ()->getKind () == Type::tBool);
}

Quantifier::~Quantifier ()
{
  myCondition->destroy ();
  delete myVariable;
}

#ifdef EXPR_CACHE
enum Quantifier::Result
Quantifier::isAcceptable (const class Valuation& valuation) const
{
  if (myCondition) {
    if (const class Value* v = myCondition->const_eval (&valuation)) {
      assert (v->getType ()->getKind () == Type::tBool);
      return static_cast<const class BoolValue*>(v)->getValue () ? pass : fail;
    }
    else
      return undefined;
  }

  return pass;
}
#else // EXPR_CACHE
enum Quantifier::Result
Quantifier::isAcceptable (const class Valuation& valuation) const
{
  if (myCondition) {
    if (class Value* v = myCondition->eval (&valuation)) {
      assert (v->getType ()->getKind () == Type::tBool);
      enum Result result =
	static_cast<const class BoolValue*>(v)->getValue () ? pass : fail;
      delete v;
      return result;
    }
    else
      return undefined;
  }

  return pass;
}
#endif // EXPR_CACHE

QuantifierList::QuantifierList () :
  myList ()
{
}

QuantifierList::~QuantifierList ()
{
  for (iterator i = begin (); i != end (); i++)
    delete *i;
}

enum Quantifier::Result
QuantifierList::augment (class Valuation& valuation) const
{
  bool skipFirst = false;

  for (const_iterator i = begin (); i != end (); i++) {
    valuation.setValue ((*i)->getVariable (),
			(*i)->getVariable ()->getType ()->getFirstValue ());
    switch ((*i)->isAcceptable (valuation)) {
    case Quantifier::pass:
      break;
    case Quantifier::fail:
      skipFirst = true;
      break;
    case Quantifier::undefined:
      return Quantifier::undefined;
    }
  }

  return skipFirst ? Quantifier::fail : Quantifier::pass;
}

enum Quantifier::Result
QuantifierList::advance (class Valuation& valuation) const
{
  next:
  for (const_iterator i = begin (); i != end (); i++) {
    bool noWrap = valuation.increment ((*i)->getVariable ());
    switch ((*i)->isAcceptable (valuation)) {
    case Quantifier::pass:
      if (!noWrap)
	continue;
      else {
	const_iterator j = i;
	while (++j != end ()) {
	  switch ((*j)->isAcceptable (valuation)) {
	  case Quantifier::fail:
	    goto next;
	  case Quantifier::pass:
	    continue;
	  case Quantifier::undefined:
	    return Quantifier::undefined;
	  }
	}

	return Quantifier::pass;
      }
      case Quantifier::fail:
	if (noWrap)
	  goto next;
	return Quantifier::fail;
      case Quantifier::undefined:
	return Quantifier::undefined;
    }
  }

  return Quantifier::fail;
}

#ifdef DEBUG
# include "util.h"

void
Quantifier::prettyPrint (unsigned indent) const
{
  print (indent, "%s %s\n",
	 myVariable->getType ()->getName ()
	 ? myVariable->getType ()->getName ()
	 : myVariable->getType ()->getSyntacticName (),
	 myVariable->getName ());
  if (myCondition) {
    print (indent, "(\n");
    myCondition->prettyPrint (indent + 1);
    print (indent, ")\n");
  }
  print (indent, ":\n");
}

void
QuantifierList::prettyPrint (unsigned indent) const
{
  for (const_iterator i = begin (); i != end (); i++)
    (*i)->prettyPrint (indent);
}

#endif // DEBUG
