// Quantifier expression -*- c++ -*-

#ifndef QUANTIFIER_H_
# define QUANTIFIER_H_

# include "Expression.h"
# include <cassert>
# include <slist>

/** Quantifier variable */
class Quantifier {
public:

  /** Result of a check */
  enum Result { undefined, pass, fail };

  // Constructors and destructors

  /** Constructs a Quantifier
   * @param name	variable name
   * @param type	variable type
   * @param condition	optional condition for the quantifier
   */
  Quantifier (char* name, const class Type* type, class Expression* condition);

  /** Constructs a Quantifier
   * @param variable	variable
   * @param condition	optional condition for the quantifier
   */
  Quantifier (class VariableDefinition* variable, class Expression* condition);

private:
  /** Copy constructor */
  Quantifier (const Quantifier& old);
  /** Assignment operator */
  class Quantifier& operator= (const class Quantifier& old);
public:
  /** Destructs Quantifier */
  ~Quantifier ();

  // Accessors

  /** Determine the variable */
  const class VariableDefinition* getVariable () const { return myVariable; }
  /** Determine the condition expression */
  const class Expression* getCondition () const { return myCondition; }
  /** Set the condition expression */
  void setCondition (class Expression* condition) {
    assert (!myCondition); myCondition = condition;
  }

  /** Determine whether the given valuation is accepted by submarking criteria
   * @param valuation	Valuation to be checked
   * @return		(@see Result)
   */
  enum Result isAcceptable (const class Valuation& valuation) const;

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class Quantifier& other) const {
    return myVariable == other.myVariable &&
      *myCondition == *other.myCondition;
  }

  /** Ordering comparison operator */
  bool operator< (const class Quantifier& other) const {
    if (myVariable < other.myVariable) return true;
    if (other.myVariable < myVariable) return false;
    return *myCondition < *other.myCondition;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the condition expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const {
    return myCondition && myCondition->invalidate (variable);
  }
# endif // EXPR_CACHE

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** Variable */
  class VariableDefinition* myVariable;
  /** Quantifier condition */
  class Expression* myCondition;
};

/** Quantifier variable list */
class QuantifierList {
public:
  // Typedefs
  typedef std::slist<class Quantifier*> List;
  typedef List::iterator iterator;
  typedef List::const_iterator const_iterator;

  // Constructors and destructors

  /** Constructor */
  QuantifierList ();
private:
  /** Copy constructor */
  QuantifierList (const QuantifierList& old);
  /** Assignment operator */
  class QuantifierList& operator= (const class QuantifierList& old);
public:
  /** Destructor */
  ~QuantifierList ();

  // Accessors

  /** Prepend a quantifier to the list
   * @param quantifier	quantifier to be prepended
   */
  void prepend (class Quantifier* quantifier) {
    myList.push_front (quantifier);
  }
  /** Append a quantifier to the list
   * @param quantifier	quantifier to be appended
   */
  void append (class Quantifier* quantifier) {
    myList.insert (end (), quantifier);
  }

  bool empty () const { return myList.empty (); }
  size_t size () const { return myList.size (); }
  const_iterator begin () const { return myList.begin (); }
  const_iterator end () const { return myList.end (); }
  iterator begin () { return myList.begin (); }
  iterator end () { return myList.end (); }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class QuantifierList& other) const {
    if (size () != other.size ())
      return false;
    for (const_iterator i = begin (), j = other.begin ();
	 i != end (); i++, j++)
      if (!(**i == **j))
	return false;
    return true;
  }

  /** Ordering comparison operator */
  bool operator< (const class QuantifierList& other) const {
    if (size () < other.size ())
      return true;
    if (other.size () < size ())
      return false;
    for (const_iterator i = begin (), j = other.begin ();
	 i != end (); i++, j++)
      if (**i < **j)
	return true;
      else if (**j < **i)
	return false;
    return false;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of each condition expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const {
    bool invalidated = false;
    for (const_iterator i = begin (); i != end (); i++)
      if ((*i)->invalidate (variable))
	invalidated = true;
    return invalidated;
  }
# endif // EXPR_CACHE

  /** Augment a valuation with the first values of the quantifiers
   * @param valuation	valuation to be augmented
   * @return		pass, fail (invalid valuation) or undefined (error)
   */
  enum Quantifier::Result augment (class Valuation& valuation) const;

  /** Step to the next values of the quantifiers
   * @return		pass, fail (wrapped) or undefined (error)
   */
  enum Quantifier::Result advance (class Valuation& valuation) const;

# ifdef EXPR_CACHE
  /** Get an undefined condition expression
   * (diagnostics for the augment () and advance () methods)
   * @return		an undefined condition expression
   */
  const class Expression* getUndefined () const {
    for (const_iterator i = begin (); i != end (); i++)
      if (const class Expression* e = (*i)->getCondition ())
	if (!e->isCached ())
	  return e;
    assert (false);
    return NULL;
  }
# else // EXPR_CACHE
  /** Get an undefined condition expression
   * (diagnostics for the augment () and advance () methods)
   * @param v		variable substitutions
   * @return		an undefined condition expression
   */
  const class Expression* getUndefined (const class Valuation& v) const {
    for (const_iterator i = begin (); i != end (); i++)
      if (const class Expression* e = (*i)->getCondition ())
	if (class Value* value = e->eval (&v))
	  delete value;
	else
	  return e;
    assert (false);
    return NULL;
  }
# endif // EXPR_CACHE

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The list of quantifiers */
  List myList;
};

#endif // QUANTIFIER_H_
