// relation operator class -*- c++ -*-

#include "RelopExpression.h"
#include "Net.h"
#include "BoolType.h"
#include "BoolValue.h"
#include "Constant.h"

RelopExpression::RelopExpression (enum Op op,
				  class Expression* left,
				  class Expression* right) :
  Expression (),
  myOp (op), myLeft (left), myRight (right)
{
  setType (Net::getBoolType ());
  assert (myLeft->getType ()->isAssignable (myRight->getType ()) ||
	  myRight->getType ()->isAssignable (myLeft->getType ()));
  assert (myOp == rEqual || myOp == rInequal ||
	  (myLeft->getType ()->isOrdered () &&
	   myRight->getType ()->isOrdered ()));
  assert (myLeft->isBasic () && myRight->isBasic ());

  switch (myLeft->getType ()->getKind ()) {
  case Type::tInt:
  case Type::tBool:
  case Type::tChar:
    break;
  case Type::tEnum:
  case Type::tId:
    assert (myLeft->getType () == myRight->getType ());
    break;
  case Type::tStruct:
  case Type::tVector:
  case Type::tUnion:
  case Type::tBuffer:
    break;
  }
}

RelopExpression::~RelopExpression ()
{
  myLeft->destroy ();
  myRight->destroy ();
}

const char*
RelopExpression::getOpString () const
{
  switch (getOp ()) {
  case rLess:
    return "<";
  case rEqual:
    return "==";
  case rGreater:
    return ">";
  case rGreaterEqual:
    return ">=";
  case rInequal:
    return "!=";
  case rLessEqual:
    return "<=";
  }

  return "???";
}

#ifdef EXPR_CACHE
bool
RelopExpression::invalidate (const class VariableDefinition* variable) const
{
  bool invalidatedLeft = myLeft->invalidate (variable);

  if (myRight->invalidate (variable) || invalidatedLeft) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
RelopExpression::do_clearValuation (const class Valuation* valuation) const
{
  myLeft->clearValuation (valuation);
  myRight->clearValuation (valuation);
}

class Value*
RelopExpression::do_eval (const class Valuation* valuation) const
{
  const class Value* left = myLeft->const_eval (valuation);
  if (!left)
    return NULL;
  const class Value* right = myRight->const_eval (valuation);
  if (!right)
    return NULL;

  bool result;
  switch (getOp ()) {
  case rLess:
    result = *left < *right;
    break;
  case rEqual:
    result = *left == *right;
    break;
  case rGreater:
    result = *right < *left;
    break;
  case rGreaterEqual:
    result = *right <= *left;
    break;
  case rInequal:
    result = !(*left == *right);
    break;
  case rLessEqual:
    result = *left <= *right;
    break;
  default:
    assert (false);
    return NULL;
  }

  return constrain (valuation, new class BoolValue (getType (), result));
}
#else // EXPR_CACHE
class Value*
RelopExpression::do_eval (const class Valuation* valuation) const
{
  class Value* left = myLeft->eval (valuation);
  if (!left)
    return NULL;
  class Value* right = myRight->eval (valuation);
  if (!right) {
    delete left;
    return NULL;
  }

  bool result;
  switch (getOp ()) {
  case rLess:
    result = *left < *right;
    break;
  case rEqual:
    result = *left == *right;
    break;
  case rGreater:
    result = *right < *left;
    break;
  case rGreaterEqual:
    result = *right <= *left;
    break;
  case rInequal:
    result = !(*left == *right);
    break;
  case rLessEqual:
    result = *left <= *right;
    break;
  default:
    assert (false);
    return NULL;
  }

  delete left, delete right;
  return constrain (valuation, new class BoolValue (getType (), result));
}
#endif // EXPR_CACHE

class Expression*
RelopExpression::ground (const class Valuation& valuation)
{
  class Expression* left = myLeft->ground (valuation);
  class Expression* right = myRight->ground (valuation);

  if (!left || !right) {
    left->destroy ();
    right->destroy ();
    return NULL;
  }

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class RelopExpression (myOp, left, right);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
RelopExpression::substitute (class Substitution& substitution)
{
  class Expression* left = myLeft->substitute (substitution);
  class Expression* right = myRight->substitute (substitution);

  if (left == myLeft && right == myRight) {
    left->destroy ();
    right->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class RelopExpression (myOp, left, right);
    expr->setType (getType ());
    return expr;
  }
}

#ifdef EXPR_CACHE
void
RelopExpression::associateVariables (const class Expression& expr,
				     bool assoc) const
{
  myLeft->associateVariables (expr, assoc);
  myRight->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
RelopExpression::forVariables (bool (*operation)
			       (const class VariableDefinition&)) const
{
  return myLeft->forVariables (operation) && myRight->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

void
RelopExpression::prettyPrint (unsigned indent) const
{
  print (indent, "(\n");
  myLeft->prettyPrint (indent + 1);
  print (indent, ")%s(\n", getOpString ());
  myRight->prettyPrint (indent + 1);
  print (indent, ")\n");
}

#endif // DEBUG
