// set expression class -*- c++ -*-

#include "SetExpression.h"
#include "Net.h"
#include "BoolType.h"
#include "BoolValue.h"
#include "PlaceMarking.h"

SetExpression::SetExpression (enum Op op,
			      class Expression* left,
			      class Expression* right) :
  Expression (),
  myOp (op), myLeft (left), myRight (right)
{
  assert (myLeft->getType () == myRight->getType () ||
	  myLeft->getKind () == Expression::eEmptySet ||
	  myRight->getKind () == Expression::eEmptySet);
  assert (!myLeft->isTemporal () && !myRight->isTemporal ());
  switch (myOp) {
  case sSubset:
  case sEquals:
    setType (Net::getBoolType ());
    break;
  case sIntersection:
  case sMinus:
  case sUnion:
    setType (myLeft->getType ());
    break;
  }
}

SetExpression::~SetExpression ()
{
  myLeft->destroy ();
  myRight->destroy ();
}

class Value*
SetExpression::do_eval (const class Valuation* valuation) const
{
  class PlaceMarking* l = myLeft->meval (*valuation);
  if (!l)
    return NULL;
  class PlaceMarking* r = myRight->meval (*valuation);
  if (!r) {
    delete l;
    return NULL;
  }

  bool result;

  switch (myOp) {
  case sSubset:
    result = *l <= *r;
    break;
  case sEquals:
    result = *l == *r;
    break;
  default:
    assert (false);
    return NULL;
  }

  return constrain (valuation, new class BoolValue (getType (), result));
}

class PlaceMarking*
SetExpression::meval (const class Valuation& valuation) const
{
  class PlaceMarking* l = myLeft->meval (valuation);
  if (!l)
    return NULL;
  class PlaceMarking* r = myRight->meval (valuation);
  if (!r) {
    delete l;
    return NULL;
  }

  switch (myOp) {
  case sIntersection:
    if (l->size () < r->size ()) {
      class PlaceMarking* p = r; r = l; l = p;
    }
    *l &= *r;
    break;
  case sMinus:
    *l -= *r;
    break;
  case sUnion:
    if (l->size () < r->size ()) {
      class PlaceMarking* p = r; r = l; l = p;
    }
    *l += *r;
    break;
  default:
    assert (false);
    return NULL;
  }

  delete r;
  return l;
}

inline static const char*
getOpString (enum SetExpression::Op op)
{
  switch (op) {
  case SetExpression::sSubset:
    return "subset";
  case SetExpression::sEquals:
    return "equals";
  case SetExpression::sIntersection:
    return "intersect";
  case SetExpression::sMinus:
    return "minus";
  case SetExpression::sUnion:
    return "union";
  }

  return "???";
}

#ifdef DEBUG
# include "util.h"

void
SetExpression::prettyPrint (unsigned indent) const
{
  print (indent, "(\n");
  myLeft->prettyPrint (indent + 1);
  print (indent, ")%s(\n", getOpString (myOp));
  myRight->prettyPrint (indent + 1);
  print (indent, ")\n");
}

#endif // DEBUG
