// set expression class -*- c++ -*-

#ifndef SETEXPRESSION_H_
# define SETEXPRESSION_H_

# include "Expression.h"

/** Set expression */
class SetExpression : public Expression {
public:
  /** Set operator */
  enum Op { sSubset, sEquals, sIntersection, sMinus, sUnion };

  // Constructors and destructors

  /** Constructs a new SetExpression
   * @param op		Operator
   * @param left	Left-hand expression of the operator
   * @param right	Right-hand expression of the operator
   */
  SetExpression (enum Op op,
		 class Expression* left, class Expression* right);

private:
  /** Copy constructor */
  SetExpression (const class SetExpression& old);
  /** Assignment operator */
  class SetExpression& operator= (const class SetExpression& old);
protected:
  /** Destructs SetExpression */
  ~SetExpression ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eSet; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return false; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Determine the operator */
  enum Op getOp () const { return myOp; }
  /** Determine the left-hand expression */
  const class Expression* getLeft () const { return myLeft; }
  /** Determine the right-hand expression */
  const class Expression* getRight () const { return myRight; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class SetExpression& other) const {
    return myOp == other.myOp &&
      *myLeft == *other.myLeft &&
      *myRight == *other.myRight;
  }

  /** Ordering comparison operator */
  bool operator< (const class SetExpression& other) const {
    if (myOp < other.myOp) return true;
    if (other.myOp < myOp) return false;
    if (*myLeft < *other.myLeft) return true;
    if (*other.myLeft < *myLeft) return false;
    return *myRight < *other.myRight;
  }

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Evaluate the multiset expression
   * @param valuation	variable substitutions and the global marking
   * @return		the filtered marking
   */
  class PlaceMarking* meval (const class Valuation& valuation) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The operator */
  enum Op myOp;
  /** Left-hand expression */
  class Expression* myLeft;
  /** Right-hand expression */
  class Expression* myRight;
};

#endif // SETEXPRESSION_H_
