// Maria structure component reference expression class -*- c++ -*-

#include "StructComponent.h"
#include "StructType.h"
#include "StructValue.h"
#include "Constant.h"
#include "Variable.h"
#include "VariableDefinition.h"

StructComponent::StructComponent (class Expression* structure,
				  card_t index) :
  myStructure (structure), myIndex (index)
{
  assert (myStructure && myStructure->getType ()->getKind () == Type::tStruct);
  const class StructType* type =
    static_cast<const class StructType*>(myStructure->getType ());
  assert (myStructure->isBasic ());
  setType ((*type)[myIndex]);
}

StructComponent::~StructComponent ()
{
  myStructure->destroy ();
}

#ifdef EXPR_CACHE
bool
StructComponent::invalidate (const class VariableDefinition* variable) const
{
  if (myStructure->invalidate (variable)) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
StructComponent::do_clearValuation (const class Valuation* valuation) const
{
  myStructure->clearValuation (valuation);
}

class Value*
StructComponent::do_eval (const class Valuation* valuation) const
{
  const class Value* v = myStructure->const_eval (valuation);
  if (!v)
    return NULL;

  assert (v->getKind () == Value::vStruct);
  v = (*static_cast<const class StructValue*>(v))[myIndex];
  return v ? constrain (valuation, v->copy ()) : NULL;
}
#else // EXPR_CACHE
class Value*
StructComponent::do_eval (const class Valuation* valuation) const
{
  class Value* v = myStructure->eval (valuation);
  if (!v)
    return NULL;

  assert (v->getKind () == Value::vStruct);
  class Value* u = NULL;
  if (const class Value* w = (*static_cast<class StructValue*>(v))[myIndex])
    u = constrain (valuation, w->copy ());
  delete v;
  return u;
}
#endif // EXPR_CACHE

class Expression*
StructComponent::ground (const class Valuation& valuation)
{
  class Expression* structure = myStructure->ground (valuation);
  if (!structure)
    return NULL;
  if (structure == myStructure) {
    structure->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class StructComponent (structure, myIndex);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
StructComponent::substitute (class Substitution& substitution)
{
  class Expression* structure = myStructure->substitute (substitution);
  if (structure == myStructure) {
    structure->destroy ();
    return copy ();
  }
  else
    return new class StructComponent (structure, myIndex);
}

#ifdef EXPR_CACHE
void
StructComponent::associateVariables (const class Expression& expr,
				     bool assoc) const
{
  myStructure->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
StructComponent::forVariables (bool (*operation)
			       (const class VariableDefinition&)) const
{
  return myStructure->forVariables (operation);
}

const class VariableDefinition*
StructComponent::getLvalue (const class Value*& value,
			    class Valuation& valuation) const
{
  if (myStructure->getKind () == Expression::eVariable &&
#ifdef EXPR_CACHE
      !myStructure->isCached () &&
#endif // EXPR_CACHE
      !valuation.getValue (static_cast<const class Variable*>
			   (myStructure)->getVariable ())) {
    const class VariableDefinition* var =
      static_cast<const class Variable*>(myStructure)
      ->getVariable ()->getChild (myIndex);
    if (!valuation.getValue (var))
      return var;
  }

  return NULL;
}

bool
StructComponent::isLvalue (const class VariableDefinition* variable) const
{
  return myStructure->isLvalue (variable);
}

#ifdef DEBUG
# include "util.h"

void
StructComponent::prettyPrint (unsigned indent) const
{
  myStructure->prettyPrint (indent);
  print (indent, ".%s\n",
	 static_cast<const class StructType*>(myStructure->getType ())
	 ->getComponentName (myIndex));
}

#endif // DEBUG
