// Maria structure component reference expression class -*- c++ -*-

#ifndef STRUCTCOMPONENT_H_
# define STRUCTCOMPONENT_H_

# include "Expression.h"
# include "typedefs.h"

/** Structure component reference expression */
class StructComponent : public Expression {
public:
  // Constructors and destructors

  /** Constructs StructComponent
   * @param structure	Expression evaluating to the structure
   * @param index	Index of the referenced structure component
   */
  StructComponent (class Expression* structure, card_t index);
private:
  /** Copy constructor */
  StructComponent (const class StructComponent& old);
  /** Assignment operator */
  class StructComponent& operator= (const class StructComponent& old);
protected:
  /** Destructs StructComponent */
  ~StructComponent ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eStructComponent; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class StructComponent& other) const {
    return myIndex == other.myIndex && *myStructure == *other.myStructure;
  }

  /** Ordering comparison operator */
  bool operator< (const class StructComponent& other) const {
    if (myIndex < other.myIndex) return true;
    if (other.myIndex < myIndex) return false;
    return *myStructure < *other.myStructure;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Get an unbound lvalue and a corresponding value
   * @param value	value the expression should evaluate to
   * @param valuation	variable substitutions
   * @return		the lvalue; set value to the suggested value
   */
  const class VariableDefinition* getLvalue
  (const class Value*& value, class Valuation& valuation) const;

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The structure being referenced */
  class Expression* myStructure;
  /** Index of the referenced structure component */
  card_t myIndex;
};

#endif // STRUCTCOMPONENT_H_
