// Maria structure (or array) expression class -*- c++ -*-

#include "StructExpression.h"
#include "StructType.h"
#include "StructValue.h"
#include "Constant.h"

StructExpression::StructExpression (const class Type* type) :
  Expression (),
  myComponents ()
{
  assert (type && type->getKind () == Type::tStruct);
  Expression::setType (type);
}

StructExpression::~StructExpression ()
{
  for (iterator i = begin (); i != end (); i++)
    (*i)->destroy ();
}

void
StructExpression::setType (const class Type* type)
{
  Expression::setType (type);
  assert (type->getKind () == Type::tStruct);
  const class StructType* t = static_cast<const class StructType*>(type);

  assert (t->getSize () == size ());
  card_t i;
  const_iterator j;

  for (i = 0, j = begin (); i < t->getSize (); i++, j++)
    (*j)->setType ((*t)[i]);
}

void
StructExpression::append (class Expression* expr)
{
  assert (expr && expr->isBasic ());
  expr->setType (getNextType ());
  myComponents.insert (end (), expr);
}

const class Type*
StructExpression::getNextType () const
{
  assert (getType () && getType ()->getKind () == Type::tStruct);
  const class StructType* type =
    static_cast<const class StructType*>(getType ());
  assert (size () <= type->getSize ());
  if (size () == type->getSize ())
    return NULL;
  else
    return (*type)[size ()];
}

#ifdef EXPR_CACHE
bool
StructExpression::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = false;

  for (const_iterator i = begin (); i != end (); i++)
    if ((*i)->invalidate (variable))
      invalidated = true;

  if (invalidated)
    Expression::invalidate ();

  return invalidated;
}

void
StructExpression::do_clearValuation (const class Valuation* valuation) const
{
  for (const_iterator i = begin (); i != end (); i++)
    (*i)->clearValuation (valuation);
}
#endif // EXPR_CACHE

class Value*
StructExpression::do_eval (const class Valuation* valuation) const
{
  class StructValue* value = new class StructValue (getType ());
  const_iterator i;
  card_t j;
  for (i = begin (), j = 0; i != end (); i++, j++) {
    if (class Value* v = (*i)->eval (valuation))
      (*value)[j] = v;
    else {
      delete value;
      return NULL;
    }
  }

  return constrain (valuation, value);
}

class Expression*
StructExpression::ground (const class Valuation& valuation)
{
  bool same = true;
  class StructExpression* expr = new class StructExpression (getType ());

  for (iterator i = begin (); i != end (); i++) {
    class Expression* e = (*i)->ground (valuation);
    if (!e) {
      expr->destroy ();
      return NULL;
    }

    if (e != *i)
      same = false;
    expr->append (e);
  }

  if (same) {
    expr->destroy ();
    return copy ();
  }
  else if (valuation.isOKExcept (Valuation::undefinedValue)) {
    if (class Value* v = expr->eval (&valuation)) {
      class Expression* c = new class Constant (v);
      expr->destroy ();
      return c;
    }
    else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
      expr->destroy ();
      return NULL;
    }
  }

  return expr;
}

class Expression*
StructExpression::substitute (class Substitution& substitution)
{
  bool same = true;
  class StructExpression* expr = new class StructExpression (getType ());

  for (iterator i = begin (); i != end (); i++) {
    class Expression* e = (*i)->substitute (substitution);
    if (e != *i)
      same = false;
    expr->append (e);
  }

  if (same) {
    expr->destroy ();
    return copy ();
  }
  else
    return expr;
}

#ifdef EXPR_CACHE
void
StructExpression::associateVariables (const class Expression& expr,
				      bool assoc) const
{
  for (const_iterator i = begin (); i != end (); i++)
    (*i)->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
StructExpression::forVariables (bool (*operation)
				(const class VariableDefinition&)) const
{
  for (const_iterator i = begin (); i != end (); i++)
    if (!(*i)->forVariables (operation))
      return false;
  return true;
}

bool
StructExpression::isCompatible (const class Value& value,
				const class Valuation& valuation) const
{
  assert (getType () && value.getType ());

  if (value.getKind () != Value::vStruct)
    return false;

  assert (value.getType ()->getKind () == Type::tStruct);

  const class StructValue& sv = static_cast<const class StructValue&>(value);

  if (!sv.getType ()->isAssignable (getType ()))
    return false;

  const_iterator i;
  card_t j;

  for (i = begin (), j = 0; i != end (); i++, j++) {
#ifdef EXPR_CACHE
    const class Value* comp = (*i)->const_eval (&valuation);
    if (comp && !(*comp == *sv[j]))
      return false;
#else // EXPR_CACHE
    class Value* comp = (*i)->eval (&valuation);
    if (comp && !(*comp == *sv[j])) {
      delete comp;
      return false;
    }
    delete comp;
#endif // EXPR_CACHE
  }

  return true;
}

const class VariableDefinition*
StructExpression::getLvalue (const class Value*& value,
			     class Valuation& valuation) const
{
  assert (value->getKind () == Value::vStruct);
  const class StructValue* sv = static_cast<const class StructValue*>(value);

  const_iterator i;
  card_t j;

  for (i = begin (), j = 0; i != end (); i++, j++) {
    const class Value* val = (*sv)[j];
    const class VariableDefinition* v =
      (*i)->getLvalue (val, valuation);
    if (v) {
      value = val;
      return v;
    }
  }

  return NULL;
}

bool
StructExpression::isLvalue (const class VariableDefinition* variable) const
{
  for (const_iterator i = begin (); i != end (); i++)
    if ((*i)->isLvalue (variable))
      return true;
  return false;
}

#ifdef DEBUG
# include "util.h"

void
StructExpression::prettyPrint (unsigned indent) const
{
  print (indent, "{\n");
  for (const_iterator i = begin (); i != end (); ) {
    (*i)->prettyPrint (indent + 1);
    if (++i == end ())
      break;
    print (indent + 1, ",\n");
  }
  print (indent, "}\n");
}

#endif // DEBUG
