// Maria structure expression class -*- c++ -*-

#ifndef STRUCTEXPRESSION_H_
# define STRUCTEXPRESSION_H_

# include "Expression.h"
# include <slist>

/** Structure expression */
class StructExpression : public Expression {
public:
  // Typedefs
  typedef std::slist<class Expression*> List;
  typedef List::iterator iterator;
  typedef List::const_iterator const_iterator;

  // Constructors and destructors

  /** Constructs StructExpression
   * @param type	type of the struct
   */
  StructExpression (const class Type* type);
private:
  /** Copy constructor */
  StructExpression (const class StructExpression& old);
  /** Assignment operator */
  class StructExpression& operator= (const class StructExpression& old);
protected:
  /** Destructs StructExpression */
  ~StructExpression ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eStruct; }

  /** Set the type of the expression
   * @param type	Type of the expression
   */
  void setType (const class Type* type);

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Determine the type of the next component affected by append ()
   * @return		type of the next component;
   *			NULL if the expression is fully initialized
   */
  const class Type* getNextType () const;

  bool empty () const { return myComponents.empty (); }
  size_t size () const { return myComponents.size (); }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class StructExpression& other) const {
    if (size () != other.size ()) return false;
    for (const_iterator i = begin (), j = other.begin ();
	 i != end (); i++, j++) {
      if (!(**i == **j))
	return false;
    }
    return true;
  }

  /** Ordering comparison operator */
  bool operator< (const class StructExpression& other) const {
    if (size () < other.size ()) return true;
    if (other.size () < size ()) return false;
    for (const_iterator i = begin (), j = other.begin ();
	 i != end (); i++, j++) {
      if (**i < **j)
	return true;
      if (**j < **i)
	return false;
    }
    return false;
  }

  /** Append an expression to the component list
   * @param expr	expression to be appended
   */
  void append (class Expression* expr);

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Determine whether the expression is compatible with the specified value,
   * neglecting subexpressions that cannot be evaluated
   * @param value	value the expression will be compared to
   * @param valuation	variable substitutions
   * @return		true if the expression is compatible with the value
   */
  bool isCompatible (const class Value& value,
		     const class Valuation& valuation) const;

  /** Get an unbound lvalue and a corresponding value
   * @param value	value the expression should evaluate to
   * @param valuation	variable substitutions
   * @return		the lvalue; set value to the suggested value
   */
  const class VariableDefinition* getLvalue
  (const class Value*& value, class Valuation& valuation) const;

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  // STL accessors
  const_iterator begin () const { return myComponents.begin (); }
  const_iterator end () const { return myComponents.end (); }
  iterator begin () { return myComponents.begin (); }
  iterator end () { return myComponents.end (); }

  /** The struct components */
  List myComponents;
};

#endif // STRUCTEXPRESSION_H_
