// Maria temporal binary operator class -*- c++ -*-

#ifndef TEMPORALBINOP_H_
# define TEMPORALBINOP_H_

# include "Expression.h"

class TemporalBinop : public Expression {
public:
  /** Temporal binary operators */
  enum Op { Until, Release };

  // Constructors and destructors

  /** Constructs a new TemporalBinop
   * @param op		Operator
   * @param qual	Transition instance qualifier expression
   * @param left	Left-hand expression of the operator
   * @param right	Right-hand expression of the operator
   */
  TemporalBinop (enum Op op,
		 class Expression* qual,
		 class Expression* left, class Expression* right);

private:
  /** Copy constructor */
  TemporalBinop (const class TemporalBinop& old);
  /** Assignment operator */
  class TemporalBinop& operator= (const class TemporalBinop& old);
protected:
  /** Destructs TemporalBinop */
  ~TemporalBinop ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eTemporalBinop; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return false; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return true; }

  /** Determine the operator */
  enum Op getOp () const { return myOp; }
  /** Determine the string representation of the operator */
  const char* getOpString () const;
  /** Determine the transition instance qualifier expression */
  const class Expression* getQual () const { return myQual; }
  /** Determine the left-hand expression */
  const class Expression* getLeft () const { return myLeft; }
  /** Determine the right-hand expression */
  const class Expression* getRight () const { return myRight; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class TemporalBinop& other) const {
    return myOp == other.myOp && *myQual == *other.myQual &&
      *myLeft == *other.myLeft && *myRight == *other.myRight;
  }

  /** Ordering comparison operator */
  bool operator< (const class TemporalBinop& other) const {
    if (myOp < other.myOp) return true;
    if (other.myOp < myOp) return false;
    if (*myQual < *other.myQual) return true;
    if (*other.myQual < *myQual) return false;
    if (*myLeft < *other.myLeft) return true;
    if (*other.myLeft < *myLeft) return false;
    return *myRight < *other.myRight;
  }

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The operator */
  enum Op myOp;
  /** Transition instance qualifier */
  class Expression* myQual;
  /** Left-hand expression */
  class Expression* myLeft;
  /** Right-hand expression */
  class Expression* myRight;
};

#endif // TEMPORALBINOP_H_
