// Unified (expanded) token -*- c++ -*-

#include "Token.h"
#include "Expression.h"
#include "VariableDefinition.h"
#include "Marking.h"
#include "Transition.h"

Token::Token (card_t cardinality, class Expression* expr,
	      class Valuation& valuation,
	      class PlaceMarking& placeMarking,
	      const class Marking& marking) :
  myStatus (New),
  myExpr (expr), myCardinality (cardinality), myValuation (valuation),
  myPlaceMarking (placeMarking), myMarking (marking),
  myQuantificationId (false),
  myIterator (placeMarking.begin ()),
  myVariables ()
{
  assert (myExpr && myCardinality);
  assert (!myPlaceMarking.empty ());
  if (myExpr->isLastCopy ())
    myExpr->associateVariables (true);
}

Token::~Token ()
{
  undefineVariables ();
  if (myExpr->isLastCopy ())
    myExpr->associateVariables (false);
  myExpr->destroy ();
  myMarking.setProcessed (Marking::unprocessed);
}

bool
Token::getConcrete ()
{
  assert (!isReserved ());
  assert (!isConcrete ());
  assert (undefined ());
  assert (!empty ());
  assert (myValuation.isOK ());

  for (myStatus = New; myIterator != end (); myIterator++)
    if (myCardinality <= myPlaceMarking.getFreeCount (myIterator))
      return true;

  return false;
}

enum Token::LvalueStatus
Token::getLvalue (const class Transition& transition,
		  const class TokenList& tokenlist)
{
  assert (myValuation.isOK ());
  assert (myIterator != end ());
  const class Value* v = &PlaceMarking::getValue (myIterator);
  if (const class VariableDefinition* var =
      myExpr->getLvalue (v, myValuation)) {
    if (!myValuation.isOKExcept (Valuation::undefinedValue)) {
      if (var->getFather ())
	delete var;
      return lvalueInconsistent;
    }
    myValuation.clearErrors ();
    assert (!myValuation.getValue (var));
    assert (v && v->getType ()->isAssignable (var->getType ()));
    if (var->getType ()->isConstrained (v)) {
      if (isNew ()) {
	myValuation.setValue (var, v->copy ());
	class Value* fv = NULL;
	if (var->getFather ())
	  if ((fv = var->getFather ()->getValue (myValuation)))
	    myValuation.setValue (var->getFather (), fv);

	const class Expression* excuse = NULL;
	if (transition.checkGates (myValuation, excuse) &&
	    isCompatible (
#ifdef EXPR_CACHE
			  var->getFather ()
			  ? *var->getFather ()
			  : *var
#endif // EXPR_CACHE
			  ) &&
	    tokenlist.isCompatible (
#ifdef EXPR_CACHE
				    var->getFather ()
				    ? *var->getFather ()
				    : *var
#endif // EXPR_CACHE
				    )) {
	  myVariables.insert (var);
	  if (fv) {
	    assert (var->getFather () != NULL);
	    myVariables.insert (var->getFather ());
	  }
	}
	else {
	  if (var->getFather ()) {
	    myValuation.undefine (var->getFather ());
	    myValuation.undefine (var);
	    delete var;
	  }
	  else
	    myValuation.undefine (var);

	  return lvalueInconsistent;
	}
      }
      else
	assert (*myValuation.getValue (var) == *v);

      if (myValuation.isUndefinedComponent () &&
	  myValuation.isOKExcept (Valuation::undefinedComponent))
	myValuation.clearErrors ();
    }
    else {
      myValuation.flagLvalueViolation ();
      if (var->getFather ())
	delete var;
    }
    return lvalueFound;
  }
  else {
    if (!myValuation.isOKExcept (Valuation::undefinedValue))
      return lvalueInconsistent;
    myValuation.clearErrors ();
  }

  assert (myValuation.isOK ());
  return lvalueNotFound;
}

#ifdef EXPR_CACHE
bool
Token::isCompatible (const class VariableDefinition& variable) const
{
  return
    !variable.isAssociated (myMarking) ||
    myExpr->isCompatible (getValue (), myValuation);
}

bool
Token::isBindable ()
{
  assert (myValuation.isOK ());
  if (const class Value* value = myExpr->const_eval (&myValuation)) {
    assert (myValuation.isOK ());
    if ((myIterator = myPlaceMarking.find (value)) != end ()) {
      if (myCardinality <= PlaceMarking::getFreeCount (myIterator))
	return true;
      myIterator = begin ();
    }
  }

  return false;
}
#else // EXPR_CACHE
bool
Token::isBindable ()
{
  assert (myValuation.isOK ());
  if (class Value* value = myExpr->eval (&myValuation)) {
    assert (myValuation.isOK ());
    myIterator = myPlaceMarking.find (value);
    delete value;
    if (myIterator != end ()) {
      if (myCardinality <= PlaceMarking::getFreeCount (myIterator))
	return true;
      myIterator = begin ();
    }
  }

  return false;
}
#endif // EXPR_CACHE

void
Token::validate ()
{
  assert (defined ());
  myMarking.setProcessed (Marking::processed);
}

#ifndef NDEBUG
bool
Token::defined () const
{
  for (Set::const_iterator i = myVariables.begin ();
       i != myVariables.end (); i++)
    if ((*i)->hasChildren () && !myValuation.getValue (*i)) {
      class Value* v = (*i)->getValue (myValuation);
      assert (v != NULL);
      delete v;
    }
    else
      assert (myValuation.getValue (*i) != NULL);
  return true;
}

bool
Token::undefined () const
{
  for (Set::const_iterator i = myVariables.begin ();
       i != myVariables.end (); i++)
    assert (!myValuation.getValue (*i));
  return true;
}
#endif // NDEBUG

void
Token::undefineVariables ()
{
  for (Set::const_iterator i = myVariables.begin ();
       i != myVariables.end (); i++) {
    const class VariableDefinition* var = *i;
    myValuation.undefine (var);
    if (var->getFather ())
      delete var;
  }
  myVariables.clear ();
}

#ifdef DEBUG
# include "util.h"
# ifndef NDEBUG
#  include "Place.h"
# endif // NDEBUG

void
Token::prettyPrint (unsigned indent) const
{
# ifndef NDEBUG
  print (indent, "%s:%u#\n", myPlaceMarking.getPlace ()->getName (),
	 myCardinality);
# else
  print (indent, "%u#\n", myCardinality);
# endif // NDEBUG
  myExpr->prettyPrint (indent + 1);
}

void
TokenList::prettyPrint (unsigned indent) const
{
  for (const_iterator i = begin (); i != end (); ) {
    (*i)->prettyPrint (indent);
    if (++i != end ())
      print (indent, ",\n");
  }
}

#endif // DEBUG
