// Unified (expanded) token -*- c++ -*-

#ifndef TOKEN_H_
# define TOKEN_H_

# include "PlaceMarking.h"
# include "Marking.h"
# include <set>
# include <slist>

/** Unified (expanded) token */

class Token {
  // Typedefs
  typedef std::set<const class VariableDefinition*> Set;

  typedef PlaceMarking::iterator iterator;
  typedef PlaceMarking::const_iterator const_iterator;

  /** Unification status of the token */
  enum Status { New, Old, Concrete, ReservedOld, ReservedConcrete };

public:
  // Constructors and destructors

  /** Constructs a Token
   * @param cardinality		cardinality of the token
   * @param expr		the token expression
   * @param valuation		variable substitutions
   * @param placeMarking	the PlaceMarking containing the actual token(s)
   * @param marking		the marking expression expanding to this token
   */
  Token (card_t cardinality, class Expression* expr,
	 class Valuation& valuation, class PlaceMarking& placeMarking,
	 const class Marking& marking);
private:
  /** Copy constructor */
  Token (const class Token& old);
  /** Assignment operator */
  class Token& operator= (const class Token& old);
public:
  /** Destructs Token */
  ~Token ();

  // Accessors

  /** Determine whether the token is new */
  bool isNew () const { return myStatus == New; }
  /** Determine whether the token is old */
  bool isOld () const {
    return myStatus == Old || myStatus == ReservedOld;
  }
  /** Determine whether the token is concrete */
  bool isConcrete () const {
    return myStatus == Concrete || myStatus == ReservedConcrete;
  }
  /** Flag the token new */
  void flagNew () {
    assert (!isReserved ());
    myStatus = New; myIterator = begin ();
  }
  /** Flag the token old */
  void flagOld () {
    assert (!isReserved ()); assert (myIterator != myPlaceMarking.end ());
    myStatus = Old;
  }
  /** Flag the token concrete */
  void flagConcrete () {
    assert (!isReserved ()); assert (myIterator != myPlaceMarking.end ());
    myStatus = Concrete;
  }

  /** Determine the cardinality of the actual token */
  card_t getCardinality () const { return myCardinality; }

  /** Determine whether the token is quantified */
  bool isQuantified () const { return myMarking.getToken () != myExpr; }

  /** Determine the quantification identifier bit */
  bool getQuantificationId () const { return myQuantificationId; }
  /** Set the quantification identifier bit
   * @param id		new value of the bit
   */
  void setQuantificationId (bool id) { myQuantificationId = id; }

  /** Find the next candidate for the concrete token
   * @return	true if a candidate was found
   */
  bool getConcrete ();

  /** Status of getLvalue () call */
  enum LvalueStatus { lvalueNotFound, lvalueFound, lvalueInconsistent };

  /**
   * Bind an unbound lvalue so that the token will match
   * the current concrete token
   * @param transition	the transition (for checking gate expressions)
   * @param tokenlist	the list of tokens this token belongs to
   * @return		(@see LvalueStatus)
   */
  enum LvalueStatus getLvalue (const class Transition& transition,
			       const class TokenList& tokenlist);

  /** Get the value of the concrete token candidate */
  const class Value& getValue () const {
    assert (myIterator != myPlaceMarking.end ());
    return PlaceMarking::getValue (myIterator);
  }

  /** Determine whether the valuation is compatible with the concrete token
   * @return	true if the token matches the value of the concrete token
   */
  bool isCompatible () const {
    return myExpr->isCompatible (getValue (), myValuation);
  }

# ifdef EXPR_CACHE
  /** Determine whether the valuation is compatible with the concrete token
   * @param variable	check the expression only if it depends on the variable
   * @return		true if the formal and the concrete token match
   */
  bool isCompatible (const class VariableDefinition& variable) const;
# endif // EXPR_CACHE

  /** Try to bind the token to a concrete token (and set myIterator to it)
   * @return	true if the token could be bound
   */
  bool isBindable ();

  // Accessors to the PlaceMarking

  bool empty () const { return myPlaceMarking.empty (); }
  size_t size () const { return myPlaceMarking.size (); }
private:
  const_iterator begin () const { return myPlaceMarking.begin (); }
  const_iterator end () const { return myPlaceMarking.end (); }
  iterator begin () { return myPlaceMarking.begin (); }
  iterator end () { return myPlaceMarking.end (); }
public:

  /** Advance the iterator to the concrete tokens
   * @return	false if the iterator reached the end
   */
  bool next () {
    assert (!isReserved ()); assert (!isConcrete ());
    assert (myIterator != end ());
    return ++myIterator != end ();
  }

  /** Determine whether the iterator is at end */
  bool atEnd () const { return myIterator == myPlaceMarking.end (); }

  /** Determine whether the token has been bound to a concrete token */
  bool isReserved () const {
    return myStatus == ReservedOld || myStatus == ReservedConcrete;
  }
  /** Bind the token to a concrete token */
  void reserve () {
    assert (!isReserved () && !isNew ());
    myPlaceMarking.reserve (myIterator, myCardinality);
    myStatus = myStatus == Old ? ReservedOld : ReservedConcrete;
  }
  /** Remove the reservation (binding) of the concrete token */
  void free () {
    assert (isReserved ());
    myPlaceMarking.free (myIterator, myCardinality);
    myStatus = myStatus == ReservedOld ? Old : Concrete;
  }

  // Other methods

  /** Re-validate the marking expression that was expanded to this token */
  void validate ();

# ifndef NDEBUG
  /** Assert that all VariableDefinitions in myVariables are defined
   * @return	true (or terminate in an assertion failure)
   */
  bool defined () const;

  /** Assert that all VariableDefinitions in myVariables are undefined
   * @return	true (or terminate in an assertion failure)
   */
  bool undefined () const;
# endif // NDEBUG
  /** Undefine all variables in myVariables */
  void undefineVariables ();

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members
private:
  /** Unification status of the token */
  enum Status myStatus;
  /** The formal token expression */
  class Expression* myExpr;
  /** Cardinality of the actual token */
  card_t myCardinality;
  /** Variable substitutions */
  class Valuation& myValuation;
  /** PlaceMarking to which the Token belongs */
  class PlaceMarking& myPlaceMarking;
  /** the Marking that expanded to the Token */
  const class Marking& myMarking;
  /** alternating bit for separating different lots of quantified tokens */
  bool myQuantificationId;
  /** Iterator to the PlaceMarking */
  iterator myIterator;
  /** Variables defined by the unification of this token */
  Set myVariables;
};

class TokenList {
public:
  // Typedefs
  typedef std::slist<class Token*> List;
  typedef List::iterator iterator;
  typedef List::const_iterator const_iterator;
  
  // Constructors and destructors

  /** Constructs a TokenList */
  TokenList () : myList (), myQuantificationId (false) {}
private:
  /** Copy constructor */
  TokenList (const class TokenList& old);
  /** Assignment operator */
  class TokenList& operator= (const class TokenList& old);
public:
  /** Destructs a TokenList */
  ~TokenList () {
    for (iterator i = begin (); i != end (); i++)
      delete *i;
  }

  // Accessors
  bool empty () const { return myList.empty (); }
  size_t size () const { return myList.size (); }
  iterator begin () { return myList.begin (); }
  iterator end () { return myList.end (); }
  const_iterator begin () const { return myList.begin (); }
  const_iterator end () const { return myList.end (); }

  // Other methods
  void insert (class Token* token) { myList.push_front (token); }
  void clear () {
    for (iterator i = begin (); i != end (); i++)
      delete *i;
    myList.clear ();
  }
  
  /** Peek at the topmost token on the stack */
  const class Token* top () const {
    assert (!empty ());
    return *begin ();
  }
  /** Pop a token from the stack */
  class Token* pop () {
    assert (!empty ());
    iterator i = begin (); class Token* token = *i; myList.erase (i);
    assert (token && token->getCardinality ());
    assert (!token->atEnd ());
    return token;
  }
  /** Push a token to the stack */
  void push (class Token* token) {
    assert (token != NULL);
    myList.push_front (token);
  }

  /** Toggle the quantification identifier for all tokens on the stack */
  void toggleQuantificationId () {
    for (const_iterator i = begin (); i != end (); i++)
      (*i)->setQuantificationId (myQuantificationId);
    myQuantificationId = !myQuantificationId;
  }
  /** Determine whether the valuations are compatible with the concrete tokens
   * @return	true if the tokens match the values of the concrete tokens
   */
  bool isCompatible () const {
    for (const_iterator i = begin (); i != end (); i++)
      if (!(*i)->isCompatible ())
	return false;
    return true;
  }

# ifdef EXPR_CACHE
  /** Determine whether the valuation is compatible with the concrete token
   * @param variable	check the expression only if it depends on the variable
   * @return		true if the formal and the concrete token match
   */
  bool isCompatible (const class VariableDefinition& variable) const {
    for (const_iterator i = begin (); i != end (); i++)
      if (!(*i)->isCompatible (variable))
	return false;
    return true;
  }
# endif // EXPR_CACHE

  /** Re-validate the marking expressions that were expanded to these tokens */
  void validate () {
    for (const_iterator i = begin (); i != end (); i++)
      (*i)->validate ();
  }

# ifndef NDEBUG
  /** Assert that all tokens are defined */
  bool defined () const {
    for (const_iterator i = begin (); i != end (); i++)
      assert ((*i)->defined ());
    return true;
  }

  /** Assert that all tokens are undefined */
  bool undefined () const {
    for (const_iterator i = begin (); i != end (); i++)
      assert ((*i)->defined ());
    return true;
  }

  /** Assert that all tokens have values */
  bool values () const {
    for (const_iterator i = begin (); i != end (); i++)
      (*i)->getValue ();
    return true;
  }
# endif // NDEBUG

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members
private:
  /** The list of tokens */
  List myList;
  /** alternating bit for separating different lots of quantified tokens */
  bool myQuantificationId;
};

#endif // TOKEN_H_
