// Maria typecast expression -*- c++ -*-

#include "Typecast.h"
#include "Type.h"
#include "allValues.h"
#include "Constant.h"

bool
Typecast::isCastable (const class Type& type)
{
  switch (type.getKind ()) {
  case Type::tInt:
  case Type::tBool:
  case Type::tChar:
  case Type::tEnum:
    return true;
  default:
    return false;
  }
}

Typecast::Typecast (const class Type* type, class Expression* expr) :
  Expression (),
  myExpr (expr)
{
  setType (type);
  assert (getType () && myExpr && myExpr->getType ());
  assert ((isCastable (*getType ()) && isCastable (*myExpr->getType ())) ||
	  Type::isCompatible (myExpr->getType (), getType ()));
}

Typecast::~Typecast ()
{
  myExpr->destroy ();
}

#ifdef EXPR_CACHE
bool
Typecast::invalidate (const class VariableDefinition* variable) const
{
  if (myExpr->invalidate (variable)) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
Typecast::do_clearValuation (const class Valuation* valuation) const
{
  myExpr->clearValuation (valuation);
}
#endif // EXPR_CACHE

class Value*
Typecast::do_eval (const class Valuation* valuation) const
{
  class Value* v = myExpr->eval (valuation);
  if (!v)
    return NULL;

  switch (getType ()->getKind ()) {
    class Value* w;

  case Type::tInt:
    switch (v->getKind ()) {
    case Value::vInt:
      v->setType (getType ());
      return constrain (valuation, v);
    case Value::vBool:
      w = new class IntValue (getType (),
			      static_cast<class BoolValue*>(v)->getValue ());
      delete v;
      return constrain (valuation, w);
    case Value::vChar:
      w = new class IntValue (getType (),
			      static_cast<class CharValue*>(v)->getValue ());
      delete v;
      return constrain (valuation, w);
    case Value::vEnum:
      w = new class IntValue (getType (),
			      static_cast<class EnumValue*>(v)->getValue ());
      delete v;
      return constrain (valuation, w);
    default:
      assert (false);
      return NULL;
    }

  case Type::tBool:
    switch (v->getKind ()) {
    case Value::vInt:
      w = new class BoolValue (getType (), static_cast<class IntValue*>(v)
			       ->getValue () != 0);
      delete v;
      return constrain (valuation, w);
    case Value::vBool:
      v->setType (getType ());
      return v;
    case Value::vChar:
      w = new class BoolValue (getType (), static_cast<class CharValue*>(v)
			       ->getValue () != 0);
      delete v;
      return constrain (valuation, w);
    case Value::vEnum:
      w = new class BoolValue (getType (), static_cast<class CharValue*>(v)
			       ->getValue () != 0);
      delete v;
      return constrain (valuation, w);
    default:
      assert (false);
      return NULL;
    }

  case Type::tChar:
    switch (v->getKind ()) {
    case Value::vInt:
      w = new class CharValue (getType (),
			       static_cast<class IntValue*>(v)->getValue ());
      delete v;
      return constrain (valuation, w);
    case Value::vBool:
      w = new class CharValue (getType (),
			       static_cast<class BoolValue*>(v)->getValue ());
      delete v;
      return constrain (valuation, w);
    case Value::vChar:
      v->setType (getType ());
      return constrain (valuation, v);
    case Value::vEnum:
      w = new class CharValue (getType (),
			       static_cast<class EnumValue*>(v)->getValue ());
      delete v;
      return constrain (valuation, w);
    default:
      assert (false);
      return NULL;
    }

  case Type::tEnum:
    switch (v->getKind ()) {
    case Value::vInt:
      w = new class EnumValue (getType (),
			       static_cast<class IntValue*>(v)->getValue ());
    checkEnum:
      delete v;
      if (!getType ()->isConstrained (v)) {
	if (valuation)
	  valuation->flagConstraintViolation ();
	delete w;
	return NULL;
      }
      return w;
    case Value::vBool:
      w = new class EnumValue (getType (),
			       static_cast<class BoolValue*>(v)->getValue ());
      goto checkEnum;
    case Value::vChar:
      w = new class EnumValue (getType (),
			       static_cast<class CharValue*>(v)->getValue ());
      goto checkEnum;
    case Value::vEnum:
      (w = v)->setType (getType ());
      v = NULL;
      goto checkEnum;
    default:
      assert (false);
      return NULL;
    }

  case Type::tId:
  case Type::tStruct:
  case Type::tVector:
  case Type::tUnion:
  case Type::tBuffer:
    if (!(v = v->cast (*getType ())) && valuation)
      valuation->flagConstraintViolation ();
    return v;
  }

  assert (false);
  return NULL;
}

class Expression*
Typecast::ground (const class Valuation& valuation)
{
  class Expression* e = myExpr->ground (valuation);
  if (!e)
    return NULL;
  if (e == myExpr) {
    e->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class Typecast (getType (), e);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
Typecast::substitute (class Substitution& substitution)
{
  class Expression* e = myExpr->substitute (substitution);
  if (e == myExpr) {
    e->destroy ();
    return copy ();
  }
  else
    return new class Typecast (getType (), e);
}

#ifdef EXPR_CACHE
void
Typecast::associateVariables (const class Expression& expr, bool assoc) const
{
  myExpr->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
Typecast::forVariables (bool (*operation)
			(const class VariableDefinition&)) const
{
  return myExpr->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

void
Typecast::prettyPrint (unsigned indent) const
{
  if (!getType () || !getType ()->getName ())
    print (indent, "(null)(\n");
  else
    print (indent, "(%s)(\n", getType ()->getName ());
  if (myExpr)
    myExpr->prettyPrint (indent + 1);
  print (indent + 1, ")\n");
}
#endif // DEBUG
