// Maria typecast expression -*- c++ -*-

#ifndef TYPECAST_H_
# define TYPECAST_H_

# include "Expression.h"

class Typecast : public Expression {
public:
  // Constructors and destructors

  /** Constructs a new Typecast
   * @param type	type reference
   * @param expr	expression to be casted
   */
  Typecast (const class Type* type, class Expression* expr);

private:
  /** Copy constructor */
  Typecast (const class Typecast& old);
  /** Assignment operator */
  class Typecast& operator= (const class Typecast& old);
protected:
  /** Destructs Typecast */
  ~Typecast ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eTypecast; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return myExpr->isBasic (); }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return myExpr->isTemporal (); }

  /** Determine the subexpression */
  const class Expression* getExpr () const { return myExpr; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class Typecast& other) const {
    return *myExpr == *other.myExpr;
  }

  /** Ordering comparison operator */
  bool operator< (const class Typecast& other) const {
    return *myExpr < *other.myExpr;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Determine whether a type is castable
   * @param type	type to be checked
   * @return		true if the type is castable
   */
  static bool isCastable (const class Type& type);

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The subexpression */
  class Expression* myExpr;
};

#endif // TYPECAST_H_
