// Maria union component reference expression class -*- c++ -*-

#include "UnionComponent.h"
#include "UnionType.h"
#include "UnionValue.h"
#include "Valuation.h"
#include "Constant.h"
#include "Variable.h"
#include "VariableDefinition.h"

UnionComponent::UnionComponent (class Expression* expr,
				card_t index) :
  myExpr (expr), myIndex (index)
{
  assert (myExpr && myExpr->getType ());
  assert (myExpr->getType ()->getKind () == Type::tUnion);
  const class UnionType* type =
    static_cast<const class UnionType*>(myExpr->getType ());
  assert (myExpr->isBasic ());
  setType ((*type)[myIndex]);
}

UnionComponent::~UnionComponent ()
{
  myExpr->destroy ();
}

#ifdef EXPR_CACHE
bool
UnionComponent::invalidate (const class VariableDefinition* variable) const
{
  if (myExpr->invalidate (variable)) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
UnionComponent::do_clearValuation (const class Valuation* valuation) const
{
  myExpr->clearValuation (valuation);
}

class Value*
UnionComponent::do_eval (const class Valuation* valuation) const
{
  const class Value* v = myExpr->const_eval (valuation);
  if (!v)
    return NULL;

  assert (v->getKind () == Value::vUnion);
  const class UnionValue* u = static_cast<const class UnionValue*>(v);
  if (u->getIndex () != myIndex) {
    if (valuation)
      valuation->flagUnionViolation ();
    return NULL;
  }
  v = u->getValue ();
  return v ? constrain (valuation, v->copy ()) : NULL;
}
#else // EXPR_CACHE
class Value*
UnionComponent::do_eval (const class Valuation* valuation) const
{
  class Value* v = myExpr->eval (valuation);
  if (!v)
    return NULL;

  assert (v->getKind () == Value::vUnion);
  class UnionValue* u = static_cast<class UnionValue*>(v);
  if (u->getIndex () != myIndex) {
    if (valuation)
      valuation->flagUnionViolation ();
    delete u;
    return NULL;
  }
  if (const class Value* w = u->getValue ())
    v = constrain (valuation, w->copy ());
  else
    v = NULL;
  delete u;
  return v;
}
#endif // EXPR_CACHE

class Expression*
UnionComponent::ground (const class Valuation& valuation)
{
  class Expression* e = myExpr->ground (valuation);
  if (!e)
    return NULL;
  if (e == myExpr) {
    e->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class UnionComponent (e, myIndex);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
UnionComponent::substitute (class Substitution& substitution)
{
  class Expression* e = myExpr->substitute (substitution);
  if (e == myExpr) {
    e->destroy ();
    return copy ();
  }
  else
    return new class UnionComponent (e, myIndex);
}

#ifdef EXPR_CACHE
void
UnionComponent::associateVariables (const class Expression& expr,
				    bool assoc) const
{
  myExpr->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
UnionComponent::forVariables (bool (*operation)
			      (const class VariableDefinition&)) const
{
  return myExpr->forVariables (operation);
}

const class VariableDefinition*
UnionComponent::getLvalue (const class Value*& value,
			   class Valuation& valuation) const
{
  if (myExpr->getKind () == Expression::eVariable &&
#ifdef EXPR_CACHE
      !myExpr->isCached () &&
#endif // EXPR_CACHE
      !valuation.getValue (static_cast<const class Variable*>
			   (myExpr)->getVariable ())) {
    const class VariableDefinition* var =
      static_cast<const class Variable*>(myExpr)
      ->getVariable ()->getChild (myIndex);
    if (!valuation.getValue (var))
      return var;
  }

  return NULL;
}

bool
UnionComponent::isLvalue (const class VariableDefinition* variable) const
{
  return myExpr->isLvalue (variable);
}

#ifdef DEBUG
# include "util.h"

void
UnionComponent::prettyPrint (unsigned indent) const
{
  myExpr->prettyPrint (indent);
  print (indent, ".%s\n",
	 static_cast<const class UnionType*>(myExpr->getType ())
	 ->getComponentName (myIndex));
}

#endif // DEBUG
