// Maria union type expression class -*- c++ -*-

#ifndef UNIONTYPEEXPRESSION_H_
# define UNIONTYPEEXPRESSION_H_

# include "Expression.h"
# include "typedefs.h"

/** Union type expression: determine the type of a union expression */
class UnionTypeExpression : public Expression {
public:
  // Constructors and destructors

  /** Constructs UnionTypeExpression
   * @param expr	Expression evaluating to the union type
   * @param index	Index of the union component
   */
  UnionTypeExpression (class Expression* expr, card_t index);
private:
  /** Copy constructor */
  UnionTypeExpression (const class UnionTypeExpression& old);
  /** Assignment operator */
  class UnionTypeExpression& operator= (const class UnionTypeExpression& old);
protected:
  /** Destructs UnionTypeExpression */
  ~UnionTypeExpression ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eUnionType; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class UnionTypeExpression& other) const {
    return myIndex == other.myIndex && *myExpr == *other.myExpr;
  }

  /** Ordering comparison operator */
  bool operator< (const class UnionTypeExpression& other) const {
    if (myIndex < other.myIndex) return true;
    if (other.myIndex < myIndex) return false;
    return *myExpr < *other.myExpr;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The union-valued expression being referenced */
  class Expression* myExpr;
  /** Index of the referenced union typeExpression */
  card_t myIndex;
};

#endif // UNIONTYPEEXPRESSION_H_
