// unary operator class -*- c++ -*-

#include "UnopExpression.h"
#include "IntType.h"
#include "IntValue.h"
#include "Net.h"
#include "Constant.h"

UnopExpression::UnopExpression (enum Op op, class Expression* expr) :
  Expression (),
  myOp (op), myExpr (expr)
{
  setType (myOp == uSucc || myOp == uPred
	   ? expr->getType () : Net::getIntType ());
  assert (myOp == uSucc || myOp == uPred
	  ? getType ()->isOrdered ()
	  : getType ()->getKind () == Type::tInt);
  assert (myExpr->isBasic ());
}

UnopExpression::~UnopExpression ()
{
  myExpr->destroy ();
}

const char*
UnopExpression::getOpString () const
{
  switch (getOp ()) {
  case uNeg:
    return "-";
  case uNot:
    return "~";
  case uSucc:
    return "+";
  case uPred:
    return "|";
  }

  return "???";
}

#ifdef EXPR_CACHE
bool
UnopExpression::invalidate (const class VariableDefinition* variable) const
{
  if (myExpr->invalidate (variable)) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
UnopExpression::do_clearValuation (const class Valuation* valuation) const
{
  myExpr->clearValuation (valuation);
}

class Value*
UnopExpression::do_eval (const class Valuation* valuation) const
{
  const class Value* v = myExpr->const_eval (valuation);
  if (!v)
    return NULL;

  switch (getOp ()) {
  case uNeg:
    assert (v->getType ()->getKind () == Type::tInt);
    if (static_cast<const class IntValue*>(v)->getValue () == INT_T_MIN) {
      if (valuation)
	valuation->flagOverflow ();
      return NULL;
    }
    return constrain (valuation,
		      new class IntValue
		      (v->getType (),
		       -static_cast<const class IntValue*>(v)->getValue ()));
  case uNot:
    assert (v->getType ()->getKind () == Type::tInt);
    return constrain (valuation,
		      new class IntValue
		      (v->getType (),
		       ~static_cast<const class IntValue*>(v)->getValue ()));
  case uSucc:
    {
      class Value* w = v->copy ();
      w->increment ();
      return w;
    }

  case uPred:
    {
      class Value* w = v->copy ();
      w->decrement ();
      return w;
    }
  }

  assert (false);
  return NULL;
}
#else // EXPR_CACHE
class Value*
UnopExpression::do_eval (const class Valuation* valuation) const
{
  class Value* v = myExpr->eval (valuation);
  if (!v)
    return NULL;

  switch (getOp ()) {
  case uNeg:
    assert (v->getKind () == Value::vInt);
    {
      class IntValue* i = static_cast<class IntValue*>(v);
      if (i->getValue () == INT_T_MIN) {
	if (valuation)
	  valuation->flagOverflow ();
	delete i;
	return NULL;
      }
      else {
	i->setValue (-i->getValue ());
	return constrain (valuation, i);
      }
    }
  case uNot:
    assert (v->getKind () == Value::vInt);
    {
      class IntValue* i = static_cast<class IntValue*>(v);
      i->setValue (~i->getValue ());
      return constrain (valuation, i);
    }
  case uSucc:
    v->increment ();
    return v;
  case uPred:
    v->decrement ();
    return v;
  }

  assert (false);
  return NULL;
}
#endif // EXPR_CACHE

class Expression*
UnopExpression::ground (const class Valuation& valuation)
{
  class Expression* e = myExpr->ground (valuation);
  if (!e)
    return NULL;
  if (e == myExpr) {
    e->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class UnopExpression (myOp, e);
    expr->setType (getType ());
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
UnopExpression::substitute (class Substitution& substitution)
{
  class Expression* e = myExpr->substitute (substitution);
  if (e == myExpr) {
    e->destroy ();
    return copy ();
  }
  else
    return new class UnopExpression (myOp, e);
}

#ifdef EXPR_CACHE
void
UnopExpression::associateVariables (const class Expression& expr,
				    bool assoc) const
{
  myExpr->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
UnopExpression::forVariables (bool (*operation)
			      (const class VariableDefinition&)) const
{
  return myExpr->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

void
UnopExpression::prettyPrint (unsigned indent) const
{
  print (indent, "%s(\n", getOpString ());
  myExpr->prettyPrint (indent + 1);
  print (indent, ")\n");
}

#endif // DEBUG
