// unary operator class -*- c++ -*-

#ifndef UNOPEXPRESSION_H_
# define UNOPEXPRESSION_H_

# include "Expression.h"

/** Unary operator expression */
class UnopExpression : public Expression {
public:
  /** Unary operators */
  enum Op { uNeg, uNot, uSucc, uPred };

  // Constructors and destructors

  /** Constructs a new UnopExpression
   * @param op		Operator
   * @param expr	Expression the operator is applied to
   */
  UnopExpression (enum Op op, class Expression* expr);

private:
  /** Copy constructor */
  UnopExpression (const class UnopExpression& old);
  /** Assignment operator */
  class UnopExpression& operator= (const class UnopExpression& old);
protected:
  /** Destructs UnopExpression */
  ~UnopExpression ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eUnop; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Determine the operator */
  enum Op getOp () const { return myOp; }
  /** Determine the string representation of the operator */
  const char* getOpString () const;
  /** Determine the expression */
  const class Expression* getExpr () const { return myExpr; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class UnopExpression& other) const {
    return myOp == other.myOp && *myExpr == *other.myExpr;
  }

  /** Ordering comparison operator */
  bool operator< (const class UnopExpression& other) const {
    if (myOp < other.myOp) return true;
    if (other.myOp < myOp) return false;
    return *myExpr < *other.myExpr;
  }

  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The operator */
  enum Op myOp;
  /** The subexpression */
  class Expression* myExpr;
};

#endif // UNOPEXPRESSION_H_
