// variable reference class -*- c++ -*-

#include "Variable.h"
#include "Valuation.h"
#include "Value.h"
#include "VariableDefinition.h"
#include "Constant.h"
#include "Substitution.h"

#include <cassert>

Variable::Variable (const class VariableDefinition* variable) :
  Expression (),
  myVariable (variable)
{
  assert (myVariable && myVariable->getType ());
  setType (myVariable->getType ());
}

Variable::~Variable ()
{
}

void
Variable::setType (const class Type* type)
{
  assert (myVariable->getType ()->isAlwaysAssignable (type));
  Expression::setType (type);
}

#ifdef EXPR_CACHE
bool
Variable::invalidate (const class VariableDefinition* variable) const
{
  if (myVariable == variable) {
    Expression::invalidate ();
    return true;
  }

  return false;
}
#endif // EXPR_CACHE

class Value*
Variable::do_eval (const class Valuation* valuation) const
{
  if (!valuation)
    return NULL;
  const class Value* v = valuation->getValue (myVariable);
  assert (!v || v->getType ()->isAssignable (getType ()));
  if (v)
    return constrain (valuation, v->copy ());
  else {
    valuation->flagUndefinedValue ();
    return NULL;
  }
}

class Expression*
Variable::ground (const class Valuation& valuation)
{
  if (!myVariable->isAggregate ()) return copy ();
  const class Value* v = valuation.getValue (myVariable);
  assert (!v || v->getType ()->isAssignable (getType ()));
  assert (v || !myVariable->isAggregate ());

  return v ? new class Constant (v->copy ()) : copy ();
}

class Expression*
Variable::substitute (class Substitution& substitution)
{
  class Expression* e = substitution.getExpr (*myVariable);
  assert (!e || e->isAssignable (getType ()));

  return e ? e->copy () : copy ();
}

#ifdef EXPR_CACHE
void
Variable::associateVariables (const class Expression& expr, bool assoc) const
{
  if (assoc)
    myVariable->associate (expr);
  else
    myVariable->disassociate (expr);
}
#endif // EXPR_CACHE

bool
Variable::forVariables (bool (*operation)
			(const class VariableDefinition&)) const
{
  return operation (*myVariable);
}

bool
Variable::isCompatible (const class Value& value,
			const class Valuation& valuation) const
{
  assert (getType () && value.getType ());

  if (const class Value* v = valuation.getValue (myVariable))
    return *v == value ||
      (myVariable->hasChildren () &&
       myVariable->isCompatible (valuation, value));
  else
    return getType ()->isConstrained (&value);
}

const class VariableDefinition*
Variable::getLvalue (const class Value*& value,
		     class Valuation& valuation) const
{
  if (valuation.getValue (myVariable))
    return NULL;
  if (myVariable->hasChildren () && value &&
      !myVariable->isCompatible (valuation, *value))
    return NULL;
  return myVariable;
}

bool
Variable::isLvalue (const class VariableDefinition* variable) const
{
  return myVariable == variable;
}

#ifdef DEBUG
# include "util.h"

void
Variable::prettyPrint (unsigned indent) const
{
  print (indent, "%s\n", myVariable->getName ());
}
#endif // DEBUG
