// variable reference class -*- c++ -*-

#ifndef VARIABLE_H_
# define VARIABLE_H_

# include "Expression.h"

/** Variable reference */
class Variable : public Expression {
public:
  // Constructors and destructors

  /** Constructs a new Variable
   * @param variable	Identifies the variable
   */
  Variable (const class VariableDefinition* variable);

private:
  /** Copy constructor */
  Variable (const class Variable& old);
  /** Assignment operator */
  class Variable& operator= (const class Variable& old);
protected:
  /** Destructs Variable */
  ~Variable ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eVariable; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  /** Determine the variable */
  const class VariableDefinition* getVariable () const { return myVariable; }

  /** Set the type of the expression and the variable */
  void setType (const class Type* type);

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class Variable& other) const {
    return myVariable == other.myVariable;
  }

  /** Ordering comparison operator */
  bool operator< (const class Variable& other) const {
    return myVariable < other.myVariable;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute the variable with an expression, if possible
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Determine whether the expression is compatible with the specified value,
   * neglecting subexpressions that cannot be evaluated
   * @param value	value the expression will be compared to
   * @param valuation	variable substitutions
   * @return		true if the expression is compatible with the value
   */
  bool isCompatible (const class Value& value,
		     const class Valuation& valuation) const;

  /** Get an unbound lvalue and a corresponding value
   * @param value	value the expression should evaluate to
   * @param valuation	variable substitutions
   * @return		the lvalue; set value to the suggested value
   */
  const class VariableDefinition* getLvalue
  (const class Value*& value, class Valuation& valuation) const;

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** Name of the variable */
  const class VariableDefinition* myVariable;
};

#endif // VARIABLE_H_
