// Vector component presence expression -*- c++ -*-

#include "VectorComponent.h"
#include "VectorType.h"
#include "VectorValue.h"
#include "Valuation.h"
#include "Net.h"
#include "BoolType.h"
#include "BoolValue.h"
#include "Constant.h"

VectorComponent::VectorComponent (class Expression* vector,
				  class Expression* index) :
  Expression (),
  myVector (vector), myIndex (index)
{
  assert (myVector && myVector->getType ());
  assert (myVector->getType ()->getKind () == Type::tVector);
  assert (myIndex && myIndex->getType ());
  assert (myIndex->getType ()->isAssignable
	  (static_cast<const class VectorType*>(myVector->getType ())
	   ->getIndexType ()));
  assert (myVector->isBasic () && !myIndex->isBasic ());
  setType (Net::getBoolType ());
}

VectorComponent::~VectorComponent ()
{
  myVector->destroy ();
  myIndex->destroy ();
}

#ifdef EXPR_CACHE
bool
VectorComponent::invalidate (const class VariableDefinition* variable) const
{
  bool invalidatedVector = myVector->invalidate (variable);

  if (myIndex->invalidate (variable) || invalidatedVector) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
VectorComponent::do_clearValuation (const class Valuation* valuation) const
{
  myVector->clearValuation (valuation);
  myIndex->clearValuation (valuation);
}

class Value*
VectorComponent::do_eval (const class Valuation* valuation) const
{
  const class Value* vector = myVector->const_eval (valuation);
  if (!vector)
    return NULL;
  assert (vector->getType () == myVector->getType ());
  const class Value* index = myIndex->const_eval (valuation);
  if (!index)
    return NULL;
  const class Type* indexType =
    static_cast<const class VectorType*>(myVector->getType ())
    ->getIndexType ();

  assert (index->getType ()->isAssignable (indexType));

  if (indexType->isConstrained (index)) {
    card_t i = indexType->convert (index);
    assert (i < CARD_T_MAX && i < indexType->getNumValues ());
    const class Value* v = (*static_cast<const class VectorValue*>(vector))[i];
    return constrain (valuation, new class BoolValue (getType (), v != NULL));
  }
  else {
    if (valuation)
      valuation->flagConstraintViolation ();
    return NULL;
  }
}
#else // EXPR_CACHE
class Value*
VectorComponent::do_eval (const class Valuation* valuation) const
{
  class Value* vector = myVector->eval (valuation);
  if (!vector)
    return NULL;
  assert (vector->getType () == myVector->getType ());
  class Value* index = myIndex->eval (valuation);
  if (!index) {
    delete vector;
    return NULL;
  }
  const class Type* indexType =
    static_cast<const class VectorType*>(myVector->getType ())
    ->getIndexType ();

  assert (index->getType ()->isAssignable (indexType));

  if (indexType->isConstrained (index)) {
    card_t i = indexType->convert (index);
    assert (i < CARD_T_MAX && i < indexType->getNumValues ());
    const class Value* v = (*static_cast<const class VectorValue*>(vector))[i];
    delete vector;
    delete index;
    return constrain (valuation, new class BoolValue (getType (), v != NULL));
  }
  else {
    if (valuation)
      valuation->flagConstraintViolation ();
    delete vector;
    delete index;
    return NULL;
  }
}
#endif // EXPR_CACHE

class Expression*
VectorComponent::ground (const class Valuation& valuation)
{
  class Expression* vector = myVector->ground (valuation);
  class Expression* index = myIndex->ground (valuation);

  if (!vector || !index) {
    vector->destroy ();
    index->destroy ();
    return NULL;
  }

  if (vector == myVector && index == myIndex) {
    vector->destroy ();
    index->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class VectorComponent (vector, index);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
VectorComponent::substitute (class Substitution& substitution)
{
  class Expression* vector = myVector->substitute (substitution);
  class Expression* index = myIndex->substitute (substitution);

  if (vector == myVector && index == myIndex) {
    vector->destroy ();
    index->destroy ();
    return copy ();
  }
  else
    return new class VectorComponent (vector, index);
}

#ifdef EXPR_CACHE
void
VectorComponent::associateVariables (const class Expression& expr,
				     bool assoc) const
{
  myVector->associateVariables (expr, assoc);
  myIndex->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
VectorComponent::forVariables (bool (*operation)
			       (const class VariableDefinition&)) const
{
  return
    myVector->forVariables (operation) &&
    myIndex->forVariables (operation);
}

#ifdef DEBUG
# include "util.h"

void
VectorComponent::prettyPrint (unsigned indent) const
{
  print (indent, "?\n");
  myVector->prettyPrint (indent);
  print (indent, "[\n");
  myIndex->prettyPrint (indent + 1);
  print (indent, "]\n");
}

#endif // DEBUG
