// Vector component presence expression -*- c++ -*-

#ifndef VECTORCOMPONENT_H_
# define VECTORCOMPONENT_H_

# include "Expression.h"

/** Vector component presence expression */
class VectorComponent : public Expression {
public:
  // Constructors and destructors

  /** Constructor
   * @param vector	vector to be queried
   * @param index	the index
   */
  VectorComponent (class Expression* vector, class Expression* index);
private:
  /** Copy constructor */
  VectorComponent (const class VectorComponent& old);
  /** Assignment operator */
  class VectorComponent& operator= (const class VectorComponent& old);
protected:
  /** Destructor */
  ~VectorComponent ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eVectorComponent; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class VectorComponent& other) const {
    return *myVector == *other.myVector && *myIndex == *other.myIndex;
  }

  /** Ordering comparison operator */
  bool operator< (const class VectorComponent& other) const {
    if (*myVector < *other.myVector) return true;
    if (*other.myVector < *myVector) return false;
    return *myIndex < *other.myIndex;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** The vector expression */
  class Expression* myVector;
  /** The index expression */
  class Expression* myIndex;
};

#endif // VECTORCOMPONENT_H_
