// Maria vector expression class -*- c++ -*-

#include "VectorExpression.h"
#include "VectorValue.h"
#include "Constant.h"
#include <cstring>

VectorExpression::VectorExpression (const class VectorType* type) :
  Expression (),
  myComponents (NULL)
{
  assert (type && type->getSize () > 0);
  myComponents = new class Expression*[type->getSize ()];
  memset (myComponents, 0, type->getSize () * sizeof *myComponents);
  setType (type);
}

VectorExpression::~VectorExpression ()
{
  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; )
    myComponents[i]->destroy ();
  delete[] myComponents;
}

void
VectorExpression::setType (const class Type* type)
{
  assert (type && type->getKind () == Type::tVector);
  const class Type* itemType =
    static_cast<const class VectorType*>(type)->getItemType ();

  for (card_t i = static_cast<const class VectorType*>(type)->getSize ();
       i--; )
    if (myComponents[i])
      myComponents[i]->setType (itemType);

  Expression::setType (type);
}

#ifdef EXPR_CACHE
bool
VectorExpression::invalidate (const class VariableDefinition* variable) const
{
  bool invalidated = false;

  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; )
    if (myComponents[i]) {
      if (myComponents[i]->invalidate (variable))
	invalidated = true;
    }
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());

  if (invalidated)
    Expression::invalidate ();

  return invalidated;
}

void
VectorExpression::do_clearValuation (const class Valuation* valuation) const
{
  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; )
    if (myComponents[i])
      myComponents[i]->clearValuation (valuation);
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
}
#endif // EXPR_CACHE

class Value*
VectorExpression::do_eval (const class Valuation* valuation) const
{
  class VectorValue* vector = new class VectorValue (getType ());

  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; ) {
    if (myComponents[i]) {
      if (class Value* v = myComponents[i]->eval (valuation)) {
	assert (v->getType () ==
		static_cast<const class VectorType*>
		(getType ())->getItemType ());
	(*vector)[i] = v;
      }
      else {
	delete vector;
	return NULL;
      }
    }
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
  }

  return constrain (valuation, vector);
}

class Expression*
VectorExpression::ground (const class Valuation& valuation)
{
  bool same = true;
  const class VectorType* const type =
    static_cast<const class VectorType*>(getType ());
  class VectorExpression* expr = new class VectorExpression (type);

  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; ) {
    if (myComponents[i]) {
      class Expression* e = myComponents[i]->ground (valuation);
      if (!e) {
	expr->destroy ();
	return NULL;
      }

      if (e != myComponents[i])
	same = false;
      expr->myComponents[i] = e;
    }
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
  }

  if (same) {
    expr->destroy ();
    return copy ();
  }
  else if (valuation.isOKExcept (Valuation::undefinedValue)) {
    if (class Value* v = expr->eval (&valuation)) {
      class Expression* c = new class Constant (v);
      expr->destroy ();
      return c;
    }
    else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
      expr->destroy ();
      return NULL;
    }
  }

  return expr;
}

class Expression*
VectorExpression::substitute (class Substitution& substitution)
{
  bool same = true;
  const class VectorType* const type =
    static_cast<const class VectorType*>(getType ());
  class VectorExpression* expr = new class VectorExpression (type);

  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; ) {
    if (myComponents[i]) {
      class Expression* e = myComponents[i]->substitute (substitution);
      if (e != myComponents[i])
	same = false;
      expr->myComponents[i] = e;
    }
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
  }
  if (same) {
    expr->destroy ();
    return copy ();
  }
  else
    return expr;
}

#ifdef EXPR_CACHE
void
VectorExpression::associateVariables (const class Expression& expr,
				      bool assoc) const
{
  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; )
    if (myComponents[i])
      myComponents[i]->associateVariables (expr, assoc);
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
}
#endif // EXPR_CACHE

bool
VectorExpression::forVariables (bool (*operation)
				(const class VariableDefinition&)) const
{
  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; )
    if (myComponents[i] && !myComponents[i]->forVariables (operation))
      return false;
    else
      assert (myComponents[i] ||
	      static_cast<const class VectorType*>(getType ())->isOptional ());
  return true;
}

bool
VectorExpression::isCompatible (const class Value& value,
				const class Valuation& valuation) const
{
  assert (getType () && value.getType ());

  if (value.getKind () != Value::vVector)
    return false;

  assert (value.getType ()->getKind () == Type::tVector);

  const class VectorValue& vv = static_cast<const class VectorValue&>(value);
  const class VectorType* vt =
    static_cast<const class VectorType*>(getType ());

  assert (vv.getSize () ==
	  static_cast<const class VectorType*>(vv.getType ())->getSize ());

  if (!vv.getType ()->isAssignable (getType ()))
    return false;

  for (card_t i = vt->getSize (); i--; ) {
    if (myComponents[i]) {
#ifdef EXPR_CACHE
      const class Value* comp = myComponents[i]->const_eval (&valuation);
      if (comp && (!vv[i] || !(*comp == *vv[i])))
	return false;
#else // EXPR_CACHE
      class Value* comp = myComponents[i]->eval (&valuation);
      if (comp && (!vv[i] || !(*comp == *vv[i]))) {
	delete comp;
	return false;
      }
      delete comp;
#endif // EXPR_CACHE
    }
    else {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (vv[i])
	return false;
    }
  }

  return true;
}

const class VariableDefinition*
VectorExpression::getLvalue (const class Value*& value,
			     class Valuation& valuation) const
{
  assert (value->getKind () == Value::vVector);
  const class VectorValue* vv = static_cast<const class VectorValue*>(value);

  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; ) {
    if (myComponents[i]) {
      if (const class Value* val = (*vv)[i]) {
	const class VariableDefinition* v =
	  myComponents[i]->getLvalue (val, valuation);
	if (v) {
	  value = val;
	  return v;
	}
      }
    }
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
  }

  return NULL;
}

bool
VectorExpression::isLvalue (const class VariableDefinition* variable) const
{
  for (card_t i = static_cast<const class VectorType*>(getType ())->getSize ();
       i--; )
    if (myComponents[i] && myComponents[i]->isLvalue (variable))
      return true;
  return false;
}

#ifdef DEBUG
# include "util.h"

void
VectorExpression::prettyPrint (unsigned indent) const
{
  const class VectorType* type =
    static_cast<const class VectorType*>(getType ());
  print (indent, "{\n");
  for (card_t i = 0;;) {
    if (myComponents[i])
      myComponents[i]->prettyPrint (indent + 1);
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
    if (++i == type->getSize ())
      break;
    print (indent + 1, ",\n");
  }
  print (indent, "}\n");
}

#endif // DEBUG
