// Maria vector indexing expression -*- c++ -*-

#include "VectorIndex.h"
#include "VectorType.h"
#include "VectorValue.h"
#include "Valuation.h"
#include "Constant.h"
#include "Variable.h"
#include "VariableDefinition.h"
#include "VectorComponent.h"

VectorIndex::VectorIndex (class Expression* vector,
			  class Expression* index) :
  Expression (),
  myVector (vector), myIndex (index)
{
  assert (myVector && myVector->getType ());
  assert (myVector->getType ()->getKind () == Type::tVector);
  assert (myIndex && myIndex->getType ());
  assert (myIndex->getType ()->isAssignable
	  (static_cast<const class VectorType*>(myVector->getType ())
	   ->getIndexType ()));
  assert (myVector->isBasic () && myIndex->isBasic ());
  setType (static_cast<const class VectorType*>(myVector->getType ())
	   ->getItemType ());
}

VectorIndex::~VectorIndex ()
{
  myVector->destroy ();
  myIndex->destroy ();
}

#ifdef EXPR_CACHE
bool
VectorIndex::invalidate (const class VariableDefinition* variable) const
{
  bool invalidatedVector = myVector->invalidate (variable);

  if (myIndex->invalidate (variable) || invalidatedVector) {
    Expression::invalidate ();
    return true;
  }

  return false;
}

void
VectorIndex::do_clearValuation (const class Valuation* valuation) const
{
  myVector->clearValuation (valuation);
  myIndex->clearValuation (valuation);
}

class Value*
VectorIndex::do_eval (const class Valuation* valuation) const
{
  const class Value* index = myIndex->const_eval (valuation);
  if (!index)
    return NULL;
  const class Type* indexType =
    static_cast<const class VectorType*>(myVector->getType ())
    ->getIndexType ();

  assert (index->getType ()->isAssignable (indexType));
  assert (indexType->isConstrained (index));

  card_t i = indexType->convert (index);
  assert (i < CARD_T_MAX && i < indexType->getNumValues ());
  const class Value* vector;
  if (valuation && myVector->getKind () == Expression::eVariable) {
    const class VariableDefinition* var =
      static_cast<const class Variable*>(myVector)->getVariable ();
    if (!(vector = valuation->getValue (var))) {
      if (var->hasChildren ())
	if (const class Value* v = var->getChildValue (*valuation, i))
	  return constrain (valuation, v->copy ());
      valuation->flagUndefinedValue ();
      return NULL;
    }
  }
  else if (!(vector = myVector->const_eval (valuation)))
    return NULL;

  assert (vector->getType () == myVector->getType ());
  const class Value* v =
    (*static_cast<const class VectorValue*>(vector))[i];
  if (!v && valuation)
    valuation->flagUndefinedComponent ();
  return v ? constrain (valuation, v->copy ()) : NULL;
}
#else // EXPR_CACHE
class Value*
VectorIndex::do_eval (const class Valuation* valuation) const
{
  class Value* index = myIndex->eval (valuation);
  if (!index)
    return NULL;
  const class Type* indexType =
    static_cast<const class VectorType*>(myVector->getType ())
    ->getIndexType ();

  assert (index->getType ()->isAssignable (indexType));
  assert (indexType->isConstrained (index));

  card_t i = indexType->convert (index);
  assert (i < CARD_T_MAX && i < indexType->getNumValues ());
  delete index;
  if (valuation && myVector->getKind () == Expression::eVariable) {
    const class VariableDefinition* var =
      static_cast<const class Variable*>(myVector)->getVariable ();
    const class Value* v;
    if (!(v = valuation->getValue (var))) {
      if (var->hasChildren ())
	if (const class Value* v = var->getChildValue (*valuation, i))
	  return constrain (valuation, v->copy ());
      valuation->flagUndefinedValue ();
      return NULL;
    }
    else {
      assert (v->getKind () == Value::vVector);
      if (!(v = (*static_cast<const class VectorValue*>(v))[i])) {
	valuation->flagUndefinedComponent ();
	return NULL;
      }
      else
	return constrain (valuation, v->copy ());
    }
  }
  else {
    class Value* vector = myVector->eval (valuation);
    if (!vector)
      return NULL;
    assert (vector->getType () == myVector->getType ());
    if (const class Value* v =
	(*static_cast<class VectorValue*>(vector))[i]) {
      class Value* w = v->copy ();
      delete vector;
      return constrain (valuation, w);
    }
    else {
      if (valuation)
	valuation->flagUndefinedComponent ();
      delete vector;
      return NULL;
    }
  }
}
#endif // EXPR_CACHE

class Expression*
VectorIndex::ground (const class Valuation& valuation)
{
  class Expression* vector = myVector->ground (valuation);
  class Expression* index = myIndex->ground (valuation);

  if (!vector || !index) {
    vector->destroy ();
    index->destroy ();
    return NULL;
  }

  if (vector == myVector && index == myIndex) {
    vector->destroy ();
    index->destroy ();
    return copy ();
  }
  else {
    class Expression* expr = new class VectorIndex (vector, index);
    if (valuation.isOKExcept (Valuation::undefinedValue)) {
      if (class Value* v = expr->eval (&valuation)) {
	class Expression* c = new class Constant (v);
	expr->destroy ();
	return c;
      }
      else if (!valuation.isOKExcept (Valuation::undefinedValue)) {
	expr->destroy ();
	return NULL;
      }
    }

    return expr;
  }
}

class Expression*
VectorIndex::substitute (class Substitution& substitution)
{
  class Expression* vector = myVector->substitute (substitution);
  class Expression* index = myIndex->substitute (substitution);

  if (vector == myVector && index == myIndex) {
    vector->destroy ();
    index->destroy ();
    return copy ();
  }
  else
    return new class VectorIndex (vector, index);
}

#ifdef EXPR_CACHE
void
VectorIndex::associateVariables (const class Expression& expr,
				 bool assoc) const
{
  myVector->associateVariables (expr, assoc);
  myIndex->associateVariables (expr, assoc);
}
#endif // EXPR_CACHE

bool
VectorIndex::forVariables (bool (*operation)
			   (const class VariableDefinition&)) const
{
  return
    myVector->forVariables (operation) &&
    myIndex->forVariables (operation);
}

bool
VectorIndex::isCompatible (const class Value& value,
			   const class Valuation& valuation) const
{
  assert (isTypeCompatible (value));

#ifdef EXPR_CACHE
  if (const class Value* v = const_eval (&valuation))
    return value == *v;
#else // EXPR_CACHE
  if (class Value* v = eval (&valuation)) {
    bool compatible = value == *v;
    delete v;
    return compatible;
  }
#endif // EXPR_CACHE

  if (myVector->getKind () == Expression::eVariable &&
#ifdef EXPR_CACHE
      myVector->isCached () &&
#else // EXPR_CACHE
      valuation.getValue (static_cast<const class Variable*>
			  (myVector)->getVariable ()) &&
#endif // EXPR_CACHE
      valuation.isUndefinedComponent () &&
      valuation.isOKExcept (Valuation::undefinedComponent))
    valuation.clearErrors ();

  return true;
}

const class VariableDefinition*
VectorIndex::getLvalue (const class Value*& value,
			class Valuation& valuation) const
{
  const class Type* const indexType =
    static_cast<const class VectorType*>(myVector->getType ())
    ->getIndexType ();
  if (myVector->getKind () == Expression::eVariable) {
#ifdef EXPR_CACHE
    if (const class Value* const index = myIndex->const_eval (&valuation)) {
      const card_t i = indexType->convert (index);
#else // EXPR_CACHE
    if (class Value* const index = myIndex->eval (&valuation)) {
      const card_t i = indexType->convert (index);
      delete index;
#endif // EXPR_CACHE

      if (const class Value* const vv =
	  valuation.getValue (static_cast<const class Variable*>
			      (myVector)->getVariable ())) {
	assert (vv->getKind () == Value::vVector);
	if (static_cast<const class VectorValue&>(*vv)[i])
	  return NULL;
      }

      const class VariableDefinition* var =
	static_cast<const class Variable*>(myVector)
	->getVariable ()->getChild (i);
      if (!valuation.getValue (var))
	return var;
    }
  }

  return NULL;
}

bool
VectorIndex::isLvalue (const class VariableDefinition* variable) const
{
  return myVector->isLvalue (variable);
}

class VectorComponent*
VectorIndex::toComponent ()
{
  class VectorComponent* v =
    new class VectorComponent (myVector->copy (), myIndex->copy ());
  destroy ();
  return v;
}

#ifdef DEBUG
# include "util.h"

void
VectorIndex::prettyPrint (unsigned indent) const
{
  myVector->prettyPrint (indent);
  print (indent, "[\n");
  myIndex->prettyPrint (indent + 1);
  print (indent, "]\n");
}

#endif // DEBUG
