// Maria vector indexing expression -*- c++ -*-

#ifndef VECTORINDEX_H_
# define VECTORINDEX_H_

# include "Expression.h"

/** Vector indexing expression */
class VectorIndex : public Expression {
public:
  // Constructors and destructors

  /** Constructs VectorIndex
   * @param vector	vector to be indexed
   * @param index	the index
   */
  VectorIndex (class Expression* vector, class Expression* index);
private:
  /** Copy constructor */
  VectorIndex (const class VectorIndex& old);
  /** Assignment operator */
  class VectorIndex& operator= (const class VectorIndex& old);
protected:
  /** Destructs VectorIndex */
  ~VectorIndex ();
public:

  // Accessors

  /** Determine the type of the expression */
  enum Expression::Kind getKind () const { return eVectorIndex; }

  /** Determine whether this is a basic expression */
  bool isBasic () const { return true; }
  /** Determine whether this is a temporal logic expression */
  bool isTemporal () const { return false; }

  // Other methods

  /** Equality comparison operator */
  bool operator== (const class VectorIndex& other) const {
    return *myVector == *other.myVector && *myIndex == *other.myIndex;
  }

  /** Ordering comparison operator */
  bool operator< (const class VectorIndex& other) const {
    if (*myVector < *other.myVector) return true;
    if (*other.myVector < *myVector) return false;
    return *myIndex < *other.myIndex;
  }

# ifdef EXPR_CACHE
  /** Conditionally invalidate the cached value of the expression,
   * provided that it depends on the value of a variable
   * @param variable	variable the value should depend on
   * @return		true if the expression was invalidated
   */
  bool invalidate (const class VariableDefinition* variable) const;

  /** Recursively clear the valuation used for computing the cached value,
   * provided that it equals with the specified one
   * @param valuation	the valuation to be cleared
   */
  void do_clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Evaluate the expression
   * @param valuation	Variable substitutions
   * @return		Value of the expression, or NULL in case of error
   */
  class Value* do_eval (const class Valuation* valuation) const;

  /** Partially evaluate the expression using a valuation
   * @param valuation	Variable substitutions
   * @return		grounded expression, or NULL in case of error
   */
  class Expression* ground (const class Valuation& valuation);

  /** Substitute some variables in the expression with expressions
   * @param substitution	Variable substitutions
   * @return			substituted expression
   */
  class Expression* substitute (class Substitution& substitution);

# ifdef EXPR_CACHE
  /** Associate the variables in the (sub)expression with the given expression
   * @param expr	expression the variables are to be associated with
   * @param assoc	false=disassociate
   */
  void associateVariables (const class Expression& expr, bool assoc) const;
# endif // EXPR_CACHE

  /** Perform an operation on all variables of the expression
   * @param operation	operation to be performed (return false on failure)
   * @return		true if all operations succeeded
   */
  bool forVariables (bool (*operation)
		     (const class VariableDefinition&)) const;

  /** Determine whether the expression is compatible with the specified value,
   * neglecting subexpressions that cannot be evaluated
   * @param value	value the expression will be compared to
   * @param valuation	variable substitutions
   * @return		true if the expression is compatible with the value
   */
  bool isCompatible (const class Value& value,
		     const class Valuation& valuation) const;

  /** Get an unbound lvalue and a corresponding value
   * @param value	value the expression should evaluate to
   * @param valuation	variable substitutions
   * @return		the lvalue; set value to the suggested value
   */
  const class VariableDefinition* getLvalue
  (const class Value*& value, class Valuation& valuation) const;

  /** Determine whether a variable can ever be bound from the expression
   * @param variable	variable to look for
   * @return		true if the variable could be bound by getLvalue ()
   */
  bool isLvalue (const class VariableDefinition* variable) const;

  /** Convert the VectorIndex expression to a VectorComponent expression */
  class VectorComponent* toComponent ();

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** The vector expression */
  class Expression* myVector;
  /** The index expression */
  class Expression* myIndex;
};

#endif // VECTORINDEX_H_
