// Buffer for packing bits and values -*- c++ -*-

#include "BitBuffer.h"
#include "StructValue.h"
#include "StructType.h"
#include "UnionValue.h"
#include "UnionType.h"
#include "VectorValue.h"
#include "VectorType.h"
#include "BufferValue.h"
#include "BufferType.h"
#include "IntValue.h"
#include "EnumValue.h"
#include "IdValue.h"

void
BitPacker::append (const class Value& value)
{
  card_t numValues = value.getType ()->getNumValues ();
  assert (numValues);
  if (numValues == 1)
    return;
  if (numValues < CARD_T_MAX)
    append (value.getType ()->convert (&value), log2 (numValues));
  else {
    switch (value.getKind ()) {
    case Value::vStruct:
      {
	const class StructValue& s =
	  static_cast<const class StructValue&>(value);
	for (unsigned i = 0; i < s.getSize (); i++)
	  append (*s[i]);
      }
      break;
    case Value::vUnion:
      {
	const class UnionValue& u =
	  static_cast<const class UnionValue&>(value);
	append (u.getIndex (),
		log2 (static_cast<const class UnionType*>(u.getType ())
		      ->getSize ()));
	append (*u.getValue ());
      }
      break;
    case Value::vVector:
      {
	const class VectorValue& v =
	  static_cast<const class VectorValue&>(value);
	for (unsigned i = 0; i < v.getSize (); i++)
	  if (static_cast<const class VectorType*>(v.getType ())
	      ->isOptional ()) {
	    append (!v[i], 1);
	    if (v[i])
	      append (*v[i]);
	  }
	  else
	    assert (v[i] != NULL), append (*v[i]);
      }
      break;
    case Value::vBuffer:
      {
	const class BufferValue& b =
	  static_cast<const class BufferValue&>(value);
	card_t i = b.getCapacity ();
	append (i, log2 (static_cast<const class BufferType*>(b.getType ())
			 ->getSize ()));
	for (; i; i--)
	  append (*b[i]);
      }
      break;
    case Value::vInt:
      append (static_cast<const class IntValue&>(value).getValue (),
	      CARD_T_BIT);
      break;
    case Value::vEnum:
      append (static_cast<const class EnumValue&>(value).getValue (),
	      CARD_T_BIT);
    case Value::vId:
      append (static_cast<const class IdValue&>(value).getValue (),
	      CARD_T_BIT);
    default:
      assert (false);
    }
  }
}

class Value*
BitUnpacker::extract (const class Type& type)
{
  card_t numValues = type.getNumValues ();
  assert (numValues);
  if (numValues == 1)
    return type.convert (card_t (0));
  if (numValues < CARD_T_MAX)
    return type.convert (extract (log2 (numValues)));
  else {
    switch (type.getKind ()) {
    case Type::tStruct:
      {
	const class StructType& s =
	  static_cast<const class StructType&>(type);
	class StructValue* sv = new class StructValue (&type);
	for (unsigned i = 0; i < s.getSize (); i++)
	  (*sv)[i] = extract (*s[i]);
	return sv;
      }
    case Type::tUnion:
      {
	const class UnionType& u =
	  static_cast<const class UnionType&>(type);
	card_t index = extract (log2 (u.getSize ()));
	return new class UnionValue (&type, index, extract (*u[index]));
      }
    case Type::tVector:
      {
	const class VectorType& v =
	  static_cast<const class VectorType&>(type);
	class VectorValue* vv = new class VectorValue (&type);
	for (unsigned i = 0; i < v.getSize (); i++)
	  if (!v.isOptional () || !extract (1))
	    (*vv)[i] = extract (*v.getItemType ());
	return vv;
      }
    case Type::tBuffer:
      {
	const class BufferType& b =
	  static_cast<const class BufferType&>(type);
	class BufferValue* bv = new class BufferValue (&type);
	for (card_t i = extract (log2 (b.getSize ())); i; i--)
	  (*bv)[i] = extract (*b.getItemType ());
	return bv;
      }
    case Type::tInt:
      return new class IntValue (&type, extract (CARD_T_BIT));
    case Type::tEnum:
      return new class EnumValue (&type, extract (CARD_T_BIT));
    case Type::tId:
      return new class IdValue (&type, extract (CARD_T_BIT));
    default:
      assert (false);
      return NULL;
    }
  }
}
