// Buffer for packing bits and values -*- c++ -*-

#ifndef BITBUFFER_H_
# define BITBUFFER_H_

# include "typedefs.h"
# include <cstring>
# include <cassert>

/** Machine word */
typedef unsigned long word_t;
# define WORD_T_BIT (CHAR_BIT * sizeof (word_t))
# define WORD_T_MAX ULONG_MAX

/** Logarithm of base 2
 * @param num		number whose logarithm is to be computed
 * @return		number of bits required to represent num
 */
inline static unsigned
log2 (card_t num) {
  assert (num > 0);
  unsigned low = 0, high = CARD_T_BIT;

  for (;;) {
    unsigned log = (low + high) >> 1;

    card_t exp = card_t (1) << log;
    if (exp < num) {
      assert (high >= log);
      if (low == high) {
	assert (log == low);
	assert (log == CARD_T_BIT);
	return log;
      }
      low = log + 1;
    }
    else if ((exp >> 1) >= num) {
      assert (low < log);
      high = log - 1;
    }
    else
      return log;
  }
}

/** Buffer for packing bits */
class BitPacker {
public:
  /** Constructor
   * @param increment	Buffer allocation granularity (in WORD_T_BITs)
   */
  explicit BitPacker (card_t increment = 1) :
    myBuf (0), myAllocated (0), myLength (0), myIncrement (increment) {}
private:
  /** Copy constructor */
  explicit BitPacker (const class BitPacker& old);
  /** Assignment operator */
  class BitPacker& operator= (const class BitPacker& old);
public:
  /** Destructor */
  ~BitPacker () { delete[] myBuf; }

  /** Append data to to the buffer
   * @param data	Data to be appended
   * @param bits	Length of data in bits
   */
  void append (card_t data, unsigned bits) {
    assert (myIncrement);
    assert (bits && bits <= CARD_T_BIT);
    assert (bits == CARD_T_BIT || !(data >> bits));
    // enlarge the buffer if necessary
    if (myLength + bits > myAllocated * WORD_T_BIT) {
      do
	myAllocated += myIncrement;
      while (myLength + bits > myAllocated * WORD_T_BIT);
      word_t* buf = new word_t[myAllocated];
      size_t offset = (myLength + WORD_T_BIT - 1) / WORD_T_BIT;
      memcpy (buf, myBuf, offset * sizeof *buf);
      memset (buf + offset, 0, (myAllocated - offset) * sizeof *buf);
      delete[] myBuf;
      myBuf = buf;
    }

    // append the data to the buffer
    unsigned wordpos = myLength / WORD_T_BIT, bitpos = myLength % WORD_T_BIT;
    assert (!(myBuf[wordpos] >> bitpos));
    myLength += bits;
    myBuf[wordpos] |= word_t (data) << bitpos;
    while (bitpos > WORD_T_BIT - bits) {
      unsigned gotbits = WORD_T_BIT - bitpos;
      myBuf[++wordpos] = data >>= gotbits;
      bits -= gotbits;
      bitpos = 0;
    }
  }
  /** Append a value to the buffer
   * @param value	Value to be appended
   */
  void append (const class Value& value);

  /** Clear the buffer */
  void clear () {
    myLength = 0;
    memset (myBuf, 0, myAllocated * sizeof *myBuf);
  }
  /** Determine whether the buffer is empty */
  bool empty () const { return myLength == 0; }
  /** Determine the length of the encoded data in bits */
  unsigned getLength () const { return myLength; }
  /** Determine the length of the encoded data in words */
  unsigned getNumWords () const {
    return (myLength + WORD_T_BIT - 1) / WORD_T_BIT;
  }
  /** Calculate a hash value of the buffer */
  size_t getHashValue () const {
    size_t hashValue = 0;
    for (unsigned i = getNumWords (); i--;
	 hashValue ^= (hashValue << 2) ^ myBuf[i]);
    return hashValue;
  }
  /** Get the buffer contents */
  const word_t* getBuf () const { return myBuf; }
private:
  /** The buffer */
  word_t* myBuf;
  /** Actual size of the buffer in WORD_T_BITs */
  unsigned myAllocated;
  /** Number of bits used of the buffer */
  unsigned myLength;
  /** Amount of WORD_T_BITs to allocate when more memory is needed */
  unsigned myIncrement;
};

class BitUnpacker {
public:
  /** Constructor for read-only access
   * @param buf		a previously encoded string (read-only)
   */
  explicit BitUnpacker (const word_t* buf) : myBuf (buf), myOffset (0) {}
private:
  /** Copy constructor */
  explicit BitUnpacker (const class BitUnpacker& old);
  /** Assignment operator */
  class BitUnpacker& operator= (const class BitUnpacker& old);
public:
  /** Destructor */
  ~BitUnpacker () {}

  /** Extract data from the encoded string
   * @param bits	length of data in bits
   * @return		the data
   */
  card_t extract (unsigned bits) {
    assert (bits && bits <= CARD_T_BIT);
    unsigned wordpos = myOffset / WORD_T_BIT, bitpos = myOffset % WORD_T_BIT;
    myOffset += bits;

    const card_t mask =
      bits < CARD_T_BIT ? (card_t (1) << bits) - 1 : card_t (-1);
    card_t data = (myBuf[wordpos] >> bitpos) & mask;
    while (bitpos > WORD_T_BIT - bits) {
      unsigned gotbits = WORD_T_BIT - bitpos;
      bits -= gotbits;
      data |= myBuf[++wordpos] << gotbits;
      data &= mask;
      bitpos = 0;
    }

    return data;
  }

  /** Extract a value from the encoded string
   * @param type	type of the value
   * @return		the value
   */
  class Value* extract (const class Type& type);

private:
  /** the encoded bit string */
  const word_t* myBuf;
  /** offset to the encoded string */
  unsigned myOffset;
};

#endif // BITBUFFER_H_
