// Reachability graph -*- c++ -*-

#include "Graph.h"
#include "Event.h"
#include "GlobalMarking.h"
#include "util.h"

#include <cerrno>

/** file offset to state directory slot
 * @param s	number of the state
 * @return	file offset of the corresponding directory slot
 */
#define DIRSLOT(s) (myDirectoryOffset + sizeof (card_t) + \
  (s) * (sizeof (size_t) + sizeof (long)))
/** file offset to the second component of a state directory slot
 * @param s	number of the state
 * @return	file offset to the second component of the slot
 */
#define DIRSLOT2(s) (DIRSLOT (s) + sizeof (size_t))

#define ERROR_BIT ((1L) << (sizeof (long) * CHAR_BIT - 1))

Graph::Graph (const class Net& net) :
  myNet (net), myStates (), myBuf (), mySearchList (),
  myDirectory (0), myStateStorage (0), myArcStorage (0), myDirectoryOffset (0),
  myNumArgs (0), myArgs (0), myFilename (0), myFilebase (0),
  myNumArcs (0), myIsComplete (false)
{
}

Graph::Graph (const class Net& net, unsigned argc, const char*const* argv,
	      const char* filename, const char* filebase) :
  myNet (net), myStates (), myBuf (), mySearchList (),
  myDirectory (0), myStateStorage (0), myArcStorage (0), myDirectoryOffset (0),
  myNumArgs (argc), myArgs (0), myFilename (0), myFilebase (0),
  myNumArcs (0), myIsComplete (false)
{
  unsigned i;
  if (myNumArgs) {
    char** args = new char*[myNumArgs];
    for (i = 0; i < myNumArgs; i++)
      args[i] = newString (argv[i]);
    myArgs = args;
  }
  assert (filename || filebase);
  if (filename)
    myFilename = newString (filename);
  if (filebase)
    myFilebase = newString (filebase);
  else {
    // remove directory component
    for (i = 0; filename[i]; i++);
    while (i > 0 && filename[i - 1] != '/') i--;
    filebase = filename + i;

    // remove the last suffix starting with '.'
    for (i = 0; filebase[i]; i++);
    while (i > 0 && filebase[i] != '.') i--;
    if (!i) // no suffix
      for (i = 0; filebase[i]; i++);
    char* name = new char[i + 1];
    memcpy (name, filebase, i);
    name[i] = 0;
    myFilebase = name;
  }
}

Graph::~Graph ()
{
  if (myDirectory) {
    fseek (myDirectory, myDirectoryOffset, SEEK_SET);
    card_t w = myStates.size ();
    fwrite (&w, sizeof w, 1, myDirectory);
    fseek (myDirectory, DIRSLOT (w), SEEK_SET);
    w = mySearchList.size ();
    fwrite (&w, sizeof w, 1, myDirectory);
    while (!mySearchList.empty ()) {
      class Search s = mySearchList.pop (false);
      w = s.getNumber ();
      fwrite (&w, sizeof w, 1, myDirectory);
      size_t h = s.getHash ();
      fwrite (&h, sizeof h, 1, myDirectory);
    }
    fclose (myDirectory);
    fclose (myStateStorage);
    fclose (myArcStorage);
  }

  for (unsigned i = 0; i < myNumArgs; i++)
    delete[] myArgs[i];
  delete[] myArgs;
  delete[] myFilename;
  delete[] myFilebase;
}

/** Determine the length of a string
 * @param s	null-terminated string whose length is to be determined
 * @param n	maximum number of bytes to look for
 * @return	length of the string, or NULL
 */
inline static size_t
strnlen (const char* s, size_t n)
{
  size_t i;
  for (i = 0; i < n && s[i]; i++);
  return i;
}

bool
Graph::openFiles (bool regenerate)
{
  assert (!myDirectory && !myStateStorage && !myArcStorage);
  unsigned i;
  for (i = 0; myFilebase[i]; i++);

  // file access modes
  static const char reading[] = "r+", writing[] = "w+";
  const char* mode = regenerate ? writing : reading;

  {
    char* basename = new char[i + 5];
    memcpy (basename, myFilebase, i);

    memcpy (basename + i, ".rgd", 5);
    if (!(myDirectory = fopen (basename, mode))) {
      if (errno == ENOENT && myFilename) {
	mode = writing;
	if (!(myDirectory = fopen (basename, mode))) {
	  perror (basename);
	  delete[] basename;
	  return false;
	}
      }
      else {
	perror (basename);
	delete[] basename;
	return false;
      }
    }
    memcpy (basename + i, ".rgs", 5);
    if (!(myStateStorage = fopen (basename, mode))) {
      perror (basename);
      fclose (myDirectory); myDirectory = 0;
      delete[] basename;
      return false;
    }
    memcpy (basename + i, ".rga", 5);
    if (!(myArcStorage = fopen (basename, mode))) {
      perror (basename);
      fclose (myDirectory); myDirectory = 0;
      fclose (myStateStorage); myStateStorage = 0;
      delete[] basename;
      return false;
    }
    delete[] basename;
  }

  static const char magic[] = "Reachability Graph Directory";

  if (mode == reading) {
    const size_t capacity = 4096;
    char* buf = new char[capacity];
    assert (capacity >= sizeof magic);

    // check the magic cookie
    if ((fread (buf, 1, sizeof magic, myDirectory) != sizeof magic) ||
	memcmp (buf, magic, sizeof magic)) {
      fprintf (stderr, "%s: graph: wrong magic cookie\n",
	       myFilebase);
    wrongFormat:
      if (feof (myDirectory))
	fprintf (stderr, "%s: premature EOF of graph directory\n",
		 myFilebase);
    wrongFormatNoEOF:
      fclose (myDirectory); myDirectory = 0;
      fclose (myStateStorage); myStateStorage = 0;
      fclose (myArcStorage); myArcStorage = 0;
      delete[] buf;
      return false;
    }

    {
      /** original argument count (from the graph directory) */
      card_t oldc;
      if (fread (&oldc, sizeof oldc, 1, myDirectory) != 1)
	goto wrongFormat;
      /** original argument vector (from the graph directory) */
      char** oldv = new char*[oldc + 1];
      if (!oldv) {
	fprintf (stderr, "%s: graph: out of memory for %u arguments\n",
		 myFilebase, oldc + 1);
	goto wrongFormat;
      }

      /** flag: are the command line arguments equal? */
      bool sameArgs = oldc == myNumArgs;
      long pos = ftell (myDirectory);

      for (i = 0; i <= oldc; i++) {
	size_t got = fread (buf, 1, capacity, myDirectory);
	size_t len = strnlen (buf, got);
	if (!len) {
	  fprintf (stderr, "%s: graph: argument %u is empty\n",
		   myFilebase, i);
	argError:
	  while (i--)
	    delete[] oldv[i];
	  delete[] oldv;
	  goto wrongFormatNoEOF;
	}
	else if (len == got) {
	  fprintf (stderr, "%s: graph: argument %u is too long\n",
		   myFilebase, i);
	  goto argError;
	}
	if (!(oldv[i] = new char[++len])) {
	  fprintf (stderr, "%s: graph: out of memory for argument %u\n",
		   myFilebase, i);
	  goto argError;
	}
	memcpy (oldv[i], buf, len);
	if (sameArgs && i < myNumArgs && strcmp (oldv[i], getArg (i)))
	  sameArgs = false;
	fseek (myDirectory, pos += len, SEEK_SET);
      }

      if (!myNumArgs) {
	myNumArgs = oldc;
	assert (!myFilename);
	assert (!myArgs);
	myFilename = oldv[oldc];
	myArgs = oldv;
	myNumArgs = oldc;
      }
      else {
	if (sameArgs && strcmp (oldv[oldc], myFilename))
	  sameArgs = false;

	if (!sameArgs) {
	  fprintf (stderr, "%s: arguments for graph generation differ:\n",
		   myFilebase);
	  fprintf (stderr, "%s: original command line:\n",
		   myFilebase);
	  for (i = 0; i <= oldc; i++)
	    fprintf (stderr, "%s: argv[%u]=\"%s\"\n", myFilebase, i, oldv[i]);
	  fprintf (stderr, "%s: current command line:\n",
		   myFilebase);
	  for (i = 0; i < myNumArgs; i++)
	    fprintf (stderr, "%s: argv[%u]=\"%s\"\n", myFilebase,
		     i, myArgs[i]);
	  fprintf (stderr, "%s: argv[%u]=\"%s\"\n", myFilebase, myNumArgs,
		   myFilename);
	}

	for (i = 0; i <= oldc; i++)
	  delete[] oldv[i];
	delete[] oldv;
      }
    }

    myDirectoryOffset = ftell (myDirectory);
    // restore the states
    card_t numStates;
    if (fread (&numStates, sizeof numStates, 1, myDirectory) != 1)
      goto wrongFormat;
    for (card_t state = 0; state < numStates; state++) {
      size_t h;
      long pos;
      if (fread (&h, sizeof h, 1, myDirectory) != 1 ||
	  fread (&pos, sizeof pos, 1, myDirectory) != 1)
	goto wrongFormat;
      StateHash::value_type v (h, State (state));
      if (pos & ERROR_BIT)
	v.second.flagErroneous ();
      myStates.insert (v);
    }
    // restore the search list
    if (fread (&numStates, sizeof numStates, 1, myDirectory) != 1)
      goto wrongFormat;
    myIsComplete = !numStates;
    while (numStates--) {
      card_t num;
      size_t h;
      if (fread (&num, sizeof num, 1, myDirectory) != 1 ||
	  fread (&h, sizeof h, 1, myDirectory) != 1)
	goto wrongFormat;
      if (const class State* s = find (num, h))
	mySearchList.push (Search (*s, h));
      else {
	fprintf (stderr, "non-existing state %u (%x) on search list\n",
		 num, h);
	goto wrongFormat;
      }
    }
    // restore the arcs
    do {
      long pos = ftell (myArcStorage);
      card_t s[2];
      if (fread (s, sizeof s, 1, myArcStorage) != 1)
	break;
      size_t h[2];
      if (fread (h, sizeof h, 1, myArcStorage) != 1)
	goto wrongFormat;
      unsigned numWords;
      if (fread (&numWords, sizeof numWords, 1, myArcStorage) != 1)
	goto wrongFormat;
      fseek (myArcStorage, numWords * sizeof (word_t), SEEK_CUR);
      const class State* source = find (s[0], h[0]);
      const class State* target = find (s[1], h[1]);
      if (!source || !target) {
	fprintf (stderr, "%s: could not find the states for arc %u->%u\n",
		 myFilebase, s[0], s[1]);
	goto wrongFormat;
      }
      const_cast<class State*>(source)->connect
	(*new class Event (*target, pos));
      myNumArcs++;
    } while (!feof (myArcStorage));

    delete[] buf;
  }
  else { // create the directory afresh
    // magic cookie
    fwrite (magic, sizeof magic, 1, myDirectory);
    // argument count
    card_t w = myNumArgs;
    fwrite (&w, sizeof w, 1, myDirectory);
    // command line arguments
    for (i = 0; i < myNumArgs; i++)
      fwrite (myArgs[i], 1, 1 + strlen (myArgs[i]), myDirectory);
    // file name
    fwrite (myFilename, 1, 1 + strlen (myFilename), myDirectory);
    myDirectoryOffset = ftell (myDirectory);
    // number of states generated
    w = 0;
    fwrite (&w, sizeof w, 1, myDirectory);
    // size of search list
    fwrite (&w, sizeof w, 1, myDirectory);
    fflush (myDirectory);
  }

  return true;
}

const class Graph::AddStatus
Graph::add (const class GlobalMarking& m)
{
  myBuf.clear ();
  State::encode (m, myBuf);
#ifndef NDEBUG
# if 0 // this will really slow down the graph generation
  {
    class GlobalMarking* m2 = State::decode (myNet, myBuf.getBuf ());
    assert (m2 && *m2 == m);
    delete m2;
  }
# endif
#endif // NDEBUG
  const size_t h = myBuf.getHashValue ();
  std::pair<StateHash::const_iterator,StateHash::const_iterator> p =
    myStates.equal_range (h);
  StateHash::const_iterator i;

  if (p.first != p.second) {
    const unsigned numWords = myBuf.getNumWords ();
    word_t* buf = new word_t[numWords];

    for (i = p.first; i != p.second; i++) {
      fseek (myDirectory, DIRSLOT2 (i->second.getNumber ()), SEEK_SET);
      long num;
      if (fread (&num, sizeof num, 1, myDirectory) != 1)
	assert (false);
      assert (((num & ERROR_BIT) && true) == i->second.isErroneous ());

      if (fseek (myStateStorage, num & ~ERROR_BIT, SEEK_SET)) {
	perror ("Graph::add (): fseek");
	assert (false);
      }

      size_t got = fread (buf, sizeof (word_t), numWords, myStateStorage);
      if (got == numWords && !memcmp (buf, myBuf.getBuf (),
				      got * sizeof (word_t))) {
	// the state already exists on disk
	if (m.isErroneous () && !i->second.isErroneous ()) {
	  // the state was not previously marked as erroneous
	  const_cast<class State&>(i->second).flagErroneous ();
	  num |= ERROR_BIT;
	  fseek (myDirectory, DIRSLOT2 (i->second.getNumber ()), SEEK_SET);
	  if (fwrite (&num, sizeof num, 1, myDirectory) != 1)
	    assert (false);
	}
	delete[] buf;
	return AddStatus (false, i->second, h);
      }
    }

    delete[] buf;
  }

  // a new state
  StateHash::value_type v (h, State (myStates.size ()));
  if (m.isErroneous ())
    v.second.flagErroneous ();
  i = myStates.insert (v);
  // store the new state on disk
  {
    fseek (myDirectory, DIRSLOT (v.second.getNumber ()), SEEK_SET);
    if (fseek (myStateStorage, 0, SEEK_END))
      assert (false);
    long num = ftell (myStateStorage);
    if (v.second.isErroneous ()) num |= ERROR_BIT;
    if ((fwrite (&h, sizeof h, 1, myDirectory) != 1) ||
	(fwrite (&num, sizeof num, 1, myDirectory) != 1))
      assert (false);
    if (fwrite (myBuf.getBuf (), sizeof (word_t), myBuf.getNumWords (),
		myStateStorage) != myBuf.getNumWords ())
      assert (false);
  }

  return AddStatus (true, i->second, h);
}

void
Graph::add (const class State& source, const class State& target,
	    size_t sourcehash, size_t targethash,
	    const class Transition& transition,
	    const class Valuation& valuation)
{
  assert (&source == find (source.getNumber (), sourcehash));
  assert (&target == find (target.getNumber (), targethash));
  myBuf.clear ();
  Event::encode (myNet, transition, valuation, myBuf);
  long pos = ftell (myArcStorage);
  const_cast<class State&>(source).connect (*new class Event (target, pos));
  card_t s[2] = { source.getNumber (), target.getNumber () };
  size_t h[2] = { sourcehash, targethash };
  fwrite (s, sizeof s, 1, myArcStorage);
  fwrite (h, sizeof h, 1, myArcStorage);
  unsigned numWords = myBuf.getNumWords ();
  fwrite (&numWords, sizeof numWords, 1, myArcStorage);
  fwrite (myBuf.getBuf (), sizeof (word_t), numWords, myArcStorage);
  myNumArcs++;
}

const class State*
Graph::find (const class GlobalMarking& m)
{
  myBuf.clear ();
  State::encode (m, myBuf);
  size_t h = myBuf.getHashValue ();
  std::pair<StateHash::const_iterator,StateHash::const_iterator> p =
    myStates.equal_range (h);

  const unsigned numWords = myBuf.getNumWords ();
  word_t* buf = new word_t[numWords];

  for (StateHash::const_iterator i = p.first; i != p.second; i++) {
    fseek (myDirectory, DIRSLOT2 (i->second.getNumber ()), SEEK_SET);
    long num;
    if (fread (&num, sizeof num, 1, myDirectory) != 1)
      assert (false);
    assert (((num & ERROR_BIT) && true) == i->second.isErroneous ());

    if (fseek (myStateStorage, num & ~ERROR_BIT, SEEK_SET))
      assert (false);
    else
    {
      size_t got = fread (buf, sizeof (word_t), numWords, myStateStorage);
      bool isOld = got == numWords && !memcmp (buf, myBuf.getBuf (),
					       got * sizeof (word_t));
      if (isOld) {
	delete[] buf;
	return &i->second;
      }
    }
  }

  delete[] buf;
  return NULL;
}

const class State*
Graph::find (card_t number, size_t h) const
{
  std::pair<StateHash::const_iterator,StateHash::const_iterator> p =
    myStates.equal_range (h);
  for (StateHash::const_iterator i = p.first; i != p.second; i++)
    if (i->second.getNumber () == number)
      return &i->second;
  return NULL;
}

class Search
Graph::find (card_t number) const
{
  assert (number < size ());
  fseek (myDirectory, DIRSLOT (number), SEEK_SET);
  size_t h;
  if (fread (&h, sizeof h, 1, myDirectory) != 1)
    assert (false);
  const class State* s = find (number, h);
  assert (s != NULL);
  return Search (*s, h);
}

class GlobalMarking*
Graph::fetchState (card_t number) const
{
  assert (number < size ());
  fseek (myDirectory, DIRSLOT2 (number), SEEK_SET);
  long num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
  long length;
  if (number == size () - 1) {
    fseek (myStateStorage, 0, SEEK_END);
    length = ftell (myStateStorage) - (num & ~ERROR_BIT);
  }
  else {
    fseek (myDirectory, DIRSLOT2 (number + 1), SEEK_SET);
    if (fread (&length, sizeof length, 1, myDirectory) != 1)
      assert (false);
    length = (length & ~ERROR_BIT) - (num & ~ERROR_BIT);
  }

  assert (!(length % sizeof (word_t)));

  if (fseek (myStateStorage, num & ~ERROR_BIT, SEEK_SET))
    assert (false);
  word_t* buf = new word_t[length / sizeof (word_t)];
  if (fread (buf, 1, length, myStateStorage) != size_t (length))
    assert (false);
  class GlobalMarking* m = State::decode (myNet, buf);
  if (num & ERROR_BIT)
    m->flagErroneous ();
  delete[] buf;
  return m;
}

class Valuation*
Graph::fetchEvent (long offset, const class Transition*& transition) const
{
#ifndef NDEBUG
  fseek (myArcStorage, 0, SEEK_END);
  assert (offset < ftell (myArcStorage));
#endif // NDEBUG
  fseek (myArcStorage, offset + sizeof (card_t[2]) + sizeof (size_t[2]),
	 SEEK_SET);
  unsigned numWords;
  if (fread (&numWords, sizeof numWords, 1, myArcStorage) != 1)
    return NULL;
  if (word_t* buf = new word_t[numWords]) {
    if (fread (buf, sizeof *buf, numWords, myArcStorage) == numWords) {
      class Valuation* v = Event::decode (myNet, buf, transition);
      delete[] buf;
      return v;
    }
    delete[] buf;
  }
  return NULL;
}

void
Graph::flagErroneous (const class State& state)
{
  if (state.isErroneous ())
    return;
  fseek (myDirectory, DIRSLOT2 (state.getNumber ()), SEEK_SET);
  long num;
  if (fread (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
  assert (!(num & ERROR_BIT));
  fseek (myDirectory, DIRSLOT2 (state.getNumber ()), SEEK_SET);
  const_cast<class State&>(state).flagErroneous ();
  num |= ERROR_BIT;
  if (fwrite (&num, sizeof num, 1, myDirectory) != 1)
    assert (false);
}

#ifdef DEBUG
# include "Transition.h"
# include "Valuation.h"

void
Graph::prettyPrint (unsigned indent) const
{
  for (StateHash::const_iterator i = myStates.begin ();
       i != myStates.end (); i++) {
    const class State& state = i->second;
    print (indent, "State %u (%x):\n",
	   state.getNumber (), i->first);
    class GlobalMarking* m = fetchState (state.getNumber ());
    m->prettyPrint (indent + 1);
    delete m;
    for (unsigned e = 0; e < state.getNumEvents (); e++) {
      const class Event& event = state[e];
      print (indent + 1, "Event->%u:\n",
	     event.getSuccessor ().getNumber ());
      const class Transition* transition;
      if (class Valuation* valuation =
	  fetchEvent (event.getOffset (), transition)) {
	print (indent + 2, "trans %s\n", transition->getName ());
	valuation->prettyPrint (indent + 2);
	delete valuation;
      }
      else
	print (indent + 2, "(could not decode)\n");
    }
  }
}

#endif // DEBUG
