// Reachability graph -*- c++ -*-

#ifndef GRAPH_H_
# define GRAPH_H_

# include "BitBuffer.h"
# include "Search.h"

# ifdef HAS_HASH_SET
#  include <hash_map>
# else
#  include <map>
# endif // HAS_HASH_SET
# include <cassert>
# include <cstdio>

/** Reachability graph */
class Graph {
  // Typedefs

# ifdef HAS_HASH_SET
  typedef std::hash_multimap<size_t,class State> StateHash;
# else
  typedef std::multimap<size_t,class State> StateHash;
# endif // HAS_HASH_SET
public:
  /** Addition status: isNew=new state?, state=encoded state, h=hash value */
  class AddStatus {
  public:
    /** Constructor
     * @param isNew_	flag: was the state just added to the graph?
     * @param state_	the state
     * @param h_	hash value of the state
     */
    AddStatus (bool isNew_, const class State& state_, size_t h_) :
      isNew (isNew_), state (state_), h (h_) {
    }
    /** Copy constructor */
    AddStatus (const class AddStatus& old) :
      isNew (old.isNew), state (old.state), h (old.h) {
    }
  private:
    /** Assignment operator */
    class AddStatus& operator= (const class AddStatus& old);
  public:
    /** Destructor */
    ~AddStatus () {}
    /** flag: was the state just added to the graph? */
    const bool isNew;
    /** the state */
    const class State& state;
    /** hash value of the state */
    const size_t h;
  };

  /** The constructor
   * @param net		the net whose reachability graph this is
   */
  explicit Graph (const class Net& net);

  /** The constructor
   * @param net		the net whose reachability graph this is
   * @param argc	number of command line arguments
   * @param argv	the command line arguments
   * @param filename	name of the Petri Net file (NULL=read from graph file)
   * @param filebase	base name for the graph files (generated if NULL)
   */
  Graph (const class Net& net, unsigned argc, const char*const* argv,
	 const char* filename, const char* filebase);
private:
  /** Copy constructor */
  Graph (const class Graph& old);
  /** Assignment operator */
  class Graph& operator= (const class Graph& old);
public:
  /** Destructor */
  ~Graph ();

  /** Determine the net whose reachability graph this is */
  const class Net& getNet () const { return myNet; }

  /** Determine the number of command line arguments */
  unsigned getNumArgs () const { return myNumArgs; }
  /** Determine a command line argument
   * @param index	zero-based index of the argument
   * @return		the command-line argument
   */
  const char* getArg (unsigned index) const {
    assert (index < myNumArgs);
    return myArgs[index];
  }
  /** Determine the name of the Petri Net */
  const char* getFilename () const { return myFilename; }
  /** Determine the base name for the graph files */
  const char* getFilebase () const { return myFilebase; }
  /** Determine the number of arcs in the graph */
  card_t getNumArcs () const { return myNumArcs; }
  /** Determine whether the graph is complete */
  bool isComplete () const { return myIsComplete; }

  /** Open the files
   * @param regenerate	flag: true=ignore existing graph files
   * @return		true if the files could be opened; false on error
   */
  bool openFiles (bool regenerate);

  /** Add a state to the graph
   * @param m		state (marking) to be added
   * @return		first=true if new state, second=encoded state
   */
  const class AddStatus add (const class GlobalMarking& m);

  /** Add an event to the graph
   * @param source	the source state
   * @param target	the target state
   * @param sourcehash	hash value of the source state
   * @param targethash	hash value of the target state
   * @param transition	the transition
   * @param valuation	the binding of the transition
   */
  void add (const class State& source, const class State& target,
	    size_t sourcehash, size_t targethash,
	    const class Transition& transition,
	    const class Valuation& valuation);

  /** Check whether a state exists in the graph
   * @param m		state (marking) to be sought
   * @return		the encoded state corresponding to m, or NULL
   */
  const class State* find (const class GlobalMarking& m);

  /** Find a state by its number and hash value
   * @param number	the state number
   * @param h		the hash value
   * @return		the encoded state, or NULL
   */
  const class State* find (card_t number, size_t h) const;

  /** Find a state by a search list item
   * @param s		the search list item identifying the state
   * @return		the encoded state, or NULL
   */
  const class State* find (const class Search& s) const {
    return find (s.getNumber (), s.getHash ());
  }

  /** Find a search list item by state number
   * @param number	the state number
   * @return		the state and its hash value
   */
  class Search find (card_t number) const;

  /** Copy and clear the search list
   * @param searchList	the list to receive the copy
   */
  void getSearchList (class SearchList& searchList) {
    assert (searchList.empty ());
    searchList = mySearchList;
    mySearchList.clear ();
  }

  /** Copy the search list
   * @param searchList	the list to be copied
   */
  void setSearchList (const class SearchList& searchList) {
    assert (mySearchList.empty ());
    mySearchList = searchList;
  }

  /** Get the number of states in the reachability graph */
  size_t size () const { return myStates.size (); }

  /** Get a state by number
   * @param number	number of the state
   * @return		the decoded state
   */
  class GlobalMarking* fetchState (card_t number) const;

  /** Get an event by offset
   * @param offset	offset to the event file
   * @param transition	(output parameter) the transition
   * @return		the variable bindings for the transition
   */
  class Valuation* fetchEvent (long offset,
			       const class Transition*& transition) const;
 
  /** Flag a state erroneous
   * @param state	state to be flagged erroneous
   */
  void flagErroneous (const class State& state);

# ifdef DEBUG
  /** Print this object
   * @param indent	level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The net whose reachability graph this is */
  const class Net& myNet;
  /** The states */
  StateHash myStates;
  /** Buffer for encoding the states */
  class BitPacker myBuf;
  /** The search list */
  class SearchList mySearchList;
  /** The graph directory file */
  FILE* myDirectory;
  /** The state storage file */
  FILE* myStateStorage;
  /** The arc storage file */
  FILE* myArcStorage;
  /** Offset of the state directory in myDirectory */
  long myDirectoryOffset;
  /** Number of command line arguments used when creating the graph */
  unsigned myNumArgs;
  /** The command line arguments */
  const char*const* myArgs;
  /** Name of the Petri Net */
  const char* myFilename;
  /** Base name for the graph files */
  const char* myFilebase;
  /** Number of arcs in the graph */
  card_t myNumArcs;
  /** Flag: is the graph complete? */
  bool myIsComplete;
};

#endif // GRAPH_H_
