// State in a reachability graph -*- c++ -*-

#ifndef STATE_H_
# define STATE_H_

# include "Event.h"

/** error flag embedded in the number of successor arcs */
# define STATE_ERROR (card_t (1) << (CARD_T_BIT - 1))
/** visit flag embedded in the number of successor arcs */
# define STATE_VISIT (card_t (1) << (CARD_T_BIT - 2))
/** mask for getting rid of the flags */
# define STATE_MASK (~(STATE_ERROR | STATE_VISIT))

/** State in a reachability graph */
class State {
public:
  /** Default constructor */
  State () : myNumber (CARD_T_MAX), myNumEvents (0), myEvents (0) {}
  /** Constructor
   * @param number	number of the state
   */
  State (card_t number) : myNumber (number), myNumEvents (0), myEvents (0) {}
  /** Copy constructor */
  State (const class State& old) : myNumber (old.myNumber),
    myNumEvents (old.myNumEvents), myEvents (0) {
    assert (!old.getNumEvents ()); assert (!old.myEvents);
  }
private:
  /** Assignment operator */
  class State& operator= (const class State& old);
public:
  /** Destructor */
  ~State () {
    for (card_t i = getNumEvents (); i--; )
      delete myEvents[i];
    delete[] myEvents;
  }

  /** Equality comparison operator
   * @param other	state to be compared to
   * @return		true if the states are equal
   */
  bool operator== (const class State& other) const {
    return myNumber == other.myNumber;
  }
  /** Less-than comparison operator
   * @param other	state to be compared to
   * @return		true if this is smaller than other
   */
  bool operator< (const class State& other) const {
    return myNumber < other.myNumber;
  }

  /** Flag the state erroneous */
  void flagErroneous () {
    myNumEvents |= STATE_ERROR;
  }
  /** Determine whether the state is erroneous */
  bool isErroneous () const {
    return bool (myNumEvents & STATE_ERROR);
  }
  /** Set the "visited" flag */
  void visit () { myNumEvents |= STATE_VISIT; }
  /** Clear the "visited" flag */
  void unvisit () { myNumEvents &= ~STATE_VISIT; }
  /** Read the "visited flag */
  bool isVisited () const { return bool (myNumEvents & STATE_VISIT); }

  /** Determine the state number */
  card_t getNumber () const { return myNumber; }

  /** Connect the state to another state through an event
   * @param event	an event leading to another state
   */
  void connect (class Event& event);

  /** Encode a state
   * @param m		Marking to be encoded
   * @param buf		Buffer to receive the encoded marking
   */
  static void encode (const class GlobalMarking& m, class BitPacker& buf);

  /** Decode the state
   * @param net		the Petri Net from which the state was encoded
   * @param buf		the encoded data
   * @return		a GlobalMarking object corresponding to buf
   */
  static class GlobalMarking* decode (const class Net& net, const word_t* buf);

  /** Determine the number of successor events */
  card_t getNumEvents () const { return myNumEvents & STATE_MASK; }
  /** Get an event by index */
  const class Event& operator[] (unsigned index) const {
    assert (index < getNumEvents ());
    return *myEvents[index];
  }

  // Members

private:
  /** Number of the state */
  const card_t myNumber;
  /** Number of events leading to successor states,
   * possibly ORed with some flags in ~STATE_MASK */
  card_t myNumEvents;
  /** The events leading to successor states */
  class Event** myEvents;
};

#endif // STATE_H_
