// Global marking (PlaceMarking for each place) -*- c++ -*-

#include "GlobalMarking.h"
#include "Net.h"
#include "Place.h"
#include "Transition.h"
#include "Valuation.h"
#include "Expression.h"
#include "Token.h"

#include <cassert>

GlobalMarking::GlobalMarking (const class Net& net) :
#ifndef NDEBUG
  myNet (net),
#endif // NDEBUG
  myErroneous (false),
  mySize (net.getNumPlaces ()),
  myPlaceMarkings (NULL)
{
  if (mySize)
    myPlaceMarkings = new class PlaceMarking[mySize];
  for (unsigned i = 0; i < mySize; i++)
    myPlaceMarkings[i].setPlace (net.getPlace (i));
}

GlobalMarking::GlobalMarking (const class GlobalMarking& old) :
#ifndef NDEBUG
  myNet (old.myNet),
#endif // NDEBUG
  myErroneous (old.myErroneous),
  mySize (old.mySize),
  myPlaceMarkings (NULL)
{
  if (mySize)
    myPlaceMarkings = new class PlaceMarking[mySize];
  for (unsigned i = 0; i < mySize; i++)
    myPlaceMarkings[i] = old.myPlaceMarkings[i];
}

GlobalMarking::~GlobalMarking ()
{
  delete[] myPlaceMarkings;
}

class PlaceMarking&
GlobalMarking::operator[] (const class Place* place)
{
  assert (place && myNet.getPlace (place->getIndex ()) == place &&
	  place->getIndex () < mySize);
  return (*this)[place->getIndex ()];
}

/** Copy an unacceptable valuation to the valuation set and clear the errors
 * @param status	default return value
 * @param valuation	valuation to be checked
 * @param errors	valuation set to receive erroneous valuations
 * @return		status, or false if there is an error in the valuation
 */
inline static bool
willDo (bool status, class Valuation& valuation, class ValuationSet& errors)
{
  if (!valuation.isOKExcept (Valuation::undefinedValue) &&
      !valuation.isOKExcept (Valuation::undefinedComponent)) {
    status = false;
    errors.insert (new class Valuation (valuation));
  }

  valuation.clearErrors ();
  return status;
}

inline static class Token*
getNextToken (const class Transition& transition,
	      class Valuation& valuation,
	      const class Expression*& excuse,
	      class TokenList& tokens,
	      class GlobalMarking& m)
{
  assert (valuation.isOK ());
  assert (!excuse);
  if (tokens.empty ()) {
    if ((excuse = transition.getNextTokens (tokens, valuation, m))) {
      tokens.clear ();
      return NULL;
    }
    else
      tokens.toggleQuantificationId ();
  }

  return tokens.empty () ? NULL : tokens.pop ();
}

void
GlobalMarking::analyze (class ValuationSet& vs, class ValuationSet& errors,
			const class Transition& transition)
{
  assert (vs.empty ());
  assert (errors.empty ());
  assert (myNet.getTransition (transition.getIndex ()) == &transition);
  assert (unreserved ());

  class Valuation valuation;

  // Check constant gates (which should have been
  // eliminated by static analysis).
  const class Expression* excuse = NULL;
  if (!willDo (transition.checkGates (valuation, excuse), valuation, errors))
    return;
  excuse = NULL;

  /** processed tokens */
  class TokenList tokens;
  /** buffer for getNextToken () */
  class TokenList tokenlist;
  for (class Token* token = NULL;;) {
    if (token) {
      /* Check whether or not token can be bound to a new value. */
      token->undefineVariables ();
      assert (tokens.defined ());

      if (!token->getConcrete ()) /* All value candidates have been checked. */
	goto cont;

      assert (token->undefined ());
      if (!willDo (token->isCompatible (), valuation, errors)) {
      incompatible:
	token->next ();
	continue;
      }

      for (enum Token::LvalueStatus lvalueStatus;
	   (lvalueStatus = token->getLvalue (transition, tokens)) !=
	     Token::lvalueNotFound; ) {
	excuse = NULL;
	if (!willDo (lvalueStatus == Token::lvalueFound &&
		     valuation.isOKExcept (Valuation::undefinedValue),
		     valuation, errors))
	  goto incompatible;
	excuse = NULL;
      }

      token->flagOld ();
    tokenBinding:
      token->reserve ();
      tokens.push (token);
      tokens.validate ();
    }

    assert (!excuse);
    assert (valuation.isOK ());

    if ((token =
	 getNextToken (transition, valuation, excuse, tokenlist, *this))) {
      assert (!excuse);
      assert (valuation.isOK ());
      assert (token->isNew ());

      if (token->isBindable ()) {
	assert (valuation.isOK ());
	token->flagConcrete ();
	goto tokenBinding;
      }

      if ((valuation.isUndefinedValue () &&
	   valuation.isOKExcept (Valuation::undefinedValue)) ||
	  (valuation.isUndefinedComponent () &&
	   valuation.isOKExcept (Valuation::undefinedComponent))) {
	valuation.clearErrors ();
	continue;
      }

      willDo (true, valuation, errors);
      goto cont;
    }
    else {
      assert (excuse || (valuation.isOK () && tokenlist.empty ()));

      if ((valuation.isUndefinedValue () &&
	   valuation.isOKExcept (Valuation::undefinedValue)) ||
	  (excuse && excuse->getKind () == Expression::eMarking)) {
	assert (excuse != NULL);
	valuation.clearErrors ();
	excuse = NULL;
	goto cont;
      }

      if (!willDo (true, valuation, errors))
	goto cont;

      if (excuse)
	goto unificationError;
    }

    assert (!excuse);

    if (!tokens.empty ()) {
      if (!transition.checkGates (valuation, excuse))
	assert (false);
      assert (valuation.isOKExcept (Valuation::undefinedValue) ||
	      valuation.isOKExcept (Valuation::undefinedComponent));

      if (excuse) {
      unificationError:
	valuation.flagUnificationError ();
	willDo (false, valuation, errors);
	// TO DO: display an error message about excuse?
	goto cont;
      }

      valuation.clearErrors ();

      if (transition.isDefined (valuation)) {
	assert (valuation.isOK ());
	vs.insert (new class Valuation (valuation));
      }
      else {
	valuation.flagUnificationError ();
	errors.insert (new class Valuation (valuation));
	valuation.clearErrors ();
      }
    }

    assert (tokenlist.empty ());
  cont:
    if (token && token->isQuantified ()) {
      token->flagNew ();
      tokenlist.push (token);
    }
    else {
      delete token;
      tokenlist.clear ();
    }
    assert (tokens.defined ());

    if (tokens.empty ()) /* all bindings have been checked */
      break;

    token = tokens.pop ();

    if (!tokenlist.empty () &&
	token->getQuantificationId () !=
	tokenlist.top ()->getQuantificationId ())
      tokenlist.clear ();

    if (token->isReserved ())
      token->free ();

    // We step to the next candidate for the value of the popped
    // token. Observe that we had to wait for the possible
    // release operation above.
    if (token->isConcrete () || !token->next ())
      goto cont;
  }

  assert (unreserved ());
  return;
}

#include "Printer.h"
void
GlobalMarking::display (const class Printer& printer) const
{
  printer.printRaw (isErroneous () ? "erroneous state " : "state ");
  printer.delimiter ('(')++;

  for (unsigned i = 0; i < mySize; i++) {
    if (!myPlaceMarkings[i].empty ()) {
      printer.linebreak ();
      myPlaceMarkings[i].display (printer);
    }
  }

  printer--.linebreak ();
  printer.delimiter (')');
}

#ifdef DEBUG
# include "util.h"

void
GlobalMarking::prettyPrint (unsigned indent) const
{
  for (unsigned i = 0; i < mySize; i++)
    myPlaceMarkings[i].prettyPrint (indent);
  if (isErroneous ())
    print (indent, "erroneous state\n");
}

#endif // DEBUG
