// Net class -*- c++ -*-

#include "Net.h"
#include "Place.h"
#include "Transition.h"
#include "Type.h"
#include "BoolType.h"
#include "IntType.h"
#include "CharType.h"
#include "GlobalMarking.h"
#include "Marking.h"
#include "Valuation.h"
#include "Graph.h"
#include "Search.h"
#include "Function.h"
#include "Quantifier.h"
#include "Printer.h"

static class BoolType boolType;
static class IntType intType;
static class CharType charType;

Net::Net () :
  myNumPlaces (0), myPlaces (0), myNumTransitions (0), myTransitions (0),
  myFunctionMap (),
  myPlaceMap (), myTransitionMap (), myTypeMap (), myTypeSet (),
  myInitMarking (0), myTesterKind (NoTester), myTrigger (0), myFormula (0)
{
}

Net::~Net ()
{
  while (myNumPlaces)
    delete myPlaces[--myNumPlaces];
  delete myPlaces;

  while (myNumTransitions)
    delete myTransitions[--myNumTransitions];
  delete myTransitions;

  for (FunctionMap::iterator f = myFunctionMap.begin ();
       f != myFunctionMap.end (); f++)
    delete f->second;

  for (TypeSet::iterator t = myTypeSet.begin ();
       t != myTypeSet.end (); t++)
    if (*t != &boolType && *t != &intType && *t != &charType)
      delete *t;

  delete myInitMarking;
  myTrigger->destroy ();
  myFormula->destroy ();
}

void
Net::addFunction (class Function& function)
{
  assert (!getFunction (function.getName ()));
  myFunctionMap[function.getName ()] = &function;
}

void
Net::setFormula (class Expression* formula)
{
  assert (!myFormula);
  assert (formula && formula->getType ());
  assert (formula->getType ()->getKind () == Type::tBool);
  myFormula = formula;
}

class Place*
Net::addPlace (char* name, const class Type* type)
{
  assert (!myInitMarking);
  assert (name && myPlaceMap.find (name) == myPlaceMap.end ());

  class Place* place = new class Place (myNumPlaces, name, type);

  myPlaceMap[place->getName ()] = place;

  if (myPlaces) {
    class Place** p = new class Place*[myNumPlaces + 1];
    memcpy (p, myPlaces, sizeof (*myPlaces) * myNumPlaces);
    delete myPlaces;
    myPlaces = p;
  }
  else {
    assert (!myNumPlaces);
    myPlaces = new class Place*[1];
  }

  myPlaces[myNumPlaces++] = place;
  return place;
}

class Transition*
Net::addTransition (char* name)
{
  assert (!myInitMarking);
  assert (name && myTransitionMap.find (name) == myTransitionMap.end ());

  class Transition* transition = new class Transition (myNumTransitions, name);
  myTransitionMap[transition->getName ()] = transition;
  if (myTransitions) {
    class Transition** t = new class Transition*[myNumTransitions + 1];
    memcpy (t, myTransitions, sizeof (*myTransitions) * myNumTransitions);
    delete myTransitions;
    myTransitions = t;
  }
  else {
    assert (!myNumTransitions);
    myTransitions = new class Transition*[1];
  }

  myTransitions[myNumTransitions++] = transition;
  return transition;
}

void
Net::addType (class Type* type, char* name)
{
  assert (!myInitMarking);
  assert (type != NULL);

  if (name) {
    if (type->getName ()) {
      assert (strcmp (type->getName (), name));
      type = type->copy ();
    }
    type->setName (name);
    assert (myTypeMap.find (type->getName ()) == myTypeMap.end ());
    myTypeMap[type->getName ()] = type;
  }

  myTypeSet.insert (type);
}

void
Net::computeInitMarking (const class Printer& printer)
{
  assert (!myInitMarking);
  if (myInitMarking)
    return;
  myInitMarking = new class GlobalMarking (*this);
  class Valuation v;
  for (unsigned i = 0; i < myNumPlaces; i++) {
    const class Marking* m = myPlaces[i]->getInitMarking ();
    if (m) {
      v.clearErrors ();
      if (!m->add ((*myInitMarking)[i], 1u, v)) {
	assert (!v.isOK ());
	printer.printRaw ("place ");
	printer.print (myPlaces[i]->getName ());
	printer.printRaw (": error in the initial marking expression:");
	printer.finish ();
	v.display (printer);
	printer.finish ();
      }
    }
  }
}

const class BoolType*
Net::getBoolType ()
{
  return &boolType;
}

const class IntType*
Net::getIntType ()
{
  return &intType;
}

const class CharType*
Net::getCharType ()
{
  return &charType;
}

/** Flag a state erroneous and display an error message about it
 * @param printer	The printer object for diagnostic output
 * @param graph		The reachability graph the state belongs to
 * @param state		The state to be flagged erroneous
 * @param transition	The transition whose binding causes the error
 * @param numErrors	Number of errors (0 = fatal error encountered)
 * @param context	Context where the error occurred
 */
inline static void
flagErroneous (const class Printer& printer,
	       class Graph& graph, const class State& state,
	       const class Transition& transition,
	       card_t numErrors, const char* context)
{
  if (!numErrors) {
    printer.printRaw ("fatal expression caught in state @");
    printer.print (state.getNumber ());
    printer.printRaw (", transition ");
    printer.print (transition.getName ());
    printer.printRaw (", ");
    printer.printRaw (context);
    printer.finish ();
  }
  else if (!state.isErroneous ()) {
    printer.print (numErrors);
    if (numErrors == 1)
      printer.printRaw (" error in state @");
    else
      printer.printRaw (" errors in state @");
    printer.print (state.getNumber ());
    printer.printRaw (", transition ");
    printer.print (transition.getName ());
    printer.printRaw (", ");
    printer.printRaw (context);
    printer.finish ();
  }
  graph.flagErroneous (state);
}

/** Flag: has the analysis been interrupted? */
extern volatile bool interrupted;

void
Net::analyze (const class Printer& printer,
	      class Graph& graph, class SearchList& states,
	      bool breadth, bool onthefly) const
{
  class ValuationSet errors, vs;
  class SearchList oldstates, *search;
  if (onthefly)
    search = &(oldstates = states), states.clear ();
  else
    search = &states;

  while (!interrupted && !search->empty ()) {
    class Search source = search->pop (breadth);

    class GlobalMarking* m = graph.fetchState (source.getNumber ());

    for (unsigned i = getNumTransitions (); i--; ) {
      const class Transition& transition = *getTransition (i);

      if (transition.hasInputs ())
	m->analyze (vs, errors, transition);
      else
	vs.insert (new class Valuation);

      if (!errors.empty ()) {
	flagErroneous (printer, graph, source.getState (), transition,
		       errors.size (), "input:");
	for (ValuationSet::iterator e = errors.begin ();
	     e != errors.end (); e++) {
	  if ((*e)->isFatalExpression ()) {
	    flagErroneous (printer, graph, source.getState (), transition,
			   0, "input");
	    delete m;
	    return;
	  }
	  else {
	    (*e)->display (printer);
	    printer.finish ();
	  }
	}
	errors.clear ();
      }

      for (ValuationSet::iterator v = vs.begin (); v != vs.end (); v++) {
	if (const class QuantifierList* q = transition.getOutputVariables ()) {
	  switch (q->augment (**v)) {
	  case Quantifier::pass:
	    break;
	  case Quantifier::fail:
	    switch (q->advance (**v)) {
	    case Quantifier::pass:
	      break;
	    case Quantifier::fail:
	      if (!source.getState ().isErroneous ()) {
		printer.printRaw ("empty output in state @");
		printer.print (source.getNumber ());
		printer.printRaw (", transition ");
		printer.print (transition.getName ());
		printer.printRaw (":");
		printer.finish ();
		(*v)->display (printer);
		printer.finish ();
		graph.flagErroneous (source.getState ());
	      }
	      continue;
	    case Quantifier::undefined:
	      goto undefined;
	    }
	    break;
	  undefined:
	  case Quantifier::undefined:
	    flagErroneous (printer, graph, source.getState (), transition,
			   !(*v)->isFatalExpression (), "first output:");
	    (*v)->display (printer);
	    printer.finish ();
	    if ((*v)->isFatalExpression ()) {
	      delete m;
	      return;
	    }
	    continue;
	  }
	}
	do {
	loop:
	  class GlobalMarking successor (*m);
	  if (transition.fire (successor, **v)) {
	    Graph::AddStatus target = graph.add (successor);
	    graph.add (source.getState (), target.state,
		       source.getHash (), target.h,
		       transition, **v);
	    if (target.isNew || onthefly)
	      states.push (Search (target.state, target.h));
	  }
	  else {
	    flagErroneous (printer, graph, source.getState (), transition,
			   !(*v)->isFatalExpression (), "firing:");
	    (*v)->display (printer);
	    printer.finish ();
	    if ((*v)->isFatalExpression ()) {
	      delete m;
	      return;
	    }
	    (*v)->clearErrors ();
	  }

	  if (const class QuantifierList* q =
	      transition.getOutputVariables ()) {
	    switch (q->advance (**v)) {
	    case Quantifier::pass:
	      goto loop;
	    case Quantifier::fail:
	      break;
	    case Quantifier::undefined:
	      flagErroneous (printer, graph, source.getState (), transition,
			     !(*v)->isFatalExpression (), "next output:");
	      (*v)->display (printer);
	      printer.finish ();
	      if ((*v)->isFatalExpression ()) {
		delete m;
		return;
	      }
	      break;
	    }
	  }
	} while (0);
      }
      vs.clear ();
    }

    delete m;
  }
}

void
Net::analyze (const class Printer& printer,
	      class Graph& graph, const class GlobalMarking& marking,
	      bool breadth) const
{
  class SearchList states;
  graph.getSearchList (states);
  class Graph::AddStatus state = graph.add (marking);
  if (state.isNew)
    states.push (Search (state.state, state.h));
  analyze (printer, graph, states, breadth, false);
  assert (states.empty () || interrupted);
  graph.setSearchList (states);
}

#ifdef DEBUG
# include "util.h"

void
Net::analyze (class GlobalMarking& marking) const
{
  unsigned indent = 0;
  class ValuationSet vs;
  class ValuationSet errors;

  for (unsigned i = 0; i < myNumTransitions; i++) {
    print (indent, "Transition %s:\n", myTransitions[i]->getName ());
    marking.analyze (vs, errors, *myTransitions[i]);
    vs.prettyPrint (indent + 1);
    if (!errors.empty ()) {
      print (indent + 1, "errors:\n");
      errors.prettyPrint (indent + 2);
      errors.clear ();
    }
    vs.clear ();
  }
}

void
Net::prettyPrint (unsigned indent) const
{
  for (TypeSet::const_iterator t = myTypeSet.begin ();
       t != myTypeSet.end (); t++) {
    if (*t == &boolType || *t == &intType || *t == &charType)
      continue;

    if ((*t)->getName ()) {
      print (indent, "typedef\n");
      (*t)->prettyPrint (indent + 1);
      print (indent, "%s\n", (*t)->getName ());
    }
  }

  for (FunctionMap::const_iterator f = myFunctionMap.begin ();
       f != myFunctionMap.end (); f++)
    f->second->prettyPrint (indent);

  for (unsigned p = 0; p < myNumPlaces; p++)
    myPlaces[p]->prettyPrint (indent);

  for (unsigned tr = 0; tr < myNumTransitions; tr++)
    myTransitions[tr]->prettyPrint (indent);

  if (myTrigger) {
    static const char* kind[InfiniteTester + 1] = {
      "???", "reject", "deadlock", "livelock", "infinite"
    };

    print (indent, "tester %s %s\n",
	   myTrigger->getPlace ()->getName (),
	   kind[myTesterKind]);
    myTrigger->prettyPrint (indent + 1);
  }

  if (myFormula) {
    print (indent, "verify\n");
    myFormula->prettyPrint (indent + 1);
  }
}

#endif // DEBUG
