// Net class -*- c++ -*-

#ifndef NET_H_
# define NET_H_

# include <cassert>
# include <cstring>
# include <map>
# include <set>

/** Petri Net class */
class Net {
  /** String comparator */
  struct ltstr {
    bool operator () (const char* s1, const char* s2) const {
      return strcmp (s1, s2) < 0;
    }
  };
  typedef std::map<const char*,class Function*,struct ltstr> FunctionMap;
  typedef std::map<const char*,class Place*,struct ltstr> PlaceMap;
  typedef std::map<const char*,class Transition*,struct ltstr> TransitionMap;
  typedef std::map<const char*,class Type*,struct ltstr> TypeMap;
  typedef std::set<class Type*> TypeSet;

public:
  /** Tester kinds */
  enum TesterKind {
    NoTester, RejectTester, DeadlockTester, LivelockTester, InfiniteTester
  };

  // Constructors and destructors

  /** Constructor */
  Net ();
private:
  /** Copy constructor */
  Net (const class Net& old);
  /** Assignment operator */
  class Net& operator= (const class Net& old);
public:
  /** Destructor */
  ~Net ();

  // Accessors

  /** Get a function definition by name
   * @param name	Name of the function
   * @return		The function definition
   */
  class Function* getFunction (const char* name) {
    FunctionMap::const_iterator i = myFunctionMap.find (name);
    return i == myFunctionMap.end () ? NULL : i->second;
  }

  /** Add a function definition
   * @param function	The function
   */
  void addFunction (class Function& function);

  /** Get the number of places in the net */
  unsigned getNumPlaces () const { return myNumPlaces; }
  /** Get a place
   * @param i		Index number of the place
   */
  const class Place* getPlace (unsigned i) const {
    assert(i < myNumPlaces); return myPlaces[i];
  }
  /** Get a place
   * @param name	Name of the place
   */
  const class Place* getPlace (const char* name) const {
    PlaceMap::const_iterator i = myPlaceMap.find (name);
    return i == myPlaceMap.end () ? NULL : i->second;
  }

  /** Get the number of transitions in the net */
  unsigned getNumTransitions () const { return myNumTransitions; }
  /** Get a transition
   * @param i		Index number of the transition
   */
  const class Transition* getTransition (unsigned i) const {
    assert(i < myNumTransitions); return myTransitions[i];
  }
  /** Get a transition
   * @param name	Name of the transition
   */
  class Transition* getTransition (const char* name) {
    TransitionMap::const_iterator i = myTransitionMap.find (name);
    return i == myTransitionMap.end () ? NULL : i->second;
  }

  /** Get a type
   * @param name	Name of the type
   */
  const class Type* getType (const char* name) const {
    TypeMap::const_iterator i = myTypeMap.find (name);
    return i == myTypeMap.end () ? NULL : i->second;
  }
  /** Determine whether the net contains a type
   * @param type	Type to be sought
   */
  bool hasType (const class Type& type) const {
    return myTypeSet.find (const_cast<class Type*>(&type)) != myTypeSet.end ();
  }
  TypeMap::const_iterator begin () const { return myTypeMap.begin (); }
  TypeMap::const_iterator end () const { return myTypeMap.end (); }

  /** Get the initial marking of the net */
  const class GlobalMarking* getInitMarking () const { return myInitMarking; }
  /** Get the initial marking of the net */
  class GlobalMarking* getInitMarking () { return myInitMarking; }
  /** Compute the initial marking of the net (@see getInitMarking)
   * @param printer	The printer object for diagnostic output
   */
  void computeInitMarking (const class Printer& printer);

  /** Determine whether the net has a trigger */
  bool hasTester () const { return myTesterKind != NoTester; }
  /** Get the tester kind */
  enum TesterKind getTesterKind () const { return myTesterKind; }
  /** Get the triggering expression for the tester */
  const class Marking* getTesterTrigger () const { return myTrigger; }
  /** Set the tester for the net
   * @param testerKind	kind of the tester
   * @param trigger	the marking expression for triggering the tester
   */
  void setTester (enum TesterKind testerKind, class Marking* trigger) {
    assert (myTesterKind == NoTester && !myTrigger);
    assert (testerKind > NoTester && testerKind <= InfiniteTester);
    myTesterKind = testerKind; myTrigger = trigger;
  }

  /** Get the on-the-fly verification formula */
  const class Expression* getFormula () const { return myFormula; }
  /** Set the on-the-fly verification formula
   * @param formula	Temporal logic formula to be used
   */
  void setFormula (class Expression* formula);

  /** Get the boolean type */
  static const class BoolType* getBoolType ();
  /** Get the integer type */
  static const class IntType* getIntType ();
  /** Get the character type */
  static const class CharType* getCharType ();

  // Other methods

  /** Add a place to the net
   * @param name	Name of the place
   * @param type	Type of the place
   * @return		pointer to the place
   */
  class Place* addPlace (char* name, const class Type* type);

  /** Add a transition to the net
   * @param name	Name of the transition
   * @return		pointer to the transition
   */
  class Transition* addTransition (char* name);

  /** Add a type to the net
   * @param type	Type to be added
   * @param name	Name of the type
   */
  void addType (class Type* type, char* name = 0);

  /** Perform reachability analysis of the net
   * @param printer	The printer object for diagnostic output
   * @param graph	The reachability graph generated so far
   * @param states	List of states to be analyzed
   * @param breadth	Flag: search breadth-first (instead of depth-first)
   * @param onthefly	Flag: do not analyze new states on the search list
   */
  void analyze (const class Printer& printer,
		class Graph& graph, class SearchList& states,
		bool breadth, bool onthefly) const;

  /** Perform reachability analysis of the net
   * @param printer	The printer object for diagnostic output
   * @param graph	The reachability graph generated so far
   * @param marking	Marking to be analyzed
   * @param breadth	Flag: search breadth-first (instead of depth-first)
   */
  void analyze (const class Printer& printer,
		class Graph& graph, const class GlobalMarking& marking,
		bool breadth) const;

# ifdef DEBUG
  /** Determine all enabled instances of all transitions for the marking
   * @param marking	Marking to be analyzed
   */
  void analyze (class GlobalMarking& marking) const;

  /**
   * Prints this Net.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** Number of places in the net */
  unsigned myNumPlaces;
  /** Places in the net */
  class Place** myPlaces;
  /** Number of transitions in the net */
  unsigned myNumTransitions;
  /** Transitions in the net */
  class Transition** myTransitions;

  /** Functions indexed by name */
  FunctionMap myFunctionMap;
  /** Places indexed by name */
  PlaceMap myPlaceMap;
  /** Transitions indexed by name */
  TransitionMap myTransitionMap;
  /** Types indexed by name */
  TypeMap myTypeMap;
  /** All types in the net (also nameless ones) */
  TypeSet myTypeSet;
  /** Initial marking of the net */
  class GlobalMarking* myInitMarking;
  /** Tester kind */
  enum TesterKind myTesterKind;
  /** Triggering expression for tester */
  class Marking* myTrigger;
  /** Formula for on-the-fly verification */
  class Expression* myFormula;
};

#endif // NET_H_
