// Place marking class -*- c++ -*-

#ifndef PLACEMARKING_H_
# define PLACEMARKING_H_

# include "Value.h"
# include <map>
# include "typedefs.h"

/** Place marking */
class PlaceMarking {
public:
  /** Comparison operator for values */
  struct vcmp {
    /** Less-than comparison */
    bool operator() (const class Value* v1, const class Value* v2) const {
      return *v1 < *v2;
    }
  };
  /** Pair of cardinalities */
  class card_pair {
  public:
    /** Default constructor */
    card_pair () : total (0), reserved (0) {}
    /** Constructor
     * @param tot	total number of tokens
     * @param res	number of reserved tokens
     */
    card_pair (card_t tot, card_t res) : total (tot), reserved (res) {}
    /** Destructor */
    ~card_pair () {}
    /** Integrity check
     * @return		true if the check is passed
     */
    bool ok () const { return total && total >= reserved; }
    /** Total number of tokens */
    card_t total;
    /** Number of reserved tokens */
    card_t reserved;
    /** Equality comparison */
    bool operator== (const class card_pair& other) const {
      return total == other.total && reserved == other.reserved;
    }
  };
  /** The token map */
  typedef std::map<class Value*,class card_pair,struct vcmp> TokenMap;
  typedef TokenMap::iterator iterator;
  typedef TokenMap::const_iterator const_iterator;

  // Constructors and destructors

  /** Constructor */
  PlaceMarking () :
    myPlace (NULL),
# ifndef NDEBUG
    myType (NULL),
# endif // NDEBUG
    myTokens () {}
  /** Copy constructor */
  explicit PlaceMarking (const class PlaceMarking& old) :
    myPlace (old.myPlace),
# ifndef NDEBUG
    myType (old.myType),
# endif // NDEBUG
    myTokens () {
    for (const_iterator i = begin (); i != end (); i++)
      myTokens.insert (TokenMap::value_type (i->first->copy (), i->second));
  }
  /** Assignment operator */
  class PlaceMarking& operator= (const class PlaceMarking& other) {
    assert (!myPlace);
    myPlace = other.myPlace;
    assert (!myType);
#ifndef NDEBUG
    myType = other.myType;
#endif // NDEBUG
    assert (myTokens.empty ());
    for (const_iterator i = other.myTokens.begin ();
	 i != other.myTokens.end (); i++) {
      assert (i->first && i->second.ok ());
      myTokens.insert (TokenMap::value_type (i->first->copy (), i->second));
    }
    return *this;
  }
  /** Destructor */
  ~PlaceMarking () {
    for (const_iterator i = begin (); i != end (); i++)
      delete i->first;
  }

  // Accessors

# ifndef NDEBUG
  /** Set the Place associated with the PlaceMarking
   * @param place	The place whose marking this is
   */
  void setPlace (const class Place* place);

  /** Determine the Place associated with the PlaceMarking
   */
  const class Place* getPlace () const { return myPlace; }

  /** Set the Type associated with the PlaceMarking
   * @param type	The type of this marking
   */
  void setType (const class Type* type);

  /** Determine the Type associated with the PlaceMarking
   */
  const class Type* getType () const { return myType; }
# else // NDEBUG
  void setPlace (const class Place* place) { myPlace = place; }
# endif // NDEBUG

  /** Get the value of tokens pointed by an iterator
   * @param i	an iterator pointing to the value
   */
  static const class Value& getValue (const_iterator i) {
    return *i->first;
  }
  /** Get the number of tokens having the specified value
   * @param i	an iterator pointing to the value
   */
  static card_t getCount (const_iterator i) {
    return i->second.total;
  }
  /** Get the number of reserved tokens having the specified value
   * @param i	an iterator pointing to the value
   */
  static card_t getReservedCount (const_iterator i) {
    return i->second.reserved;
  }
  /** Get the number of non-reserved tokens having the specified value
   * @param i	an iterator pointing to the value
   */
  static card_t getFreeCount (const_iterator i) {
    return i->second.total - i->second.reserved;
  }

  // Iterators

  const_iterator begin () const { return myTokens.begin (); }
  const_iterator end () const { return myTokens.end (); }
  iterator begin () { return myTokens.begin (); }
  iterator end () { return myTokens.end (); }
  const_iterator find (const class Value* value) const {
    return myTokens.find (const_cast<class Value*>(value));
  }
  iterator find (const class Value* value) {
    return myTokens.find (const_cast<class Value*>(value));
  }
  size_t size () const { return myTokens.size (); }
  bool empty () const { return myTokens.empty (); }

  // Other methods

  /** Equality comparison */
  bool operator== (const class PlaceMarking& other) const {
    return myTokens == other.myTokens;
  }
  /** Subset comparison
   * @param other	a superset candidate
   * @return		true if this is a subset of other
   */
  bool operator<= (const class PlaceMarking& other) const {
    const_iterator i, j;
    assert (getType () == other.getType ());
    for (i = begin (); i != end (); i++) {
      assert (!i->second.reserved);
      if ((j = other.find (i->first)) != other.end ()) {
	assert (!j->second.reserved);
	if (j->second.total < i->second.total)
	return false;
      }
      else
      return false;
    }
    return true;
  }
  /** Multiset intersection
   * @param other	a set to be intersected with this
   * @return		the intersection
   */
  class PlaceMarking& operator&= (const class PlaceMarking& other) {
    assert (getType () == other.getType ());
    iterator i;
    const_iterator j;
    for (j = other.begin (); j != other.end (); j++) {
      assert (!j->second.reserved);
      if ((i = find (j->first)) != end ()) {
	assert (!i->second.reserved);
	if (i->second.total > j->second.total)
	  i->second.total = j->second.total;
      }
    }
    for (i = begin (); i != end (); i++) {
      assert (!i->second.reserved);
      if ((j = other.find (i->first)) == other.end ())
	myTokens.erase (i);
    }
    return *this;
  }
  /** Multiset difference
   * @param other	a set to be subtracted from this
   * @return		the difference
   */
  class PlaceMarking& operator-= (const class PlaceMarking& other) {
    assert (getType () == other.getType ());
    iterator i;
    const_iterator j;
    for (j = other.begin (); j != other.end (); j++) {
      assert (!j->second.reserved);
      if ((i = find (j->first)) != end ()) {
	assert (!i->second.reserved);
	if (i->second.total > j->second.total)
	  i->second.total -= j->second.total;
	else
	  myTokens.erase (i);
      }
    }
    return *this;
  }
  /** Multiset union (or sum)
   * @param other	a set to be added to this
   * @return		the sum
   */
  class PlaceMarking& operator+= (const class PlaceMarking& other) {
    assert (getType () == other.getType ());
    iterator i;
    const_iterator j;
    for (j = other.begin (); j != other.end (); j++) {
      assert (!j->second.reserved);
      if ((i = find (j->first)) != end ()) {
	assert (!i->second.reserved);
	i->second.total += j->second.total;
      }
    }
    return *this;
  }

# ifndef NDEBUG
  /** Assert that no tokens have been reserved
   * @return	true (or terminate in an assertion failure)
   */
  bool unreserved () const {
    for (const_iterator i = begin (); i != end (); i++)
      assert (!getReservedCount (i));
    return true;
  }
# endif // NDEBUG

  /** Add a number of specified tokens to the place marking.
   * @param value	Value of the tokens
   * @param amount	Number of tokens to be added
   */
  void add (class Value* value, card_t amount);

  /** Remove a number of specified tokens from the place marking.
   * @param value	Value of the tokens
   * @param amount	Number of tokens to be removed
   */
  void remove (const class Value* value, card_t amount);

  /** Remove a value from the place marking
   * @param i		Iterator to the value to be removed
   */
  void remove (iterator i) {
    assert (i != end ());
    assert (!i->second.reserved);
    myTokens.erase (i);
  }

  /** Reserve a number of tokens having the specified value.
   * @param i		Iterator to the value to be reserved
   * @param amount	Amount of tokens to reserve
   */
  void reserve (iterator i, card_t amount) {
    assert (i != end ());
    assert (i->second.total >= i->second.reserved + amount);
    i->second.reserved += amount;
  }

  /** Free a number of previously reserved tokens
   * @param i		Iterator to the value to be freed
   * @param amount	Amount of tokens to be freed
   */
  void free (iterator i, card_t amount) {
    assert (i != end ());
    assert (i->second.total >= i->second.reserved);
    assert (i->second.reserved >= amount);
    i->second.reserved -= amount;
  }

  /** Display this object
   * @param printer	The printer object
   */
  void display (const class Printer& printer) const;

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** The place */
  const class Place* myPlace;
# ifndef NDEBUG
  /** The type */
  const class Type* myType;
# endif // NDEBUG
  /** The tokens */
  TokenMap myTokens;
};

#endif // PLACEMARKING_H_
