// Transition class -*- c++ -*-

#include "Transition.h"
#include "Place.h"
#include "Arc.h"
#include "Marking.h"
#include "Expression.h"
#include "VariableDefinition.h"
#include "BoolValue.h"
#include "Token.h"
#include "GlobalMarking.h"
#include "BooleanBinop.h"
#include "Function.h"
#include "Quantifier.h"
#include <cstdio>

Transition::Transition (unsigned index, char* name) :
  myIndex (index), myName (name),
  myNumInputs (0), myInputs (0), myNumOutputs (0), myOutputs (0),
  myGates (), myOutputVariables (0), myVariables (), myFunctions ()
{
}

Transition::~Transition ()
{
  delete[] myName;

  while (myNumInputs)
    delete myInputs[--myNumInputs];
  delete myInputs;

  while (myNumOutputs)
    delete myOutputs[--myNumOutputs];
  delete myOutputs;

  for (GateList::iterator g = myGates.begin (); g != myGates.end (); g++)
    (*g)->destroy ();

  for (FunctionMap::iterator f = myFunctions.begin ();
       f != myFunctions.end (); f++)
    delete f->second;

  if (myOutputVariables) {
    for (QuantifierList::const_iterator q = myOutputVariables->begin ();
	 q != myOutputVariables->end (); q++) {
      iterator v = myVariables.find ((*q)->getVariable ()->getName ());
      assert (v != end () && v->second == (*q)->getVariable ());
      myVariables.erase (v);
    }

    delete myOutputVariables;
  }

  for (iterator v = myVariables.begin (); v != myVariables.end (); v++)
    delete v->second;
}

void
Transition::addGate (class Expression* gate)
{
  assert (gate && gate->getType ());
  assert (gate->getType ()->getKind () == Type::tBool);
  if (!gate->isAtomic () && gate->getKind () == Expression::eBooleanBinop) {
    class BooleanBinop* b = static_cast<class BooleanBinop*>(gate);
    if (b->getOp () == BooleanBinop::bAnd) {
      addGate (const_cast<class Expression*>(b->getLeft ())->copy ());
      addGate (const_cast<class Expression*>(b->getRight ())->copy ());
      b->destroy ();
      return;
    }
  }

  myGates.insert (myGates.end (), gate);
  gate->associateVariables (true);
}

#ifdef EXPR_CACHE
bool
Transition::isInput (const class VariableDefinition* variable) const
{
  for (unsigned i = 0; i < myNumInputs; i++)
    if (myInputs[i]->getExpr ()->invalidate (variable))
      return true;
  return false;
}

bool
Transition::isOutput (const class VariableDefinition* variable) const
{
  for (unsigned i = 0; i < myNumOutputs; i++)
    if (myOutputs[i]->getExpr ()->invalidate (variable))
      return true;
  return false;
}

bool
Transition::isGate (const class VariableDefinition* variable) const
{
  for (GateList::const_iterator i = myGates.begin (); i != myGates.end (); i++)
    if ((*i)->invalidate (variable))
      return true;
  return false;
}
#endif // EXPR_CACHE

bool
Transition::isLvalue (const class VariableDefinition* variable) const
{
  for (unsigned i = 0; i < myNumInputs; i++)
    if (myInputs[i]->getExpr ()->isLvalue (variable))
      return true;
  return false;
}

inline static unsigned
log16 (card_t num)
{
  assert (num > 0);
  assert (!(CARD_T_BIT % 4));
  unsigned low = 0, high = CARD_T_BIT / 4;

  for (;;) {
    unsigned log = (low + high) >> 1;

    card_t exp = card_t (1) << (4 * log);
    if (exp < num) {
      assert (high > log);
      low = log + 1;
    }
    else if ((exp >> 4) >= num) {
      assert (low < log);
      high = log - 1;
    }
    else
      return log;
  }
}

const class VariableDefinition*
Transition::addOutputVariable (const class Type* type, char* name,
			       class Expression* expr)
{
  assert (type != NULL);

  if (!myOutputVariables)
    myOutputVariables = new class QuantifierList;

  if (name) {
    const class VariableDefinition* d = getVariable (name);
    assert (d && d->getType () == type);
    myOutputVariables->append (new class Quantifier
			       (const_cast<class VariableDefinition*>(d),
				expr));
    return d;
  }
  else {
    for (card_t index = myOutputVariables ? myOutputVariables->size () : 0;
	 ; index++) {
      name = new char[index ? 2 + log16 (index) : 3];
      sprintf (name, ":%x", index);
      if (myVariables.find (name) == myVariables.end ())
	break;
      delete[] name;
    }
    class Quantifier* q = new class Quantifier (name, type, expr);
    myOutputVariables->append (q);
    const class VariableDefinition* d = q->getVariable ();
    assert (myVariables.find (name) == myVariables.end ());
    myVariables[d->getName ()] = const_cast<class VariableDefinition*>(d);
    return d;
  }
}

bool
Transition::isOutputVariable (const class VariableDefinition* variable) const
{
  assert (variable && getVariable (variable->getName ()) == variable);
  if (myOutputVariables)
    for (QuantifierList::const_iterator i = myOutputVariables->begin ();
	 i != myOutputVariables->end (); i++)
      if ((*i)->getVariable () == variable)
	return true;
  return false;
}

const class VariableDefinition*
Transition::addVariable (char* name, const class Type* type)
{
  assert (name && myVariables.find (name) == myVariables.end ());
  class VariableDefinition* d =
    new class VariableDefinition (name, type);
  myVariables[d->getName ()] = d;
  return d;
}

void
Transition::addFunction (class Function& function)
{
  assert (!getFunction (function.getName ()));
  myFunctions[function.getName ()] = &function;
}

#ifdef EXPR_CACHE
bool
Transition::checkGates (const class Valuation& valuation,
			const class Expression*& excuse) const
{
  excuse = NULL;
  for (GateList::const_iterator i = myGates.begin ();
       i != myGates.end (); i++) {
    const class Value* v = (*i)->const_eval (&valuation);
    if (!v) {
      excuse = *i;
      continue;
    }
    assert (v->getKind () == Value::vBool);
    if (!static_cast<const class BoolValue*>(v)->getValue ())
      return false;
  }

  return true;
}
#else // EXPR_CACHE
bool
Transition::checkGates (const class Valuation& valuation,
			const class Expression*& excuse) const
{
  excuse = NULL;
  for (GateList::const_iterator i = myGates.begin ();
       i != myGates.end (); i++) {
    class Value* v = (*i)->eval (&valuation);
    if (!v) {
      excuse = *i;
      continue;
    }
    assert (v->getKind () == Value::vBool);
    bool accepted = static_cast<const class BoolValue*>(v)->getValue ();
    delete v;
    if (!accepted)
      return false;
  }

  return true;
}
#endif // EXPR_CACHE

bool
Transition::isDefined (const class Valuation& valuation) const
{
  assert (valuation.isOK ());

  for (const_iterator vd = begin (); vd != end (); vd++) {
    const class VariableDefinition* const var = vd->second;
    if (var->hasChildren () && !valuation.getValue (var)) {
      if (class Value* value = var->getValue (valuation))
	delete value;
      else {
	valuation.flagUndefinedComponent ();
	return false;
      }
    }
  }

  for (unsigned index = 0; index < myNumInputs; index++) {
    if (!myInputs[index]->getExpr ()->isDefined (valuation)) {
      assert ((valuation.isUndefinedValue () &&
	       valuation.isOKExcept (Valuation::undefinedValue)) ||
	      (valuation.isUndefinedComponent () &&
	       valuation.isOKExcept (Valuation::undefinedComponent)));
      return false;
    }
  }

  for (GateList::const_iterator i = myGates.begin ();
       i != myGates.end (); i++) {
#ifdef EXPR_CACHE
    if (const class Value* v = (*i)->const_eval (&valuation)) {
      assert (v->getKind () == Value::vBool);
      assert (static_cast<const class BoolValue*>(v)->getValue () == true);
    }
#else // EXPR_CACHE
    if (class Value* v = (*i)->eval (&valuation)) {
      assert (v->getKind () == Value::vBool);
      assert (static_cast<const class BoolValue*>(v)->getValue () == true);
      delete v;
    }
#endif // EXPR_CACHE
    else {
      assert (valuation.isOKExcept (Valuation::undefinedValue));
      assert (valuation.isUndefinedValue ());
      return false;
    }
  }

  return true;
}

void
Transition::addArc (class Arc* arc)
{
  assert (arc != NULL);

  if (arc->isOutput ()) {
    if (myOutputs) {
      class Arc** a = new class Arc*[myNumOutputs + 1];
      memcpy (a, myOutputs, sizeof (*myOutputs) * myNumOutputs);
      delete[] myOutputs;
      myOutputs = a;
    }
    else {
      assert (!myNumOutputs);
      myOutputs = new class Arc*[1];
    }

    myOutputs[myNumOutputs++] = arc;
  }
  else {
    if (myInputs) {
      class Arc** a = new class Arc*[myNumInputs + 1];
      memcpy (a, myInputs, sizeof (*myInputs) * myNumInputs);
      delete[] myInputs;
      myInputs = a;
    }
    else {
      assert (!myNumInputs);
      myInputs = new class Arc*[1];
    }

    myInputs[myNumInputs++] = arc;
  }
}

class Arc*
Transition::getArc (bool output, const class Place* place)
{
  if (output) {
    for (unsigned i = 0; i < myNumOutputs; i++)
      if (myOutputs[i]->getPlace () == place)
	return myOutputs[i];
  }
  else {
    for (unsigned i = 0; i < myNumInputs; i++)
      if (myInputs[i]->getPlace () == place)
	return myInputs[i];
  }
  return NULL;
}

const class Expression*
Transition::getNextTokens (class TokenList& tokens,
			   class Valuation& valuation,
			   class GlobalMarking& m) const
{
  assert (tokens.empty ());
  const class Expression* firstexcuse = NULL;

  for (unsigned index = 0; index < myNumInputs && tokens.empty (); index++) {
    const class Arc* arc = myInputs[index];

    assert (arc && arc->getExpr ());
    assert (tokens.empty ());
    assert (valuation.isOK ());

    class PlaceMarking& place = m[arc->getPlace ()];
    firstexcuse = NULL;

    for (const class Marking* m = arc->getExpr (); m; m = m->getNext ()) {
      if (m->isQuantified ()) {
	if (!firstexcuse) {
	  if (!(firstexcuse = m->getNextTokens (tokens, valuation, place)))
	    break;
	  else if (firstexcuse->getKind () != Expression::eMarking &&
		   (valuation.isOKExcept (Valuation::undefinedValue) ||
		    valuation.isOKExcept (Valuation::undefinedComponent)))
	    valuation.clearErrors ();
	  else
	    return firstexcuse;
	}
      }
      else {
	const class Expression* excuse = NULL;
	if (class Token* token =
	    m->getNextToken (valuation, place, excuse)) {
	  assert (!excuse);
	  tokens.push (token);
	  return NULL;
	}
	else if (excuse)
	  return excuse;
      }
    }
  }

  return firstexcuse;
}

bool
Transition::fire (class GlobalMarking& marking,
		  const class Valuation& valuation) const
{
  assert (valuation.isOK ());

  for (unsigned in = 0; in < myNumInputs; in++) {
    const class Arc* arc = myInputs[in];
    class PlaceMarking& place = marking[arc->getPlace ()];
    if (!arc->getExpr ()->remove (place, 1, valuation))
      assert (false);
  }

  for (unsigned out = 0; out < myNumOutputs; out++) {
    const class Arc* arc = myOutputs[out];
    class PlaceMarking& place = marking[arc->getPlace ()];
    if (!arc->getExpr ()->add (place, 1, valuation)) {
      assert (!valuation.isOK ());
      return false;
    }
  }

  return true;
}

#ifdef DEBUG
# include "util.h"

void
Transition::prettyPrint (unsigned indent) const
{
  print (indent, "trans %s\n", myName);
  if (!myVariables.empty () || !myFunctions.empty ()) {
    print (indent, "var {\n");

    for (const_iterator v = myVariables.begin (); v != myVariables.end (); ) {
      (*v).second->prettyPrint (indent + 1);
      if (++v == myVariables.end ())
	break;
      print (indent + 1, ",\n");
    }

    if (!myVariables.empty () && !myFunctions.empty ())
      print (indent + 1, ",\n");

    for (FunctionMap::const_iterator f = myFunctions.begin ();
	 f != myFunctions.end (); ) {
      f->second->prettyPrint (indent + 1);
      if (++f == myFunctions.end ())
	break;
      print (indent + 1, ",\n");
    }

    print (indent, "}\n");
  }

  if (myNumInputs) {
    print (indent, "in {\n");
    for (unsigned i = 0; i < myNumInputs; ) {
      print (indent, "%s:\n", myInputs[i]->getPlace ()->getName ());
      myInputs[i]->getExpr ()->prettyPrint (indent + 1);
      if (++i < myNumInputs)
	print (indent, ";\n");
    }
    print (indent, "}\n");
  }

  if (myNumOutputs) {
    print (indent, "out {\n");
    for (unsigned i = 0; i < myNumOutputs; ) {
      print (indent, "%s:\n", myOutputs[i]->getPlace ()->getName ());
      myOutputs[i]->getExpr ()->prettyPrint (indent + 1);
      if (++i < myNumOutputs)
	print (indent, ";\n");
    }
    print (indent, "}\n");
  }

  if (!myGates.empty ()) {
    print (indent, "gate\n");
    for (GateList::const_iterator i = myGates.begin ();
	 i != myGates.end (); ) {
      (*i)->prettyPrint (indent);
      if (++i != myGates.end ())
	print (indent, ",\n");
    }
  }
}

#endif // DEBUG
