// Transition class -*- c++ -*-

#ifndef TRANSITION_H_
# define TRANSITION_H_

# include <cstring>
# include <map>
# include <slist>

/** Transition class */
class Transition {
public:
  // Typedefs

  typedef std::slist<class Expression*> GateList;
  /** String comparator */
  struct ltstr {
    bool operator () (const char* s1, const char* s2) const {
      return strcmp (s1, s2) < 0;
    }
  };
  typedef std::map<const char*,class VariableDefinition*,
    struct ltstr> VariableMap;
  typedef VariableMap::iterator iterator;
  typedef VariableMap::const_iterator const_iterator;
  typedef std::map<const char*,class Function*,struct ltstr> FunctionMap;

  // Constructors and destructors

  /** Constructs a Transition
   * @param index	Index number of the transition (ask Net)
   * @param name	Name of the transition
   */
  Transition (unsigned index, char* name);
private:
  /** Copy constructor */
  Transition (const Transition& old);
  /** Assignment operator */
  Transition& operator= (const Transition& old);
public:
  /** Destructs a Transition */
  ~Transition ();

  // Accessors

  /** Get the index number of the transition */
  unsigned getIndex () const { return myIndex; }
  /** Get the name of the transition */
  const char* getName () const { return myName; }

  /** Add a gate expression
   * @param gate	The gate expression
   */
  void addGate (class Expression* gate);

  /** Get a variable by name
   * @param name	Name of the variable
   * @return		the VariableDefinition object; NULL if there is none
   */
  const class VariableDefinition* getVariable (const char* name) const {
    const_iterator i = myVariables.find (name);
    return i != myVariables.end () ? i->second : NULL;
  }

# ifdef EXPR_CACHE
  /** Determine whether a variable is present on any of the input arcs
   * @param variable	variable to look for
   * @return		true if it is an input variable
   */
  bool isInput (const class VariableDefinition* variable) const;

  /** Determine whether a variable is present on any of the output arcs
   * @param variable	variable to look for
   * @return		true if it is an output variable
   */
  bool isOutput (const class VariableDefinition* variable) const;

  /** Determine whether a variable is present on a gate expression
   * @param variable	variable to look for
   * @return		true if the variable is used in a gate expression
   */
  bool isGate (const class VariableDefinition* variable) const;
# endif // EXPR_CACHE

  /** Determine whether a variable can ever be bound from an input arc
   * @param variable	variable to look for
   * @return		true if the variable can be bound from an input arc
   */
  bool isLvalue (const class VariableDefinition* variable) const;

  // Iterators to the variable map
  const_iterator begin () const { return myVariables.begin (); }
  const_iterator end () const { return myVariables.end (); }
  iterator begin () { return myVariables.begin (); }
  iterator end () { return myVariables.end (); }
  size_t size () const { return myVariables.size (); }

  /** Add an output variable
   * @param type	Type of the variable
   * @param name	Name of the variable (may be NULL)
   * @param expr	Condition expression for the quantification
   * @return		pointer to the VariableDefinition object
   */
  const class VariableDefinition* addOutputVariable (const class Type* type,
						     char* name,
						     class Expression* expr);

  /** Get the quantifier list for the output variables of the transition */
  const class QuantifierList* getOutputVariables () const {
    return myOutputVariables;
  }

  /** Determine whether a variable is an output variable
   * @param variable	variable to look for
   * @return		true if the variable is an output variable
   */
  bool isOutputVariable (const class VariableDefinition* variable) const;

  /** Add a variable
   * @param name	Name of the variable
   * @param type	Type of the variable
   * @return		pointer to the VariableDefinition object
   */
  const class VariableDefinition* addVariable (char* name,
					       const class Type* type);

  /** Get a function definition by name
   * @param name	Name of the function
   * @return		The function definition
   */
  class Function* getFunction (const char* name) {
    FunctionMap::const_iterator i = myFunctions.find (name);
    return i == myFunctions.end () ? NULL : i->second;
  }

  /** Add a function definition
   * @param function	The function
   */
  void addFunction (class Function& function);

  // Other methods

  /** Check the gates
   * @param valuation	Valuation for evaluating the gate expressions
   * @param excuse	placeholder for returning an expression that
   *			cannot be evaluated because of unbound variables,
   *			or NULL if all encountered gates were evaluated
   * @return		true if the gates allow the transition to be fired
   */
  bool checkGates (const class Valuation& valuation,
		   const class Expression*& excuse) const;

  /** Determine whether all input arcs and gates are defined
   * @param valuation	Valuation for evaluating the expressions
   * @return		true if everything is defined (except the outputs)
   */
  bool isDefined (const class Valuation& valuation) const;

  /** Add an arc
   * @param arc		The arc to be added
   */
  void addArc (class Arc* arc);
  /** Determine whether the transition has an arc from/to a place
   * @param output	specifies the type of arcs to look for
   * @param place	identifies the place
   * @return		the arc, or NULL
   */
  class Arc* getArc (bool output, const class Place* place);
  /** Determine whether the transition has input arcs */
  bool hasInputs () const { return myNumInputs != 0; }
  /** Determine whether the transition has output arcs */
  bool hasOutputs () const { return myNumOutputs != 0; }

  /** Get next formal tokens that can be assigned to
   * @param tokens	container for the tokens to be retrieved
   * @param valuation	variable substitutions
   * @param m		GlobalMarking containing the actual token(s)
   * @return		an expression that cannot be evaluated
   *			because of unbound variables, or NULL if all
   *			formal tokens could be retrieved
   */
  const class Expression* getNextTokens (class TokenList& tokens,
					 class Valuation& valuation,
					 class GlobalMarking& m) const;

  /** Fire the transition
   * @param marking	the state (input/output parameter)
   * @param valuation	the transition instance
   * @return		true if the transition was fired (no errors occurred)
   */
  bool fire (class GlobalMarking& marking,
	     const class Valuation& valuation) const;

# ifdef DEBUG
  /**
   * Prints this Transition.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** Index number of the transition */
  unsigned myIndex;
  /** Name of the transition */
  char* myName;
  /** Number of input arcs */
  unsigned myNumInputs;
  /** Input arcs */
  class Arc** myInputs;
  /** Number of output arcs */
  unsigned myNumOutputs;
  /** Output arcs */
  class Arc** myOutputs;

  /** Gate expressions */
  GateList myGates;

  /** Output variables */
  class QuantifierList* myOutputVariables;

  /** Transition variables indexed by name */
  VariableMap myVariables;

  /** Functions indexed by name */
  FunctionMap myFunctions;
};

#endif // TRANSITION_H_
