// Variable definition class -*- c++ -*-

#include "VariableDefinition.h"
#include "Marking.h"
#include "VectorType.h"
#include "VectorValue.h"
#include "StructType.h"
#include "StructValue.h"
#include "UnionType.h"
#include "UnionValue.h"

#include <cassert>

VariableDefinition::VariableDefinition () :
  myName (0), myType (0), myIsAggregate (false),
#ifdef EXPR_CACHE
  mySet (),
#endif // EXPR_CACHE
  myChildren (), myFather (0), myIndex (0)
{
}

VariableDefinition::VariableDefinition (char* name, const class Type* type,
					bool aggregate) :
  myName (name), myType (type), myIsAggregate (aggregate),
#ifdef EXPR_CACHE
  mySet (),
#endif // EXPR_CACHE
  myChildren (), myFather (0), myIndex (0)
{
  assert (myName && myType);
}

VariableDefinition::VariableDefinition (const class VariableDefinition* father,
					card_t index) :
  myName (0), myType (0), myIsAggregate (false),
#ifdef EXPR_CACHE
  mySet (),
#endif // EXPR_CACHE
  myChildren (), myFather (father), myIndex (index)
{
  assert (myFather && myFather->getType ());
  assert (!myFather->isAggregate ());

  switch (myFather->getType ()->getKind ()) {
  case Type::tVector:
    {
      const class VectorType* const type =
	static_cast<const class VectorType*>(myFather->getType ());

      assert (myIndex < type->getSize ());
      myType = type->getItemType ();
    }
    break;
  case Type::tStruct:
    {
      const class StructType* const type =
	static_cast<const class StructType*>(myFather->getType ());

      assert (myIndex < type->getSize ());
      myType = (*type)[myIndex];
    }
    break;
  case Type::tUnion:
    {
      const class UnionType* const type =
	static_cast<const class UnionType*>(myFather->getType ());

      assert (myIndex < type->getSize ());
      myType = (*type)[myIndex];
    }
    break;
  default:
    assert (false);
  }
}

VariableDefinition::~VariableDefinition ()
{
  delete[] myName;
  if (myFather) {
    List::iterator j = myFather->myChildren.end ();

    for (List::iterator i = myFather->myChildren.begin ();
	 i != myFather->myChildren.end () && *i != this; j = i++);
    if (j == myFather->myChildren.end ()) {
      assert (*myFather->myChildren.begin () == this);
      myFather->myChildren.erase (myFather->myChildren.begin ());
    }
    else {
#ifndef NDEBUG
      List::iterator k = j;
      assert (*++k == this);
#endif // NDEBUG
      myFather->myChildren.erase_after (j);
    }
  }
}

const class VariableDefinition*
VariableDefinition::getChild (card_t index) const
{
  const class VariableDefinition* v;

  List::iterator j = myChildren.end ();
  for (List::iterator i = myChildren.begin ();
       i != myChildren.end (); j = i++) {
    v = *i;
    assert (v->getFather () == this);
    if (v->getIndex () == index)
      return v;
    else if (v->getIndex () > index)
      break;
  }

  v = new class VariableDefinition (this, index);

  if (j == myChildren.end ())
    myChildren.insert (myChildren.begin (), v);
  else
    myChildren.insert_after (j, v);

  return v;
}

bool
VariableDefinition::isCompatible (const class Valuation& valuation,
				  const class Value& value) const
{
  assert (!myChildren.empty ());
  assert (value.getType () == getType ());
  switch (getType ()->getKind ()) {
  case Type::tVector:
    {
      const class VectorValue& vv =
	static_cast<const class VectorValue&>(value);
      for (List::const_iterator i = myChildren.begin ();
	   i != myChildren.end (); i++) {
	const class VariableDefinition* const v = *i;
	assert (v->getFather () == this);
	const class Value* const c = valuation.getValue (v);
	const class Value* const vc = vv[v->getIndex ()];
	if ((c || vc) && (!c || !vc || !(*c == *vc)))
	  return false;
      }
    }
    break;
  case Type::tStruct:
    break;
  case Type::tUnion:
    break;
  default:
    assert (false);
  }

  return true;
}

class Value*
VariableDefinition::getValue (const class Valuation& valuation) const
{
  assert (!myChildren.empty ());

  switch (getType ()->getKind ()) {
  case Type::tVector:
    {
      const class VectorType* const type =
	static_cast<const class VectorType*>(getType ());
      if (!type->isOptional () && myChildren.size () != type->getSize ())
	break;

      class VectorValue* const vv = new class VectorValue (getType ());
      for (List::const_iterator i = myChildren.begin ();
	   i != myChildren.end (); i++) {
	const class VariableDefinition* const v = *i;
	assert (v->getFather () == this);
	const class Value* const c = valuation.getValue (v);
	(*vv)[v->getIndex ()] = c ? c->copy () : NULL;
      }
      if (getType ()->isConstrained (vv))
	return vv;
      else
	delete vv;
    }
    break;
  case Type::tStruct:
    {
      const class StructType* const type =
	static_cast<const class StructType*>(getType ());
      if (myChildren.size () != type->getSize ())
	break;

      class StructValue* const sv = new class StructValue (getType ());
      for (List::const_iterator i = myChildren.begin ();
	   i != myChildren.end (); i++) {
	const class VariableDefinition* const v = *i;
	assert (v->getFather () == this);
	const class Value* const c = valuation.getValue (v);
	assert (c && !(*sv)[v->getIndex ()]);
	(*sv)[v->getIndex ()] = c->copy ();
      }
      if (getType ()->isConstrained (sv))
	return sv;
      else
	delete sv;
    }
    break;
  case Type::tUnion:
    {
      if (myChildren.empty ())
	break;
      assert (myChildren.size () == 1);

      const class VariableDefinition* const v = *myChildren.begin ();
      const class Value* const c = valuation.getValue (v);
      assert (c != NULL);

      class UnionValue* const uv =
	new class UnionValue (getType (), v->getIndex (), c->copy ());
      if (getType ()->isConstrained (uv))
	return uv;
      else
	delete uv;
    }
    break;
  default:
    assert (false);
  }

  return NULL;
}

const class Value*
VariableDefinition::getChildValue (const class Valuation& valuation,
				   card_t index) const
{
  for (List::const_iterator i = myChildren.begin ();
       i != myChildren.end (); i++) {
    const class VariableDefinition* v = *i;
    assert (v->getFather () == this);
    if (v->getIndex () == index)
      return valuation.getValue (v);
    else if (v->getIndex () > index)
      break;
  }
  return NULL;
}

#ifdef EXPR_CACHE
void
VariableDefinition::invalidate () const
{
  for (const_iterator i = mySet.begin (); i != mySet.end (); i++)
    (*i)->invalidate (this);
  for (MSet::const_iterator j = myMSet.begin (); j != myMSet.end (); j++)
    (*j)->setProcessed (Marking::unprocessed);
}

void
VariableDefinition::clearValuation (const class Valuation* valuation) const
{
  for (const_iterator i = mySet.begin (); i != mySet.end (); i++)
    (*i)->clearValuation (valuation);
}
#endif // EXPR_CACHE

void
VariableDefinition::invalidateMarking () const
{
  for (MSet::const_iterator j = myMSet.begin (); j != myMSet.end (); j++)
    (*j)->setProcessed (Marking::unprocessed);
}

#ifdef DEBUG
# include "util.h"
# include "Type.h"

void
VariableDefinition::prettyPrint (unsigned indent) const
{
  assert (myType != NULL);
  if (myFather) {
    const char* const fathername = myFather->myType->getName ()
      ? myFather->myType->getName ()
      : myFather->myType->getSyntacticName ();

    switch (myFather->getType ()->getKind ()) {
    case Type::tVector:
      {
	const class VectorType* const type =
	  static_cast<const class VectorType*>(myFather->getType ());
	print (indent, "%s %s:\n", fathername, myFather->getName ());
	class Value* value = type->getIndexType ()->convert (myIndex);
	value->prettyPrint (indent + 1);
	print (indent + 1, "->\n");
	delete value;
      }
      break;
    case Type::tStruct:
      {
	const class StructType* const type =
	  static_cast<const class StructType*>(myFather->getType ());
	print (indent, "%s %s[%s]\n", fathername, myFather->getName (),
	       type->getComponentName (myIndex));
      }
      break;
    case Type::tUnion:
      {
	const class UnionType* const type =
	  static_cast<const class UnionType*>(myFather->getType ());
	print (indent, "%s %s[%s]\n", fathername, myFather->getName (),
	       type->getComponentName (myIndex));
      }
      break;
    default:
      assert (false);
    }
  }
  else
    print (indent, "%s %s\n",
	   myType->getName ()
	   ? myType->getName ()
	   : myType->getSyntacticName (),
	   getName ());
}

#endif // DEBUG
