// Variable definition class -*- c++ -*-

#ifndef VARIABLEDEFINITION_H_
# define VARIABLEDEFINITION_H_

# include <set>
# include <slist>
# include "typedefs.h"

class VariableDefinition {
public:
  // Typedefs
# ifdef EXPR_CACHE
  typedef std::set<const class Expression*> Set;
  typedef Set::iterator iterator;
  typedef Set::const_iterator const_iterator;
# endif // EXPR_CACHE
  typedef std::set<const class Marking*> MSet;
  typedef std::slist<const class VariableDefinition*> List;

public:
  // Constructors and destructors

  /** Constructs a VariableDefinition */
  VariableDefinition ();

  /** Constructs a VariableDefinition
   * @param name	Name of the variable
   * @param type	Type of the variable
   * @param aggregate	Flag: is this an aggregate variable (quantifier)
   */
  VariableDefinition (char* name, const class Type* type,
		      bool aggregate = false);

private:
  /** Constructs a VariableDefinition as a part of another variable
   * @param variable	father variable
   * @param index	identifies the component of the father variable
   */
  VariableDefinition (const class VariableDefinition* variable,
		      card_t index);
  /** Copy constructor */
  VariableDefinition (const class VariableDefinition& old);
  /** Assignment operator */
  class VariableDefinition& operator= (const class VariableDefinition& old);
public:
  /** Destructs a VariableDefinition */
  ~VariableDefinition ();

  /** Get a child variable corresponding to a component of the variable
   * @param index	index of the variable component
   */
  const class VariableDefinition* getChild (card_t index) const;

  // Accessors

  /** Determine the name of the variable */
  const char* getName () const { return myName; }
  /** Determine the type of the variable */
  const class Type* getType () const { return myType; }
  /** Determine whether this is an aggregate variable */
  bool isAggregate () const { return myIsAggregate; }
  /** Determine the father variable */
  const class VariableDefinition* getFather () const { return myFather; }
  /** Determine the index of the father variable */
  card_t getIndex () const { return myIndex; }

  /** Determine whether there are child variables */
  bool hasChildren () const { return !myChildren.empty (); }

  /** Determine whether the children of the variable match the value
   * @param valuation	variable definitions for the children
   * @param value	value to be checked
   * @return		true if the value matches
   */
  bool isCompatible (const class Valuation& valuation,
		     const class Value& value) const;

  /** Determine the value of a father variable using the child variables
   * @param valuation	variable definitions for the child variables
   * @return		the value, or NULL if impossible to calculate
   */
  class Value* getValue (const class Valuation& valuation) const;

  /** Determine the value of a child variable
   * @param valuation	variable definitions for the child variables
   * @param index	index of the variable component
   */
  const class Value* getChildValue (const class Valuation& valuation,
				    card_t index) const;

  // Other methods
# ifdef EXPR_CACHE
  /** Associate the variable with an expression
   * so that whenever the value of the variable changes,
   * the expression will be invalidated
   * @param expression	an expression the variable is to be associated with
   */
  void associate (const class Expression& expression) const {
    mySet.insert (&expression);
  }

  /** Disassociate the variable with a previously associated expression
   * @param expression	expression to be disassociated with
   */
  void disassociate (const class Expression& expression) const {
    iterator i = mySet.find (&expression);
    if (i != mySet.end ())
      mySet.erase (i);
  }

  /** Determine whether the variable is associated with an expression
   * @param expression	expression to search for
   * @return		true if the variable is associated with the expression
   */
  bool isAssociated (const class Expression& expression) const {
    return mySet.find (&expression) != mySet.end ();
  }

  /** Invalidate all the expressions the variable is associated with */
  void invalidate () const;
  /** Clear the valuation of all expressions the variable is associated with
   * @param valuation	valuation to be cleared
   */
  void clearValuation (const class Valuation* valuation) const;
# endif // EXPR_CACHE

  /** Invalidate all the marking expressions the variable is associated with */
  void invalidateMarking () const;

  /** Associate the variable with a marking expression whose
   * quantification condition depends on the variable
   * @param marking	marking expression to be associated with the variable
   */
  void associateMarking (const class Marking& marking) const {
    myMSet.insert (&marking);
  }

  /** Disassociate the variable with a previously associated marking expression
   * @param marking	marking expression to be disassociated with
   */
  void disassociateMarking (const class Marking& marking) const {
    MSet::iterator i = myMSet.find (&marking);
    if (i != myMSet.end ())
      myMSet.erase (i);
  }

# ifdef DEBUG
  /**
   * Prints this object.
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** Name of the variable */
  char* myName;
  /** Type of the variable */
  const class Type* myType;
  /** Flag: is this an aggregate variable? */
  const bool myIsAggregate;
# ifdef EXPR_CACHE
  /** The expressions the variable is associated with */
  mutable Set mySet;
# endif // EXPR_CACHE
  /** The marking expressions the variable is associated with */
  mutable MSet myMSet;
  /** Child variables */
  mutable List myChildren;
  /** Father variable */
  const class VariableDefinition* myFather;
  /** Index of the father variable */
  const card_t myIndex;
};

#endif // VARIABLEDEFINITION_H_
