// Buffer type class -*- c++ -*-

#include "BufferType.h"
#include "BufferValue.h"
#include "Constraint.h"
#include "Range.h"

#include <cstring>

BufferType::BufferType (const class Type* itemType, card_t size,
			bool stack) :
  Type (),
  myItemType (itemType), mySize (size), myStack (stack)
{
  assert (myItemType && size);
}

BufferType::BufferType (const BufferType& old) :
  Type (),
  myItemType (old.myItemType), mySize (old.mySize),
  myStack (old.myStack)
{
}

BufferType::~BufferType ()
{
}

class Value*
BufferType::getFirstValue () const
{
  const class Value* v = myConstraint ? myConstraint->getFirstValue () : NULL;
  return v ? v->copy () : new class BufferValue (this);
}

class Value*
BufferType::getLastValue () const
{
  if (const class Value* v = myConstraint ? myConstraint->getLastValue () : 0)
    return v->copy ();
  else {
    class Value* const w = new class BufferValue (this);
    w->top ();
    return w;
  }
}

bool
BufferType::isAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return Type::isAssignable (type);

  const class BufferType* v = static_cast<const class BufferType*>(type);

  return
    myItemType->isAssignable (v->myItemType);
}

bool
BufferType::isAlwaysAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return false;

  const class BufferType* v = static_cast<const class BufferType*>(type);

  return
    mySize == v->mySize &&
    myItemType->isAlwaysAssignable (v->myItemType);
}

bool
BufferType::isConstrained (const class Value* value) const
{
  assert (value && value->getType ());
  assert (value->getType ()->isAssignable (this));
  if (value->getKind () != Value::vBuffer)
    return false;

  const class BufferValue* b = static_cast<const class BufferValue*>(value);

  for (card_t i = 0; i < getSize () && (*b)[i]; i++)
    if (!myItemType->isConstrained ((*b)[i]))
      return false;

  return Type::isConstrained (value);
}

card_t
BufferType::do_getNumValues () const
{
  card_t numValues;

  if (myConstraint) {
    numValues = 0;
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      assert (!low || !high || *low <= *high);
      card_t num = 0;
      if (low && high)
	num += *high - *low;
      else if (high) {
	class Value* v = getFirstValue ();
	num += *high - *v;
	delete v;
      }
      else {
	assert (low != NULL);
	class Value* v = low->copy ();
	while (v->increment ())
	  num++;
	delete v;
      }

      if (num != CARD_T_MAX && ++num < CARD_T_MAX - numValues)
	numValues += num;
      else
	return CARD_T_MAX;
    }
  }
  else {
    numValues = 1;
    card_t numItemValues = myItemType->getNumValues ();

    if (numItemValues == CARD_T_MAX)
      return CARD_T_MAX;

    for (card_t i = 1; i <= mySize; i++) {
      card_t product = 1;
      for (card_t j = 0; j < i; j++) {
	if (numItemValues < CARD_T_MAX / product)
	  product *= numItemValues;
	else
	  return CARD_T_MAX;
      }
      if (product < CARD_T_MAX - numValues)
	numValues += product;
      else
	return CARD_T_MAX;
    }
  }

  return numValues;
}

card_t
BufferType::convert (const class Value* value) const
{
  assert (value && value->getKind () == Value::vBuffer);
  assert (isConstrained (value));
  assert (getNumValues () < CARD_T_MAX);

  card_t number = 0;

  if (myConstraint) {
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      class Value* v = NULL;
      if (!low)
	low = v = getFirstValue ();

      assert (!high || *low <= *high);
      if (!high || *value <= *high) {
	number += *value - *low;
	delete v;
	break;
      }
      else {
	number += *high - *low + 1;
	delete v;
      }
    }
  }
  else {
    const class BufferValue* v = static_cast<const class BufferValue*>(value);

    card_t numItemValues = myItemType->getNumValues ();
    card_t offset = 0, i = 0;
    for (card_t mult = 1; i < mySize && (*v)[i]; mult *= numItemValues, i++)
      offset += mult;

    while (i--)
      (number *= numItemValues) += myItemType->convert ((*v)[i]);

    number += offset;
  }

  assert (number < getNumValues ());
  return number;
}

class Value*
BufferType::convert (card_t number) const
{
  assert (number < getNumValues ());
  assert (getNumValues () < CARD_T_MAX);

  class BufferValue* value = NULL;
  if (myConstraint) {
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      class Value* v = NULL;
      if (!low)
	low = v = getFirstValue ();

      assert (!high || *low <= *high);
      if (high) {
	card_t diff = *high - *low;
	if (number <= diff)
	  goto found;
	else
	  number -= diff;
      }
      else {
      found:
	value = static_cast<class BufferValue*>(low->copy ());
	delete v;
	while (number--)
	  if (!value->increment ())
	    assert (false);
	break;
      }

      delete v;
    }
  }
  else {
    value = new class BufferValue (this);
    if (!number)
      return value;

    card_t numItemValues = myItemType->getNumValues ();
    card_t size = 0;
    for (card_t mult = 1; number >= mult; size++, mult *= numItemValues)
      number -= mult;

    for (card_t i = 0; i < size; i++) {
      card_t num = number % numItemValues;
      number /= numItemValues;

      (*value)[i] = myItemType->convert (num);
    }
  }

  assert (isConstrained (value));
  return value;
}

#ifdef DEBUG
# include "util.h"

void
BufferType::prettyPrint (unsigned indent) const
{
  const char* const type = myStack ? "stack" : "buffer";

  if (myItemType->getName ())
    print (indent, "%s[%s %u]\n", type, myItemType->getName (), mySize);
  else {
    myItemType->prettyPrint (indent);
    print (indent, "[%s %u]\n", type, mySize);
  }
  if (getConstraint ())
    getConstraint ()->prettyPrint (indent);
}

#endif // DEBUG
