// Buffer (queue or stack) type class -*- c++ -*-

#ifndef BUFFERTYPE_H_
# define BUFFERTYPE_H_

# include "Type.h"

/** Buffer (variable-length array with queue or stack discipline) */
class BufferType : public Type {
public:
  // Constructors and destructors

  /** Constructor
   * @param itemType	The buffer item type
   * @param size	The buffer size
   * @param stack	Flag: true=stack, false=queue
   */
  BufferType (const class Type* itemType, card_t size, bool stack);
private:
  /** Copy constructor */
  BufferType (const class BufferType& old);
  /** Assignment operator */
  class BufferType& operator= (const class BufferType& old);
public:
  /** Destructs BufferType */
  ~BufferType ();
  /** Copy the Type */
  class Type* copy () const { return new class BufferType (*this); }

  // Accessors

  /** Determine the kind of the type */
  enum Kind getKind () const { return tBuffer; }

  /** Get the item type */
  const class Type* getItemType () const { return myItemType; }
  /** Get the size */
  card_t getSize () const { return mySize; }
  /** Determine whether this is a stack */
  bool isStack () const { return myStack; }

  /** Determine whether the type is ordered */
  bool isOrdered () const { return myItemType->isOrdered (); }
  /** Get the first value of this type */
  class Value* getFirstValue () const;
  /** Get the last value of this type */
  class Value* getLastValue () const;

  // Other methods

  /** Convert a value of this type to a number
   * @param value	value to be converted
   * @return		number between 0 and getNumValues () - 1
   */
  card_t convert (const class Value* value) const;

  /** Convert a number to a value of this type
   * @param number	number between 0 and getNumValues () - 1
   * @return		the corresponding value
   */
  class Value* convert (card_t number) const;

  /** See if the type is assignable to another type
   * @param type	the type this type should be assignable to
   * @return		true if this type is assignable to the type
   */
  bool isAssignable (const class Type* type) const;

  /** See if the type is always assignable to another type
   * @param type	the type this type should always be assignable to
   * @return		true if this type always is assignable to the type
   */
  bool isAlwaysAssignable (const class Type* type) const;

  /** Determine whether a value is compatible with the constraints of this type
   * @param value	value to check
   * @return		true if the value passes the constraint check
   */
  bool isConstrained (const class Value* value) const;

  /** Get the number of possible values for this type */
  card_t do_getNumValues () const;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The item type */
  const class Type* myItemType;
  /** The size */
  card_t mySize;
  /** Flag: true=stack, false=queue */
  bool myStack;
};

#endif // BUFFERTYPE_H_
