// Type constraint class -*- c++ -*-

#include "Constraint.h"
#include "Range.h"
#include "Value.h"
#ifndef NDEBUG
# include "Type.h"
#endif // NDEBUG

Constraint::Constraint () :
  myList ()
{
}

Constraint::Constraint (const class Constraint& old, const class Type* type) :
  myList ()
{
  assert (type && (!type->getConstraint () || this == type->getConstraint ()));

  for (const_iterator i = old.begin (); i != old.end (); i++) {
    class Range* range = new class Range (**i);
    myList.push_back (range);
    range->setType (type);
  }
}

Constraint::~Constraint ()
{
  for (iterator i = begin (); i != end (); i++)
    delete *i;
}

void
Constraint::setType (const class Type* type)
{
  assert (type != NULL);
  for (iterator i = begin (); i != end (); i++)
    (*i)->setType (type);
}

const class Value*
Constraint::getFirstValue () const
{
  const_iterator i = begin ();
  return i == end () ? NULL : (*i)->getLow ();
}

const class Value*
Constraint::getLastValue () const
{
  const_iterator i = end ();
  return --i == end () ? NULL : (*i)->getHigh ();
}

const class Value*
Constraint::getNextHigh (const class Value* value) const
{
  for (const_iterator i = begin (); i != end (); i++)
    if ((*i)->check (value) == Range::pass)
      return (*i)->getHigh ();

  assert (false);
  return NULL;
}

const class Value*
Constraint::getNextLow (const class Value* value) const
{
  for (const_iterator i = begin (); i != end (); i++)
    if (*(*i)->getHigh () == *value)
      return ++i == end () ? NULL : (*i)->getLow ();

  assert (false);
  return NULL;
}

const class Value*
Constraint::getPrevLow (const class Value* value) const
{
  for (const_iterator i = begin (); i != end (); i++)
    if ((*i)->check (value) == Range::pass)
      return (*i)->getLow ();

  assert (false);
  return NULL;
}

const class Value*
Constraint::getPrevHigh (const class Value* value) const
{
  for (const_iterator i = begin (), j = end (); i != end (); j = i++)
    if (*(*i)->getLow () == *value)
      return i == begin () ? NULL : (*j)->getHigh ();

  assert (false);
  return NULL;
}

void
Constraint::append (class Range* range)
{
  assert (range != NULL);
  iterator i;
  for (i = begin (); i != end (); ) {
    if (class Range* r = (*i)->combine (range)) {
      delete *i, delete range;
      range = r;

      if (i == begin ()) {
	myList.erase (i);
	if ((i = begin ()) == end ())
	  break;
	else
	  continue;
      }
      else {
	iterator j = i;
	j--; myList.erase (i);
	i = j;
      }
    }
    else if (range->getHigh () && (!(*i)->getHigh () ||
				   *range->getHigh () <= *(*i)->getHigh ()))
      break;

    i++;
  }

  myList.insert (i, range);
}

void
Constraint::restrict (const class Constraint* constraint,
		      const class Type* type)
{
  assert (type && this == type->getConstraint ());
  assert (constraint && !isDisjoint (constraint));
  List l = myList;
  myList.clear ();

  for (iterator j = l.begin (); j != l.end (); j++) {
    for (const_iterator k = constraint->begin ();
	 k != constraint->end (); k++) {
      if (class Range* range = (*j)->cut (*k)) {
	range->setType (type);
	myList.push_back (range);
      }
    }

    delete *j;
  }
}

bool
Constraint::isSuperset (const class Constraint* other) const
{
  assert (other != NULL);

  // This is an ugly O(n^2) algorithm,
  // but it's only used during static analysis.

  // The algorithm does not consider (1..3,4,5) to be a superset of (2..4).
  // Neighbouring ranges will have been combined by Constraint::append ().
  for (const_iterator i = other->begin (); i != other->end (); i++) {
    const class Value* low = (*i)->getLow ();
    const class Value* high = (*i)->getHigh ();

    if (low && high) {
      for (const_iterator j = begin (); j != end (); j++)
	if (*(*j)->getLow () <= *low && *high <= *(*j)->getHigh ())
	  goto match;
    }
    else if (low) {
      for (const_iterator j = begin (); j != end (); j++)
	if (!(*j)->getHigh () && *(*j)->getLow () <= *low)
	  goto match;
    }
    else {
      assert (high != NULL);
      for (const_iterator j = begin (); j != end (); j++)
	if (!(*j)->getLow () && *high <= *(*j)->getHigh ())
	  goto match;
    }

    return false;
  match:
    continue;
  }

  return true;
}

bool
Constraint::isDisjoint (const class Constraint* other) const
{
  assert (other != NULL);

  // This is an ugly O(n^2) algorithm,
  // but it's only used during static analysis.
  for (const_iterator i = other->begin (); i != other->end (); i++) {
    const class Value* low = (*i)->getLow ();
    const class Value* high = (*i)->getHigh ();

    if (low && high) {
      for (const_iterator j = begin (); j != end (); j++)
	if ((*j)->check (low) == Range::pass ||
	    (*j)->check (high) == Range::pass)
	  return false;
    }
    else if (low) {
      for (const_iterator j = begin (); j != end (); j++)
	if (!(*j)->getHigh () || *low <= *(*j)->getHigh ())
	  return false;
    }
    else {
      assert (high != NULL);
      for (const_iterator j = begin (); j != end (); j++)
	if (!(*j)->getLow () || *(*j)->getLow () <= *high)
	  return false;
    }
  }

  return true;
}

bool
Constraint::isConstrained (const class Value* value) const
{
  assert (value != NULL);

  for (const_iterator i = begin (); i != end (); i++) {
    switch ((*i)->check (value)) {
    case Range::undefined:
      assert (false);
    case Range::fail:
      break;
    case Range::pass:
      return true;
    }
  }

  return false;
}

#ifdef DEBUG
# include "util.h"

void
Constraint::prettyPrint (unsigned indent) const
{
  print (indent, "(\n");

  const_iterator i = begin ();

  for (;;) {
    if (*i)
      (*i)->prettyPrint (indent + 1);

    if (++i == end ())
      break;

    print (indent + 1, ",\n");
  }

  print (indent, ")\n");
}
#endif // DEBUG
