// Type constraint class -*- c++ -*-

#ifndef CONSTRAINT_H_
# define CONSTRAINT_H_

# include <list>

/** Value constraint */
class Constraint {
public:
  // Typedefs
  typedef std::list<class Range*> List;
  typedef List::iterator iterator;
  typedef List::const_iterator const_iterator;

  // Constructors and destructors

  /** Constructs a new Constraint */
  Constraint ();

  /** Copy constructor
   * @param old		constraint to be copied
   * @param type	type of the new constraint
   */
  Constraint (const class Constraint& old, const class Type* type);
private:
  /** Copy constructor */
  Constraint (const class Constraint& old);
  /** Assignment operator */
  class Constraint& operator= (const class Constraint& old);
public:
  /** Destructs Constraint */
  ~Constraint ();

  // Accessors

  /** Set the type of the constraint */
  void setType (const class Type* type);

  /** Get the first value of the constraint
   * @return		the smallest value, or NULL if there is no lower bound
   */
  const class Value* getFirstValue () const;
  /** Get the last value of the constraint
   * @return		the largest value, or NULL if there is no upper bound
   */
  const class Value* getLastValue () const;
  /** Get the high bound of the value
   * @param value	value whose high bound is to be sought
   * @return		the high bound, or NULL if it is unbounded
   */
  const class Value* getNextHigh (const class Value* value) const;
  /** Get the smallest high bound greater than the value
   * @param value	value whose successor is to be sought
   * @return		the successor of the value, or NULL
   */
  const class Value* getNextLow (const class Value* value) const;
  /** Get the low bound of the value
   * @param value	value whose high bound is to be sought
   * @return		the high bound, or NULL if it is unbounded
   */
  const class Value* getPrevLow (const class Value* value) const;
  /** Get the biggest high bound smaller than the value
   * @param value	value whose predecessor is to be sought
   * @return		the predecessor of the value, or NULL
   */
  const class Value* getPrevHigh (const class Value* value) const;

  bool empty () const { return myList.empty (); }
  size_t size () const { return myList.size (); }
  const_iterator begin () const { return myList.begin (); }
  const_iterator end () const { return myList.end (); }
  iterator begin () { return myList.begin (); }
  iterator end () { return myList.end (); }

  // Other methods

  /** Append a range to the constraint
   * @param range	Range to be appended
   */
  void append (class Range* range);

  /** Restrict the constraint by another constraint
   * @param constraint	Constraint to be restricted by
   * @param type	Type of this constraint
   */
  void restrict (const class Constraint* constraint, const class Type* type);

  /** Determine whether the constraint encloses another constraint
   * @param other	the other constraint to be investigated
   * @return		true if all values accepted by the other constraint
   *			are accepted by this constraint
   */
  bool isSuperset (const class Constraint* other) const;

  /** Determine whether the two constraints are disjoint
   * @param other	the other constraint
   * @return		true if the constraints are disjoint
   */
  bool isDisjoint (const class Constraint* other) const;

  /** Determine whether the constraint accepts a value
   * @param value	value to be checked
   * @return		true if the value passes the check
   */
  bool isConstrained (const class Value* value) const;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members
private:
  /** List of ranges belonging to the constraint */
  List myList;
};

#endif // CONSTRAINT_H_
