// Enumerated type class -*- c++ -*-

#include "EnumType.h"
#include "EnumValue.h"
#include "util.h"
#include "Constraint.h"
#include "Range.h"

EnumType::EnumType () :
  Type (),
  myEnums (),
  myReverseMap (),
  myNextValue (0)
{
}

EnumType::EnumType (const EnumType& old) :
  Type (),
  myEnums (),
  myReverseMap (),
  myNextValue (old.myNextValue)
{
  if (old.getConstraint ())
    setConstraint (new class Constraint (*old.getConstraint (), this));

  for (EnumMap::const_iterator i = old.myEnums.begin ();
       i != old.myEnums.end (); i++) {
    char* comp = newString (i->first);
    myEnums[comp] = i->second;
    myReverseMap[i->second] = comp;
  }
}

EnumType::~EnumType ()
{
  for (EnumMap::iterator i = myEnums.begin (); i != myEnums.end (); i++)
    delete i->first;
}

void
EnumType::addEnumeration (char* name, int_t value)
{
  assert (!getValue (name));
  assert (!hasValue (value));

  myEnums[name] = value;
  myReverseMap[value] = name;
  assert (myEnums.size () == myReverseMap.size ());

  myNextValue = value + 1;
}

const int_t*
EnumType::getValue (const char* name) const
{
  EnumMap::const_iterator i = myEnums.find (const_cast<char*>(name));
  return i == myEnums.end () ? NULL : &i->second;
}

class Value*
EnumType::getFirstValue () const
{
  const class Value* v = myConstraint ? myConstraint->getFirstValue () : NULL;
  assert (!v || v->getType () == this);
  return v ? v->copy () : new class EnumValue (this, getFirstEnum ());
}

class Value*
EnumType::getLastValue () const
{
  const class Value* v = myConstraint ? myConstraint->getLastValue () : NULL;
  assert (!v || v->getType () == this);
  return v ? v->copy () : new class EnumValue (this, getLastEnum ());
}

int_t
EnumType::getFirstEnum () const
{
  assert (!myReverseMap.empty ());
  return myReverseMap.begin ()->first;
}

int_t
EnumType::getLastEnum () const
{
  assert (!myReverseMap.empty ());
  return (--myReverseMap.end ())->first;
}

bool
EnumType::getNextEnum (int_t& value) const
{
  assert (hasValue (value));

  ReverseMap::const_iterator i = myReverseMap.find (value);
  assert (i != myReverseMap.end ());

  if (++i == myReverseMap.end ()) {
    value = getFirstEnum ();
    return false;
  }
  else {
    value = i->first;
    return true;
  }
}

bool
EnumType::getPrevEnum (int_t& value) const
{
  assert (hasValue (value));

  ReverseMap::const_iterator i = myReverseMap.find (value);
  assert (i != myReverseMap.end ());

  if (i == myReverseMap.begin ()) {
    value = getLastEnum ();
    return false;
  }
  else {
    value = (--i)->first;
    return true;
  }
}

bool
EnumType::hasValue (int_t value) const
{
  return myReverseMap.find (value) != myReverseMap.end ();
}

const char*
EnumType::getEnumName (int_t value) const
{
  ReverseMap::const_iterator i = myReverseMap.find (value);
  return (i != myReverseMap.end ()) ? i->second : NULL;
}

bool
EnumType::isConstrained (const class Value* value) const
{
  assert (value && value->getType ());
  assert (value->getType ()->isAssignable (this));

  if (!getConstraint ()) {
    const class EnumValue* v = static_cast<const class EnumValue*>(value);
    return myReverseMap.find (v->getValue ()) != myReverseMap.end ();
  }
  else
    return Type::isConstrained (value);
}

card_t
EnumType::do_getNumValues () const
{
  if (!myConstraint)
    return myEnums.size ();
  card_t numValues = 0;

  for (Constraint::const_iterator i = myConstraint->begin ();
       !(i == myConstraint->end ()); i++) {
    const class Range* r = *i;
    int_t lowLimit = getFirstEnum ();
    int_t highLimit = getLastEnum ();

    if (const class Value* low = r->getLow ())
      lowLimit = static_cast<const class EnumValue*>(low)->getValue ();
    if (const class Value* high = r->getHigh ())
      highLimit = static_cast<const class EnumValue*>(high)->getValue ();
    assert (lowLimit <= highLimit);

    if (int_t diff = highLimit - lowLimit + 1) {
      assert (CARD_T_MAX - numValues >= static_cast<card_t>(diff));
      numValues += static_cast<card_t>(diff);
    }
    else { // overflow
      assert (!numValues);
      return CARD_T_MAX;
    }
  }

  return numValues;
}

card_t
EnumType::convert (const class Value* value) const
{
  assert (value && value->getKind () == Value::vEnum);
  int_t v = static_cast<const class EnumValue*>(value)->getValue ();
  if (!myConstraint) {
    ReverseMap::const_iterator i = myReverseMap.find (v);
    assert (i != myReverseMap.end ());
    card_t number = getNumValues ();
    while (i != myReverseMap.end ()) number--, i++;
    assert (number < getNumValues ());
    return number;
  }
  else {
    assert (isConstrained (value));
    card_t number = 0;

    for (Constraint::const_iterator i = myConstraint->begin ();
	 !(i == myConstraint->end ()); i++) {
      const class Range* r = *i;
      int_t lowLimit = getFirstEnum ();
      int_t highLimit = getLastEnum ();

      if (const class Value* low = r->getLow ())
	lowLimit = static_cast<const class EnumValue*>(low)->getValue ();
      if (const class Value* high = r->getHigh ())
	highLimit = static_cast<const class EnumValue*>(high)->getValue ();
      assert (lowLimit <= highLimit);

      if (v < lowLimit || v > highLimit) {
	if (int_t diff = highLimit - lowLimit + 1)
	  number += diff;
	else {
	  assert (false);
	  return CARD_T_MAX;
	}
      }
      else {
	number += static_cast<int_t>(v) - lowLimit;
	assert (number <= getNumValues ());
	return number;
      }
    }

    assert (false);
    return CARD_T_MAX;
  }
}

class Value*
EnumType::convert (card_t number) const
{
  assert (number < getNumValues ());
  if (!myConstraint) {
    assert (number < myReverseMap.size ());
    ReverseMap::const_iterator i = myReverseMap.begin ();
    while (number--)
      i++;
    assert (i != myReverseMap.end ());
    return new class EnumValue (this, i->first);
  }
  else {
    for (Constraint::const_iterator i = myConstraint->begin ();
	 !(i == myConstraint->end ()); i++) {
      const class Range* r = *i;
      int_t lowLimit = getFirstEnum ();
      int_t highLimit = getLastEnum ();

      if (const class Value* low = r->getLow ())
	lowLimit = static_cast<const class EnumValue*>(low)->getValue ();
      if (const class Value* high = r->getHigh ())
	highLimit = static_cast<const class EnumValue*>(high)->getValue ();
      assert (lowLimit <= highLimit);

      card_t diff = highLimit - lowLimit;
      if (number <= diff)
	return new class EnumValue (this, lowLimit + number);
      else
	number -= diff, number--;
    }

    assert (false);
    return NULL;
  }
}

#ifdef DEBUG

void
EnumType::prettyPrint (unsigned indent) const
{
  print (indent, "enum {\n");

  ReverseMap::const_iterator i = myReverseMap.begin ();

  if (i != myReverseMap.end ())
    for (;;) {
      ReverseMap::const_iterator j = i++;
      if (i == myReverseMap.end ()) {
	print (indent + 1, "%s = %d\n", (*j).second, (*j).first);
	break;
      }
      else
	print (indent + 1, "%s = %d,\n", (*j).second, (*j).first);
    }

  print (indent, "}\n");
  if (getConstraint ())
    getConstraint ()->prettyPrint (indent);
}

#endif // DEBUG
