// Enumerated type class -*- c++ -*-

#ifndef ENUMTYPE_H_
# define ENUMTYPE_H_

# include <map>
# include "Type.h"

/** Boolean type */
class EnumType : public Type {
public:
  // Constructors and destructors

  /** Constructs EnumType */
  EnumType ();
private:
  /** Copy constructor */
  EnumType (const class EnumType& old);
  /** Assignment operator */
  class EnumType& operator= (const class EnumType& old);
public:
  /** Destructs EnumType */
  ~EnumType ();
  /** Copy the Type */
  class Type* copy () const { return new class EnumType (*this); }

  // Accessors

  /** Add an enumeration symbol
   * @param name	name of the symbol
   * @param value	numeric value of the symbol
   */
  void addEnumeration (char* name, int_t value);
  /** Add an enumeration symbol
   * @param name	name of the symbol
   */
  void addEnumeration (char* name) { addEnumeration (name, myNextValue); }

  /** Get the value of an enumeration symbol
   * @param name	name of the symbol
   * @return		pointer to the value, or NULL if not found
   */
  const int_t* getValue (const char* name) const;

  /** Get the first enumeration value */
  int_t getFirstEnum () const;
  /** Get the last enumeration value */
  int_t getLastEnum () const;
  /** Get next enumeration value
   * @param value	value to be incremented
   * @return		false if the value wrapped around, otherwise true
   */
  bool getNextEnum (int_t& value) const;
  /** Get previous enumeration value
   * @param value	value to be decremented
   * @return		false if the value wrapped around, otherwise true
   */
  bool getPrevEnum (int_t& value) const;

  /** Determine whether the given value exists in the enumeration
   * @param value	the enumeration value
   * @return		true if the value exists in the enumeration
   */
  bool hasValue (int_t value) const;

  /** Get the name of an enumeration constant
   * @param value	numeric value of the constant
   * @return		the name, or NULL if not found
   */
  const char* getEnumName (int_t value) const;

  /** Determine the kind of the type */
  enum Kind getKind () const { return tEnum; }

  /** Determine whether the type is ordered */
  bool isOrdered () const { return true; }
  /** Get the first value of this type */
  class Value* getFirstValue () const;
  /** Get the last value of this type */
  class Value* getLastValue () const;

  // Other methods

  /** See if the type is assignable to another type
   * @param type	the type this type should be assignable to
   * @return		true if this type is assignable to the type
   */
  bool isAssignable (const class Type* type) const { return type == this; }

  /** Determine whether a value is compatible with the constraints of this type
   * @param value	value to check
   * @return		true if the value passes the constraint check
   */
  bool isConstrained (const class Value* value) const;

  /** Get the number of possible values for this type */
  card_t do_getNumValues () const;

  /** Convert a value of this type to a number
   * @param value	value to be converted
   * @return		number between 0 and getNumValues () - 1
   */
  card_t convert (const class Value* value) const;

  /** Convert a number to a value of this type
   * @param number	number between 0 and getNumValues () - 1
   * @return		the corresponding value
   */
  class Value* convert (card_t number) const;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  typedef std::map<char*,int_t,struct ltstr> EnumMap;
  /** Enumeration symbols */
  EnumMap myEnums;
  typedef std::map<int_t,const char*> ReverseMap;
  /** Reverse map from constants to symbols */
  ReverseMap myReverseMap;
  /** Next enumeration value */
  int_t myNextValue;
};

#endif // ENUMTYPE_H_
