// Struct type class -*- c++ -*-

#include "StructType.h"
#include "ComponentList.h"
#include "StructValue.h"
#include "Constraint.h"
#include "Range.h"

StructType::StructType (class ComponentList* components) :
  Type (),
  myComponents (*components)
{
  delete components;
}

StructType::StructType (const class StructType& old) :
  Type (),
  myComponents (old.myComponents)
{
}

StructType::~StructType ()
{
}

bool
StructType::isOrdered () const
{
  for (card_t i =  0; i < getSize (); i++)
    if (!(*this)[i]->isOrdered ())
      return false;
  return true;
}

class Value*
StructType::getFirstValue () const
{
  assert (isOrdered ());
  if (const class Value* w = myConstraint ? myConstraint->getFirstValue () : 0)
    return w->copy ();
  else {
    class StructValue* v = new class StructValue (this);
    for (card_t i = 0; i < getSize (); i++)
      (*v)[i] = (*this)[i]->getFirstValue ();
    return v;
  }
}

class Value*
StructType::getLastValue () const
{
  assert (isOrdered ());
  if (const class Value* w = myConstraint ? myConstraint->getLastValue () : 0)
    return w->copy ();
  else {
    class StructValue* v = new class StructValue (this);
    for (card_t i = 0; i < getSize (); i++)
      (*v)[i] = (*this)[i]->getLastValue ();
    return v;
  }
}

bool
StructType::isAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return Type::isAssignable (type);

  return myComponents.isAssignable
    (static_cast<const class StructType*>(type)->myComponents);
}

bool
StructType::isAlwaysAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return false;

  return myComponents.isAlwaysAssignable
    (static_cast<const class StructType*>(type)->myComponents);
}

bool
StructType::isConstrained (const class Value* value) const
{
  assert (value && value->getType ());
  assert (value->getType ()->isAssignable (this));
  if (value->getKind () != Value::vStruct)
    return false;

  const class StructValue* v = static_cast<const class StructValue*>(value);
  assert (v->getSize () == getSize ());

  for (card_t i = getSize (); i--; )
    if (!(*this)[i]->isConstrained ((*v)[i]))
      return false;

  return Type::isConstrained (value);
}

card_t
StructType::do_getNumValues () const
{
  card_t numValues;
  if (myConstraint) {
    numValues = 0;
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      assert (!low || !high || *low <= *high);
      card_t num = 0;
      if (low && high)
	num += *high - *low;
      else if (high) {
	class Value* v = getFirstValue ();
	num += *high - *v;
	delete v;
      }
      else {
	assert (low != NULL);
	class Value* v = low->copy ();
	while (v->increment ())
	  num++;
	delete v;
      }

      if (num != CARD_T_MAX && ++num < CARD_T_MAX - numValues)
	numValues += num;
      else
	return CARD_T_MAX;
    }
  }
  else {
    numValues = 1;
    for (card_t i = 0; i < myComponents.getSize (); i++) {
      card_t num = myComponents[i]->getNumValues ();
      if (num != CARD_T_MAX && num < CARD_T_MAX / numValues)
	numValues *= num;
      else
	return CARD_T_MAX;
    }
  }

  return numValues;
}

card_t
StructType::convert (const class Value* value) const
{
  assert (value && value->getKind () == Value::vStruct);
  assert (isConstrained (value));
  assert (getNumValues () < CARD_T_MAX);

  if (!myComponents.getSize ())
    return 0;

  card_t number;

  if (myConstraint) {
    assert (isConstrained (value));
    number = 0;

    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      class Value* v = NULL;
      if (!low)
	low = v = getFirstValue ();

      assert (!high || *low <= *high);
      if (!high || *value <= *high) {
	number += *value - *low;
	delete v;
	break;
      }
      else {
	number += *high - *low + 1;
	delete v;
      }
    }
  }
  else {
    const class StructValue* v = static_cast<const class StructValue*>(value);
    number = myComponents[card_t (0)]->convert ((*v)[0]);
    for (card_t i = 1; i < myComponents.getSize (); i++) {
      number *= myComponents[i]->getNumValues ();
      number += myComponents[i]->convert ((*v)[i]);
    }
  }

  assert (number < getNumValues ());
  return number;
}

class Value*
StructType::convert (card_t number) const
{
  assert (number < getNumValues ());
  assert (getNumValues () < CARD_T_MAX);

  class StructValue* value = NULL;

  if (myConstraint) {
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      class Value* v = NULL;
      if (!low)
	low = v = getFirstValue ();

      assert (!high || *low <= *high);
      if (high) {
	card_t diff = *high - *low;
	if (number <= diff)
	  goto found;
	else
	  number -= diff;
      }
      else {
      found:
	value = static_cast<class StructValue*>(low->copy ());
	delete v;
	while (number--)
	  if (!value->increment ())
	    assert (false);
	break;
      }

      delete v;
    }
  }
  else {
    value = new class StructValue (this);
    if (myComponents.getSize ()) {
      for (card_t i = myComponents.getSize (); --i; ) {
	card_t card = myComponents[i]->getNumValues ();
	card_t num = number % card;
	number /= card;

	(*value)[i] = myComponents[i]->convert (num);
      }

      (*value)[0] = myComponents[card_t (0)]->convert (number);
    }
  }

  assert (isConstrained (value));
  return value;
}

#ifdef DEBUG
# include "util.h"

void
StructType::prettyPrint (unsigned indent) const
{
  print (indent, "struct {\n");
  myComponents.prettyPrint (indent + 1);
  print (indent, "}\n");
  if (getConstraint ())
    getConstraint ()->prettyPrint (indent);
}

#endif // DEBUG
