// Data type base class -*- c++ -*-

#ifndef TYPE_H_
# define TYPE_H_

# include "typedefs.h"
# include <cstring>
# include <cassert>

/** Data type base class */
class Type {
public:
  /** Type kinds (@see getKind) */
  enum Kind {
    tInt, tBool, tChar, tEnum, tId, tStruct, tVector, tUnion, tBuffer
  };

  // Constructors and destructors

  /** Constructs Type */
  Type ();
protected:
  /** Copy constructor */
  Type (const class Type& old);
private:
  /** Assignment operator */
  class Type& operator= (const class Type& old);
public:
  /** Destructs Type */
  virtual ~Type ();
  /** Copy Type */
  virtual class Type* copy () const = 0;

  // Accessors

  /** Determine the kind of the type */
  virtual enum Kind getKind () const = 0;

  /** Set the name of the type */
  void setName (char* name) { assert (!myName); myName = name; }
  /** Determine the name of the type */
  const char* getName () const { return myName; }

  /** Determine the syntactic name of the type */
  const char* getSyntacticName () const;

  /** Determine whether the type is ordered */
  virtual bool isOrdered () const = 0;
  /** Get the first value of this type (only for ordered types) */
  virtual class Value* getFirstValue () const = 0;
  /** Get the last value of this type (only for ordered types) */
  virtual class Value* getLastValue () const = 0;

  /** See if the type is assignable to another type
   * @param type	the type this type should be assignable to
   * @return		true if this type is assignable to the type
   */
  virtual bool isAssignable (const class Type* type) const;

  /** See if the type is always assignable to another type
   * @param type	the type this type should always be assignable to
   * @return		true if this type always is assignable to the type
   */
  virtual bool isAlwaysAssignable (const class Type* type) const;

  /** See if two types are assignable to each other
   * @param type1	first type to be checked
   * @param type2	second type to be checked
   * @return 		true if the types are compatible with each other
   */
  static bool isCompatible (const class Type* type1, const class Type* type2) {
    return type1->isAssignable (type2) && type2->isAssignable (type1);
  }

  /** See if two types always are assignable to each other
   * @param type1	first type to be checked
   * @param type2	second type to be checked
   * @return 		true if the types always are compatible with each other
   */
  static bool isAlwaysCompatible (const class Type* type1,
				  const class Type* type2) {
    return
      type1->isAlwaysAssignable (type2) && type2->isAlwaysAssignable (type1);
  }

  /** Determine whether a value is compatible with the constraints of this type
   * @param value	value to check
   * @return		true if the value passes the constraint check
   */
  virtual bool isConstrained (const class Value* value) const;

  /**
   * Determine the number of possible values for this type or CARD_T_MAX,
   * if there are >= CARD_T_MAX possible values
   * @return            number of possible values
   */
  card_t getNumValues () const {
    if (!myNumValues)
      myNumValues = do_getNumValues ();
    assert (myNumValues);
    return myNumValues;
  }

private:
  /** Determine the number of possible values for this type or CARD_T_MAX,
   * if there are >= CARD_T_MAX possible values
   * @return		number of possible values
   */
  virtual card_t do_getNumValues () const = 0;
public:

  /** Set the constraint */
  void setConstraint (class Constraint* constraint);

  /** Get the constraint */
  const class Constraint* getConstraint () const { return myConstraint; }
  /** Get the constraint */
  class Constraint* getConstraint () { return myConstraint; }

  // Other methods

  /** Convert a value of this type to a number
   * @param value	value to be converted
   * @return		number between 0 and getNumValues () - 1
   */
  virtual card_t convert (const class Value* value) const = 0;

  /** Convert a number to a value of this type
   * @param number	number between 0 and getNumValues () - 1
   * @return		the corresponding value
   */
  virtual class Value* convert (card_t number) const = 0;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  virtual void prettyPrint (unsigned indent) const;
# endif // DEBUG

  /** The string comparison operator */
  struct ltstr {
    bool operator () (const char* s1, const char* s2) const {
      return strcmp (s1, s2) < 0;
    }
  };

private:
  /** Number of possible values in the type */
  mutable card_t myNumValues;
  /** Name of the type (may be NULL) */
  char* myName;
protected:
  /** Constraint associated with the type */
  class Constraint* myConstraint;
};

#endif // TYPE_H_
