// Union type class -*- c++ -*-

#include "UnionType.h"
#include "UnionValue.h"
#include "ComponentList.h"
#include "Constraint.h"

UnionType::UnionType (class ComponentList* components) :
  Type (),
  myComponents (*components), myNumValues (0)
{
  delete components;
}

UnionType::UnionType (const class UnionType& old) :
  Type (),
  myComponents (old.myComponents), myNumValues (0)
{
}

UnionType::~UnionType ()
{
  delete[] myNumValues;
}

bool
UnionType::isOrdered () const
{
  for (card_t i =  0; i < getSize (); i++)
    if (!(*this)[i]->isOrdered ())
      return false;
  return true;
}

class Value*
UnionType::getFirstValue () const
{
  assert (isOrdered ());
  return new class UnionValue (this, 0, (*this)[card_t (0)]->getFirstValue ());
}

class Value*
UnionType::getLastValue () const
{
  assert (isOrdered ());
  return new class UnionValue (this, getSize () - 1,
			       (*this)[card_t (getSize () - 1)]
			       ->getLastValue ());
}

bool
UnionType::isAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return myComponents.isAssignable (*type);

  return myComponents.isAssignable
    (static_cast<const class UnionType*>(type)->myComponents);
}

bool
UnionType::isAlwaysAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return false;

  return myComponents.isAlwaysAssignable
    (static_cast<const class UnionType*>(type)->myComponents);
}

bool
UnionType::isConstrained (const class Value* value) const
{
  assert (value && value->getType ());
  assert (value->getType ()->isAssignable (this));
  if (value->getKind () != Value::vUnion)
    return false;

  const class UnionValue* v = static_cast<const class UnionValue*>(value);
  assert (v->getIndex () < getSize ());

  return (*this)[v->getIndex ()]->isConstrained (v->getValue ());
}

card_t
UnionType::do_getNumValues () const
{
  card_t numValues = 0;
  assert (!myNumValues && myComponents.getSize ());
  myNumValues = new card_t[myComponents.getSize ()];

  for (card_t i = 0; i < myComponents.getSize (); i++) {
    card_t num = myComponents[i]->getNumValues ();
    if (num != CARD_T_MAX && num < CARD_T_MAX - numValues)
      myNumValues[i] = numValues += num;
    else
      return CARD_T_MAX;
  }

  return numValues;
}

card_t
UnionType::convert (const class Value* value) const
{
  assert (value && value->getKind () == Value::vUnion);
  assert (isConstrained (value));

  const class UnionValue* v = static_cast<const class UnionValue*>(value);
  card_t i = v->getIndex ();
  card_t number = i ? myNumValues[i - 1] : 0;
  number += (*this)[i]->convert (v->getValue ());

  assert (number < getNumValues ());
  return number;
}

class Value*
UnionType::convert (card_t number) const
{
  assert (number < getNumValues ());

  card_t i = 0;
  while (myNumValues[i] <= number)
    i++;

  class Value* v =
    (*this)[i]->convert (number - (i ? myNumValues[i - 1] : 0));

  return new class UnionValue (this, i, v);
}

class Value*
UnionType::cast (class Value& value) const
{
  for (card_t i = getSize (); i--; ) {
    if (value.getType ()->isAssignable ((*this)[i])) {
      if ((*this)[i]->isConstrained (&value)) {
	value.setType ((*this)[i]);
	return new class UnionValue (this, i, &value);
      }
      else
	break;
    }
  }
  delete &value;
  return NULL;
}

#ifdef DEBUG
# include "util.h"

void
UnionType::prettyPrint (unsigned indent) const
{
  print (indent, "union {\n");
  myComponents.prettyPrint (indent + 1);
  print (indent, "}\n");
  if (getConstraint ())
    getConstraint ()->prettyPrint (indent);
}

#endif // DEBUG
