// Union type class -*- c++ -*-

#ifndef UNIONTYPE_H_
# define UNIONTYPE_H_

# include "ComponentList.h"

/** Union type */
class UnionType : public Type {
public:
  // Constructors and destructors

  /** Constructs UnionType
   * @param components	Components of the union
   */
  UnionType (class ComponentList* components);
private:
  /** Copy constructor */
  UnionType (const class UnionType& old);
  /** Assignment operator */
  class UnionType& operator= (const class UnionType& old);
public:
  /** Destructs UnionType */
  ~UnionType ();
  /** Copy the Type */
  class Type* copy () const { return new class UnionType (*this); }

  // Accessors

  /** Determine the kind of the type */
  enum Kind getKind () const { return tUnion; }

  /** Determine whether the type is ordered */
  bool isOrdered () const;
  /** Get the first value of this type */
  class Value* getFirstValue () const;
  /** Get the last value of this type */
  class Value* getLastValue () const;

  /** Determine the number of components in the type */
  card_t getSize () const { return myComponents.getSize (); }

  /** Get a component by name
   * @param name	name of the component
   * @return		pointer to the component, or NULL if not found
   */
  const class Type* operator[] (const char* name) const {
    return myComponents[name];
  }

  /** Get a component by index number
   * @param index	index to the component
   * @return		pointer to the component
   */
  const class Type* operator[] (card_t index) const {
    return myComponents[index];
  }

  /** Get the index number of a component
   * @param name	name of the component
   * @return		index of the component
   */
  card_t getIndex (const char* name) const {
    return myComponents.getIndex (name);
  }

  /** Get the name of a component
   * @param index	index of the component
   * @return		name of the component
   */
  const char* getComponentName (card_t index) const {
    return myComponents.getName (index);
  }

  /** Determine the cumulated number of values for a component
   * @param index	index of the component
   * @return		number of values of all components up to index
   */
  card_t getCumulatedValues (card_t index) const {
    assert (index < myComponents.getSize ());
    return index ? myNumValues[index - 1] : 0;
  }

  // Other methods

  /** Convert a value of this type to a number
   * @param value	value to be converted
   * @return		number between 0 and getNumValues () - 1
   */
  card_t convert (const class Value* value) const;

  /** Convert a number to a value of this type
   * @param number	number between 0 and getNumValues () - 1
   * @return		the corresponding value
   */
  class Value* convert (card_t number) const;

  /** See if the type is assignable to another type
   * @param type	the type this type should be assignable to
   * @return		true if this type is assignable to the type
   */
  bool isAssignable (const class Type* type) const;

  /** See if the type is always assignable to another type
   * @param type	the type this type should always be assignable to
   * @return		true if this type always is assignable to the type
   */
  bool isAlwaysAssignable (const class Type* type) const;

  /** See if the component list is assignable from a type
   * @param type	the type this list should be assignable from
   * @return		true if this list is assignable from the type
   */
  bool isAssignableFrom (const class Type& type) const {
    return myComponents.isAssignableFrom (type);
  }
  /** See if the component list is always assignable from a type
   * @param type	the type the list should always be assignable from
   * @return		true if this list always is assignable from the type
   */
  bool isAlwaysAssignableFrom (const class Type& type) const {
    return myComponents.isAlwaysAssignableFrom (type);
  }

  /** Determine whether a value is compatible with the constraints of this type
   * @param value	value to check
   * @return		true if the value passes the constraint check
   */
  bool isConstrained (const class Value* value) const;

  /** Get the number of possible values for this type */
  card_t do_getNumValues () const;

  /** Convert a value to this type, if possible
   * @param value	value to be converted
   * @return		the converted value, or NULL
   */
  class Value* cast (class Value& value) const;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The union components */
  class ComponentList myComponents;
  /** Cumulated numbers of possible values of the components */
  mutable card_t* myNumValues;
};

#endif // UNIONTYPE_H_
