// Vector type class -*- c++ -*-

#include "VectorType.h"
#include "VectorValue.h"
#include "Constraint.h"
#include "Range.h"

#include <cstring>

VectorType::VectorType (const class Type* itemType,
			const class Type* indexType,
			bool optional) :
  Type (),
  myItemType (itemType), myIndexType (indexType), myOptional (optional)
{
  assert (myItemType && myIndexType);
  assert (myIndexType->isOrdered ());
#ifndef NDEBUG
  card_t size = myIndexType->getNumValues ();
  assert (size > 0 && size < CARD_T_MAX);
#endif // NDEBUG
}

VectorType::VectorType (const VectorType& old) :
  Type (),
  myItemType (old.myItemType), myIndexType (old.myIndexType),
  myOptional (old.myOptional)
{
#ifndef NDEBUG
  card_t size = myIndexType->getNumValues ();
  assert (size > 0 && size < CARD_T_MAX);
#endif // NDEBUG
}

VectorType::~VectorType ()
{
}

class Value*
VectorType::getFirstValue () const
{
  if (const class Value* w = myConstraint ? myConstraint->getFirstValue () : 0)
    return w->copy ();
  if (myOptional)
    return new class VectorValue (this);

  class Value* comp = myItemType->getFirstValue ();
  if (!comp)
    return NULL;

  class VectorValue* v = new class VectorValue (this);

  for (card_t i = getSize (); --i; (*v)[i] = comp->copy ());
  (*v)[0] = comp;

  return v;
}

class Value*
VectorType::getLastValue () const
{
  if (const class Value* w = myConstraint ? myConstraint->getLastValue () : 0)
    return w->copy ();

  class Value* comp = myItemType->getLastValue ();
  if (!comp)
    return NULL;

  class VectorValue* v = new class VectorValue (this);

  for (card_t i = getSize (); --i; (*v)[i] = comp->copy ());
  (*v)[0] = comp;

  return v;
}

bool
VectorType::isAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return Type::isAssignable (type);

  const class VectorType* v = static_cast<const class VectorType*>(type);

  return
    myIndexType->isAlwaysAssignable (v->myIndexType) &&
    myItemType->isAssignable (v->myItemType);
}

bool
VectorType::isAlwaysAssignable (const class Type* type) const
{
  if (type == this)
    return true;

  if (type->getKind () != getKind ())
    return false;

  const class VectorType* v = static_cast<const class VectorType*>(type);

  return
    myOptional == v->myOptional &&
    myIndexType->isAlwaysAssignable (v->myIndexType) &&
    myItemType->isAlwaysAssignable (v->myItemType);
}

bool
VectorType::isConstrained (const class Value* value) const
{
  assert (value && value->getType ());
  assert (value->getType ()->isAssignable (this));
  if (value->getKind () != Value::vVector)
    return false;

  const class VectorValue* v = static_cast<const class VectorValue*>(value);
  assert (v->getSize () == getSize ());

  for (card_t i = getSize (); i--; ) {
    if (const class Value* comp = (*v)[i]) {
      if (!myItemType->isConstrained (comp))
	return false;
    }
    else if (!myOptional)
      return false;
  }

  return Type::isConstrained (value);
}

card_t
VectorType::do_getNumValues () const
{
  card_t numValues;
  if (myConstraint) {
    numValues = 0;
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      assert (!low || !high || *low <= *high);
      card_t num = 0;
      if (low && high)
	num += *high - *low;
      else if (high) {
	class Value* v = getFirstValue ();
	num += *high - *v;
	delete v;
      }
      else {
	assert (low != NULL);
	class Value* v = low->copy ();
	while (v->increment ())
	  num++;
	delete v;
      }

      if (num != CARD_T_MAX && ++num < CARD_T_MAX - numValues)
	numValues += num;
      else
	return CARD_T_MAX;
    }
  }
  else {
    card_t numItemValues = myItemType->getNumValues ();
    numValues = 1;
    if (myOptional && numItemValues++ == CARD_T_MAX)
      return CARD_T_MAX;

    if (numItemValues == CARD_T_MAX)
      return CARD_T_MAX;

    for (card_t i = 0, size = getSize (); i < size; i++) {
      if (numItemValues < CARD_T_MAX / numValues)
	numValues *= numItemValues;
      else
	return CARD_T_MAX;
    }
  }

  return numValues;
}

card_t
VectorType::convert (const class Value* value) const
{
  assert (value && value->getKind () == Value::vVector);
  assert (isConstrained (value));
  assert (getNumValues () < CARD_T_MAX);

  card_t number = 0;

  if (myConstraint) {
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      class Value* v = NULL;
      if (!low)
	low = v = getFirstValue ();

      assert (!high || *low <= *high);
      if (!high || *value <= *high) {
	number += *value - *low;
	delete v;
	break;
      }
      else {
	number += *high - *low + 1;
	delete v;
      }
    }
  }
  else {
    const class VectorValue* v = static_cast<const class VectorValue*>(value);
    card_t i = getSize (), numItemValues = myItemType->getNumValues ();
    if (myOptional)
      numItemValues++;
    while (i--) {
      card_t num = myOptional
	? ((*v)[i] ? myItemType->convert ((*v)[i]) + 1 : 0)
	: myItemType->convert ((*v)[i]);
      (number *= numItemValues) += num;
    }
  }

  assert (number < getNumValues ());
  return number;
}

class Value*
VectorType::convert (card_t number) const
{
  assert (number < getNumValues ());
  assert (getNumValues () < CARD_T_MAX);

  class VectorValue* value = NULL;
  if (myConstraint) {
    for (Constraint::const_iterator i = myConstraint->begin ();
	 i != myConstraint->end (); i++) {
      const class Range* r = *i;
      const class Value* low = r->getLow ();
      const class Value* high = r->getHigh ();
      class Value* v = NULL;
      if (!low)
	low = v = getFirstValue ();

      assert (!high || *low <= *high);
      if (high) {
	card_t diff = *high - *low;
	if (number <= diff)
	  goto found;
	else
	  number -= diff;
      }
      else {
      found:
	value = static_cast<class VectorValue*>(low->copy ());
	delete v;
	while (number--)
	  if (!value->increment ())
	    assert (false);
	break;
      }

      delete v;
    }
  }
  else {
    value = new class VectorValue (this);
    card_t numItemValues = myItemType->getNumValues ();
    if (myOptional)
      numItemValues++;
    card_t size = getSize ();

    for (card_t i = 0; i < size; i++) {
      card_t num = number % numItemValues;
      number /= numItemValues;

      if (!myOptional || num--)
	(*value)[i] = myItemType->convert (num);
    }
  }

  assert (isConstrained (value));
  return value;
}

#ifdef DEBUG
# include "util.h"

void
VectorType::prettyPrint (unsigned indent) const
{
  const char* const opt = myOptional ? "?" : "";

  if (myItemType->getName () && myIndexType->getName ())
    print (indent, "%s[%s%s]\n", myItemType->getName (), opt,
	   myIndexType->getName ());
  else if (myIndexType->getName ()) {
    myItemType->prettyPrint (indent);
    print (indent, "[%s%s]\n", myIndexType->getName (), opt);
  }
  else if (myItemType->getName ()) {
    print (indent, "%s[%s\n", myItemType->getName (), opt);
    myIndexType->prettyPrint (indent + 1);
    print (indent, "]\n");
  }
  else {
    myItemType->prettyPrint (indent);
    print (indent, "[%s\n", opt);
    myIndexType->prettyPrint (indent + 1);
    print (indent, "]\n");
  }
  if (getConstraint ())
    getConstraint ()->prettyPrint (indent);
}

#endif // DEBUG
