// Vector type class -*- c++ -*-

#ifndef VECTORTYPE_H_
# define VECTORTYPE_H_

# include "Type.h"

/** Vector type (integer-indexed array) */
class VectorType : public Type {
public:
  // Constructors and destructors

  /** Constructs VectorType
   * @param itemType	The vector item type
   * @param indexType	The vector index type
   * @param optional	Flag: are the vector components optional?
   */
  VectorType (const class Type* itemType,
	      const class Type* indexType,
	      bool optional);
private:
  /** Copy constructor */
  VectorType (const class VectorType& old);
  /** Assignment operator */
  class VectorType& operator= (const class VectorType& old);
public:
  /** Destructs VectorType */
  ~VectorType ();
  /** Copy the Type */
  class Type* copy () const { return new class VectorType (*this); }

  // Accessors

  /** Determine the kind of the type */
  enum Kind getKind () const { return tVector; }

  /** Get the item type */
  const class Type* getItemType () const { return myItemType; }
  /** Get the index type */
  const class Type* getIndexType () const { return myIndexType; }
  /** Determine the vector size */
  card_t getSize () const { return myIndexType->getNumValues (); }
  /** Determine whether the components are optional */
  bool isOptional () const { return myOptional; }

  /** Determine whether the type is ordered */
  bool isOrdered () const { return myItemType->isOrdered (); }
  /** Get the first value of this type */
  class Value* getFirstValue () const;
  /** Get the last value of this type */
  class Value* getLastValue () const;

  // Other methods

  /** Convert a value of this type to a number
   * @param value	value to be converted
   * @return		number between 0 and getNumValues () - 1
   */
  card_t convert (const class Value* value) const;

  /** Convert a number to a value of this type
   * @param number	number between 0 and getNumValues () - 1
   * @return		the corresponding value
   */
  class Value* convert (card_t number) const;

  /** See if the type is assignable to another type
   * @param type	the type this type should be assignable to
   * @return		true if this type is assignable to the type
   */
  bool isAssignable (const class Type* type) const;

  /** See if the type is always assignable to another type
   * @param type	the type this type should always be assignable to
   * @return		true if this type always is assignable to the type
   */
  bool isAlwaysAssignable (const class Type* type) const;

  /** Determine whether a value is compatible with the constraints of this type
   * @param value	value to check
   * @return		true if the value passes the constraint check
   */
  bool isConstrained (const class Value* value) const;

  /** Get the number of possible values for this type */
  card_t do_getNumValues () const;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

private:
  /** The item type */
  const class Type* myItemType;
  /** The index type */
  const class Type* myIndexType;
  /** Flag: are the vector components optional? */
  bool myOptional;
};

#endif // VECTORTYPE_H_
