// Buffer value class -*- c++ -*-

#include "BufferValue.h"
#include "BufferType.h"
#include "Constraint.h"

BufferValue::BufferValue (const class Type* type) :
  Value (type),
  myComponents (static_cast<const class BufferType*>(type)->getSize ())
{
  assert (getType ()->getKind () == Type::tBuffer);
}

BufferValue::BufferValue (const class BufferValue& old) :
  Value (old.getType ()),
  myComponents (old.myComponents)
{
  assert (getType ()->getKind () == Type::tBuffer);
}

BufferValue::~BufferValue ()
{
}

bool
BufferValue::operator< (const class BufferValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  card_t size = getCapacity (), osize = other.getCapacity ();
  if (size < osize)
    return true;
  else if (osize < size)
    return false;

  while (size--)
    if (*myComponents[size] < *other.myComponents[size])
      return true;
    else if (*other.myComponents[size] < *myComponents[size])
      return false;

  return false;
}

bool
BufferValue::operator<= (const class BufferValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  card_t size = getCapacity (), osize = other.getCapacity ();
  if (size < osize)
    return true;
  else if (osize < size)
    return false;

  while (size--)
    if (*myComponents[size] < *other.myComponents[size])
      return true;
    else if (*other.myComponents[size] < *myComponents[size])
      return false;

  return true;
}

bool
BufferValue::operator== (const class BufferValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = 0; i < myComponents.getSize (); i++) {
    if (!myComponents[i])
      return !other.myComponents[i];
    if (!other.myComponents[i] ||
	!(*myComponents[i] == *other.myComponents[i]))
      return false;
  }

  return true;
}

card_t
BufferValue::operator- (const class BufferValue& other) const
{
  if (!getSize ())
    return 0;

  const class Type* itemType =
    static_cast<const class BufferType*>(getType ())->getItemType ();
  card_t numItemValues = itemType->getNumValues ();
  card_t offset = 0, i = 0;
  for (card_t mult = 1; i < getSize () && myComponents[i];
       mult *= numItemValues, i++)
    if (!other.myComponents[i])
      offset += mult;
  assert (i == getSize () || !other.myComponents[i]);

  card_t diff = 0;
  while (i--) {
    diff *= numItemValues;
    if (!other.myComponents[i])
      diff += itemType->convert (myComponents[i]);
    else if (*other.myComponents[i] < *myComponents[i])
      diff += *myComponents[i] - *other.myComponents[i];
    else
      diff -= *other.myComponents[i] - *myComponents[i];
  }
  return diff + offset;
}

void
BufferValue::bottom ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class BufferType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    if (const class Value* v =
	getType ()->getConstraint ()->getFirstValue ()) {
      assert (v->getKind () == getKind ());
      const class BufferValue& bv = static_cast<const class BufferValue&>(*v);
      for (card_t i = getSize (); i--; ) {
	delete myComponents[i];
	myComponents[i] = bv[i] ? bv[i]->copy () : NULL;
      }

      return;
    }
  }

  for (card_t i = 0; i < getSize (); i++)
    delete myComponents[i], myComponents[i] = NULL;
}

void
BufferValue::top ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class BufferType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    if (const class Value* v =
	getType ()->getConstraint ()->getLastValue ()) {
      assert (v->getKind () == getKind ());
      const class BufferValue& bv = static_cast<const class BufferValue&>(*v);
      for (card_t i = getSize (); i--; ) {
	delete myComponents[i];
	myComponents[i] = bv[i] ? bv[i]->copy () : NULL;
      }

      return;
    }
  }

  for (card_t i = 0; i < getSize (); i++)
    delete myComponents[i], myComponents[i] =
      static_cast<const class BufferType*>(getType ())->getItemType ()
      ->getLastValue ();
}

bool
BufferValue::increment ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class BufferType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    const class Value* v = getType ()->getConstraint ()->getNextHigh (this);
    if (v) {
      assert (v->getKind () == getKind ());
      if (*this == static_cast<const class BufferValue&>(*v)) {
	v = getType ()->getConstraint ()->getNextLow (this);
	if (!v) {
	  bottom ();
	  return false;
	}
	assert (v->getKind () == getKind ());
	const class BufferValue* bv = static_cast<const class BufferValue*>(v);
	for (card_t i = getSize (); i--; ) {
	  delete myComponents[i];
	  myComponents[i] = (*bv)[i] ? (*bv)[i]->copy () : NULL;
	}
	return true;
      }
    }
  }

  for (card_t i = 0; i < getSize (); i++) {
    if (!myComponents[i]) {
      myComponents[i] =
	static_cast<const class BufferType*>(getType ())->getItemType ()
	->getFirstValue ();
      return true;
    }

    if (myComponents[i]->increment ())
      return true;
  }

  bottom ();
  return false;
}

bool
BufferValue::decrement ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class BufferType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    const class Value* v = getType ()->getConstraint ()->getPrevLow (this);
    if (v) {
      assert (v->getKind () == getKind ());
      if (*this == static_cast<const class BufferValue&>(*v)) {
	v = getType ()->getConstraint ()->getPrevHigh (this);
	if (!v) {
	  top ();
	  return false;
	}
	assert (v->getKind () == getKind ());
	const class BufferValue* bv = static_cast<const class BufferValue*>(v);
	for (card_t i = getSize (); i--; ) {
	  delete myComponents[i];
	  myComponents[i] = (*bv)[i] ? (*bv)[i]->copy () : NULL;
	}
	return true;
      }
    }
  }

  card_t i;
  for (i = 0; i < getSize (); i++)
    if (!myComponents[i])
      break;
    else if (myComponents[i]->decrement ())
      return true;

  if (i--) {
    delete myComponents[i];
    myComponents[i] = NULL;
    return true;
  }
  else {
    top ();
    return false;
  }
}

class Value*
BufferValue::cast (const class Type& type)
{
  assert (getType ()->isAssignable (&type));
  if (type.getKind () != Type::tBuffer)
    return Value::cast (type);
  const class BufferType& bt = static_cast<const class BufferType&>(type);
  const class Type& itemType = *bt.getItemType ();
  assert (getSize () == bt.getSize ());

  for (card_t i = 0; i < getSize () && (*this)[i]; i++) {
    if (!((*this)[i] = (*this)[i]->cast (itemType))) {
      delete this;
      return NULL;
    }
  }
  return Value::cast (type);
}

#include "Printer.h"
#include "CharValue.h"

void
BufferValue::display (const class Printer& printer) const
{
  card_t i;
  const class BufferType* const type =
    static_cast<const class BufferType*>(getType ());
  if (type->getItemType ()->getKind () == Type::tChar) {
    char* buf = new char[i = getSize ()];
    for (i = 0; myComponents[i] && i < getSize (); i++) {
      assert (myComponents[i]->getKind () == Value::vChar);
      buf[i] =
	static_cast<const class CharValue*>(myComponents[i])->getValue ();
    }
    printer.print (buf, i);
    delete[] buf;
  }
  else {
    printer.delimiter ('{')++;
    for (i = 0; myComponents[i]; ) {
      myComponents[i]->display (printer);
      if (++i == getSize () || !myComponents[i])
	break;
      else
	printer.delimiter (',');
    }
    printer--.delimiter ('}');
  }
}

#ifdef DEBUG
# include "util.h"

void
BufferValue::prettyPrint (unsigned indent) const
{
  print (indent, "{\n");

  for (card_t i = 0; myComponents[i]; ) {
    myComponents[i]->prettyPrint (indent + 1);
    if (++i == getSize () || !myComponents[i])
      break;
    else
      print (indent + 1, ",\n");
  }

  print (indent, "}\n");
}

#endif // DEBUG
