// Structure value class -*- c++ -*-

#include "StructValue.h"
#include "StructType.h"
#include "Constraint.h"

StructValue::StructValue (const class Type* type) :
  Value (type),
  myComponents (static_cast<const class StructType*>(type)->getSize ())
{
  assert (getType ()->getKind () == Type::tStruct);
}

StructValue::StructValue (const class StructValue& old) :
  Value (old.getType ()),
  myComponents (old.myComponents)
{
  assert (getType ()->getKind () == Type::tStruct);
}

StructValue::~StructValue ()
{
}

bool
StructValue::operator< (const class StructValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = myComponents.getSize (); i--; )
    if (*myComponents[i] < *other.myComponents[i])
      return true;
    else if (*other.myComponents[i] < *myComponents[i])
      return false;

  return false;
}

bool
StructValue::operator<= (const class StructValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = myComponents.getSize (); i--; )
    if (*myComponents[i] < *other.myComponents[i])
      return true;
    else if (*other.myComponents[i] < *myComponents[i])
      return false;

  return true;
}

bool
StructValue::operator== (const class StructValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = myComponents.getSize (); i--; )
    if (!(*myComponents[i] == *other.myComponents[i]))
      return false;

  return true;
}

void
StructValue::bottom ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class StructType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    if (const class Value* const v =
	getType ()->getConstraint ()->getFirstValue ()) {
      assert (v->getKind () == getKind ());
      const class StructValue& sv = static_cast<const class StructValue&>(*v);

      for (card_t i = getSize (); i--; ) {
	delete myComponents[i];
	myComponents[i] = sv[i]->copy ();
      }

      return;
    }
  }

  for (card_t i = 0; i < getSize (); i++)
    myComponents[i]->bottom ();
}

void
StructValue::top ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class StructType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    if (const class Value* const v =
	getType ()->getConstraint ()->getLastValue ()) {
      assert (v->getKind () == getKind ());
      const class StructValue& sv = static_cast<const class StructValue&>(*v);

      for (card_t i = getSize (); i--; ) {
	delete myComponents[i];
	myComponents[i] = sv[i]->copy ();
      }

      return;
    }
  }

  for (card_t i = 0; i < getSize (); i++)
    myComponents[i]->top ();
}

card_t
StructValue::operator- (const class StructValue& other) const
{
  if (!getSize ())
    return 0;

  card_t i, numValues;

  for (i = numValues = 1; i < getSize (); i++)
    numValues *= myComponents[i - 1]->getType ()->getNumValues ();

  card_t diff = 0;

  for (i = getSize (); --i;
       numValues /= myComponents[i - 1]->getType ()->getNumValues ())
    if (*myComponents[i] < *other.myComponents[i])
      diff -= (*other.myComponents[i] - *myComponents[i]) * numValues;
    else
      diff += (*myComponents[i] - *other.myComponents[i]) * numValues;

  assert (numValues == 1);

  if (*myComponents[0] < *other.myComponents[0])
    diff -= *other.myComponents[0] - *myComponents[0];
  else
    diff += *myComponents[0] - *other.myComponents[0];

  return diff;
}

bool
StructValue::increment ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class StructType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    const class Value* v = getType ()->getConstraint ()->getNextHigh (this);
    if (v) {
      assert (v->getKind () == getKind ());
      if (*this == static_cast<const class StructValue&>(*v)) {
	v = getType ()->getConstraint ()->getNextLow (this);
	if (!v) {
	  bottom ();
	  return false;
	}
	assert (v->getKind () == getKind ());
	const class StructValue* sv = static_cast<const class StructValue*>(v);
	for (card_t i = getSize (); i--; ) {
	  delete myComponents[i];
	  myComponents[i] = (*sv)[i]->copy ();
	}
	return true;
      }
    }
  }

  for (card_t i = 0; i < getSize (); i++)
    if (myComponents[i]->increment ())
      return true;

  if (getType ()->getConstraint ())
    bottom ();

  return false;
}

bool
StructValue::decrement ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class StructType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    const class Value* v = getType ()->getConstraint ()->getPrevLow (this);
    if (v) {
      assert (v->getKind () == getKind ());
      if (*this == static_cast<const class StructValue&>(*v)) {
	v = getType ()->getConstraint ()->getPrevHigh (this);
	if (!v) {
	  top ();
	  return false;
	}
	assert (v->getKind () == getKind ());
	const class StructValue* sv = static_cast<const class StructValue*>(v);
	for (card_t i = getSize (); i--; ) {
	  delete myComponents[i];
	  myComponents[i] = (*sv)[i]->copy ();
	}
	return true;
      }
    }
  }

  for (card_t i = 0; i < getSize (); i++)
    if (myComponents[i]->decrement ())
      return true;

  if (getType ()->getConstraint ())
    top ();

  return false;
}

class Value*
StructValue::cast (const class Type& type)
{
  assert (getType ()->isAssignable (&type));
  if (type.getKind () != Type::tStruct)
    return Value::cast (type);
  const class StructType& st = static_cast<const class StructType&>(type);
  assert (getSize () == st.getSize ());

  for (card_t i = getSize (); i--; ) {
    if (!((*this)[i] = (*this)[i]->cast (*st[i]))) {
      delete this;
      return NULL;
    }
  }
  return Value::cast (type);
}

#include "Printer.h"

void
StructValue::display (const class Printer& printer) const
{
  if (getSize ()) {
    printer.delimiter ('{')++;
    for (card_t i = 0;;) {
      myComponents[i]->display (printer);
      if (++i == getSize ())
	break;
      else
	printer.delimiter (',');
    }
    printer--.delimiter ('}');
  }
  else
    printer.delimiter ('{').delimiter ('}');
}

#ifdef DEBUG
# include "util.h"

void
StructValue::prettyPrint (unsigned indent) const
{
  if (getSize ()) {
    print (indent, "{\n");

    for (card_t i = 0;;) {
      myComponents[i]->prettyPrint (indent + 1);
      if (++i == getSize ())
	break;
      else
	print (indent + 1, ",\n");
    }

    print (indent, "}\n");
  }
  else
    print (indent, "{}\n");
}

#endif // DEBUG
