// Union value class -*- c++ -*-

#include "UnionValue.h"
#include "UnionType.h"

UnionValue::UnionValue (const class Type* type,
			card_t index, class Value* value) :
  Value (type),
  myIndex (index), myValue (value)
{
  assert (getType ()->getKind () == Type::tUnion);
  assert (myValue && myValue->getType () ==
	  (*static_cast<const class UnionType*>(getType ()))[myIndex]);
}

UnionValue::UnionValue (const class UnionValue& old) :
  Value (old.getType ()),
  myIndex (old.myIndex), myValue (old.myValue->copy ())
{
  assert (getType ()->getKind () == Type::tUnion);
  assert (myValue && myValue->getType () ==
	  (*static_cast<const class UnionType*>(getType ()))[myIndex]);
}

UnionValue::~UnionValue ()
{
  delete myValue;
}

bool
UnionValue::operator< (const class UnionValue& other) const
{
  assert (getType () == other.getType ());

  if (myIndex < other.myIndex)
    return true;
  else if (other.myIndex < myIndex)
    return false;
  else
    return *getValue () < *other.getValue ();
}

bool
UnionValue::operator<= (const class UnionValue& other) const
{
  assert (getType () == other.getType ());

  if (myIndex < other.myIndex)
    return true;
  else if (other.myIndex < myIndex)
    return false;
  else
    return *getValue () <= *other.getValue ();
}

bool
UnionValue::operator== (const class UnionValue& other) const
{
  assert (getType () == other.getType ());

  return myIndex == other.myIndex && *getValue () == *other.getValue ();
}

card_t
UnionValue::operator- (const class UnionValue& other) const
{
  const class UnionType* type =
    static_cast<const class UnionType*>(getType ());
  card_t diff =
    type->getCumulatedValues (myIndex) -
    type->getCumulatedValues (other.myIndex);
  if (myIndex == other.myIndex)
    diff += *getValue () - *other.getValue ();
  else
    diff +=
      (*type)[myIndex]->convert (getValue ()) -
      (*type)[other.myIndex]->convert (other.getValue ());

  assert (diff < getType ()->getNumValues ());
  return diff;
}

void
UnionValue::bottom ()
{
  delete myValue;
  myValue = (*static_cast<const class UnionType*>(getType ()))[myIndex = 0]
    ->getFirstValue ();
}

void
UnionValue::top ()
{
  delete myValue;
  const class UnionType& type =
    *static_cast<const class UnionType*>(getType ());
  myValue = type[myIndex = type.getSize () - 1]->getLastValue ();
}

bool
UnionValue::increment ()
{
  if (myValue->increment ())
    return true;

  if (++myIndex == static_cast<const class UnionType*>(getType ())->getSize ())
    myIndex = 0;

  delete myValue;
  myValue = (*static_cast<const class UnionType*>(getType ()))[myIndex]
    ->getFirstValue ();

  return myIndex != 0;
}

bool
UnionValue::decrement ()
{
  if (myValue->decrement ())
    return true;

  const class UnionType& type =
    *static_cast<const class UnionType*>(getType ());

  if (!myIndex) {
    top ();
    return false;
  }
  else {
    delete myValue;
    myValue = type[--myIndex]->getLastValue ();
    return true;
  }
}

class Value*
UnionValue::cast (const class Type& type)
{
  if (type.getKind () == Type::tUnion)
    return Value::cast (type);
  const class UnionType& ut =
    *static_cast<const class UnionType*>(getType ());
  if (!ut[myIndex]->isAssignable (&type)) {
    delete this;
    return NULL;
  }
  class Value* v = myValue;
  myValue = NULL;
  delete this;
  return v;
}

#include "Printer.h"

void
UnionValue::display (const class Printer& printer) const
{
  printer.print (static_cast<const class UnionType*>(getType ())
		 ->getComponentName (myIndex));
  printer.delimiter ('=')++;
  myValue->display (printer);
  printer--;
}

#ifdef DEBUG
# include "util.h"

void
UnionValue::prettyPrint (unsigned indent) const
{
  print (indent, "%s = \n",
	 static_cast<const class UnionType*>(getType ())
	 ->getComponentName (myIndex));
  myValue->prettyPrint (indent + 1);
}

#endif // DEBUG
