// Valuation (value distribution for variables) -*- c++ -*-

#include "Valuation.h"
#include "Value.h"
#include "VariableDefinition.h"
#ifndef NDEBUG
# include "Type.h"
#endif // NDEBUG

Valuation::Valuation () :
  myValues (), myErrors (noErrors), myGlobalMarking (0)
{
}

Valuation::Valuation (const class Valuation& old) :
  myValues (), myErrors (old.myErrors), myGlobalMarking (old.myGlobalMarking)
{
  for (const_iterator i = old.myValues.begin (); i != old.myValues.end ();
       i++) {
    assert (i->second != NULL);
    if (const class VariableDefinition* father = i->first->getFather ()) {
      if (old.myValues.find (father) == old.myValues.end () &&
	  myValues.find (father) == myValues.end ()) {
	if (class Value* fv = father->getValue (old))
	  myValues[father] = fv;
      }
    }
    else
      myValues[i->first] = i->second->copy ();
  }
}

Valuation::~Valuation ()
{
  for (iterator i = myValues.begin (); i != myValues.end (); i++) {
#ifdef EXPR_CACHE
    i->first->clearValuation (this);
#endif // EXPR_CACHE
    delete i->second;
  }
}

const class Value*
Valuation::getValue (const class VariableDefinition* var) const
{
  assert (var != NULL);

  const_iterator i = myValues.find (var);
  if (i == myValues.end ())
    return NULL;

  return i->second;
}

void
Valuation::setValue (const class VariableDefinition* var, class Value* value)
{
  assert (var && value);
  assert (var->getType ()->isConstrained (value));
  Map::value_type p (var, value);
  std::pair<iterator, bool> status = myValues.insert (p);
  if (!status.second) {
    delete status.first->second;
    myValues.erase (status.first);
    status = myValues.insert (p);
    assert (status.second);
  }
  if (!var->getFather ()) {
#ifdef EXPR_CACHE
    var->invalidate ();
#endif // EXPR_CACHE
    var->invalidateMarking ();
  }
}

bool
Valuation::increment (const class VariableDefinition* var)
{
  assert (var && !var->getFather ());
  const_iterator i = myValues.find (var);
  assert (i != myValues.end ());
#ifdef EXPR_CACHE
  var->invalidate ();
#endif // EXPR_CACHE
  var->invalidateMarking ();
  return i->second->increment ();
}

bool
Valuation::decrement (const class VariableDefinition* var)
{
  assert (var && !var->getFather ());
  const_iterator i = myValues.find (var);
  assert (i != myValues.end ());
#ifdef EXPR_CACHE
  var->invalidate ();
#endif // EXPR_CACHE
  var->invalidateMarking ();
  return i->second->decrement ();
}

void
Valuation::undefine (const class VariableDefinition* var)
{
  assert (var != NULL);

  iterator i = myValues.find (var);
  assert (i != myValues.end () || var->getFather () || var->hasChildren ());
  if (i != myValues.end ()) {
    delete i->second;
    myValues.erase (i);
  }
  if (!var->getFather ()) {
#ifdef EXPR_CACHE
    var->invalidate ();
#endif // EXPR_CACHE
    var->invalidateMarking ();
  }
}

bool
Valuation::operator< (const class Valuation& other) const
{
  const_iterator i, j;

  if (myErrors < other.myErrors)
    return true;
  if (other.myErrors < myErrors)
    return false;

  for (i = myValues.begin (), j = other.myValues.begin ();
       i != myValues.end (); i++, j++) {
    if (j == other.myValues.end () || i->first > j->first)
      return false;
    if (i->first < j->first)
      return true;

    if (!i->second) {
      if (j->second)
	return true;
    }
    else if (!j->second)
      return false;
    else if (*i->second < *j->second)
      return true;
    else if (*j->second < *i->second)
      return false;
  }

  return false;
}

#include "Printer.h"
void
Valuation::display (const class Printer& printer) const
{
  printer.delimiter ('<')++;
  if (isDivisionByZero ())
    printer.printRaw ("divisionByZero"), printer.delimiter (',');
  if (isOverflow ())
    printer.printRaw ("overflow"), printer.delimiter (',');
  if (isShiftError ())
    printer.printRaw ("shiftError"), printer.delimiter (',');
  if (isConstraintViolation ())
    printer.printRaw ("constraintViolation"), printer.delimiter (',');
  if (isUnionViolation ())
    printer.printRaw ("unionViolation"), printer.delimiter (',');
  if (isNegativeCardinality ())
    printer.printRaw ("negativeCardinality"), printer.delimiter (',');
  if (isLvalueViolation ())
    printer.printRaw ("lvalueViolation"), printer.delimiter (',');
  if (isUnificationError ())
    printer.printRaw ("unificationError"), printer.delimiter (',');
  if (isCapacityViolation ())
    printer.printRaw ("capacityViolation"), printer.delimiter (',');
  if (isUndefinedValue ())
    printer.printRaw ("undefinedValue"), printer.delimiter (',');
  if (isUndefinedComponent ())
    printer.printRaw ("undefinedComponent"), printer.delimiter (',');
  if (isBufferViolation ())
    printer.printRaw ("bufferViolation"), printer.delimiter (',');
  if (isUndefinedExpression ())
    printer.printRaw ("undefinedExpression"), printer.delimiter (',');
  if (isFatalExpression ())
    printer.printRaw ("fatalExpression"), printer.delimiter (',');

  for (const_iterator i = myValues.begin (); i != myValues.end (); i++) {
    assert (!i->first->isAggregate ());
    if (!i->first->getFather ()) {
      printer.linebreak ();
      printer.print (i->first->getName ());
      printer.delimiter (':')++;
      i->second->display (printer);
      printer--;
    }
  }

  printer--.linebreak ();
  printer.delimiter ('>');
}

#ifdef DEBUG
# include "util.h"
# include "GlobalMarking.h"

void
Valuation::prettyPrint (unsigned indent) const
{
  print (indent, "Valuation {\n");
  if (isDivisionByZero ())
    print (indent + 1, "divisionByZero\n");
  if (isOverflow ())
    print (indent + 1, "overflow\n");
  if (isShiftError ())
    print (indent + 1, "shiftError\n");
  if (isConstraintViolation ())
    print (indent + 1, "constraintViolation\n");
  if (isUnionViolation ())
    print (indent + 1, "unionViolation\n");
  if (isNegativeCardinality ())
    print (indent + 1, "negativeCardinality\n");
  if (isLvalueViolation ())
    print (indent + 1, "lvalueViolation\n");
  if (isUnificationError ())
    print (indent + 1, "unificationError\n");
  if (isCapacityViolation ())
    print (indent + 1, "capacityViolation\n");
  if (isUndefinedValue ())
    print (indent + 1, "undefinedValue\n");
  if (isUndefinedComponent ())
    print (indent + 1, "undefinedComponent\n");
  if (isBufferViolation ())
    print (indent + 1, "bufferViolation\n");
  if (isUndefinedExpression ())
    print (indent + 1, "undefinedExpression\n");
  if (isFatalExpression ())
    print (indent + 1, "fatalExpression\n");

  for (const_iterator i = myValues.begin (); i != myValues.end (); i++) {
    i->first->prettyPrint (indent);
    i->second->prettyPrint (indent + 1);
  }
  if (myGlobalMarking)
    myGlobalMarking->prettyPrint (indent + 1);
  print (indent, "}\n");
}

void
ValuationSet::prettyPrint (unsigned indent) const
{
  for (const_iterator i = myValuations.begin (); i != myValuations.end (); i++)
    (*i)->prettyPrint (indent);
}

#endif // DEBUG

ValuationSet::ValuationSet () :
  myValuations ()
{
}

ValuationSet::~ValuationSet ()
{
  for (iterator i = myValuations.begin (); i != myValuations.end (); i++)
    delete *i;
}

void
ValuationSet::insert (class Valuation* v)
{
  std::pair<iterator, bool> status = myValuations.insert (v);
  if (!status.second)
    delete v;
}
