// Valuation (value distribution for variables) -*- c++ -*-

#ifndef VALUATION_H_
# define VALUATION_H_

# include <map>

/** Valuation (value distribution for variables) */
class Valuation {
  /** Pointer comparator */
  struct vptr {
    bool operator () (const class VariableDefinition* v1,
		      const class VariableDefinition* v2) const {
      return v1 < v2;
    }
  };
  typedef std::map<const class VariableDefinition*,class Value*,vptr> Map;
  typedef Map::iterator iterator;
  typedef Map::const_iterator const_iterator;

public:
  /** Error codes (combined by bitwise OR) */
  enum Error {
    noErrors = 0,
    divisionByZero = 1, overflow = 2, shiftError = 4,
    constraintViolation = 8, unionViolation = 16,
    negativeCardinality = 32, lvalueViolation = 64, unificationError = 128,
    capacityViolation = 256, undefinedValue = 512, undefinedComponent = 1024,
    bufferViolation = 2048,
    undefinedExpression = 4096, fatalExpression = 8192
  };

  // Constructors and destructors

  /** Constructs Valuation */
  Valuation ();
  /** Copy constructor */
  Valuation (const class Valuation& old);
private:
  /** Assignment operator */
  class Valuation& operator= (const class Valuation& old);
public:
  /** Destructs Valuation */
  ~Valuation ();

  /** Copies Valuation */
  class Valuation* copy () const { return new class Valuation (*this); }

  // Accessors

  /** Get the value of a variable
   * @param var		identifies the variable
   * @return		the value, or NULL if there is none
   */
  const class Value* getValue (const class VariableDefinition* var) const;

  /** Assign a value to a variable.  If the variable already has a value,
   * delete the old value.
   * @param var		identifies the variable
   * @param value	value to be assigned
   */
  void setValue (const class VariableDefinition* var, class Value* value);

  /** Get the global marking */
  const class GlobalMarking* getGlobalMarking () const {
    return myGlobalMarking;
  }
  /** Set the global marking
   * @param m		the new global marking
   */
  void setGlobalMarking (const class GlobalMarking* m) {
    myGlobalMarking = m;
  }

  /** Increase the value of a variable in the valuation
   * @param var		identifies the variable
   * @return		false if the value wrapped around; otherwise true
   */
  bool increment (const class VariableDefinition* var);
  /** Decrease the value of a variable in the valuation
   * @param var		identifies the variable
   * @return		false if the value wrapped around; otherwise true
   */
  bool decrement (const class VariableDefinition* var);

  /** Remove the value of a variable from the valuation
   * @param var		identifies the variable
   */
  void undefine (const class VariableDefinition* var);

  /** Determine whether there are any errors in the valuation */
  bool isOK () const { return myErrors == noErrors; }
  /** Determine whether there are any other errors than the one(s) specified */
  bool isOKExcept (enum Error errors) const { return !(myErrors & ~errors); }
  /** Clear the errors associated with the valuation */
  void clearErrors () const { myErrors = noErrors; }
  /** Copy the errors from one valuation to another */
  void copyErrors (const class Valuation& v) const { myErrors = v.myErrors; }

  /** Flag a division-by-zero error */
  void flagDivisionByZero () const {
    myErrors = static_cast<enum Error>(myErrors | divisionByZero);
  }
  /** Flag an overflow error */
  void flagOverflow () const {
    myErrors = static_cast<enum Error>(myErrors | overflow);
  }
  /** Flag a shift error */
  void flagShiftError () const {
    myErrors = static_cast<enum Error>(myErrors | shiftError);
  }
  /** Flag a constraint violation */
  void flagConstraintViolation () const {
    myErrors = static_cast<enum Error>(myErrors | constraintViolation);
  }
  /** Flag a union violation */
  void flagUnionViolation () const {
    myErrors = static_cast<enum Error>(myErrors | unionViolation);
  }
  /** Flag a negative cardinality error */
  void flagNegativeCardinality () const {
    myErrors = static_cast<enum Error>(myErrors | negativeCardinality);
  }
  /** Flag an lvalue violation */
  void flagLvalueViolation () const {
    myErrors = static_cast<enum Error>(myErrors | lvalueViolation);
  }
  /** Flag a unification error */
  void flagUnificationError () const {
    myErrors = static_cast<enum Error>(myErrors | unificationError);
  }
  /** Flag a capacity violation */
  void flagCapacityViolation () const {
    myErrors = static_cast<enum Error>(myErrors | capacityViolation);
  }
  /** Flag an "undefined value referenced" error */
  void flagUndefinedValue () const {
    myErrors = static_cast<enum Error>(myErrors | undefinedValue);
  }
  /** Flag an "undefined array component referenced" error */
  void flagUndefinedComponent () const {
    myErrors = static_cast<enum Error>(myErrors | undefinedComponent);
  }
  /** Flag a buffer violation */
  void flagBufferViolation () const {
    myErrors = static_cast<enum Error>(myErrors | bufferViolation);
  }
  /** Flag an UndefinedExpression error */
  void flagUndefinedExpression () const {
    myErrors = static_cast<enum Error>(myErrors | undefinedExpression);
  }
  /** Flag a fatal UndefinedExpression error */
  void flagFatalExpression () const {
    myErrors = static_cast<enum Error>(myErrors | fatalExpression);
  }

  /** Determine whether there is a division-by-zero error */
  bool isDivisionByZero () const {
    return true && myErrors & divisionByZero;
  }
  /** Determine whether there is an overflow error */
  bool isOverflow () const {
    return true && myErrors & overflow;
  }
  /** Determine whether there is a shift error */
  bool isShiftError () const {
    return true && myErrors & shiftError;
  }
  /** Determine whether there is a constraint violation */
  bool isConstraintViolation () const {
    return true && myErrors & constraintViolation;
  }
  /** Determine whether there is a union violation */
  bool isUnionViolation () const {
    return true && myErrors & unionViolation;
  }
  /** Determine whether there is a negative cardinality error */
  bool isNegativeCardinality () const {
    return true && myErrors & negativeCardinality;
  }
  /** Determine whether there is an lvalue violation */
  bool isLvalueViolation () const {
    return true && myErrors & lvalueViolation;
  }
  /** Determine whether there is a unification error */
  bool isUnificationError () const {
    return true && myErrors & unificationError;
  }
  /** Determine whether there is a capacity violation */
  bool isCapacityViolation () const {
    return true && myErrors & capacityViolation;
  }
  /** Determine whether there is an "undefined value referenced" error */
  bool isUndefinedValue () const {
    return true && myErrors & undefinedValue;
  }
  /** Determine whether there is an "undefined component referenced" error */
  bool isUndefinedComponent () const {
    return true && myErrors & undefinedComponent;
  }
  /** Determine whether there is a buffer violation */
  bool isBufferViolation () const {
    return true && myErrors & bufferViolation;
  }
  /** Determine whether there is an UndefinedExpression error */
  bool isUndefinedExpression () const {
    return true && myErrors & undefinedExpression;
  }
  /** Determine whether there is a fatal UndefinedExpression error */
  bool isFatalExpression () const {
    return true && myErrors & fatalExpression;
  }

  // Other methods

  /** Less-than comparison operator
   * @param other	object to compare to
   * @return		the comparison result
   */
  bool operator< (const class Valuation& other) const;

  /** Display this object
   * @param printer	The printer object
   */
  void display (const class Printer& printer) const;

# ifdef DEBUG
  /** Print this object
   * @param indent	level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members
private:
  /** The values sorted by variable names */
  Map myValues;
  /** Errors occurred while processing the valuation */
  mutable enum Error myErrors;
  /** The global marking (if available) */
  const class GlobalMarking* myGlobalMarking;
};

# include <set>

/** Valuation set */
class ValuationSet {
  /** Pointer comparator */
  struct vptr {
    bool operator () (const class Valuation* v1,
		      const class Valuation* v2) const {
      return *v1 < *v2;
    }
  };

  typedef std::set<class Valuation*,vptr> Set;
public:
  typedef Set::iterator iterator;
  typedef Set::const_iterator const_iterator;

  /** Constructs ValuationSet */
  ValuationSet ();
  /** Destructs ValuationSet */
  ~ValuationSet ();
private:
  /** Copy constructor */
  ValuationSet (const ValuationSet& old);
  /** Assignment operator */
  class ValuationSet& operator= (const class ValuationSet& old);
public:
  iterator begin () { return myValuations.begin (); }
  iterator end () { return myValuations.end (); }
  const_iterator begin () const { return myValuations.begin (); }
  const_iterator end () const { return myValuations.end (); }
  size_t size () const { return myValuations.size (); }
  bool empty () const { return myValuations.empty (); }
  void clear () {
    for (iterator i = begin (); i != end (); i++)
      delete *i;
    myValuations.clear ();
  }

  /** Insert a valuation to the set, provided that it does not already exist
   * @param v	valuation to be added (or deleted if it already exists)
   */
  void insert (class Valuation* v);

# ifdef DEBUG
  /** Print this object
   * @param indent	level of indentation
   */
  void prettyPrint (unsigned indent) const;
# endif // DEBUG

  // Members

private:
  /** The set of valuations */
  Set myValuations;
};

#endif // VALUATION_H_
