// Value base class -*- c++ -*-

#ifndef VALUE_H_
# define VALUE_H_
# include <cassert>
# include "typedefs.h"

/** Value base class */
class Value {
public:
  /** Value kinds (@see getKind) */
  enum Kind {
    vInt, vBool, vChar, vEnum, vId, vStruct, vUnion, vVector, vBuffer
  };

  // Constructors and destructors

  /** Constructs Value
   * @param type	Type of the Value
   */
  Value (const class Type* type);
private:
  /** Copy constructor */
  Value (const Value& old);
  /** Assignment operator */
  Value& operator= (const Value& old);
public:
  /** Destructs Value */
  virtual ~Value ();
  /** Copies Value */
  virtual class Value* copy () const = 0;

  // Accessors

  /** Determine the kind of the value */
  virtual enum Kind getKind () const = 0;
  /** Determine the type of the value */
  const class Type* getType () const { return myType; }
  /** Set the type of the value */
  void setType (const class Type* type) { myType = type; }

  // Other methods

  /** Less-than comparison */
  bool operator< (const class Value& other) const;
  /** Less-or-equal comparison */
  bool operator<= (const class Value& other) const;
  /** Equality comparison */
  bool operator== (const class Value& other) const;

  /** Difference operator
   * @param other	value to be substracted from this (may not be less)
   * @return		the difference
   */
  card_t operator- (const class Value& other) const;

  /** Reset the value to the first value */
  virtual void bottom () = 0;
  /** Reset the value to the last value */
  virtual void top () = 0;
  /** Get the next value
   * @return	false if the value wrapped around; otherwise true
   */
  virtual bool increment () = 0;
  /** Get the previous value
   * @return	false if the value wrapped around; otherwise true
   */
  virtual bool decrement () = 0;

  /** Convert the value to another type if possible
   * @param type	type to be casted to
   * @return		the converted value, or NULL
   */
  virtual class Value* cast (const class Type& type);

  /** Display this object
   * @param printer	The printer object
   */
  virtual void display (const class Printer& printer) const = 0;

# ifdef DEBUG
  /**
   * Print this object
   * @param indent	Level of indentation
   */
  virtual void prettyPrint (unsigned indent) const = 0;
# endif // DEBUG

  // Members

private:
  /** Type of the Value */
  const class Type* myType;
};

#endif // VALUE_H_
