// Vector value class -*- c++ -*-

#include "VectorValue.h"
#include "VectorType.h"
#include "Constraint.h"

VectorValue::VectorValue (const class Type* type) :
  Value (type),
  myComponents (static_cast<const class VectorType*>(type)->getSize ())
{
  assert (getType ()->getKind () == Type::tVector);
}

VectorValue::VectorValue (const class VectorValue& old) :
  Value (old.getType ()),
  myComponents (old.myComponents)
{
  assert (getType ()->getKind () == Type::tVector);
}

VectorValue::~VectorValue ()
{
}

bool
VectorValue::operator< (const class VectorValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = myComponents.getSize (); i--; )
    if (!myComponents[i]) {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (other.myComponents[i])
	return true;
    }
    else if (!other.myComponents[i]) {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (myComponents[i])
	return false;
    }
    else if (*myComponents[i] < *other.myComponents[i])
      return true;
    else if (*other.myComponents[i] < *myComponents[i])
      return false;

  return false;
}

bool
VectorValue::operator<= (const class VectorValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = myComponents.getSize (); i--; )
    if (!myComponents[i]) {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (other.myComponents[i])
	return true;
    }
    else if (!other.myComponents[i]) {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (myComponents[i])
	return false;
    }
    else if (*myComponents[i] < *other.myComponents[i])
      return true;
    else if (*other.myComponents[i] < *myComponents[i])
      return false;

  return true;
}

bool
VectorValue::operator== (const class VectorValue& other) const
{
  assert (other.myComponents.getSize () == myComponents.getSize ());

  for (card_t i = myComponents.getSize (); i--; )
    if (!myComponents[i]) {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (other.myComponents[i])
	return false;
    }
    else if (!other.myComponents[i]) {
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
      if (myComponents[i])
	return false;
    }
    else if (!(*myComponents[i] == *other.myComponents[i]))
      return false;

  return true;
}

card_t
VectorValue::operator- (const class VectorValue& other) const
{
  if (!getSize ())
    return 0;

  bool optional =
    static_cast<const class VectorType*>(getType ())->isOptional ();
  const class Type* itemType =
    static_cast<const class VectorType*>(getType ())->getItemType ();
  card_t numItemValues = itemType->getNumValues ();
  if (optional)
    numItemValues++;
  card_t diff = 0, i = getSize ();

  while (i--) {
    card_t d;
    if (myComponents[i] && other.myComponents[i])
      d = *myComponents[i] < *other.myComponents[i]
	? -(*other.myComponents[i] - *myComponents[i])
	: *myComponents[i] - *other.myComponents[i];
    else if (optional) {
      if (!myComponents[i] && !other.myComponents[i])
	d = 0;
      else if (myComponents[i])
	d = itemType->convert (myComponents[i]) + 1;
      else
	d = -(itemType->convert (other.myComponents[i]) + 1);
    }
    else
      assert (false), d = 0;
    (diff *= numItemValues) += d;
  }

  return diff;
}

void
VectorValue::bottom ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class VectorType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    if (const class Value* const v =
	getType ()->getConstraint ()->getFirstValue ()) {
      assert (v->getKind () == getKind ());
      const class VectorValue& vv = static_cast<const class VectorValue&>(*v);

      for (card_t i = getSize (); i--; ) {
	assert (vv[i] ||
		static_cast<const class VectorType*>(getType ())
		->isOptional ());
	delete myComponents[i];
	myComponents[i] = vv[i] ? vv[i]->copy () : NULL;
      }

      return;
    }
  }

  if (static_cast<const class VectorType*>(getType ())->isOptional ())
    for (card_t i = getSize (); i--; )
      delete myComponents[i], myComponents[i] = NULL;
  else
    for (card_t i = 0; i < getSize (); i++)
      myComponents[i]->bottom ();
}

void
VectorValue::top ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class VectorType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    if (const class Value* const v =
	getType ()->getConstraint ()->getLastValue ()) {
      assert (v->getKind () == getKind ());
      const class VectorValue& vv = static_cast<const class VectorValue&>(*v);

      for (card_t i = getSize (); i--; ) {
	assert (vv[i] ||
		static_cast<const class VectorType*>(getType ())
		->isOptional ());
	delete myComponents[i];
	myComponents[i] = vv[i] ? vv[i]->copy () : NULL;
      }

      return;
    }
  }

  for (card_t i = getSize (); i--; )
    delete myComponents[i], myComponents[i] =
      static_cast<const class VectorType*>(getType ())->getItemType ()
      ->getLastValue ();
}

bool
VectorValue::increment ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class VectorType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    const class Value* v = getType ()->getConstraint ()->getNextHigh (this);
    if (v) {
      assert (v->getKind () == getKind ());
      if (*this == static_cast<const class VectorValue&>(*v)) {
	v = getType ()->getConstraint ()->getNextLow (this);
	if (!v) {
	  bottom ();
	  return false;
	}
	assert (v->getKind () == getKind ());
	const class VectorValue* vv = static_cast<const class VectorValue*>(v);
	for (card_t i = getSize (); i--; ) {
	  assert ((*vv)[i] ||
		  static_cast<const class VectorType*>(getType ())
		  ->isOptional ());
	  delete myComponents[i];
	  myComponents[i] = (*vv)[i] ? (*vv)[i]->copy () : NULL;
	}
	return true;
      }
    }
  }

  if (static_cast<const class VectorType*>(getType ())->isOptional ()) {
    for (card_t i = 0; i < getSize (); i++) {
      if (!myComponents[i]) {
	myComponents[i] =
	  static_cast<const class VectorType*>(getType ())->getItemType ()
	  ->getFirstValue ();
	assert (myComponents[i] != NULL);
	return true;
      }
      if (myComponents[i]->increment ())
	return true;
      else
	delete myComponents[i], myComponents[i] = NULL;
    }
  }
  else
    for (card_t i = 0; i < getSize (); i++)
      if (myComponents[i]->increment ())
	return true;

  if (getType ()->getConstraint ())
    bottom ();

  return false;
}

bool
VectorValue::decrement ()
{
  assert (getType ()->isOrdered ());
  assert (getSize () ==
	  static_cast<const class VectorType*>(getType ())->getSize ());

  if (getType ()->getConstraint ()) {
    const class Value* v = getType ()->getConstraint ()->getPrevLow (this);
    if (v) {
      assert (v->getKind () == getKind ());
      if (*this == static_cast<const class VectorValue&>(*v)) {
	v = getType ()->getConstraint ()->getPrevHigh (this);
	if (!v) {
	  top ();
	  return false;
	}
	assert (v->getKind () == getKind ());
	const class VectorValue* vv = static_cast<const class VectorValue*>(v);
	for (card_t i = getSize (); i--; ) {
	  assert ((*vv)[i] ||
		  static_cast<const class VectorType*>(getType ())
		  ->isOptional ());
	  delete myComponents[i];
	  myComponents[i] = (*vv)[i] ? (*vv)[i]->copy () : NULL;
	}
	return true;
      }
    }
  }

  if (static_cast<const class VectorType*>(getType ())->isOptional ()) {
    for (card_t i = 0; i < getSize (); i++) {
      if (!myComponents[i]) {
	myComponents[i] =
	  static_cast<const class VectorType*>(getType ())->getItemType ()
	  ->getLastValue ();
	assert (myComponents[i] != NULL);
	return true;
      }
      if (myComponents[i]->decrement ())
	return true;
      else
	delete myComponents[i], myComponents[i] = NULL;
    }
  }
  else
    for (card_t i = 0; i < getSize (); i++)
      if (myComponents[i]->decrement ())
	return true;

  if (getType ()->getConstraint ())
    top ();

  return false;
}

class Value*
VectorValue::cast (const class Type& type)
{
  assert (getType ()->isAssignable (&type));
  if (type.getKind () != Type::tVector)
    return Value::cast (type);
  const class VectorType& vt = static_cast<const class VectorType&>(type);
  const class Type& itemType = *vt.getItemType ();
  assert (getSize () == vt.getSize ());

  for (card_t i = getSize (); i--; ) {
    if ((*this)[i] && !((*this)[i] = (*this)[i]->cast (itemType))) {
      delete this;
      return NULL;
    }
  }
  return Value::cast (type);
}

#include "Printer.h"
#include "CharValue.h"

void
VectorValue::display (const class Printer& printer) const
{
  card_t i;
  const class VectorType* const type =
    static_cast<const class VectorType*>(getType ());
  if (!type->isOptional () &&
      type->getItemType ()->getKind () == Type::tChar) {
    char* buf = new char[i = getSize ()];
    for (i = 0; i < getSize (); i++) {
      assert (myComponents[i] && myComponents[i]->getKind () == Value::vChar);
      buf[i] =
	static_cast<const class CharValue*>(myComponents[i])->getValue ();
    }
    printer.print (buf, i);
    delete[] buf;
  }
  else {
    printer.delimiter ('{')++;
    for (i = 0;;) {
      if (myComponents[i])
	myComponents[i]->display (printer);
      if (++i == getSize ())
	break;
      else
	printer.delimiter (',');
    }
    printer--.delimiter ('}');
  }
}

#ifdef DEBUG
# include "util.h"

void
VectorValue::prettyPrint (unsigned indent) const
{
  print (indent, "{\n");

  for (card_t i = 0;;) {
    if (myComponents[i]) {
      class Value* value =
	static_cast<const class VectorType*>(getType ())
	->getIndexType ()->convert (i);
      value->prettyPrint (indent + 1);
      delete value;
      print (indent + 1, "->\n");
      myComponents[i]->prettyPrint (indent + 1);
    }
    else
      assert (static_cast<const class VectorType*>(getType ())->isOptional ());
    if (++i == getSize ())
      break;
    else
      print (indent + 1, ",\n");
  }

  print (indent, "}\n");
}

#endif // DEBUG
