// Pretty-printing utility -*- c++ -*-

#include "Printer.h"
#include <cassert>
#include <cstdio>
#include <cstring>
#include <cctype>

void
Printer::finish () const
{
  assert (!myNesting);
  if (myColumn) {
    fputc ('\n', stderr);
    myColumn = 0;
  }
}

void
Printer::linebreak () const
{
  fputc ('\n', stderr);
  for (myColumn = 0; myColumn < myNesting; myColumn++)
    fputc (' ', stderr);
}

const class Printer&
Printer::delimiter (char character) const
{
  linebreak (1);
  fputc (character, stderr);
  myColumn++;
  return *this;
}

void
Printer::printRaw (const char* string) const
{
  unsigned i;
  for (i = 0; string[i]; i++);
  linebreak (i);
  fputs (string, stderr);
  myColumn += i;
}

void
Printer::print (bool value) const
{
  print (value ? "true" : "false");
}

/** Compute logarithm of base 16, rounded up
 * @param num		number whose logarithm is to be computed
 * @return		16-based logarithm of num
 */
inline static unsigned
log16 (card_t num)
{
  assert (num > 0);
  assert (!(CARD_T_BIT % 4));
  unsigned low = 0, high = CARD_T_BIT / 4;

  for (;;) {
    unsigned log = (low + high) >> 1;

    card_t exp = card_t (1) << (4 * log);
    if (exp < num) {
      assert (high > log);
      low = log + 1;
    }
    else if ((exp >> 4) >= num) {
      assert (low < log);
      high = log - 1;
    }
    else
      return log;
  }
}

/** Compute logarithm of base 8, rounded up
 * @param num		number whose logarithm is to be computed
 * @return		8-based logarithm of num
 */
inline static unsigned
log8 (card_t num)
{
  assert (num > 0);
  assert (!(CARD_T_BIT % 4));
  unsigned low = 0, high = CARD_T_BIT / 3;

  if (num > (card_t (1) << (3 * high)))
    return high + 1;

  for (;;) {
    unsigned log = (low + high) >> 1;

    card_t exp = card_t (1) << (3 * log);
    if (exp < num) {
      assert (high > log);
      low = log + 1;
    }
    else if ((exp >> 3) >= num) {
      assert (low < log);
      high = log - 1;
    }
    else
      return log;
  }
}

void
Printer::print (card_t number) const
{
  switch (myRadix) {
  case Octal:
    linebreak (number ? log8 (number) + 1 : 1);
    myColumn += fprintf (stderr, number ? "%o" : "0%o", number);
  case Hexadecimal:
    linebreak (number ? log16 (number) + 2 : 3);
    myColumn += fprintf (stderr, "0x%X", number);
    break;
  case Decimal:
    linebreak (number ? log8 (number) : 1);
    myColumn += fprintf (stderr, "%u", number);
    break;
  }
}

void
Printer::print (int_t number) const
{
  switch (myRadix) {
  case Octal:
    linebreak (number ? log8 (card_t (number)) + 1 : 1);
    myColumn += fprintf (stderr, number ? "%o" : "0%o", number);
  case Hexadecimal:
    linebreak (number ? log16 (card_t (number)) + 2 : 3);
    myColumn += fprintf (stderr, "0x%X", number);
    break;
  case Decimal:
    linebreak (number ? log8 (card_t (number < 0 ? -number : number)) : 1);
    myColumn += fprintf (stderr, "%d", number);
    break;
  }
}

void
Printer::print (const char* name) const
{
  bool printable = true;
  unsigned i;
  for (i = 0; name[i]; i++)
    if (!((name[i] >= 'a' && name[i] <= 'z') ||
	  (name[i] >= 'A' && name[i] <= 'Z') ||
	  (name[i] >= '0' && name[i] <= '9') ||
	  name[i] == '_'))
      printable = false;
  if (!i || !((name[0] >= 'a' && name[0] <= 'z') ||
	      (name[0] >= 'A' && name[0] <= 'Z') ||
	      name[0] == '_'))
    printable = false;
  if (printable) {
    linebreak (i);
    fputs (name, stderr);
    myColumn += i;
  }
  else
    print (name, i);
}

void
Printer::print (const char* string, unsigned length) const
{
  *myBuffer.create (1) = '"';
  unsigned j = 1;
  for (unsigned i = 0; i < length; i++) {
    unsigned char c = static_cast<unsigned char>(string[i]);
    switch (c) {
    case '\f':
      memcpy (myBuffer.extend (2) + j, "\\f", 2), j += 2;
      break;
    case '\n':
      memcpy (myBuffer.extend (2) + j, "\\n", 2), j += 2;
      break;
    case '\r':
      memcpy (myBuffer.extend (2) + j, "\\r", 2), j += 2;
      break;
    case '\t':
      memcpy (myBuffer.extend (2) + j, "\\t", 2), j += 2;
      break;
    case '\v':
      memcpy (myBuffer.extend (2) + j, "\\v", 2), j += 2;
      break;
    case '\b':
      memcpy (myBuffer.extend (2) + j, "\\b", 2), j += 2;
      break;
    case '\a':
      memcpy (myBuffer.extend (2) + j, "\\a", 2), j += 2;
      break;
    case 0:
      memcpy (myBuffer.extend (2) + j, "\\0", 2), j += 2;
      break;
    default:
      if (iscntrl (c)) {
	unsigned l = log8 (c) + 1;
	if (l < 4 && i < length &&
	    string[i + 1] >= '0' && string[i + 1] <= '7')
	  sprintf (myBuffer.extend (4) + j, "\\%03o", c), j += 4;
	else
	  sprintf (myBuffer.extend (l) + j, "\\%o", c), j += l;
      }
      else
	myBuffer.extend (1)[j++] = c;
      break;
    }
  }

  memcpy (&myBuffer.extend (1)[j++], "\"", 2);
  linebreak (j);
  fputs (myBuffer.getString (), stderr);
  myColumn += j;
}

void
Printer::print (char_t character) const
{
  unsigned char c = static_cast<unsigned char>(character);
  unsigned j = 1;
  *myBuffer.create (1) = '\'';
  switch (c) {
  case '\f':
    memcpy (myBuffer.extend (2) + j, "\\f", 2), j += 2;
    break;
  case '\n':
    memcpy (myBuffer.extend (2) + j, "\\n", 2), j += 2;
    break;
  case '\r':
    memcpy (myBuffer.extend (2) + j, "\\r", 2), j += 2;
    break;
  case '\t':
    memcpy (myBuffer.extend (2) + j, "\\t", 2), j += 2;
    break;
  case '\v':
    memcpy (myBuffer.extend (2) + j, "\\v", 2), j += 2;
    break;
  case '\b':
    memcpy (myBuffer.extend (2) + j, "\\b", 2), j += 2;
    break;
  case '\a':
    memcpy (myBuffer.extend (2) + j, "\\a", 2), j += 2;
    break;
  case 0:
    memcpy (myBuffer.extend (2) + j, "\\0", 2), j += 2;
    break;
  default:
    if (iscntrl (c)) {
      unsigned l = log8 (c) + 1;
      sprintf (myBuffer.extend (l) + j, "\\%o", c), j += l;
    }
    else
      myBuffer.extend (1)[j++] = c;
    break;
  }
  memcpy (&myBuffer.extend (1)[j++], "'", 2);
  linebreak (j);
  fputs (myBuffer.getString (), stderr);
  myColumn += j;
}
