// Pretty-printing utility -*- c++ -*-

#ifndef PRINTER_H_
# define PRINTER_H_

# include "typedefs.h"
# include "StringBuffer.h"

/** Pretty-printing utility */
class Printer {
public:
  /** Verbosity level */
  enum Verbosity { Terse, Verbose };
  /** Radix for displaying numbers */
  enum Radix { Octal, Hexadecimal, Decimal };

  /** Constructor
   * @param width	maximum width of output in columns
   * @param radix	radix for displaying numbers
   * @param verbosity	verbosity level
   */
  Printer (unsigned width = 80,
	   enum Radix radix = Decimal,
	   enum Verbosity verbosity = Terse) :
    myWidth (width), myRadix (radix), myVerbosity (verbosity),
    myNesting (0), myColumn (0), myBuffer () {}
private:
  /** Copy constructor */
  Printer (const class Printer& old);
  /** Assignment operator */
  class Printer& operator= (const class Printer& old);
public:
  /** Destructor */
  ~Printer () { finish (); }

  /** Set the maximum width of the output */
  void setWidth (unsigned width) { myWidth = width; }
  /** Get the maximum width of the output */
  unsigned getWidth () const { return myWidth; }
  /** Set the radix */
  void setRadix (enum Radix radix) { myRadix = radix; }
  /** Get the radix */
  enum Radix getRadix () const { return myRadix; }
  /** Set the verbosity level */
  void setVerbosity (enum Verbosity verbosity) { myVerbosity = verbosity; }
  /** Get the verbosity level */
  enum Verbosity getVerbosity () const { return myVerbosity; }

  /** Increment the nesting level */
  const class Printer& operator++ () const { myNesting++; return *this; }
  /** Increment the nesting level */
  const class Printer& operator++ (int postfix) const {
    myNesting++; return *this;
  }
  /** Decrement the nesting level */
  const class Printer& operator-- () const { myNesting--; return *this; }
  /** Decrement the nesting level */
  const class Printer& operator-- (int postfix) const {
    myNesting--; return *this;
  }

  /** Finish printing (add a newline if necessary) */
  void finish () const;
  /** Add an unconditional line break */
  void linebreak () const;
  /** Break the line if necessary
   * @param margin	safety margin required
   */
  void linebreak (unsigned margin) const {
    if (myColumn + margin >= myWidth)
      linebreak ();
  }

  /** Display a delimiter character
   * @param character	delimiter character to be displayed
   * @return		*this
   */
  const class Printer& delimiter (char character) const;

  /** Display raw text
   * @param string	NUL-terminated string to be displayed
   */
  void printRaw (const char* string) const;

  /** Display a boolean value
   * @param value	value to be displayed
   */
  void print (bool value) const;
  /** Display a cardinality
   * @param number	number to be displayed
   */
  void print (card_t number) const;
  /** Display a signed number
   * @param number	number to be displayed
   */
  void print (int_t number) const;
  /** Display a name
   * @param name	NUL-terminated name to be displayed
   */
  void print (const char* name) const;
  /** Display a character string in double quotes
   * @param string	characters to be displayed
   * @param length	length of the string
   */
  void print (const char* string, unsigned length) const;
  /** Display a character in single quotes
   * @param character	character to be displayed
   */
  void print (char_t character) const;

private:
  /** maximum width of output in columns */
  unsigned myWidth;
  /** radix for displaying numbers */
  enum Radix myRadix;
  /** verbosity level */
  enum Verbosity myVerbosity;
  /** nesting depth of output */
  mutable unsigned myNesting;
  /** current output column */
  mutable unsigned myColumn;
  /** buffer for building up strings */
  mutable class StringBuffer myBuffer;
};

#endif // PRINTER_H_
