/* Command-line interface for lexically C-like languages -*- c -*- */

#include "cmdline.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAS_READLINE
# include <readline.h>
# include <history.h>
#else /* HAS_READLINE */
/** for error () */
static const char* rl_readline_name = "command:";
/** Read a line from the standard input
 * @param prompt	Prompt to be displayed
 * @return		a line of text
 */
static char* readline (const char* prompt)
{
  size_t length = 0;
  char* buf = malloc (BUFSIZ);
  fputs (prompt, stderr);
  while (fgets (buf + length, BUFSIZ, stdin)) {
    char* last = buf + length;
    while (*last) last++;
    if (last > buf + length && last[-1] == '\n') {
      last[-1] = 0;
      return buf;
    }
    else {
      char* buf2 = realloc (buf, (length += last - buf) + BUFSIZ);
      if (!buf2)
        return buf;
      buf = buf2;
    }
  }

  if (length)
    return buf;
  else {
    free (buf);
    return NULL;
  }
}

/** Add a line to the command line history
 * @param line		line to be added
 */
static void
add_history (const char* line)
{
}
#endif /* HAS_READLINE */

/** The string holding the command line */
static char* line = NULL;

/** Report a syntax error on the command line
 * @param msg		message to be reported
 */
static void error (const char* msg) {
  fprintf (stderr, "\n%s: %s\n", rl_readline_name, msg);
}

/** States of the input buffer */
enum state {
  normal, single_quote, double_quote, squote_backslash, dquote_backslash,
  backslash, comment, scomment
};

/** Determine the state of the input buffer
 * @param s		current state (input/output)
 * @param line		command line to be scanned
 * @param length	length of the command line
 * @return		state of line[length]
 */
static enum state
scanstate (enum state s, const char* line, unsigned length)
{
  unsigned i;

  for (i = 0; i < length; i++) {
    switch (line[i]) {
    case '\'':
      switch (s) {
      case normal:
	s = single_quote;
	break;
      case single_quote:
	s = normal;
	break;
      default:
	break;
      }
      break;
    case '"':
      switch (s) {
      case normal:
	s = double_quote;
	break;
      case double_quote:
	s = normal;
	break;
      default:
	break;
      }
      break;
    case '\\':
      switch (s) {
      case normal:
	s = backslash;
	break;
      case single_quote:
	s = squote_backslash;
	break;
      case double_quote:
	s = dquote_backslash;
	break;
      case squote_backslash:
	s = single_quote;
	break;
      case dquote_backslash:
	s = double_quote;
	break;
      case backslash:
	s = normal;
	break;
      case comment:
      case scomment:
	break;
      }
      break;
    case '/':
      if (s == normal) {
	if (line[i + 1] == '*') {
	  s = comment;
	  break;
	}
	else if (line[i + 1] == '/') {
	  s = scomment;
	  break;
	}
      }
      else if (s == comment && i > 0 && line[i - 1] == '*') {
	s = normal;
	break;
      }
      /* fall through */
    default:
      switch (s) {
      case squote_backslash:
	s = single_quote;
	break;
      case dquote_backslash:
	s = double_quote;
	break;
      case backslash:
	s = normal;
	break;
      default:
	break;
      }
    }
  }

  return s;
}

/** Determine the state of the input buffer
 * @param s		current state (input/output)
 * @param line		command line to be scanned (continuation or new line)
 * @return		strlen (line)
 */
static unsigned
getstate (enum state* s, const char* line) {
  unsigned length = strlen (line);
  if ((*s = scanstate (*s, line, length)) == scomment)
    *s = normal;
  return length;
}

const char* cmdline (const char* prompt, const char* prompt2)
{
  if (line) {
    free (line);
    line = NULL;
  }

  if ((line = readline (prompt)) && *line) {
    enum state s = normal;
    unsigned i = getstate (&s, line);
    line[i] = 0;
    while (s != normal) {
      char* cont = readline (prompt2);
      if (!cont) {
	switch (s) {
	case single_quote:
	case squote_backslash:
	  error ("unexpected EOF while looking for matching `''");
	  break;
	case double_quote:
	case dquote_backslash:
	  error ("unexpected EOF while looking for matching `\"'");
	  break;
	case comment:
	  error ("unexpected EOF while looking for matching `*/'");
	  break;
	case backslash:
	  error ("unexpected EOF after `\\'");
	  break;
	case normal:
	case scomment:
	  fputc ('\n', stderr);
	  break;
	}

	if (s == normal)
	  break;
	else {
	  free (line);
	  line = NULL;
	  return "";
	}
      }
      else {
	unsigned j;
	char* newline;
	switch (s) {
	case normal:
	case comment:
	case single_quote:
	case double_quote:
	  line[i++] = '\n';
	  break;

	case squote_backslash:
	  s = single_quote;
	  i--;
	  break;
	case dquote_backslash:
	  s = double_quote;
	  i--;
	  break;
	case backslash:
	  s = normal;
	  /* fall through */
	default:
	  i--;
	  break;
	}

	j = getstate (&s, cont);
	cont[j] = 0;

	if ((newline = realloc (line, i + j + 1))) {
	  memcpy ((line = newline) + i, cont, j + 1);
	  i += j;
	  free (cont);
	}
	else {
	  free (line);
	  free (cont);
	  line = NULL;
	  return "";
	}
      }
    }

    add_history (line);
  }
  else if (!line)
    fputc ('\n', stderr);

  return line;
}

#ifdef HAS_READLINE
/** Determine whether a character in the command line is quoted
 * @param text		the command line
 * @param index		index to the character of interest
 */
static int
char_is_quoted (const char* text, int index)
{
  return scanstate (normal, text, index) != normal;
}

/** Quote a string using double quotes
 * @param string	string to be quoted
 * @param rtype		flag: SINGLE_MATCH or MULT_MATCH
 * @param qcp		pointer to the quote character
 * @return		a new string
 */
static char*
quote_filename (const char* string, int rtype, char *qcp)
{
  register int c;
  char *result, *r;

  if (*qcp != '"')
    *qcp = '"';

  result = (char *)malloc (3 + (2 * strlen (string)));
  r = result;
  *r++ = '"';

  for (; (c = *string); string++) {
    switch (c) {
    case '"':
    case '\\':
      *r++ = '\\';
      /* fall through */
    default:
      *r++ = c;
      break;
    }
  }

  if (rtype != MULT_MATCH)
    *r++ = '"';
  *r = '\0';

  return result;
}

void
init_cmdline (const char* progname,
	      char** (*attempted_completer) (const char*, int, int),
	      char* (*entry_completer) (const char*, int))
{
  rl_readline_name = (char*) progname;
  rl_attempted_completion_function = attempted_completer;
  rl_completion_entry_function = (Function*) entry_completer;
  rl_completer_quote_characters = "'\"";
  rl_filename_quote_characters =
    "\1\2\3\4\5\6\7\10\11\12\13\14\15\16\17"
    "\20\21\22\23\24\25\26\27\30\31\32\33\34\35\36\37"
    " !\"#$%&'()*+,-./[\\]^`{|}~\177"
    "\200\201\202\203\204\205\206\207"
    "\210\211\212\213\214\215\216\217"
    "\220\221\222\223\224\225\226\227"
    "\230\231\232\233\234\235\236\237"
    "\240\241\242\243\244\245\246\247"
    "\250\251\252\253\254\255\256\257"
    "\260\261\262\263\264\265\266\267"
    "\270\271\272\273\274\275\276\277"
    "\300\301\302\303\304\305\306\307"
    "\310\311\312\313\314\315\316\317"
    "\320\321\322\323\324\325\326\327"
    "\330\331\332\333\334\335\336\337"
    "\340\341\342\343\344\345\346\347"
    "\350\351\352\353\354\355\356\357"
    "\360\361\362\363\364\365\366\367"
    "\370\371\372\373\374\375\376\377";
  rl_basic_word_break_characters =
    " \t\n!\"#$%&'()*+,-./:;<=>?@[]^`{|}~";
  rl_basic_quote_characters = "\"'";
  rl_char_is_quoted_p = char_is_quoted;
  rl_filename_quoting_function = quote_filename;
}
#endif /* HAS_READLINE */
