#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <clocale>

#include "version.h"

#include <getopt.h>
#ifdef HAS_GETOPT_LONG
# define getopt(argc,argv,optstring) getopt_long(argc,argv,optstring,opts,0)
static const struct option opts[] = {
  { "include", 1, 0, 'I' },
  { "help", 0, 0, '?' },
  { "version", 0, 0, 'V' },
  { "warnings", 0, 0, 'W' },
  { "no-warnings", 0, 0, 'w' },
  { "execute", 1, 0, 'e' },
  { "file", 1, 0, 'f' },
  { "width", 1, 0, 'i' },
  { "radix", 1, 0, 'x' },
  { 0, 0, 0, 0 }
};
#endif // HAS_GETOPT_LONG

/** Option string for getopts(3) */
static const char optstring[] = "I:?hVwWe:f:i:x:";

#include "util.h"
#include "NameList.h"
static class NameList includePath;
#include "Printer.h"

#include "Net.h"
#include "Graph.h"

#ifndef YYDEBUG
# define YYDEBUG 0
#endif

#if YYDEBUG
/** Debugging flag for the Bison-generated parser */
extern int dedebug;
#endif // YYDEBUG

/** The Bison-generated Petri Net parser
 * @return	zero if exiting due to end-of-file; 1 if parse error occurred
 */
extern int pnparse ();

/** The Bison-generated scripting language parser
 * @return	zero if exiting due to end-of-file; 1 if parse error occurred
 */
extern int deparse ();
/** Read a command line */
extern const char* deline ();
/** Reset the scripting language parser */
extern void dereset ();

/** warning flag (true=print warnings) */
extern bool dewarnings;
/** number of errors encountered while parsing the Petri Net */
extern unsigned pnerrors;
/** number of errors encountered while parsing the script */
extern unsigned deerrors;

/** The printer object */
class Printer thePrinter;
/** The Petri Net being processed */
class Net* net = NULL;
/** The reachability graph being processed */
class Graph* graph = NULL;

/** Name of the Petri Net being parsed */
const char* netname;
/** Name of Petri Nets parsed from the command line */
const char* commandline = "(command line)";
/** Petri Net model to be parsed from the command line */
const char* netstring;

/** Name of the script being parsed */
const char* scriptname;
/** Name of scripts parsed from the command line */
const char* scriptcommand = commandline;
/** The script to be parsed from the command line */
const char* scriptstring;

/** Open an input file
 * @param name		name of the input file
 * @return		a descriptor to the file, or NULL on error
 */
FILE* openfile (const char* name);
/** Open a script
 * @param name		name of the script
 * @return		a descriptor to the file, or NULL on error
 */
FILE* openscript (const char* name);

/** Execute a script
 * @param filename	name of the script to be executed
 * @return		false on a fatal error, true otherwise
 */
static bool execute (const char* filename);
/** Flag: has the analysis been interrupted? */
bool interrupted = false;
/** Handle a signal
 * @param num		number of the signal
 */
extern "C" void sig (int num) {
  interrupted = true;
}
#include <csignal>

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful;
 *		1 on parse error
 */
int main (int argc, char** argv);

int main (int argc, char** argv)
{
  /** Name of the reachability graph file (NULL=use default) */
  const char* graphfile = NULL;
  setlocale (LC_ALL, "");
  signal (SIGINT, sig);

#if YYDEBUG
  {
    char *debug = getenv ("DEBUG");
    if (debug) dedebug = atoi (debug);
  }
#endif // YYDEBUG

  for (;;) {
    int c = getopt (argc, argv, optstring);
    if (c == EOF)
      break;
    switch (c) {
    case ':':
      // fall through
    default:
    case '?':
    case 'h':
      fprintf (stderr, "Usage: %s [options] file...\n", *argv);
      return 0;
    case 'V':
      fprintf (stderr, "%s version " VERSION "\n", *argv);
      return 0;
    case 'w':
      dewarnings = false;
      break;
    case 'W':
      dewarnings = true;
      break;
    case 'i':
      {
	char* end = NULL;
	unsigned width = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in output width limit: `%s'\n", optarg);
	  return 1;
	}
	else
	  thePrinter.setWidth (width);
      }
      break;
    case 'x':
      {
	if (!strcmp (optarg, "oct") || !strcmp (optarg, "octal") ||
	    !strcmp (optarg, "8"))
	  thePrinter.setRadix (Printer::Octal); 
	else if (!strcmp (optarg, "hex") || !strcmp (optarg, "hexadecimal") ||
		 !strcmp (optarg, "16"))
	  thePrinter.setRadix (Printer::Hexadecimal);
	else if (!strcmp (optarg, "dec") || !strcmp (optarg, "decimal") ||
		 !strcmp (optarg, "10"))
	  thePrinter.setRadix (Printer::Decimal);
	else
	  fprintf (stderr, "ignoring unknown radix setting: `%s'\n", optarg);
      }
      break;
    case 'I':
      includePath.append (newString (optarg));
      break;
    case 'e':
      scriptstring = optarg;
      break;
    case 'f':
      graphfile = optarg;
      break;
    }
  }

  if (!graphfile) {
    fprintf (stderr, "%s:no graph file specified\n", *argv);
    return 1;
  }
  
  /** The Petri Net */
  class Net theNet;
  /** The reachability graph */
  class Graph theGraph (theNet, 0, 0, 0, graphfile);
  if (!theGraph.openFiles (false)) {
    fprintf (stderr, "%s:could not open the graph files\n", *argv);
    return 2;
  }

  pnerrors = 0;
  netname = theGraph.getFilename ();
  net = &theNet;
  graph = &theGraph;
  if (pnparse () || pnerrors)
    return 3;

  for (int i = optind; i < argc; i++)
    if (!execute (argv[i]))
      return 4;

  dereset ();
  extern bool exiting;
  do if (scriptstring) execute (scriptcommand);
  while (!exiting && (scriptstring = deline ()));
}

static bool
execute (const char* filename)
{
  extern bool exiting;
  scriptname = filename;
  dereset ();
  return !deparse () && !deerrors && !exiting;
}

/** Open a file
 * @param name		name of the file
 * @param origfile	name of the original file
 * @param includes	include directories
 * @return		handle to the file, or NULL in case of error
 */
inline static FILE*
openfile (const char* name,
	  const char* origfile,
	  const class NameList& includes)
{
  FILE* f;
  // first try from the current directory
  if (!(f = fopen (name, "r"))) {
    // next, try from the directory of the original net

    size_t dirlength;
    char* fullname;
    size_t namelength = strlen (name);
    for (dirlength = 0; origfile[dirlength]; dirlength++);
    while (dirlength > 0 && origfile[dirlength] != '/')
      dirlength--;
    if (dirlength > 0) {
      fullname = new char[namelength + dirlength + 2];
      memcpy (fullname, origfile, dirlength + 1);
      memcpy (fullname + dirlength + 1, name, namelength + 1);
      f = fopen (fullname, "r");
      delete[] fullname;
      if (f)
	return f;
    }

    for (NameList::const_iterator i = includes.begin ();
	 i != includes.end (); i++) {
      const char* dirname = *i;
      size_t dirlength = strlen (dirname);
      char* fullname = new char[namelength + dirlength + 2];
      memcpy (fullname, dirname, dirlength);
      fullname[dirlength] = '/';
      memcpy (fullname + dirlength + 1, name, namelength + 1);
      f = fopen (fullname, "r");
      delete[] fullname;
      if (f)
	break;
    }
  }

  return f;
}

FILE*
openfile (const char* name)
{
  return openfile (name, netname, includePath);
}

FILE*
openscript (const char* name)
{
  return openfile (name, scriptname, includePath);
}
