/* -*- c++ -*- Scanner for the Maria Debugger */

%option nostdinit
%option stack
%option prefix="de"
%option 8bit
%option caseful
%option warn
%option pointer
%option never-interactive
%option nodefault
%option noinput
%option nounput
%option noyywrap
%option noyy_top_state
%option noyy_scan_buffer
%option noyy_scan_bytes
%option noyy_scan_string

%{
# include <cstdio>
# include <climits>
# include <cstring>
# include <cstdarg>
# include <cassert>

# include "typedefs.h"
# include "marde.tab.h"

# include "StringBuffer.h"
# include "util.h"

# include <list>
# include <stack>
  /* Stack of include files */
  struct input {
    /** name of the input file */
    class StringBuffer name;
    /** input line number */
    unsigned lineno;
    /** handle to the input file */
    FILE* file;
    /** buffer state */
    YY_BUFFER_STATE bs;
  };
  static std::stack<struct input, std::list<struct input> > in;

  /* Dynamically growing string buffer */
  static class StringBuffer sbuf;
  /* String handling macros.  All of these use sbuf. */
# define CREAT memcpy (sbuf.create (yyleng), yytext, yyleng + 1)
# define INITSTR sbuf.create (0)
# define EXTEND memcpy (sbuf.extend (yyleng) + sbuf.getLength () - yyleng, \
			yytext, yyleng + 1)
# define APPEND(c) (&((sbuf.extend (1))[sbuf.getLength () - 1] = c))[1] = 0
# define COPY delval.s = newString (sbuf.getString ())
# define COPYBUF delval.sbuf = new class StringBuffer (sbuf)

  /* Start condition stack handling */
# define SC_FLUSH	yy_start_stack_ptr = 0, BEGIN (INITIAL)
# define SC_POP		yy_pop_state ()
# define SC_POP_B	do { SC_POP; yyless (0); } while (0)
# define SC_PUSH(s)	yy_push_state (s)
# define SC_SWITCH(s)	BEGIN (s)
# define SC_SWITCH_B(s)	do { yyless (0); SC_SWITCH (s); } while (0)
# define SC_CALL(s,r)	do { rt = r; SC_PUSH (ret); SC_PUSH (s); } while (0)
# define SC_USE(s,r)	do { SC_SWITCH (r); SC_PUSH (s); } while (0)

  /** initial source file name */
  extern const char* scriptname;
  /** Name of Petri Nets parsed from the command line */
  extern const char* scriptcommand;
  /** Petri Net model to be parsed from the command line */
  extern const char* scriptstring;
  /** Cursor to scriptstring */
  static const char* stringptr = NULL;

  /** flag: expecting a command keyword */
  extern bool command;

# define RETURN(token)	return command = false, token
# define RETURNC(token)							\
  return command ? (command = false, token) : (CREAT, COPY, NAME)
  /** source file line number */
  static unsigned lineno = 1;
  /** source file name */
  static class StringBuffer filename;

  /** Token to be returned to the parser */
  static int rt;
  /** Flag: input constant overflow */
  static bool overflow = false;
  /** Number of digits read in an octal or hexadecimal character constant */
  static unsigned digits = 0;
  /** Character constant being parsed in a string */
  static char character;

  /** Reset the lexical analyzer */
  void delexreset ();

  /** warning flag (true=print warnings) */
  bool dewarnings = true;
  /** number of errors encountered */
  unsigned deerrors = 0;

  /**
   * generic warning reporting function
   * @param s		warning message
   */
  void dewarn (const char* s);

  /**
   * generic printf-like warning reporting function
   * @param fmt		printf format string
   * @param ...		variable arguments
   */
  void dewarnf (const char* fmt, ...)
# ifdef __GNUC__
    __attribute__ ((format (printf, 1, 2)))
# endif /* __GNUC__ */
    ;

  /**
   * generic parser error reporting function
   * @param s		error message
   */
  void deerror (const char* s);

  /**
   * generic printf-like parser error reporting function
   * @param fmt		printf format string
   * @param ...		variable arguments
   */
  void deerrorf (const char* fmt, ...)
# ifdef __GNUC__
    __attribute__ ((format (printf, 1, 2)))
# endif /* __GNUC__ */
    ;

  /** Open a script
   * @param name	name of the script
   * @return		a descriptor to the file, or NULL on error
   */
  extern FILE* openscript (const char* name);
  /** Switch files on EOF (return to the file that included the current file)
   * @return	true if there was a parent file
   */
  static bool popInclude ();

# define NEXTBUF do if (!popInclude ()) { SC_FLUSH; yyterminate (); } while (0)
# define YY_USER_INIT NEXTBUF

# define YY_INPUT(buf,result,max_size)					  \
  if (!dein || dein == stdin) {						  \
    if (!stringptr)							  \
      result = 0;							  \
    else if (*stringptr) {						  \
      if ((result = strlen (stringptr)) > max_size)			  \
	result = max_size;						  \
      memcpy (buf, stringptr, result);					  \
      stringptr += result;						  \
    }									  \
    else if (max_size >= 1)						  \
      *buf = '\n', stringptr = NULL, result = 1;			  \
    else								  \
      result = 0;							  \
  }									  \
  else if (!(result = fread (buf, 1, max_size, dein)) && ferror (dein)) { \
    deerror ("cannot read file"); NEXTBUF;				  \
  }
%}

/* Start conditions */
/* returning to the parser */
%x ret
/* inside a comment */
%x comment

/* parsing a preprocessor directive */
%x pr
/* parsing a preprocessor directive, ignoring the rest of the line */
%x pr_ignore
/* parsing a preprocessor directive "line" */
%x pr_line
/* parsing a preprocessor directive "line", got line number */
%x pr_line_number
/* parsing a preprocessor directive "line", got file name */
%x pr_line_name
/* parsing a preprocessor directive "include" */
%x pr_incl
/* parsing a preprocessor directive "include", got file name */
%x pr_incl_name

/* parsing an octal constant */
%x octal
/* parsing a hexadecimal constant */
%x hexadecimal
/* parsing a decimal constant */
%x decimal

/* parsing a name */
%x dename
/* parsing a character constant */
%x dechar
/* parsing an octal character constant */
%x dechar_oct
/* parsing a hexadecimal character constant */
%x dechar_hex
/* parsing a string constant */
%x destring
/* parsing an octal character constant in a string */
%x destring_oct
/* parsing a hexadecimal character constant in a string */
%x destring_hex

ws		[ \t\r\v\f]+
nn		[^\\A-Za-z0-9_]

%%

<ret>.|\n	{ SC_POP_B; if (rt == STRING) COPYBUF; RETURN (rt); }
<ret><<EOF>>	{ if (rt == STRING) COPYBUF; SC_POP; NEXTBUF; RETURN (rt); }

<comment>{
  [^*\n]*
  [^*\n]*\n		lineno++;
  "*"+[^*/\n]*
  "*"+[^*/\n]*\n	lineno++;
  "*"+"/"		SC_POP;
  <<EOF>>		{ deerror ("EOF in comment"); SC_POP; NEXTBUF; }
}

<pr,pr_line,pr_line_number,pr_line_name,pr_incl,pr_incl_name>{
  {ws}+
  "//"[^\n]*
  "/*"		SC_PUSH (comment);
  <<EOF>>	{ deerror ("EOF in preprocessor directive"); SC_POP; NEXTBUF; }
}

<pr>{
  [A-Za-z_][A-Za-z0-9_]* {
    if (!strcmp (yytext, "line"))
      SC_SWITCH (pr_line);
    else if (!strcmp (yytext, "include"))
      SC_SWITCH (pr_incl);
    else {
      deerrorf ("unknown preprocessor directive #%s", yytext);
      SC_SWITCH (pr_ignore);
    }
  }
  !			{
    SC_SWITCH (pr_ignore); /* #! interpreter command */
  }
  .			{
    deerror ("error in preprocessor directive"); SC_SWITCH (pr_ignore);
  }
  \n			{
    deerror ("expected a preprocessor directive"); SC_POP; lineno++;
  }
}

<pr_ignore>{
  [^\n]+		SC_POP;
  \n			{ SC_POP; lineno++; }
  <<EOF>>		{ SC_POP; NEXTBUF; }
}

<pr_line>{
  0x		{
    delval.i = 0, overflow = false;
    SC_USE (hexadecimal, pr_line_number);
  }
  0		{
    delval.i = 0, overflow = false;
    SC_USE (octal, pr_line_number);
  }
  [1-9]		{
    delval.i = *yytext - '0', overflow = false;
    SC_USE (decimal, pr_line_number);
  }
  \"			{ INITSTR; SC_USE (destring, pr_line_name); }
  \n			{ lineno++; SC_POP; }
  .			{
    deerror ("error in preprocessor directive #line");
    SC_SWITCH (pr_ignore);
  }
}

<pr_line_number>.|\n	{
  SC_SWITCH_B (pr_line);
  lineno = delval.i - 1;
}

<pr_line_name>.|\n	{ SC_SWITCH_B (pr_line); filename = sbuf; }

<pr_incl>{
  \"			{ INITSTR; SC_USE (destring, pr_incl_name); }
  \n			{ lineno++; SC_POP; }
  .			{
    deerror ("error in preprocessor directive #include");
    SC_SWITCH_B (pr_ignore);
  }
}

<pr_incl_name>.|\n	{
  SC_POP_B;
  if (FILE* f = openscript (sbuf.getString ())) {
    struct input is = { filename, lineno, dein, YY_CURRENT_BUFFER };
    in.push (is);
    filename = sbuf;
    lineno = 1;
    dein = f;
    yy_switch_to_buffer (yy_create_buffer (dein, YY_BUF_SIZE));
  }
  else
    deerrorf ("file `%s' not found", sbuf.getString ());
}

{ws}+
\n		lineno++;
"/*"		SC_PUSH (comment);
"//"[^\n]*
^#		SC_PUSH (pr);

"union"/{nn}		RETURN (UNION);
"is"/{nn}		RETURN (IS);
"always"/{nn}		RETURN (ALWAYS);
"exists"/{nn}		RETURN (EXISTS);
"until"/{nn}		RETURN (UNTIL);
"release"/{nn}		RETURN (RELEASE);
"subset"/{nn}		RETURN (SUBSET);
"intersect"/{nn}	RETURN (INTERSECT);
"equals"/{nn}		RETURN (EQUALS);
"minus"/{nn}		RETURN (MINUS);
"atom"/{nn}		RETURN (ATOM);
"empty"/{nn}		RETURN (EMPTY);

"place"/{nn}		RETURN (PLACE);
"trans"/{nn}		RETURN (TRANS);
"true"/{nn}		RETURN (TRUE_);
"false"/{nn}		RETURN (FALSE_);
"undefined"/{nn}	RETURN (UNDEFINED);
"fatal"/{nn}		RETURN (FATAL);

"cardinality"/{nn}	RETURN (CARDINALITY);

"=>"		RETURN (IMPL);
"^^"		RETURN (XOR);
"<=>"		RETURN (EQUIV);
"||"		RETURN (OR);
"&&"		RETURN (AND);
"<>"		RETURN (FINALLY);
"[]"		RETURN (GLOBALLY);
"()"		RETURN (NEXT);

"=="		RETURN (EQ);
"!="		RETURN (NE);
">="		RETURN (GE);
"<="		RETURN (LE);
"<<"		RETURN (ROL);
">>"		RETURN (ROR);

"eval"/{nn}	RETURNC (EVAL);
"show"/{nn}	RETURNC (SHOW);
"succ"/{nn}	RETURNC (SUCC);
"pred"/{nn}	RETURNC (PRED);
"go"/{nn}	RETURNC (GO);
"stats"/{nn}	RETURNC (STATS);
"help"/{nn}	RETURNC (HELP);
"path"/{nn}	RETURNC (PATH);
"function"/{nn}	RETURNC (FUNCTION);
"exit"/{nn}	RETURNC (EXIT);

@?0x		{
  delval.i = 0, overflow = false;
  SC_CALL (hexadecimal, *yytext == '@' ? STATE : NUMBER);
}
@?0		{
  delval.i = 0, overflow = false;
  SC_CALL (octal, *yytext == '@' ? STATE : NUMBER);
}
[1-9]		{
  delval.i = *yytext - '0', overflow = false; SC_CALL (decimal, NUMBER);
}
@[1-9]		{
  delval.i = yytext[1] - '0', overflow = false; SC_CALL (decimal, STATE);
}

<hexadecimal>{
  [0-9]		{
    if (overflow);
    else if (delval.i < 0 || delval.i > INT_MAX >> 3)
      deerror ("hexadecimal constant too large"), overflow = true;
    else
      (delval.i <<= 4) |= *yytext - '0';
  }
  [A-F]		{
    if (overflow);
    else if (delval.i < 0 || delval.i > INT_MAX >> 3)
      deerror ("hexadecimal constant too large"), overflow = true;
    else
      (delval.i <<= 4) |= *yytext - 'A' + 0x0A;
  }
  [a-f]		{
    if (overflow);
    else if (delval.i < 0 || delval.i > INT_MAX >> 3)
      deerror ("hexadecimal constant too large"), overflow = true;
    else
      (delval.i <<= 4) |= *yytext - 'a' + 0x0A;
  }
}

<octal>[0-7]	{
  if (overflow);
  else if (delval.i < 0 || delval.i > INT_MAX >> 2)
    deerror ("octal constant too large"), overflow = true;
  else
    (delval.i <<= 3) |= *yytext - '0';
}

<decimal>[0-9]	{
  if (overflow);
  else if (delval.i < 0) {
    int v = delval.i * 10 - (*yytext - '0');
    if (v / 10 != delval.i)
      deerror ("decimal constant too small"), overflow = true;
    else
      delval.i = v;
  }
  else {
    int v = delval.i * 10 + (*yytext - '0');
    if (v / 10 != delval.i)
      deerror ("decimal constant too large"), overflow = true;
    else
      delval.i = v;
  }
}

<hexadecimal,octal,decimal>{
  [0-9a-zA-Z_]	{
    deerror ("improperly terminated numeric constant");
    yyless (0); SC_POP;
  }
  .|\n		{
    yyless (0); SC_POP;
  }
  <<EOF>>	{
    SC_POP; deerror ("EOF in numeric constant"); NEXTBUF;
  }
}

[A-Za-z_][A-Za-z0-9_]*	{ CREAT; SC_PUSH (dename); }
\\.			{ INITSTR; APPEND (yytext[1]); SC_PUSH (dename); }

<dename>{
  \\\n{ws}*		lineno++;
  \\.			APPEND (yytext[1]);
  [A-Za-z0-9_]+		EXTEND;
  .|\n			{ SC_POP_B; COPY; RETURN (NAME); }
}

"'"			{ overflow = false; SC_PUSH (dechar); }

<dechar>{
  \'			{
    if (!overflow) {
      deerror ("null char constant");
      delval.c = 0;
    }
    SC_POP; RETURN (CHARACTER);
  }
  \\\n{ws}*		lineno++;
  \\/[0-7]		{
    if (!overflow) {
      delval.c = 0, digits = 0, overflow = true; SC_SWITCH (dechar_oct);
    }
    else
      deerror ("char constant too large, ignoring `\\'");
  }
  \\x/[0-9a-fA-F]	{
    if (!overflow) {
      delval.c = 0, digits = 0, overflow = true; SC_SWITCH (dechar_hex);
    }
    else
      deerror ("char constant too large, ignoring `\\x'");
  }
  \\.			{
    if (!overflow) {
      overflow = true;
      switch (yytext[1]) {
      case 'a': delval.c = '\a'; break;
      case 'b': delval.c = '\b'; break;
      case 't': delval.c = '\t'; break;
      case 'n': delval.c = '\n'; break;
      case 'v': delval.c = '\v'; break;
      case 'f': delval.c = '\f'; break;
      case 'r': delval.c = '\r'; break;
      default:
	delval.c = yytext[1];
      }
    }
    else
      deerrorf ("char constant too large, ignoring `\\%c'", yytext[1]);
  }
  .			{
    if (!overflow) {
      overflow = true;
      delval.c = *yytext;
    }
    else
      deerrorf ("char constant too large, ignoring `%c'", *yytext);
  }
  \n			{
    lineno++;
    deerror ("unterminated char constant");
    if (!overflow)
      delval.c = 0;
    SC_POP; RETURN (CHARACTER);
  }
}

<dechar_oct>[0-7]	{
  (delval.c <<= 3) |= *yytext - '0';
  if (++digits >= 3 || (delval.c >> (CHAR_T_BIT - 3)))
    SC_SWITCH (dechar);
}

<dechar_hex>{
  [0-9] {
    (delval.c <<= 4) |= *yytext - '0';
    if (++digits >= 2)
      SC_SWITCH (dechar);
  }
  [a-f] {
    (delval.c <<= 4) |= *yytext - 'a' + 0x0a;
    if (++digits >= 2)
      SC_SWITCH (dechar);
  }
  [A-F] {
    (delval.c <<= 4) |= *yytext - 'A' + 0x0a;
    if (++digits >= 2)
      SC_SWITCH (dechar);
  }
}

<dechar_oct,dechar_hex>{
  .|\n			SC_SWITCH_B (dechar);
}

<dechar,dechar_oct,dechar_hex><<EOF>>		{
  deerror ("EOF in char constant"); SC_POP; NEXTBUF;
}

\"			{ INITSTR; SC_CALL (destring, STRING); }

<destring>{
  \"			SC_POP;
  \\\n{ws}*		lineno++;
  \\/[0-7]		{
    character = 0, digits = 0; SC_SWITCH (destring_oct);
  }
  \\x/[0-9a-fA-F]	{
    character = 0, digits = 0; SC_SWITCH (destring_hex);
  }
  \\x
  \\.			{
    switch (yytext[1]) {
    case 'a': APPEND ('\a'); break;
    case 'b': APPEND ('\b'); break;
    case 't': APPEND ('\t'); break;
    case 'n': APPEND ('\n'); break;
    case 'v': APPEND ('\v'); break;
    case 'f': APPEND ('\f'); break;
    case 'r': APPEND ('\r'); break;
    default:
      APPEND (yytext[1]);
    }
  }
  \n			{ EXTEND; lineno++; }
  [^\\\"\n]+		EXTEND;
  \\			{
    deerror ("string constant terminated by `\\'"); EXTEND; SC_POP;
  }
  <<EOF>>		{
    deerror ("EOF in string constant"); SC_POP; NEXTBUF;
  }
}

<destring_oct>[0-7]	{
  (character <<= 3) |= *yytext - '0';
  if (++digits >= 3 || (character >> (CHAR_BIT - 3))) {
    APPEND (character);
    SC_SWITCH (destring);
  }
}

<destring_hex>{
  [0-9] {
    (character <<= 4) |= *yytext - '0';
    if (++digits >= 2) {
      APPEND (character);
      SC_SWITCH (destring);
    }
  }
  [a-f] {
    (character <<= 4) |= *yytext - 'a' + 0x0a;
    if (++digits >= 2) {
      APPEND (character);
      SC_SWITCH (destring);
    }
  }
  [A-F] {
    (character <<= 4) |= *yytext - 'A' + 0x0a;
    if (++digits >= 2) {
      APPEND (character);
      SC_SWITCH (destring);
    }
  }
}

<destring_oct,destring_hex>{
  .|\n			{ APPEND (character); SC_SWITCH_B (destring); }
  <<EOF>>		{
    APPEND (character); SC_POP;
    deerror ("EOF at escaped character in string constant"); NEXTBUF;
  }
}

;				return *yytext;
[!#%&()*+,\-./:<=>?\[\]^{|}~]	RETURN (*yytext);
\\\n			lineno++;
.			{
  deerrorf ("unknown character 0x%02x",
	    static_cast<unsigned char>(*yytext));
}

<INITIAL><<EOF>>	NEXTBUF;

%%

static const char* lastscriptname;
static const char* lastscriptstring;

void
delexreset ()
{
  lineno = 1;
  deerrors = 0;
  filename.create (0);
  lastscriptname = 0;
  stringptr = lastscriptstring = 0;
  overflow = false;
  SC_FLUSH;
  assert (in.empty ());
}

void
dewarn (const char* s)
{
  if (dewarnings)
    fprintf (stderr, "%s:%u:Warning:%s\n", filename.getString (), lineno, s);
}

void
dewarnf (const char* fmt, ...)
{
  if (dewarnings) {
    va_list ap;
    va_start (ap, fmt);
    fprintf (stderr, "%s:%u:Warning:", filename.getString (), lineno);
    vfprintf (stderr, fmt, ap);
    putc ('\n', stderr);
    va_end (ap);
  }
}

void
deerror (const char* s)
{
  if (!strcmp (s, "parse error"))
    command = true;
  fprintf (stderr, "%s:%u:%s\n", filename.getString (), lineno, s);
  deerrors++;
}

void
deerrorf (const char* fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  fprintf (stderr, "%s:%u:", filename.getString (), lineno);
  vfprintf (stderr, fmt, ap);
  putc ('\n', stderr);
  va_end (ap);
  deerrors++;
}

static bool
popInclude ()
{
  if (in.empty ()) {
    if (lastscriptname == scriptname && lastscriptstring == scriptstring) {
      filename = scriptname;
      return false;
    }
    else if (filename = lastscriptname = scriptname,
	     scriptname == scriptcommand) {
      dein = NULL;
      if (lastscriptstring == scriptstring) {
	yy_init = 1;
	return false;
      }
      stringptr = lastscriptstring = scriptstring;
      return true;
    }
    else if (!(dein = openscript (lastscriptname))) {
      perror (lastscriptname);
      yy_init = 1;
      return false;
    }
    else {
      lineno = 1;
      return true;
    }
  }
  else {
    if (dein && dein != stdin) fclose (dein);
    filename = in.top ().name;
    lineno = in.top ().lineno;
    dein = in.top ().file;
    yy_delete_buffer (YY_CURRENT_BUFFER);
    yy_switch_to_buffer (in.top ().bs);
    in.pop ();
    return true;
  }
}
