/* -*- c++ -*- Parser for the Maria Debugger */

%{
#include <alloca.h>

#include "util.h"
#include "StringBuffer.h"
#include "VariableStackMap.h"
#include "Printer.h"

#include "allExpressions.h"
#include "ExpressionSet.h"

#include "Transition.h"
#include "Place.h"
#include "Net.h"
#include "Graph.h"
#include "GlobalMarking.h"
#include "VariableDefinition.h"
#include "Function.h"
#include "Property.h"
#include "Product.h"

#include "Quantifier.h"

#include "TransitionQualifier.h"

#include "typedefs.h"
#include "allTypes.h"
#include "allValues.h"
#include "Valuation.h"

# include <set>
# include <map>
# include <list>
# include <stack>

# ifndef NDEBUG
#  define ASSERT1(x) x
# else
#  define ASSERT1(x)
# endif

  /** flag: exiting */
  bool exiting = false;
  /** warning flag (true=print warnings) */
  extern bool dewarnings;

  /** read a command line */
  const char* deline ();
  /** reset the parser and the lexical analyzer */
  void dereset ();
  /** reset the lexical analyzer */
  extern void delexreset ();

  /** lexical analyzer function
   * @return		a terminal symbol, and set yylval to its semantic value
   */
  extern int delex ();

  /** flag: expecting a command keyword */
  bool command = true;
  /** The state being examined */
  static class Search currentState;

  /** The printer object */
  extern class Printer thePrinter;
  /** The Petri Net */
  extern class Net* net;
# define theNet (*net)
  /** The reachability graph */
  extern class Graph* graph;
# define theGraph (*graph)
  /** Parameters of the function that is currently being processed */
  static class VariableDefinition** currentParameters = NULL;
  /** Arity of the function that is currently being processed */
  static unsigned currentArity = 0;
  /** Function being invoked */
  static std::stack< class Function*, std::list<class Function*> >
    functionStack;
  /** The context transition */
  static class Transition* currentContextTransition = NULL;
  /** The type definition that is currently being processed */
  static const class Type* currentType = NULL;
  /** The context type of currentType (StructType or VectorType or NULL) */
  static const class Type* currentContextType = NULL;
  /** Struct or vector component number currently being parsed */
  static card_t currentComponent;
  /** Type context */
  struct TypeContext {
    /** Context type (@see currentContextType), may be NULL */
    const class Type* context;
    /** Component type (@see currentContext) */
    const class Type* component;
    /** Number of current component (@see currentComponent)
     * , meaningless if context==NULL */
    card_t numComponent;
  };
  /** Stack of types, used e.g. when parsing quantifier conditions */
  static std::stack< struct TypeContext, std::list<struct TypeContext> >
    typeStack;

  /** Flag: has an empty list component been parsed? */
  static bool emptyParsed = false;

  typedef std::pair<const class Type*,card_t> StructLevel;
  /** Stack for parsing structs and vectors */
  static std::stack< StructLevel, std::list<StructLevel> > structLevels;

  /** Go to a state
   * @param state	number of the state
   */
  static void go (card_t state);

  /** Show a state
   * @param state	number of the state
   */
  static void show (card_t state);

  /** Show the successors of a state
   * @param state	number of the state
   */
  static void succ (card_t state);

  /** Evaluate a formula in the current state
   * @param state	state in which to evaluate the formula
   * @param expr	formula to be evaluated
   */
  static void eval (card_t state, class Expression* expr);

  /** Convert a StringBuffer to a NUL-terminated string
   * @param sbuf	the string buffer
   * @return		the corresponding NUL-terminated string
   */
  static char* convertString (class StringBuffer* sbuf);

  /** Push the type context on structLevels
   * @param type	new type context
   */
  static void pushTypeContext (const class Type* type);
  /** Restore the type context from structLevels */
  static void popTypeContext ();

  /** Get the parameter of the currently parsed function by name
   * @param name	name of the parameter
   * @return		the corresponding VariableDefinition, or NULL
   */
  static const class VariableDefinition* getParameter (char* name);

  /** Add a parameter to the function definition being parsed
   * @param type	type of the parameter
   * @param name	name of the parameter
   */
  static void addParameter (const class Type* type, char* name);

  /** Get a Function by name
   * @param name	the function name
   * @return		a Function object
   */
  static class Function* getFunction (const char* name);

  /** Get a zero-arity Function by name
   * @param name	the function name
   * @return		a Function object
   */
  static class Function* getConstantFunction (const char* name);

  /** Get a VariableDefinition by name
   * @param name	the variable name
   * @return		a VariableDefinition object
   */
  static const class VariableDefinition* getVariable (char* name);

  /** Resolve a name to a constant or variable
   * @param name	the name
   * @return		the resolved expression, or NULL
   */
  static class Expression* resolveName (char* name);

  /** Determine whether a formula is a set-valued expression
   * @param expr	formula to be checked
   * @return		true if the formula is a marking expression
   */
  static bool isSet (const class Expression* expr);

  /** Ensure that the formula is a set-valued expression
   * @param expr	formula to be checked
   * @return		expr, pointer to a Marking object, or NULL
   */
  static class Expression* ensureSet (class Expression* expr);

  /** Ensure that the formula is a marking expression, converting if needed
   * @param expr	formula to be checked
   * @return		pointer to a Marking object, or NULL
   */
  static class Marking* ensureMarking (class Expression* expr);

  /** Ensure that the formula is not a marking expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* noMarking (class Expression* expr);

  /** Ensure that the formula is a basic expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureExpr (class Expression* expr);

  /** Ensure that the formula is a boolean expression
   * @param expr	formula to be checked
   * @param undefined	flag: allow the formula to be Undefined
   * @return		expr or NULL
   */
  static class Expression* ensureBool (class Expression* expr,
				       bool undefined = false);

  /** Ensure that the formula is an integer expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureInt (class Expression* expr);

  /** Ensure that the formula is a buffer expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureBuffer (class Expression* expr);

  /** Ensure that the formula is a basic expression of ordered type
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureOrdered (class Expression* expr);

  /** Convert an integer to an enumeration value if possible
   * @param type	type of the enumeration value
   * @param value	numerical value to be converted to the enumeration type
   * @return		a constant having the enumeration value or NULL
   */
  static class Expression* enumValue (const class EnumType* type, int_t value);

  /** Ensure that a named component matches a struct being parsed
   * @param name	name of the component
   * @param expr	initialization expression for the component
   * @return		expr or NULL
   */
  static class Expression* namedComponent (char* name, class Expression* expr);

  /** Begin parsing a quantifier expression
   * @param name	name of the quantifier variable
   * @param type	type of the quantifier variable
   */
  static void beginQuantifier (char* name, const class Type* type);

  /** Continue parsing a quantifier expression
   *  (after parsing an optional condition)
   * @param type	type of the quantifier variable
   */
  static void continueQuantifier (ASSERT1 (const class Type* type));

  /** End parsing a quantifier expression
   * @param name	name of the quantifier variable
   * @param type	type of the quantifier variable
   * @param condition	quantification condition (optional)
   * @param expr	marking expression
   * @return		expr augmented with the quantifier, or NULL if failed
   */
  static class Marking* endQuantifier (char* name, const class Type* type,
				       class Expression* condition,
				       class Marking* expr);

  /** Try to make expr compatible with type if expr is a Constant
   * @param expr	expression to be made compatible
   * @param type	type to be made compatible with
   * @param casting	flag: typecasting in progress (suppress a warning)
   * @return		true if expr is not a Constant or
   *			if the conversion succeeded
   */
  static bool makeCompatible (class Expression*& expr,
			      const class Type* type,
			      bool casting = false);

  /** Determine whether an expression is compatible with a type
   * @param expr	expression to be checked
   * @param type	the type
   * @return		true if the expression is compatible with the type
   */
  static bool isCompatible (class Expression*& expr, const class Type* type);

  /** Determine whether an expression is compatible with
   * the struct component being parsed
   * @param expr	expression to be checked
   * @param s		the struct whose current component is considered
   * @return		true if the expression is compatible with the struct
   */
  static bool isCompatible (class Expression*& expr,
			    const class StructExpression* s);

  /** Determine whether two expressions are type compatible
   * @param expr1	first expression
   * @param expr2	second expression
   * @return		true if the expressions are compatible
   */
  static bool isCompatible (class Expression*& expr1,
			    class Expression*& expr2);

  /** Ensure that the expressions have an ordered type and are compatible
   * @param expr1	first expression to be checked
   * @param expr2	second expression to be checked
   * @return		true if the expressions have an ordered type
   */
  static bool isOrdered (class Expression*& expr1,
			 class Expression*& expr2);

  /** Perform constant folding
   * @param expr	expression to be folded
   * @return		expr or a new Constant, or NULL in case of error
   */
  static class Expression* fold (class Expression* expr);

  /** Create a RelopExpression or optimize it
   * @param op		operator
   * @param left	left-hand-side of RelopExpression
   * @param right	right-hand-side of RelopExpression
   * @return		RelopExpression or a Boolean constant
   */
  static class Expression* newRelopExpression (enum RelopExpression::Op op,
					       class Expression* left,
					       class Expression* right);

  /** Create a BooleanBinop or optimize it
   * @param op		operator
   * @param left	left-hand-side of BooleanBinop
   * @param right	right-hand-side of BooleanBinop
   * @return		left, right, Boolean constant, or a BooleanBinop
   */
  static class Expression* newBooleanBinop (enum BooleanBinop::Op op,
					    class Expression* left,
					    class Expression* right);

  /** Create a SetExpression or optimize it
   * @param op		operator
   * @param left	left-hand-side of SetExpression
   * @param right	right-hand-side of SetExpression
   */
  static class Expression* newSetExpression (enum SetExpression::Op op,
					     class Expression* left,
					     class Expression* right);

  /** Create a BufferUnop from a buffer expression
   * @param op		operator
   * @param expr	the buffer expression
   * @return		a BufferUnop object, or NULL
   */
  static class Expression* newBufferUnop (enum BufferUnop::Op op,
					  class Expression* expr);

  /** Variable stack map to be used with variable scoping */
  static class VariableStackMap currentVariables;

  /** Set of subexpressions for common subexpression elimination */
  static class ExpressionSet currentExpressions;

  /** Eliminate common subexpressions
   * @param expr	expression to be searched in currentExpressions
   * @return		expr or an equivalent expression
   */
  inline static class Expression* CSE (class Expression* expr) {
    return expr ? currentExpressions.insert (expr) : NULL;
  }

  /**
   * generic compiler warning reporting function
   * @param s		warning message
   */
  extern void dewarn (const char* s);

  /**
   * generic printf-like compiler warning reporting function
   * @param fmt		printf format string
   * @param ...		variable arguments
   */
  extern void dewarnf (const char* fmt, ...)
# ifdef __GNUC__
    __attribute__ ((format (printf, 1, 2)))
# endif /* __GNUC__ */
    ;

  /**
   * generic parser error reporting function
   * @param s		error message
   */
  extern void deerror (const char* s);

  /**
   * generic printf-like parser error reporting function
   * @param fmt		printf format string
   * @param ...		variable arguments
   */
  extern void deerrorf (const char* fmt, ...)
# ifdef __GNUC__
    __attribute__ ((format (printf, 1, 2)))
# endif /* __GNUC__ */
    ;
%}

%union{
  int_t i;
  bool flag;
  char_t c;
  char* s;
  class StringBuffer* sbuf;
  class Expression* expr;
  class ExpressionList* exprList;
  const class Transition* trans;
  const class Place* place;
  class Marking* marking;
  const class Type* type;
}

%right '?' ':'
%left ALWAYS EXISTS UNTIL RELEASE
%left EQUALS
%left SUBSET
%left MINUS UNION
%left INTERSECT
%left IMPL EQUIV
%left OR
%left XOR
%left AND
%left FINALLY GLOBALLY NEXT
%left '|'
%left '^'
%left '&'
%left NE EQ
%left GE LE '<' '>'
%left ROL ROR
%left '+' '-'
%left '*' '/' '%'
%right '#' IS
%left '~' '!'
%right CARDINALITY
%left ATOM
%left '.' '[' '('

%token <i>
	NUMBER
	STATE

%token <c>
	CHARACTER

%token <s>
	NAME
%token <sbuf>
	STRING

%token
	TRUE_ FALSE_ UNDEFINED FATAL EMPTY
	PLACE TRANS

%token	EVAL SHOW SUCC PRED GO STATS HELP PATH FUNCTION EXIT

%type <i>
	state

%type <s>
	name
	name_

%type <expr>
	_opt_expr
	_opt__expr
	expr_
	expr
	formula
	marking_list_
	qualifier
	qual_expr
	struct_expr

%type <exprList>
	arg_list
	colon_formula

%type <marking>
	marking

%type <place>
	placename

%type <trans>
	transname

%type <type>
	typereference

%type <flag>
	_opt_trans/* true if present */

%%

script:
	|
	statements
	;

statements:
	';'
	{ command = true; dewarn ("extraneous semicolon"); }
	|
	';'
	{ command = true; dewarn ("extraneous semicolon"); }
	statement
	{ command = true; }
	|
	statement
	{ command = true; }
	|
	statements ';' statement
	{ command = true; }
	|
	statements ';'
	;

statement:
	eval
	|
	EVAL eval
	|
	EVAL
	{ eval (currentState.getNumber (), resolveName (newString ("eval"))); }
	|
	SHOW state
	{ show ($2); }
	|
	SUCC state
	{ succ ($2); }
	|
	PRED state
	{ fprintf (stderr, "Predecessors are not yet supported.\n"); }
	|
	go
	|
	STATS
	{
	  fprintf (stderr, "%s: %u states, %u arcs (%s)\n",
		   theGraph.getFilebase (), theGraph.size (),
		   theGraph.getNumArcs (),
		   theGraph.isComplete () ? "complete" : "incomplete");
	}
	|
	HELP
	{ fprintf (stderr, "No help is available.  Tough!\n"); }
	|
/*	path
	|
*/	FUNCTION function
	|
	EXIT
	{ exiting = true; YYABORT; }
	;

eval:
	STATE formula
	{
	  if (card_t ($1) >= theGraph.size ())
	    deerrorf ("the graph only has %u states", theGraph.size ());
	  else
	    eval ($1, $2);
	}
	|
	formula
	{ eval (currentState.getNumber (), $1); }
	;

state:
	{ $$ = currentState.getNumber (); }
	|
	state STATE
	{
	  if (card_t ($$ = $1) != currentState.getNumber ())
	    deerrorf ("ignoring extraneous state @%u", $2);
	  else if (card_t ($$ = $2) >= theGraph.size ())
	    deerrorf ("the graph only has %u states", theGraph.size ());
	}
	|
	state name_
	{
	  deerror ("expecting a state number");
	  delete[] $2; $$ = theGraph.size ();
	}
	;

go:
	GO state
	{ go ($2); }
	|
	STATE
	{
	  if (card_t ($1) < theGraph.size ())
	    go ($1);
	  else
	    deerrorf ("the graph only has %u states", theGraph.size ());
	}
	;

typereference:
	name_
	{
	  if (!$1) $$ = NULL;
	  else if (($$ = theNet.getType ($1)));
	  else if (!strcmp ($1, "bool"))
	    $$ = Net::getBoolType ();
	  else if (!strcmp ($1, "int"))
	    $$ = Net::getIntType ();
	  else if (!strcmp ($1, "char"))
	    $$ = Net::getCharType ();
	  else
	    deerrorf ("undefined type name `%s'", $1);
	  delete[] $1;
	}
	;

name_:
	NAME
	{ $$ = $1; }
	|
	STRING
	{ $$ = convertString ($1); }
	|
	error
	{ $$ = NULL; }
	;

name:
	{ deerror ("missing name"); $$ = NULL; }
	|
	name_
	{ $$ = $1; }
	;

delim:	';' | ',' ;

next_eq:
	NEXT
	|
	'='
	;

function:
	typereference name next_eq
	{ assert (!currentType); currentType = $1; }
	formula
	{
	  assert (currentType == $1); currentType = NULL;
	  if ($1 && $2 && $5) {
	    if (theNet.getFunction ($2)) {
	      deerrorf ("ignoring redefinition of function `%s'", $2);
	      delete[] $2, $5->destroy ();
	    }
	    else if (isCompatible ($5, $1)) {
	      theNet.addFunction (*new class Function ($2, 0, 0, $5));
	      if (dewarnings && theNet.getType ($2))
		dewarnf ("function `%s' shadows a type name", $2);
	    }
	    else
	      delete[] $2, $5->destroy ();
	  }
	  else
	    delete[] $2, $5->destroy ();
	}
	|
	typereference name '(' param_list ')'
	{
	  assert (!currentType); currentType = $1;
	  if (emptyParsed && currentArity)
	    deerror ("extraneous delimiter after formal parameter list");
	  emptyParsed = false;
	}
	formula
	{
	  assert (currentType == $1); currentType = NULL;
	  if ($1 && $2 && $7) {
	    if (theNet.getFunction ($2)) {
	      deerrorf ("ignoring redefinition of function `%s'", $2);
	      delete[] $2, delete[] currentParameters, $7->destroy ();
	    }
	    else if (isCompatible ($7, $1)) {
	      theNet.addFunction (*new class Function
				  ($2, currentArity, currentParameters, $7));
	      if (dewarnings) {
		if (theNet.getType ($2))
		  dewarnf ("function `%s' shadows a type name", $2);
#ifdef EXPR_CACHE
		for (unsigned i = 0; i < currentArity; i++)
		  if (!$7->invalidate (currentParameters[i]))
		    dewarnf ("function `%s' does not use parameter `%s'",
			     $2, currentParameters[i]->getName ());
#endif // EXPR_CACHE
	      }
	    }
	    else
	      delete[] $2, delete[] currentParameters, $7->destroy ();
	  }
	  else
	    delete[] $2, delete[] currentParameters, $7->destroy ();
	  currentArity = 0, currentParameters = 0;
	  emptyParsed = false;
	}
	;

param_list:
	typereference name
	{
	  assert (!currentArity && !currentParameters);
	  addParameter ($1, $2);
	  emptyParsed = false;
	}
	|
	param_list delim typereference name
	{
	  if ($3 && $4 && emptyParsed)
	    deerror ("extraneous delimiter in parameter list");
	  addParameter ($3, $4);
	  emptyParsed = false;
	}
	|
	param_list delim
	{
	  if (emptyParsed)
	    deerror ("extraneous delimiter in parameter list");
	  emptyParsed = true;
	}
	|
	{ assert (!currentArity && !currentParameters); emptyParsed = true; }
	;

marking:
	formula
	{
	  class Marking* m = ensureMarking ($1);
	  if (($$ = m)) m->associate ();
	}
	;

marking_list_:
	formula
	{
	  if (($$ = $1) && $$->getKind () == Expression::eMarking)
	    static_cast<class Marking*>($$)->associate ();
	}
	|
	marking_list_ ','
	{
	  if (($1 = ensureMarking ($1)) &&
	      $1->getKind () == Expression::eMarking)
	    static_cast<class Marking*>($1)->associate ();
	}
	marking
	{
	  assert (!$1 || $1->getKind () == Expression::eMarking);
	  class Marking* m = static_cast<class Marking*>($1);
	  if (($$ = m) && $4) {
	    if (m->isQuantified ())
	      $$ = m = new class Marking (m->getPlace (), m);
	    m->append ($4);
	  }
	  else
	    $4->destroy ();
	}
	;

_opt_expr:
	{ $$ = NULL; }
	|
	'(' expr_ ')'
	{ $$ = $2; }
	;

_opt__expr:
	{ $$ = NULL; }
	|
	expr
	{ $$ = $1; }
	|
	NAME ':'
	{ $$ = namedComponent ($1, NULL); }
	|
	STRING ':'
	{ $$ = namedComponent (convertString ($1), NULL); }
	|
	NAME ':' expr
	{ $$ = namedComponent ($1, $3); }
	|
	STRING ':' expr
	{ $$ = namedComponent (convertString ($1), $3); }
	;

struct_expr:
	_opt__expr
	{
	  if (currentContextType && !YYRECOVERING ()) {
	    assert (!currentComponent);

	    switch (currentContextType->getKind ()) {
	    case Type::tVector:
	      {
		const class VectorType* type =
		  static_cast<const class VectorType*>(currentContextType);
		class VectorExpression* v = new class VectorExpression (type);
		assert (currentType == type->getItemType ());

		if (!$1) {
		  if (!type->isOptional ()) {
		    deerror ("missing expression for first vector component");
		    v->destroy (), $$ = NULL;
		  }
		  else
		    $$ = v;
		}
		else if (isCompatible ($1, currentType))
		  $$ = v, v->setComponent (currentComponent, $1);
		else
		  v->destroy (), $1->destroy (), $$ = NULL;
	      }
	      break;
	    case Type::tBuffer:
	      {
		const class BufferType* type =
		  static_cast<const class BufferType*>(currentContextType);
		class BufferExpression* b = new class BufferExpression (type);
		assert (currentType == type->getItemType ());

		if (!$1)
		  $$ = b;
		else if (isCompatible ($1, currentType))
		  $$ = b, b->append ($1);
		else
		  b->destroy (), $1->destroy (), $$ = NULL;
	      }
	      break;
	    case Type::tStruct:
	      {
		if (!$1 &&
		    !static_cast<const class StructType*>(currentContextType)
		    ->getSize ())
		  $$ = CSE (new class Constant
			    (new class StructValue (currentContextType)));
		else {
		  class StructExpression* s =
		    new class StructExpression (currentContextType);
		  assert (currentType == s->getNextType ());

		  if (isCompatible ($1, s))
		    $$ = s, s->append ($1), currentType = s->getNextType ();
		  else
		    s->destroy (), $1->destroy (),
		      $$ = NULL, currentType = NULL;
		}
	      }
	      break;
	    default:
	      assert (false);
	    }
	  }
	  else
	    $1->destroy (), $$ = NULL;

	  currentComponent++;
	}
	|
	struct_expr ',' _opt__expr
	{
	  if (($$ = $1)) {
	    assert (currentContextType != NULL);
	    assert (currentComponent > 0);
	    switch (currentContextType->getKind ()) {
	    case Type::tVector:
	      {
		assert ($$->getKind () == Expression::eVector);
		class VectorExpression* v =
		  static_cast<class VectorExpression*>($$);
		const class VectorType* type =
		  static_cast<const class VectorType*>(currentContextType);
		assert (currentType == type->getItemType ());

		if (currentComponent < type->getSize ()) {
		  if (!$3) {
		    if (!type->isOptional ()) {
		      deerrorf ("missing expression for vector component %u",
				currentComponent);
		      $$->destroy (), $$ = NULL;
		    }
		  }
		  else if (isCompatible ($3, currentType))
		    v->setComponent (currentComponent, $3);
		  else
		    $$->destroy (), $3->destroy (), $$ = NULL;
		}
		else {
		  if (type->getName ())
		    deerrorf ("array `%s' only has %u elements",
			      type->getName (), type->getSize ());
		  else
		    deerrorf ("array only has %u elements", type->getSize ());

		  $$->destroy (), $3->destroy (), $$ = NULL;
		}
	      }
	      break;
	    case Type::tBuffer:
	      {
		assert ($$->getKind () == Expression::eBuffer);
		class BufferExpression* b =
		  static_cast<class BufferExpression*>($$);
		const class BufferType* type =
		  static_cast<const class BufferType*>(currentContextType);
		assert (currentType == type->getItemType ());

		if (b->hasCapacity ()) {
		  if (!$3 || !b->getSize ()) {
		    deerrorf ("missing expression for buffer component %u",
			      $3 ? currentComponent : 0);
		    $$->destroy (), $3->destroy (), $$ = NULL;
		  }
		  else if (isCompatible ($3, currentType))
		    b->append ($3);
		  else
		    $$->destroy (), $3->destroy (), $$ = NULL;
		}
		else {
		  if (type->getName ())
		    deerrorf ("%s `%s' only has %u slots",
			      type->isStack () ? "stack" : "queue",
			      type->getName (), type->getSize ());
		  else
		    deerrorf ("%s only has %u slots",
			      type->isStack () ? "stack" : "queue",
			      type->getSize ());
		  $$->destroy (), $3->destroy (), $$ = NULL;
		}
	      }
	      break;
	    case Type::tStruct:
	      {
		assert ($$->getKind () == Expression::eStruct);
		class StructExpression* s =
		  static_cast<class StructExpression*>($$);
		assert (currentType == s->getNextType ());

		if (isCompatible ($3, s))
		  s->append ($3), currentType = s->getNextType ();
		else
		  $$->destroy (), $3->destroy (),
		    $$ = NULL, currentType = NULL;
	      }
	      break;
	    default:
	      assert (false);
	    }
	  }
	  else
	    $$->destroy (), $3->destroy (), $$ = NULL;

	  currentComponent++;
	}
	;

expr_:
	{ $$ = NULL; deerror ("missing expression"); }
	|
	expr
	{ $$ = $1; }
	;

expr:
	formula
	{ $$ = ensureExpr ($1); }
	;

formula:
	error
	{ $$ = NULL; }
	|
	name_ NEXT
	{
	  $$ = NULL;
	  if ($1) {
	    if (class Function* f = getFunction ($1)) {
	      if (!($$ = fold (f->expand (NULL))))
		deerrorf ("could not expand function `%s'", f->getName ());
	    }
	    else
	      deerrorf ("there is no function `%s'", $1);
	  }
	}
	|
	name_ '('
	{
	  class Function* f = NULL;
	  if ($1 && !(f = getFunction ($1)))
	    deerrorf ("there is no function `%s'", $1);
	  delete[] $1;
	  functionStack.push (f);
	}
	arg_list ')'
	{
	  assert (!functionStack.empty ());
	  if (class Function* f = functionStack.top ()) {
	    $$ = NULL;
	    if ($4 && !($$ = fold (f->expand ($4))))
	      deerrorf ("could not expand function `%s'", f->getName ());
	  }
	  else
	    $$ = NULL;
	  $4->destroy ();
	  functionStack.pop ();
	}
	|
	TRUE_
	{
	  $$ = CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), true)));
	}
	|
	FALSE_
	{
	  $$ = CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), false)));
	}
	|
	placename
	{ $$ = $1 ? CSE (new class PlaceContents ($1)) : NULL; }
	|
	NAME
	{ $$ = resolveName ($1); }
	|
	NUMBER
	{
	  $$ = CSE (new class Constant
		    (new class IntValue (Net::getIntType (), $1)));
	}
	|
	CHARACTER
	{
	  $$ = CSE (new class Constant
		    (new class CharValue (Net::getCharType (), $1)));
	}
	|
	STRING
	{
	  if (currentType && currentType->getKind () == Type::tVector) {
	    const class VectorType* type =
	      static_cast<const class VectorType*>(currentType);
	    if (type->getItemType ()->getKind () == Type::tChar) {
	      const class CharType* charType =
		static_cast<const class CharType*>(type->getItemType ());
	      card_t length = $1->getLength ();
	      if (length == type->getSize ()) {
		class VectorValue* v = new class VectorValue (type);
		const char* const str = $1->getString ();
		for (card_t i = 0; i < length; i++) {
		  class CharValue* c = new class CharValue (charType, str[i]);
		  if (!charType->isConstrained (c)) {
		    deerrorf ("%s constant 0x%02x out of range for type %s",
			      charType->getSyntacticName (),
			      static_cast<unsigned char>(c->getValue ()),
			      type->getName ()
			      ? type->getName ()
			      : type->getSyntacticName ());
		    delete c;
		    delete v;
		    v = NULL;
		    break;
		  }
		  else
		    (*v)[i] = c;
		}

		$$ = v ? CSE (new class Constant (v)) : NULL;
	      }
	      else {
		deerrorf ("expecting an array of %u characters, got %u",
			  type->getSize (), length);
		$$ = NULL;
	      }

	      delete $1;
	    }
	    else
	      $$ = resolveName (convertString ($1));
	  }
	  else
	    $$ = resolveName (convertString ($1));
	}
	|
	UNDEFINED
	{ $$ = CSE (new class Undefined (Undefined::fError)); }
	|
	FATAL
	{ $$ = CSE (new class Undefined (Undefined::fFatalError)); }
	|
	'{'
	{
	  structLevels.push (StructLevel
			     (currentContextType, currentComponent));

	  if (!currentType) {
	    deerror ("cannot determine type of structured expression");
	    currentContextType = NULL, currentComponent = 0;
	  }
	  else {
	    assert (!structLevels.empty () || !currentComponent);
	    currentContextType = currentType;
	    currentComponent = 0;

	    switch (currentType->getKind ()) {
	    case Type::tStruct:
	      currentType =
		static_cast<const class StructType*>(currentContextType)
		->getSize ()
		? (*static_cast<const class StructType*>(currentContextType))
		[currentComponent]
		: NULL;
	      break;
	    case Type::tVector:
	      currentType =
		static_cast<const class VectorType*>(currentContextType)
		->getItemType ();

	      if (dewarnings) {
		const class Type* indexType =
		  static_cast<const class VectorType*>(currentContextType)
		  ->getIndexType ();
		assert (indexType != NULL);

		if (indexType->getKind () == Type::tInt) {
		  const class Constraint* constraint =
		    static_cast<const class IntType*>(indexType)
		    ->getConstraint ();
		  if (!constraint || constraint->size () <= 1)
		    break;
		}
		if (indexType->getKind () == Type::tBool)
		  break;

		dewarn ("initializing an array indexed by something else"
			" than a boolean or a simple integer range");
	      }

	      break;
	    case Type::tBuffer:
	      currentType =
		static_cast<const class BufferType*>(currentContextType)
		->getItemType ();

	      break;
	    default:
	      deerror ("cannot determine type of structured expression");
	      currentType = currentContextType = NULL;
	    }
	  }
	}
	struct_expr
	'}'
	{
	  $$ = $3;
	  if (currentContextType) {
	    if ($$) {
	      switch ($$->getKind ()) {
	      case Expression::eStruct:
		{
		  class StructExpression* s =
		    static_cast<class StructExpression*>($$);
		  if (s->getNextType ()) {
		    if (currentContextType->getName ())
		      deerrorf ("struct `%s' not fully initialized",
				currentContextType->getName ());
		    else
		      deerror ("struct not fully initialized");

		    $$->destroy (), $$ = NULL;
		  }
		}
		break;

	      case Expression::eConstant:
	      case Expression::eBuffer:
		break;

	      case Expression::eVector:
		{
		  card_t diff =
		    static_cast<const class VectorType*>(currentContextType)
		    ->getSize () - currentComponent;

		  if (diff) {
		    if (currentContextType->getName ())
		      deerrorf ("%u %s of array `%s' left uninitialized",
				diff, diff == 1 ? "element" : "elements",
				currentContextType->getName ());
		    else
		      deerrorf ("%u %s of array left uninitialized",
				diff, diff == 1 ? "element" : "elements");

		    $$->destroy (), $$ = NULL;
		  }
		}
		break;

	      default:
		assert (false);
	      }
	    }

	    StructLevel p = structLevels.top ();
	    currentType = currentContextType;
	    currentContextType = p.first;
	    currentComponent = p.second;
	  }
	  else {
	    assert (!currentType);
	    currentComponent = 0;
	  }

	  structLevels.pop ();

	  if ($$)
	    $$ = fold ($$);
	}
	|
	IS typereference
	{ pushTypeContext ($2); }
	formula %prec ATOM
	{
	  if (($$ = $4) && $$->getKind () == Expression::eMarking &&
	      $$->getType () == currentType);
	  else if (($$ = noMarking ($$)) && currentType) {
	    if (!$$->getType ()) {
	      deerror ("expression to cast has no type");
	      $$->destroy (), $$ = NULL;
	    }
	    else if (!makeCompatible ($$, currentType, true));
	    else if ($$->getType () == currentType);
	    else if ((Typecast::isCastable (*currentType) &&
		      Typecast::isCastable (*$$->getType ()) ||
		      Type::isCompatible ($$->getType (), currentType)))
	      $$ = fold (new class Typecast (currentType, $$));
	    else {
	      deerrorf ("cannot cast from `%s' to `%s'",
			$$->getType ()->getName ()
			? $$->getType ()->getName ()
			: $$->getType ()->getSyntacticName (),
			currentType->getName ()
			? currentType->getName ()
			: currentType->getSyntacticName ());
	      $$->destroy (), $$ = NULL;
	    }
	  }
	  popTypeContext ();
	}
	|
	name_ '='
	{
	  StructLevel p (currentContextType, currentComponent);
	  structLevels.push (p);
	  currentContextType = currentType;
	  currentType = NULL, currentComponent = 0;

	  if ($1 && currentContextType &&
	      currentContextType->getKind () == Type::tUnion) {
	    const class UnionType* u =
	      static_cast<const class UnionType*>(currentContextType);

	    if ((currentType = (*u)[$1]))
	      currentComponent = u->getIndex ($1);
	    else {
	      if (u->getName ())
		deerrorf ("union `%s' does not have component `%s'",
			  u->getName (), $1);
	      else
		deerrorf ("union does not have component `%s'", $1);
	    }
	  }
	  else if ($1) {
	    if (currentContextType && currentContextType->getName ())
	      deerrorf ("cannot reference component `%s' of type `%s'",
			$1, currentContextType->getName ());
	    else
	      deerrorf ("cannot reference component `%s' of non-union type",
			$1);
	  }

	  delete[] $1;
	}
	formula %prec ALWAYS
	{
	  assert (!currentType ||
		  (currentContextType->getKind () == Type::tUnion &&
		   currentType ==
		   (*static_cast<const class UnionType*>(currentContextType))
		   [currentComponent]));

	  if (($4 = ensureExpr ($4)) && currentType) {
	    const class UnionType* u =
	      static_cast<const class UnionType*>(currentContextType);

	    if (isCompatible ($4, currentType))
	      $$ = fold (new class UnionExpression (u, $4, currentComponent));
	    else {
	      if (u->getName ())
		deerrorf ("type mismatch for union `%s' component `%s'",
			  u->getName (),
			  u->getComponentName (currentComponent));
	      else
		deerrorf ("type mismatch for union component `%s'",
			  u->getComponentName (currentComponent));
	      $4->destroy (), $$ = NULL;
	    }
	  }
	  else
	    $4->destroy (), $$ = NULL;

	  assert (!structLevels.empty ());
	  StructLevel p = structLevels.top ();
	  currentType = currentContextType;
	  currentContextType = p.first;
	  currentComponent = p.second;
	  structLevels.pop ();

	  if ($$)
	    $$ = CSE ($$);
	}
	|
	formula IS name
	{
	  if (($1 = ensureExpr ($1)) && $3) {
	    if ($1->getType () && $1->getType ()->getKind () == Type::tUnion) {
	      const class UnionType* u =
		static_cast<const class UnionType*>($1->getType ());

	      if ((*u)[$3])
		$$ = fold (new class UnionTypeExpression ($1,
							  u->getIndex ($3)));
	      else {
		if (u->getName ())
		  deerrorf ("union `%s' does not have component `%s'",
			    u->getName (), $3);
		else
		  deerrorf ("union does not have component `%s'", $3);

		$1->destroy (), $$ = NULL;
	      }
	    }
	    else {
	      if ($1->getType () && $1->getType ()->getName ())
		deerrorf ("type `%s' is not a union",
			  $1->getType ()->getName ());
	      else
		deerror ("type is not a union");

	      $1->destroy (), $$ = NULL;
	    }
	  }
	  else
	    $1->destroy (), $$ = NULL;

	  delete[] $3;

	  if ($$)
	    $$ = CSE ($$);
	}
	|
	ATOM formula
	{ if (($$ = noMarking ($2))) { $$->setAtomic (true); $$ = CSE ($$); } }
	|
	formula '<' formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rLess, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula EQ formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isCompatible ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rEqual, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '>' formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rGreater, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula GE formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rGreaterEqual, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula NE formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isCompatible ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rInequal, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula LE formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rLessEqual, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '+' formula
	{
	  if ($1 && $1->getType () &&
	      $1->getType ()->getKind () == Type::tBuffer) {
	    const class BufferType* type =
	      static_cast<const class BufferType*>($1->getType ());
	    if (($1 = ensureExpr ($1)) && ($3 = ensureExpr ($3)) &&
		isCompatible ($3, type->getItemType ())) {
	      if ($1->getKind () == Expression::eBufferIndex) {
		class BufferIndex* b = static_cast<class BufferIndex*>($1);
		$$ = fold (new class BufferWrite (b->getBuffer (), $3,
						  b->getIndex ()));
		b->destroy ();
	      }
	      else
		$$ = fold (new class BufferWrite ($1, $3, NULL));
	    }
	    else
	      $1->destroy (), $3->destroy (), $$ = NULL;
	  }
	  else {
	    $1 = ensureInt ($1); $3 = ensureInt ($3);
	    if ($1 && $3)
	      $$ = fold (new class BinopExpression (BinopExpression::bPlus,
						    $1, $3));
	    else {
	      $1->destroy (); $3->destroy (); $$ = NULL;
	    }
	  }
	}
	|
	formula '-' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bMinus,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '/' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bDiv,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '*' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bMul,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '%' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bMod,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	'~' formula
	{
	  if (($$ = ensureInt ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uNot, $$));
	}
	|
	'#' typereference %prec '!'
	{
	  if ($2) {
	    card_t num = $2->getNumValues ();
	    assert (num > 0);
	    if (num > INT_T_MAX) {
	      deerrorf ("type `%s' has too many values to count",
			$2->getName ()
			? $2->getName ()
			: $2->getSyntacticName ());
	      $$ = NULL;
	    }
	    else
	      $$ = CSE (new class Constant
			(new class IntValue
			 (Net::getIntType (), int_t (num))));
	  }
	  else
	    $$ = NULL;
	}
	|
	'^' typereference %prec '!'
	{
	  if ($2) {
	    if ($2->isOrdered ())
	      $$ = CSE (new class Constant ($2->getFirstValue ()));
	    else {
	      deerrorf ("cannot get smallest value of non-ordered type `%s'",
			$2->getName ()
			? $2->getName ()
			: $2->getSyntacticName ());
	      $$ = NULL;
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	'&' typereference %prec '!'
	{
	  if ($2) {
	    if ($2->isOrdered ())
	      $$ = CSE (new class Constant ($2->getLastValue ()));
	    else {
	      deerrorf ("cannot get largest value of non-ordered type `%s'",
			$2->getName ()
			? $2->getName ()
			: $2->getSyntacticName ());
	      $$ = NULL;
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	'|' formula %prec '!'
	{
	  if (($$ = ensureOrdered ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uPred, $$));
	}
	|
	'+' formula %prec '!'
	{
	  if (($$ = ensureOrdered ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uSucc, $$));
	}
	|
	'-' formula %prec '!'
	{
	  if (($2 = ensureExpr ($2)) && $2->getType () &&
	      $2->getType ()->getKind () == Type::tBuffer)
	    $$ = newBufferUnop (BufferUnop::bRemove, $2);
	  else if (($$ = ensureInt ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uNeg, $$));
	}
	|
	'*' formula %prec '!'
	{ $$ = newBufferUnop (BufferUnop::bPeek, $2); }
	|
	'/' formula %prec '!'
	{ $$ = newBufferUnop (BufferUnop::bUsed, $2); }
	|
	'%' formula %prec '!'
	{ $$ = newBufferUnop (BufferUnop::bFree, $2); }
	|
	'?' formula %prec '!'
	{
	  if (($$ = ensureExpr ($2)) &&
	      $$->getKind () == Expression::eVectorIndex)
	    $$ = CSE (static_cast<class VectorIndex*>($$)->toComponent ());
	  else if ($$ && $$->getType () &&
		   $$->getType ()->getKind () == Type::tVector) {
	    const class VectorType* const type =
	      static_cast<const class VectorType*>($$->getType ());
	    if (type->isOptional ())
	      $$ = CSE (new class VectorDefined ($2));
	    else
	      deerror ("ignoring unary '?':"
		       " vector has no optional components");
	  }
	  else
	    deerror ("ignoring unary `?': no vector context");
	}
	|
	formula ROL formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bRol,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula ROR formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bRor,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '&' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bAnd,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '^' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bXor,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '|' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bOr,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '?' colon_formula
	{
	  $1 = noMarking ($1);
	  if ($1 && $3) {
	    if (!$1->getType ()) {
	      deerror ("left-hand-side of ? has no type");
	      $1->destroy (); $3->destroy (); $$ = NULL;
	    }
	    else if ($1->getType ()->getNumValues () != $3->size ()) {
	      deerrorf ("left-hand-side of ? has %u!=%u possible values",
			$1->getType ()->getNumValues (), $3->size ());
	      $1->destroy (); $3->destroy (); $$ = NULL;
	    }
	    else if ($1->getType ()->getNumValues () == 1) {
	      dewarn ("left-hand-side of ? has only one possible value");
	      $$ = (*$3)[0]->copy (); $1->destroy (); $3->destroy ();
	    }
	    else
	      $$ = CSE (new class IfThenElse ($1, $3));
	  }
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '.' name
	{
	  $1 = ensureExpr ($1);
	  if ($1 && $3) {
	    if ($1->getType () &&
		$1->getType ()->getKind () == Type::tStruct) {
	      const class StructType *s =
		static_cast<const class StructType*>($1->getType ());
	      if ((*s)[$3])
		$$ = fold (new class StructComponent ($1, s->getIndex ($3)));
	      else {
		if (s->getName ())
		  deerrorf ("struct `%s' does not have component `%s'",
			    s->getName (), $3);
		else
		  deerrorf ("struct does not have component `%s'", $3);

		$1->destroy (), $$ = NULL;
	      }
	    }
	    else if ($1->getType () &&
		     $1->getType ()->getKind () == Type::tUnion) {
	      const class UnionType *u =
		static_cast<const class UnionType*>($1->getType ());
	      if ((*u)[$3])
		$$ = fold (new class UnionComponent ($1, u->getIndex ($3)));
	      else {
		if (u->getName ())
		  deerrorf ("union `%s' does not have component `%s'",
			    u->getName (), $3);
		else
		  deerrorf ("union does not have component `%s'", $3);

		$1->destroy (), $$ = NULL;
	      }
	    }
	    else {
	      if ($1->getType () && $1->getType ()->getName ())
		deerrorf ("cannot assign to component `%s' of type `%s'",
			  $3, $1->getType ()->getName ());
	      else
		deerrorf ("cannot assign to component `%s' of expression", $3);

	      $1->destroy (), $$ = NULL;
	    }
	  }
	  else
	    $1->destroy (), $$ = NULL;

	  delete[] $3;
	}
	|
	formula '['
	{
	  if (($1 = ensureExpr ($1)) && $1->getType ()) {
	    switch ($1->getType ()->getKind ()) {
	    case Type::tVector:
	      {
		const class VectorType* type =
		  static_cast<const class VectorType*>($1->getType ());
		pushTypeContext (type->getIndexType ());
	      }
	      break;
	    case Type::tBuffer:
	      if ($1->getKind () == Expression::eBufferIndex)
		deerror ("a buffer index has already been specified");
	      break;
	    default:
	      deerror ("only vector-typed expressions can be indexed");
	      $1->destroy (), $1 = NULL;
	    }
	  }
	  else if ($1) {
	    deerror ("no type for indexed expression");
	    $1->destroy (), $1 = NULL;
	  }
	}
	expr ']'
	{
	  if ($1) {
	    switch ($1->getType ()->getKind ()) {
	    case Type::tVector:
	      {
		const class VectorType* type =
		  static_cast<const class VectorType*>($1->getType ());
		popTypeContext ();

		if (isCompatible ($4, type->getIndexType ()))
		  $$ = fold (new class VectorIndex ($1, $4));
		else
		  $$ = NULL, $1->destroy (), $4->destroy ();
	      }
	      break;
	    case Type::tBuffer:
	      if ($1->getKind () == Expression::eBufferIndex)
		$4->destroy (), $$ = $1;
	      else
		$$ = CSE (new class BufferIndex ($1, $4));
	      break;
	    default:
	      assert (false);
	    }
	  }
	  else
	    $$ = NULL, $4->destroy ();
	}
	|
	'!' formula
	{
	  if (($$ = ensureBool ($2)))
	    $$ = fold (new class NotExpression ($$));
	}
	|
	formula AND formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3, true);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bAnd, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula OR formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3, true);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bOr, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula IMPL formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3, true);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bImpl, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula XOR formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bXor, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula EQUIV formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bEquiv, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	ALWAYS _no_qualifier formula
	{
	  $3 = ensureBool ($3);
	  $$ = $3
	    ? CSE (new class TemporalQuantifier (TemporalQuantifier::Always,
						 $3))
	    : NULL;
	}
	|
	EXISTS _no_qualifier formula
	{
	  $3 = ensureBool ($3);
	  $$ = $3
	    ? CSE (new class TemporalQuantifier (TemporalQuantifier::Exists,
						 $3))
	    : NULL;
	}
	|
	FINALLY qualifier formula
	{
	  $3 = ensureBool ($3);
	  if ($3)
	    $$ = CSE (new class TemporalUnop (TemporalUnop::Finally, $2, $3));
	  else {
	    $$ = NULL; $2->destroy ();
	  }
	}
	|
	GLOBALLY qualifier formula
	{
	  $3 = ensureBool ($3);
	  if ($3)
	    $$ = CSE (new class TemporalUnop (TemporalUnop::Globally, $2, $3));
	  else {
	    $$ = NULL; $2->destroy ();
	  }
	}
	|
	NEXT qualifier formula
	{
	  $3 = ensureBool ($3);
	  if ($3)
	    $$ = CSE (new class TemporalUnop (TemporalUnop::Next, $2, $3));
	  else {
	    $$ = NULL; $2->destroy ();
	  }
	}
	|
	formula UNTIL qualifier formula
	{
	  $1 = ensureBool ($1); $4 = ensureBool ($4);
	  if ($1 && $4)
	    $$ = CSE (new class TemporalBinop (TemporalBinop::Until,
					       $3, $1, $4));
	  else {
	    $1->destroy (); $3->destroy (); $4->destroy (); $$ = NULL;
	  }
	}
	|
	formula RELEASE qualifier formula
	{
	  $1 = ensureBool ($1); $4 = ensureBool ($4);
	  if ($1 && $4)
	    $$ = CSE (new class TemporalBinop (TemporalBinop::Release,
					       $3, $1, $4));
	  else {
	    $1->destroy (); $3->destroy (); $4->destroy (); $$ = NULL;
	  }
	}
	|
	EMPTY
	{ $$ = CSE (new class EmptySet); }
	|
	formula SUBSET formula
	{ $$ = newSetExpression (SetExpression::sSubset, $1, $3); }
	|
	formula INTERSECT formula
	{ $$ = newSetExpression (SetExpression::sIntersection, $1, $3); }
	|
	formula MINUS formula
	{ $$ = newSetExpression (SetExpression::sMinus, $1, $3); }
	|
	formula UNION formula
	{ $$ = newSetExpression (SetExpression::sUnion, $1, $3); }
	|
	formula EQUALS formula
	{ $$ = newSetExpression (SetExpression::sEquals, $1, $3); }
	|
	CARDINALITY formula
	{
	  if (class Expression* m = ensureSet ($2)) {
	    if (m->getKind () == Expression::eEmptySet) {
	      m->destroy ();
	      dewarn ("the cardinality of an empty set is always zero");
	      $$ = CSE (new class Constant (new class IntValue
					    (Net::getIntType (), 0)));
	    }
	    else {
	      $$ = CSE (new class CardinalityExpression (m));
	      if (m->getKind () == Expression::eMarking)
		static_cast<class Marking*>(m)->associate ();
	      $$->associateVariables (true);
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	'(' marking_list_ ')'
	{ $$ = $2; }
	|
	formula '#' formula
	{
	  $1 = ensureInt ($1);
	  class Marking* m = ensureMarking ($3);

	  if (($$ = m) && $1) {
	    if ($1->getKind () == Expression::eConstant) {
	      const class Value* v =
		static_cast<class Constant*>($1)->getValue ();
	      assert (v->getKind () == Value::vInt);
	      int_t mult = static_cast<const class IntValue*>(v)->getValue ();
	      if (mult < 1) {
		deerrorf ("non-positive cardinality %d", mult);
		$1->destroy (), m->associate (false), m->destroy ();
		$$ = m = NULL;
	      }
	      else {
		for (;;) {
		  if (class Expression* c = m->getMultiplicity ()) {
		    class Expression* card =
		      fold (new class BinopExpression (BinopExpression::bMul,
						       $1, c));
		    if (card && card->getKind () != Expression::eConstant)
		      card->associateVariables (true);

		    m->setMultiplicity (card);
		  }
		  else
		    m->setMultiplicity ($1);

		  if ((m = const_cast<class Marking*>(m->getNext ())))
		    $1 = CSE (new class Constant
			      (new class IntValue ($1->getType (), mult)));
		  else
		    break;
		}
	      }
	    }
	    else {
	      if (m->getNext ())
		$$ = m = static_cast<class Marking*>
		  (CSE (new class Marking (m->getPlace (), m)));

	      class Expression* card = $1;

	      if (class Expression* c = m->getMultiplicity ())
		card = fold
		  (new class BinopExpression (BinopExpression::bMul, $1, c));

	      if (card && card->getKind () != Expression::eConstant)
		card->associateVariables (true);

	      m->setMultiplicity (card);
	    }
	  }
	  else
	    $1->destroy ();

	  if (m)
	    m->associate ();
	}
	|
	typereference name_
	{ beginQuantifier ($2, $1); }
	_opt_expr
	{ continueQuantifier (ASSERT1 ($1)); }
	':' formula %prec ':'
	{ $$ = endQuantifier ($2, $1, $4, ensureMarking ($7)); }
	|
	SUBSET name '{' formula '}'
	{
	  if ($4 && $4->getKind () == Expression::eEmptySet) {
	    dewarn ("any subset of an empty set is an empty set");
	    delete[] $2, $2 = NULL;
	  }
	  else if (($4 = ensureSet ($4)) && $2 && $4->getType ())
	    currentVariables.push (new class VariableDefinition
				   ($2, $4->getType ()));
	  else
	    delete[] $2, $2 = NULL;
	}
	formula %prec CARDINALITY
	{
	  $7 = ensureBool (ensureExpr ($7));
	  if (($$ = $4) && $7 && $2) {
	    ($7 = CSE ($7))->associateVariables (true);
#ifdef EXPR_CACHE
	    if (dewarnings && !$7->invalidate (currentVariables.top ()))
	      dewarnf ("subset criterion does not depend on `%s'",
		       currentVariables.top ()->getName ());
#endif // EXPR_CACHE
	    if ($7->getKind () == Expression::eConstant) {
	      const class Value* v =
		static_cast<class Constant*>($7)->getValue ();
	      assert (v->getKind () == Value::vBool);
	      bool result =
		static_cast<const class BoolValue*>(v)->getValue ();
	      $7->destroy ();
	      dewarnf ("subset criterion is always %s",
		       result ? "true" : "false");
	      if (result)
		$$ = $4;
	      else {
		$$->destroy ();
		$$ = CSE (new class EmptySet);
	      }
	    }
	    else
	      $$ = CSE (new class Submarking
			(currentVariables.top (), $4, $7));
	  }
	  else
	    $7->destroy ();
	  if ($2)
	    currentVariables.pop ();
	}
	;

arg_list:
	{
	  if (class Function* f = functionStack.top ()) {
	    if (!f->getArity ())
	      $$ = new class ExpressionList;
	    else {
	      deerrorf ("function `%s' expects %u!=0 arguments",
			f->getName (), f->getArity ());
	      $$ = NULL;
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	{
	  const class Type* type = NULL;

	  if (class Function* f = functionStack.top ()) {
	    if (f->getArity ())
	      type = (*f)[0]->getType ();
	    else
	      deerrorf ("function `%s' has no arguments", f->getName ());
	  }

	  pushTypeContext (type);
	}
	formula
	{
	  if (currentType &&
	      isCompatible (($2 = noMarking ($2)), currentType)) {
	    if (class Function* f = functionStack.top ())
	      ($$ = new class ExpressionList)->append ($2);
	    else {
	      $2->destroy (); $$ = NULL;
	    }
	  }
	  else {
	    $2->destroy (); $$ = NULL;
	  }
	  popTypeContext ();
	}
	|
	arg_list ','
	{
	  const class Type* type = NULL;

	  if (class Function* f = functionStack.top ()) {
	    if ($1) {
	      if (f->getArity () > $1->size ())
		type = (*f)[$1->size ()]->getType ();
	      else
		deerrorf ("function `%s' takes only %u arguments",
			  f->getName (), f->getArity ());
	    }
	  }

	  pushTypeContext (type);
	}
	formula
	{
	  if (($$ = $1) && currentType &&
	      isCompatible (($4 = noMarking ($4)), currentType)) {
	    if (class Function* f = functionStack.top ())
	      $$->append ($4);
	    else {
	      $$->destroy (); $4->destroy (); $$ = NULL;
	    }
	  }
	  else {
	    $$->destroy (); $4->destroy (); $$ = NULL;
	  }
	  popTypeContext ();
	}
	|
	arg_list ','
	{ if (($$ = $1)) deerrorf ("extraneous comma in argument list"); }
	;


colon_formula:
	formula %prec ':'
	{
	  if (($1 = noMarking ($1)))
	    ($$ = new class ExpressionList)->prepend ($1);
	  else {
	    $1->destroy (); $$ = NULL;
	  }
	}
	|
	colon_formula ':' formula
	{
	  $3 = noMarking ($3);
	  if (($$ = $1) && $3) {
	    if ($$->getType () && $$->getType () != $3->getType ()) {
	      deerror ("incompatible types in ?: expression");
	      $$->destroy (); $3->destroy (); $$ = NULL;
	    }
	    else
	      $$->prepend ($3);
	  }
	  else {
	    $$->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	;

_no_qualifier:
	qualifier
	{
	  if ($1) {
	    $1->destroy ();
	    deerror ("unexpected transition instance qualifier");
	  }
	}
	;

qualifier:
	{ $$ = NULL; }
	|
	qualifier '[' qual_expr ']'
	{
	  if (($$ = $1)) {
	    deerror ("transition instance qualifier already specified");
	    $3->destroy ();
	  }
	  else
	    $$ = $3;
	}
	;

qual_expr:
	transname ':'
	{ currentContextTransition = const_cast<class Transition*>($1); }
	formula
	{
	  $4 = ensureBool (ensureExpr ($4));
	  if ($1 && $4)
	    $$ = CSE (new class TransitionQualifier ($1, CSE ($4)));
	  else
	    $4->destroy (), $$ = NULL;
	  currentContextTransition = NULL;
	}
	|
	transname
	{ $$ = $1 ? CSE (new class TransitionQualifier ($1, NULL)) : NULL; }
	|
	'(' qual_expr ')'
	{ $$ = $2; }
	|
	'!' qual_expr
	{ if (($$ = $2)) $$ = fold (new class NotExpression ($$)); }
	|
	qual_expr AND qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bAnd, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr OR qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bOr, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr IMPL qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bImpl, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr XOR qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bXor, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr EQUIV qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bEquiv, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	;

transname:
	_opt_trans name
	{
	  if ($2) {
	    if (!($$ = theNet.getTransition ($2)))
	      deerrorf ("undefined transition `%s'", $2);
	  }
	  else
	    $$ = NULL;
	  delete[] $2;
	}
	;

_opt_trans:
	{ $$ = false; }
	|
	_opt_trans TRANS
	{ $$ = true; if ($1) deerror ("ignoring extraneous `trans'"); }
	|
	_opt_trans PLACE
	{ $$ = $1; deerror ("expected `trans', got `place'"); }
	;

placename:
	_nopt_place name
	{
	  if ($2) {
	    if (!($$ = theNet.getPlace ($2)))
	      deerrorf ("undefined place `%s'", $2);
	  }
	  else
	    $$ = NULL;
	  delete[] $2;
	}
	;

_nopt_place:
	PLACE
	|
	_nopt_place PLACE
	{ deerror ("ignoring extraneous `place'"); }
	|
	TRANS
	{ deerror ("expected `place', got `trans'"); }
	|
	_nopt_place TRANS
	{ deerror ("ignoring extraneous `trans'"); }
	;

%%

#include "cmdline.h"

const char*
deline ()
{
  static char prompt1[21], prompt2[21];
  card_t number = currentState.getNumber ();
  sprintf (prompt1, "@%u$", number);
  sprintf (prompt2, "@%u>", number);
  return cmdline (prompt1, prompt2);
}

#ifdef HAS_READLINE
/** Duplicate a string
 * @param s	string to be duplicated
 * @return	a copy of the string allocated with malloc ()
 */
inline static char*
dupstr (const char* s)
{
  size_t length = strlen (s) + 1;
  char* r = static_cast<char*>(malloc (length));
  if (r)
    memcpy (r, s, length);
  return r;
}

/** flag: is quoting desired? */
extern int rl_filename_completion_desired;
/** character to be appended to completed words (0 or ' ') */
extern int rl_completion_append_character;

/** Iterate through all keywords starting with given text
 * @param text		text to be completed
 * @param state		flag: if state==0, start from the beginning
 * @return		a command name
 */
extern "C" char*
marde_keywords (const char* text, int state)
{
  /** Keywords of the language */
  static const char* keywords[] = {
    "eval", "show", "succ", "pred", "go", "stats", "help", "path",
    "function", "exit", "union", "is", "always", "exists", "until",
    "release", "subset", "intersect", "equals", "minus", "atom",
    "empty", "place \"", "trans \"", "true", "false", "undefined", "fatal",
    "cardinality"
  };

  /** Number of the last keyword searched */
  static unsigned i;
  /** Length of the string */
  static size_t length;

  if (!state) {
    i = 0;
    length = strlen (text);
  }
  else {
    assert (i <= (sizeof keywords) / sizeof *keywords);
    assert (length == strlen (text));
  }

  while (i < (sizeof keywords) / sizeof *keywords) {
    const char* name = keywords[i++];
    if (!strncmp (name, text, length)) {
      if (name[strlen (name) - 1] == '"')
	rl_completion_append_character = 0;
      return dupstr (name);
    }
  }

  return NULL;
}

/** Iterate through place names starting with given text
 * @param text		text to be completed
 * @param state		flag: if state==0, start from the beginning
 * @return		a place name
 */
extern "C" char*
marde_placenames (const char* text, int state)
{
  /** Number of the last place searched */
  static unsigned place;
  /** Length of the string */
  static size_t length;

  if (!state) {
    place = 0;
    length = strlen (text);
    rl_filename_completion_desired = 1;
  }
  else {
    assert (place <= theNet.getNumPlaces ());
    assert (length == strlen (text));
    assert (rl_filename_completion_desired);
  }

  while (place < theNet.getNumPlaces ()) {
    const char* name = theNet.getPlace (place++)->getName ();
    if (!strncmp (name, text, length))
      return dupstr (name);
  }

  return NULL;
}

/** Iterate through transition names starting with given text
 * @param text		text to be completed
 * @param state		flag: if state==0, start from the beginning
 * @return		a transition name
 */
extern "C" char*
marde_transnames (const char* text, int state)
{
  /** Number of the last transition searched */
  static unsigned trans;
  /** Length of the string */
  static size_t length;

  if (!state) {
    trans = 0;
    length = strlen (text);
    rl_filename_completion_desired = 1;
  }
  else {
    assert (trans <= theNet.getNumTransitions ());
    assert (length == strlen (text));
    assert (rl_filename_completion_desired);
  }

  while (trans < theNet.getNumTransitions ()) {
    const char* name = theNet.getTransition (trans++)->getName ();
    if (!strncmp (name, text, length))
      return dupstr (name);
  }

  return NULL;
}

/** Iterate through type names starting with given text
 * @param text		text to be completed
 * @param state		flag: if state==0, start from the beginning
 * @return		a type name
 */
extern "C" char*
marde_typenames (const char* text, int state)
{
  /** Last type searched */
  static Net::TypeMap::const_iterator type;
  /** Length of the string */
  static size_t length;

  if (!state) {
    type = theNet.begin ();
    length = strlen (text);
    rl_filename_completion_desired = 1;
  }
  else {
    assert (length == strlen (text));
    assert (rl_filename_completion_desired);
  }

  while (type != theNet.end ()) {
    const char* name = (type++)->first;
    if (!strncmp (name, text, length))
      return dupstr (name);
  }

  return NULL;
}

/** Transition whose variable names are to be completed */
static const class Transition* completionTransition = NULL;

/** Iterate through variable names starting with given text
 * @param text		text to be completed
 * @param state		flag: if state==0, start from the beginning
 * @return		a variable name
 */
extern "C" char*
marde_variables (const char* text, int state)
{
  /** Last variable searched */
  static Transition::const_iterator var;
  /** Length of the string */
  static size_t length;

  assert (completionTransition != NULL);

  if (!state) {
    var = completionTransition->begin ();
    length = strlen (text);
    rl_filename_completion_desired = 1;
  }
  else {
    assert (length == strlen (text));
    assert (rl_filename_completion_desired);
  }

  while (var != completionTransition->end ()) {
    const char* name = (var++)->first;
    if (!strncmp (name, text, length))
      return dupstr (name);
  }

  return NULL;
}

/** Iterate through other names starting with given text
 * @param text		text to be completed
 * @param state		flag: if state==0, start from the beginning
 * @return		a name
 */
extern "C" char*
marde_othernames (const char* text, int state)
{
  return NULL;
}

/** Completion function of readline
 * @param word		word to be completed
 * @param completer	the completion function
 */
extern "C" char **
completion_matches (const char* word,
		    char* (*completer) (const char*, int));

/** Complete place or transition names
 * @param word		word to be completed
 * @param start		start offset of the string to be completed
 * @param end		end offset of the string to be completed
 */
extern "C" char**
marde_completer (const char* word, int start, int end)
{
  /** The line gathered so far */
  extern char* rl_line_buffer;
  assert (start >= 0 && start <= end);
  assert (size_t (end - start) == strlen (word));
  assert (size_t (end) <= strlen (rl_line_buffer));
  rl_completion_append_character = ' ';
  completionTransition = NULL;

  if ((start >= 6 && !memcmp (&rl_line_buffer[start - 6], "place ", 6)) ||
      (start >= 7 && !memcmp (&rl_line_buffer[start - 7], "place \"", 7)))
    return completion_matches (word, marde_placenames);
  if ((start >= 6 && !memcmp (&rl_line_buffer[start - 6], "trans ", 6)) ||
      (start >= 7 && !memcmp (&rl_line_buffer[start - 7], "trans \"", 7)))
    return completion_matches (word, marde_transnames);
  if ((start >= 3 && !memcmp (&rl_line_buffer[start - 3], "is ", 3)) ||
      (start >= 4 && !memcmp (&rl_line_buffer[start - 4], "is \"", 4)))
    return completion_matches (word, marde_typenames);
  else {
    // try completing transition variables
    int transname_end = start;
    int transname_start = start;
    bool quoted = false;

    for (int i = start; --i > 0; ) {
      if (rl_line_buffer[i] == '"' && rl_line_buffer[i - 1] != '\\')
	if (!(quoted = !quoted) && transname_end < transname_start)
	  transname_start = i;
      if (quoted)
	continue;

      switch (rl_line_buffer[i]) {
      case ':':
	transname_end = i - 1;
	break;
      case ' ':
      case '\t':
      case '\r':
      case '\n':
      case '\v':
      case '\f':
      case '[':
	if (transname_end <= transname_start)
	  transname_start = i + 1;
	if (transname_start < transname_end && transname_start < start) {
	  // Unquote the string.
	  // This will not process octal or hex constants after backslash.
	  char* unquoted = new char[transname_end - transname_start + 1];
	  size_t length = 0;
	  for (int j = transname_start; j <= transname_end; j++) {
	    switch (rl_line_buffer[j]) {
	    case '"':
	      continue;
	    case '\\':
	      switch (rl_line_buffer[++j]) {
	      case 'a': unquoted[length++] = '\a'; break;
	      case 'b': unquoted[length++] = '\b'; break;
	      case 't': unquoted[length++] = '\t'; break;
	      case 'n': unquoted[length++] = '\n'; break;
	      case 'v': unquoted[length++] = '\v'; break;
	      case 'f': unquoted[length++] = '\f'; break;
	      case 'r': unquoted[length++] = '\r'; break;
	      default:
		unquoted[length++] = rl_line_buffer[j];
	      }
	      break;
	    default:
	      unquoted[length++] = rl_line_buffer[j];
	    }
	  }
	  unquoted[length] = 0;

	  for (unsigned trans = theNet.getNumTransitions (); trans--; ) {
	    const char* name = theNet.getTransition (trans)->getName ();
	    if (!strcmp (name, unquoted)) {
	      completionTransition = theNet.getTransition (trans);
	      break;
	    }
	  }

	  delete[] unquoted;

	  if (completionTransition)
	    return completion_matches (word, marde_variables);
	}

	if (rl_line_buffer[i] != '[') {
	  transname_end = i - 1;
	  break;
	}
	// fall through
      case ']':
	return completion_matches (word, marde_keywords);
      }
    }

    return completion_matches (word, marde_keywords);
  }
}
#endif // HAS_READLINE

void
dereset ()
{
  command = true;
  currentExpressions.clear ();
  delete[] currentParameters;
  currentParameters = 0;
  while (!functionStack.empty ()) functionStack.pop ();
  currentContextTransition = 0;
  currentType = currentContextType = 0;
  currentComponent = 0;
  while (!typeStack.empty ()) typeStack.pop ();
  emptyParsed = false;
  while (!structLevels.empty ()) structLevels.pop ();
  delexreset ();
  if (!currentState.isState ())
    currentState = theGraph.find (0);
#ifdef HAS_READLINE
  init_cmdline ("marde", marde_completer, marde_othernames);
#endif // HAS_READLINE
}

static void
go (card_t state)
{
  if (state < theGraph.size ())
    currentState = theGraph.find (state);
}

static void
show (card_t state)
{
  if (state < theGraph.size ()) {
    class GlobalMarking* m = theGraph.fetchState (state);
    m->display (thePrinter);
    thePrinter.finish ();
    delete m;
  }
}

static void
succ (card_t state)
{
  if (state < theGraph.size ()) {
    const class State& s = theGraph.find (state).getState ();
    const class Transition* transition;
    for (unsigned i = 0; i < s.getNumEvents (); i++) {
      class Valuation* v =
	theGraph.fetchEvent (s[i].getOffset (), transition);
      thePrinter.printRaw ("transition ");
      thePrinter.print (transition->getName ());
      thePrinter.printRaw ("->@");
      thePrinter.print (s[i].getSuccessor ().getNumber ());
      thePrinter.finish ();
      v->display (thePrinter);
      thePrinter.finish ();
      delete v;
    }
  }
}

static void
eval (card_t state, class Expression* expr)
{
  class Valuation v;
  if (state >= theGraph.size () || !expr);
  else if (isSet (expr)) {
    class GlobalMarking* m = theGraph.fetchState (state);
    v.setGlobalMarking (m);
    if (class PlaceMarking* p = expr->meval (v)) {
      p->display (thePrinter);
      delete p;
    }
    else
      v.display (thePrinter);
    thePrinter.finish ();
    delete m;
  }
  else if (expr->isBasic ()) {
    if (class Value* value = expr->eval (&v)) {
      value->display (thePrinter);
      delete value;
    }
    else
      v.display (thePrinter);
    thePrinter.finish ();
  }
  else if (class Property* property = Property::create (thePrinter, *expr)) {
    class Product product (theGraph, *property);
    class GlobalMarking* m = theGraph.fetchState (state);
    if (product.analyze (*m))
      thePrinter.printRaw ("property holds");
    else {
      thePrinter.printRaw ("counterexample found");
    }
    thePrinter.finish ();
    delete m;
  }
  expr->destroy ();
}

static char*
convertString (class StringBuffer* sbuf)
{
  assert (sbuf != NULL);
  char* str = newString (sbuf->getString ());
  if (strlen (str) != sbuf->getLength ())
    deerror ("truncating name containing NUL chars");
  delete sbuf;
  return str;
}

static void pushTypeContext (const class Type* type)
{
  StructLevel p (currentContextType, currentComponent);
  structLevels.push (p);
  StructLevel q (currentType, 0);
  structLevels.push (q);

  currentType = type;
  currentContextType = NULL, currentComponent = 0;
}

static void popTypeContext ()
{
  StructLevel p = structLevels.top ();
  structLevels.pop ();
  currentType = p.first;
  assert (!p.second);
  p = structLevels.top ();
  structLevels.pop ();
  currentContextType = p.first;
  currentComponent = p.second;
}

static const class VariableDefinition*
getParameter (char* name)
{
  for (unsigned i = 0; i < currentArity; i++) {
    if (!strcmp (currentParameters[i]->getName (), name))
      return currentParameters[i];
  }

  return NULL;
}

static void
addParameter (const class Type* type, char* name)
{
  if (!type || !name) {
    delete[] name;
    return;
  }

  if (getParameter (name)) {
    deerrorf ("ignoring redefinition of parameter `%s'", name);
    delete[] name;
    return;
  }

  if (dewarnings) {
    if (getConstantFunction (name))
      dewarnf ("parameter `%s' shadows a constant", name);
    else if (theNet.getType (name))
      dewarnf ("parameter `%s' shadows a type name", name);
  }

  class VariableDefinition** v =
    new class VariableDefinition*[currentArity + 1];
  assert (!currentArity || currentParameters);
  assert (currentArity || !currentParameters);

  memcpy (v, currentParameters, currentArity * sizeof *v);
  delete[] currentParameters;
  currentParameters = v;
  currentParameters[currentArity++] =
    new class VariableDefinition (name, type);
}

static class Function*
getFunction (const char* name)
{
  class Function* f = currentContextTransition
    ? currentContextTransition->getFunction (name)
    : NULL;
  if (!f)
    f = theNet.getFunction (name);
  return f;
}

static class Function*
getConstantFunction (const char* name)
{
  class Function* f = currentContextTransition
    ? currentContextTransition->getFunction (name)
    : NULL;
  if (!f || f->getArity ())
    f = theNet.getFunction (name);
  return f && f->getArity () ? NULL : f;
}

static const class VariableDefinition*
getVariable (char* name)
{
  assert (name != NULL);

  const class VariableDefinition* v = getParameter (name);
  const class Function* f = getConstantFunction (name);

  if (v || (v = currentVariables.find (name)));
  else if (currentContextTransition &&
	   (v = currentContextTransition->getVariable (name)));
  else if (!f &&
	   !(currentType && currentType->getKind () == Type::tEnum &&
	     static_cast<const class EnumType*>(currentType)->getValue (name)))
    deerrorf ("undefined variable `%s'", name);
  assert (!v || v->getType ());
  delete[] name;
  return v;
}

static class Expression*
resolveName (char* name)
{
  class Function* f = getConstantFunction (name);
  const int_t*const i =
    (currentType && currentType->getKind () == Type::tEnum)
    ? static_cast<const class EnumType*>(currentType)->getValue (name)
    : NULL;
  if (const class VariableDefinition* d = getVariable (name))
    return static_cast<class Variable*>(CSE (new class Variable (d)));
  else if (class Expression* expr = f ? f->expand (NULL) : NULL)
    return fold (expr);
  else if (i)
    return enumValue (static_cast<const class EnumType*>(currentType), *i);
  else
    return NULL;
}

bool
isSet (const class Expression* expr)
{
  if (!expr)
    return false;
  switch (expr->getKind ()) {
  case Expression::eMarking:
  case Expression::ePlaceContents:
  case Expression::eSubmarking:
  case Expression::eEmptySet:
    return true;
  case Expression::eSet:
    switch (static_cast<const class SetExpression*>(expr)->getOp ()) {
    case SetExpression::sIntersection:
    case SetExpression::sMinus:
    case SetExpression::sUnion:
      return true;
    case SetExpression::sSubset:
    case SetExpression::sEquals:
      return false;
    }
  default:
    return false;
  }
}

static class Expression*
ensureSet (class Expression* expr)
{
  return isSet (expr) ? expr : ensureMarking (expr);
}

static class Marking*
ensureMarking (class Expression* expr)
{
  if (!expr)
    return NULL;

  assert (expr->getType () || expr->getKind () == Expression::eEmptySet);

  if (expr->getKind () == Expression::eMarking)
    return static_cast<class Marking*>(expr);
  else if (expr->isBasic ()) {
    if (!currentType || makeCompatible (expr, currentType)) {
      if (currentType) {
	if (!expr->getType () ||
	    expr->getType ()->isAlwaysAssignable (currentType))
	  expr->setType (currentType);
	else {
	  deerrorf ("marking expression has too broad type `%s'",
		    expr->getType ()->getName ()
		    ? expr->getType ()->getName ()
		    : expr->getType ()->getSyntacticName ());
	  expr->destroy ();
	  return NULL;
	}
      }
      class Marking* m = new class Marking (NULL);
      m->setToken (expr);
      return m;
    }
    else {
      expr->destroy ();
      return NULL;
    }
  }
  else {
    deerror ("marking expression expected");
    expr->destroy ();
    return NULL;
  }
}

static class Expression*
noMarking (class Expression* expr)
{
  if (expr && expr->getKind () == Expression::eMarking) {
    deerror ("marking expression not allowed here");
    expr->destroy ();
    return NULL;
  }

  return expr;
}

static class Expression*
ensureExpr (class Expression* expr)
{
  expr = noMarking (expr);

  if (expr && !expr->isBasic ()) {
    deerror ("basic expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureBool (class Expression* expr, bool undefined)
{
  expr = noMarking (expr);

  if (!expr)
    return NULL;

  if (expr->getKind () == Expression::eUndefined) {
    if (undefined)
      return expr;
    deerror ("undefined or fatal expression is not meaningful here");
    expr->destroy (); return NULL;
  }

  if (!expr->getType ()) {
    deerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (expr->getType ()->getKind () != Type::tBool) {
    deerror ("boolean expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureInt (class Expression* expr)
{
  expr = ensureExpr (expr);

  if (!expr)
    return NULL;

  if (!expr->getType ()) {
    deerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (expr->getType ()->getKind () != Type::tInt) {
    deerror ("integer expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureBuffer (class Expression* expr)
{
  expr = ensureExpr (expr);

  if (!expr)
    return NULL;

  if (!expr->getType ()) {
    deerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (expr->getType ()->getKind () != Type::tBuffer) {
    deerror ("buffer expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureOrdered (class Expression* expr)
{
  expr = ensureExpr (expr);

  if (!expr)
    return NULL;

  if (!expr->getType ()) {
    deerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (!expr->getType ()->isOrdered ()) {
    deerror ("expression of ordered type expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
enumValue (const class EnumType* type, int_t value)
{
  class EnumValue* e = new class EnumValue (type, value);
  if (type->isConstrained (e))
    return CSE (new class Constant (e));
  else {
    if (const char* c = type->getEnumName (value))
      deerrorf ("enumeration constant %s=%d out of range for %s", c, value,
		type->getName ()
		? type->getName ()
		: type->getSyntacticName ());
    else
      deerrorf ("enumeration constant %d out of range for %s", value,
		type->getName ()
		? type->getName ()
		: type->getSyntacticName ());
    delete e;
    return NULL;
  }
}

static class Expression*
namedComponent (char* name, class Expression* expr)
{
  if (!name) {
    expr->destroy ();
    return NULL;
  }

  if (!currentContextType || currentContextType->getKind () != Type::tStruct) {
    deerrorf ("named component `%s' can only be used with struct values",
	      name);
    expr->destroy ();
    delete[] name;
    return NULL;
  }

  const class StructType& type =
    static_cast<const class StructType&>(*currentContextType);

  if (currentComponent < type.getSize ()) {
    const char* comp = type.getComponentName (currentComponent);
    if (strcmp (name, comp)) {
      deerrorf ("expecting named component `%s' instead of `%s'", comp, name);
      expr->destroy ();
      delete[] name;
      return NULL;
    }
  }
  else {
    deerrorf ("extraneous named component `%s' at end of struct", name);
    expr->destroy ();
    delete[] name;
    return NULL;
  }

  delete[] name;
  return expr;
}

static void
beginQuantifier (char* name, const class Type* type)
{
  if (name && type) {
    if (dewarnings) {
      if (currentVariables.find (name))
	dewarnf ("quantifier `%s' shadows another quantifier", name);
      else if (currentContextTransition &&
	       currentContextTransition->getVariable (name))
	dewarnf ("quantifier `%s' shadows a transition variable", name);
      else if (getConstantFunction (name))
	dewarnf ("quantifier `%s' shadows a constant", name);
      else if (currentType && currentType->getKind () == Type::tEnum) {
	if (const int_t*const c =
	    static_cast<const class EnumType*>(currentType)->getValue (name))
	  dewarnf ("quantifier `%s' shadows an enumeration constant %d",
		   name, *c);
      }
      else if (theNet.getType (name))
	dewarnf ("quantifier `%s' shadows a type name", name);
    }

    currentVariables.push (new class VariableDefinition (name, type, true));
  }
  assert (!currentContextType);
  struct TypeContext tc = {
    currentContextType,
    currentType,
    currentComponent
  };
  typeStack.push (tc);

  currentContextType = NULL;
  currentComponent = 0;
  currentType = type;
}

static void
continueQuantifier (ASSERT1 (const class Type* type))
{
  assert (!currentContextType);
  assert (currentType == type);
  assert (!typeStack.empty ());
  struct TypeContext tc = typeStack.top ();
  typeStack.pop ();

  currentContextType = tc.context;
  currentType = tc.component;
  currentComponent = tc.numComponent;
}

static class Marking*
endQuantifier (char* name, const class Type* type,
	       class Expression* condition, class Marking* expr)
{
  if (name && type && expr) {
    expr->addQuantifier (new class Quantifier
			 (currentVariables.top (),
			  condition = ensureBool (condition)));
    if (condition) {
      condition->associateVariables (true);
#ifdef EXPR_CACHE
      if (!condition->invalidate (currentVariables.top ()))
	dewarnf ("condition does not depend on quantifier `%s'",
		 currentVariables.top ()->getName ());
#endif // EXPR_CACHE
    }
  }
  else
    condition->destroy (), expr->destroy (), expr = NULL;

  if (name && type)
    currentVariables.pop ();
  else
    delete[] name;

  return expr;
}

static bool
makeCompatible (class Expression*& expr, const class Type* type, bool casting)
{
  assert (expr && expr->getType ());
  assert (type != NULL);

  if (expr->getKind () == Expression::eConstant) {
    const class Value* const cv =
      static_cast<class Constant*>(expr)->getValue ();
    int_t value;
    switch (cv->getKind ()) {
    case Value::vInt:
      value = static_cast<const class IntValue*>(cv)->getValue ();
      break;
    case Value::vBool:
      value = static_cast<const class BoolValue*>(cv)->getValue () ? 1 : 0;
      break;
    case Value::vChar:
      value = static_cast<const class CharValue*>(cv)->getValue ();
      break;
    case Value::vEnum:
      value = static_cast<const class EnumValue*>(cv)->getValue ();
      break;
    default:
      if (cv->getType ()->isAssignable (type)) {
	if (type->isConstrained (cv)) {
	  expr->setType (type);
	  return true;
	}
	else {
	  deerrorf ("%s constant out of range for type %s",
		    expr->getType ()->getSyntacticName (),
		    type->getName ()
		    ? type->getName ()
		    : type->getSyntacticName ());
	  return false;
	}
      }
      else {
      incompatible:
	deerrorf ("%s constant is meaningless for type %s",
		  cv->getType ()->getSyntacticName (),
		  type->getName ()
		  ? type->getName ()
		  : type->getSyntacticName ());
	return false;
      }
    }

    class Value* v = NULL;

    switch (type->getKind ()) {
    case Type::tInt:
      if (!type->isConstrained (v = new class IntValue (type, value)))
	delete v, v = NULL;
      break;
    case Type::tBool:
      if (!(value == 0 || value == 1) ||
	  !type->isConstrained (v = new class BoolValue (type, value == 1)))
	delete v, v = NULL;
      break;
    case Type::tChar:
      if (value < CHAR_T_MIN || value > CHAR_T_MAX ||
	  !type->isConstrained (v = new class CharValue (type, value)))
	delete v, v = NULL;
      break;
    case Type::tEnum:
      if (!type->isConstrained (v = new class EnumValue (type, value)))
	delete v, v = NULL;
      break;

    default:
      goto incompatible;
    }

    if (v) {
      if (!casting && dewarnings && !expr->getType ()->isAssignable (type))
	dewarnf ("converting %s constant %d to type %s",
		 expr->getType ()->getSyntacticName (), value,
		 type->getName ()
		 ? type->getName ()
		 : type->getSyntacticName ());
      expr->destroy ();
      expr = CSE (new class Constant (v));
      return true;
    }
    else {
      const char* const name =
	type->getName () ? type->getName () : type->getSyntacticName ();

      switch (cv->getKind ()) {
      case Value::vInt:
	deerrorf ("%s constant %d out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<const class IntValue*>(cv)->getValue (),
		  name);
	break;
      case Value::vBool:
	deerrorf ("%s constant %s out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<const class BoolValue*>(cv)->getValue ()
		  ? "true" : "false", name);
	break;
      case Value::vEnum:
	deerrorf ("%s constant %d out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<const class EnumValue*>(cv)->getValue (),
		  name);
	break;
      case Type::tChar:
	deerrorf ("%s constant 0x%02x out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<unsigned char>
		  (static_cast<const class CharValue*>(cv)->getValue ()),
		  name);
	break;
      default:
	assert (false);
      }
      return false;
    }
  }
  else
    return true;
}

static bool
isCompatible (class Expression*& expr, const class Type* type)
{
  if (!type) {
    deerror ("too many components for struct");
    return false;
  }

  if (!expr)
    return false;

  if (!expr->getType ()) {
    deerror ("expression has no type");
    return false;
  }

  if (!makeCompatible (expr, type))
    return false;

  if (!expr->getType ()->isAlwaysAssignable (type)) {
    const char* reason = Type::isCompatible (expr->getType (), type)
      ? "has too broad type for" : "is incompatible with type";
    deerrorf ("expression %s `%s'", reason,
	     type->getName ()
	     ? type->getName ()
	     : type->getSyntacticName ());
    return false;
  }

  return true;
}

static bool
isCompatible (class Expression*& expr, const class StructExpression* s)
{
  assert (s != NULL);
  const class Type* type = s->getNextType ();
  const char* structName = s->getType ()->getName ();

  if (!type) {
    if (structName)
      deerrorf ("too many components for struct `%s'",
		structName);
    else
      deerror ("too many components for struct");
    return false;
  }

  const char* comp =
    static_cast<const class StructType*>(s->getType ())
    ->getComponentName (s->size ());

  if (!expr) {
    if (structName)
      deerrorf ("error in expression for struct `%s' component `%s'",
		structName, comp);
    else
      deerrorf ("error in expression for struct component `%s'", comp);
    return false;
  }

  if (!expr->getType ()) {
    if (structName)
      deerrorf ("expression for struct `%s' component `%s' has no type",
		structName, comp);
    else
      deerrorf ("expression for struct component `%s' has no type", comp);
    return false;
  }

  if (!makeCompatible (expr, type))
    return false;

  if (!expr->getType ()->isAlwaysAssignable (type)) {
    const char* reason = Type::isCompatible (expr->getType (), type)
      ? "too broad type" : "type mismatch";

    if (structName)
      deerrorf ("%s for struct `%s' component `%s'", reason,
		structName, comp);
    else
      deerrorf ("%s for struct component `%s'", reason, comp);
    return false;
  }

  return true;
}

static bool
isCompatible (class Expression*& expr1, class Expression*& expr2)
{
  if (!expr1 || !expr2)
    return false;

  if (!expr1->getType () || !expr2->getType ()) {
    deerror ("expression has no type");
    return false;
  }

  if (!makeCompatible (expr1, expr2->getType ()) ||
      !makeCompatible (expr2, expr1->getType ()))
    return false;

  if (!Type::isCompatible (expr1->getType (), expr2->getType ())) {
    deerror ("subexpressions have incompatible types");
    return false;
  }

  return true;
}

static bool
isOrdered (class Expression*& expr1, class Expression*& expr2)
{
  if (isCompatible (expr1, expr2)) {
    if (expr1->getType ()->isOrdered () && expr2->getType ()->isOrdered ())
      return true;

    deerror ("invalid comparison for unordered types");
  }

  return false;
}

static class Expression*
fold (class Expression* expr)
{
  if (expr && expr->getKind () != Expression::eConstant && expr->isBasic ()) {
    class Valuation v;
    class Value* value = expr->eval (&v);
    if (value) {
      assert (v.isOK ());
      expr->destroy ();
      return CSE (new class Constant (value));
    }
    else {
      bool ok = true;
      if (v.isDivisionByZero ())
	ok = false, deerror ("division by zero");
      if (v.isOverflow ())
	ok = false, deerror ("overflow");
      if (v.isShiftError ())
	ok = false, deerror ("shift error");
      if (v.isConstraintViolation ())
	ok = false, deerror ("constraint violation");
      if (v.isUnionViolation ())
	ok = false, deerror ("union violation");
      if (v.isUndefinedComponent ())
	ok = false, deerror ("undefined component");
      if (v.isBufferViolation ())
	ok = false, deerror ("buffer violation");
      assert (!v.isNegativeCardinality ());
      assert (!v.isUnificationError ());
      assert (!v.isCapacityViolation ());
      if (!ok) {
	expr->destroy ();
	return NULL;
      }
      else {
	if (v.isFatalExpression ()) {
	  expr->destroy ();
	  return CSE (new class Undefined (Undefined::fFatalError));
	}
	if (v.isUndefinedExpression ()) {
	  expr->destroy ();
	  return CSE (new class Undefined (Undefined::fError));
	}

	assert (v.isUndefinedValue ());
	return CSE (expr);
      }
    }
  }
  else
    return CSE (expr);
}

static class Expression*
newRelopExpression (enum RelopExpression::Op op,
		    class Expression* left,
		    class Expression* right)
{
  if (left == right) {
    bool result;
    switch (op) {
    case RelopExpression::rLess:
    case RelopExpression::rGreater:
    case RelopExpression::rInequal:
      result = false;
      break;
    case RelopExpression::rEqual:
    case RelopExpression::rGreaterEqual:
    case RelopExpression::rLessEqual:
      result = true;
      break;
    }

    dewarnf ("result of comparison is always %s", result ? "true" : "false");
    left->destroy ();
    right->destroy ();
    return CSE (new class Constant
		(new class BoolValue (Net::getBoolType (), result)));
  }

  return fold (new class RelopExpression (op, left, right));
}

static class Expression*
newBooleanBinop (enum BooleanBinop::Op op,
		 class Expression* left,
		 class Expression* right)
{
  assert (left && right);
  assert (left->getKind () == Expression::eUndefined ||
	  (left->getType () && left->getType ()->getKind () == Type::tBool));
  assert (right->getKind () == Expression::eUndefined ||
	  (right->getType () && right->getType ()->getKind () == Type::tBool));
  const class Value* v;
  bool value;
  assert (left == CSE (left));
  assert (right == CSE (right));

  if (left->getKind () == Expression::eConstant) {
    v = static_cast<class Constant*>(left)->getValue ();
    left->destroy ();
    assert (v && v->getKind () == Value::vBool);
    value = static_cast<const class BoolValue*>(v)->getValue ();

    switch (op) {
    case BooleanBinop::bAnd:
      if (value)
	return right;
      else {
	right->destroy ();
      False:
	dewarn ("subexpression is always false");
	return CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), false)));
      }
    case BooleanBinop::bOr:
      if (value) {
	right->destroy ();
      True:
	dewarn ("subexpression is always true");
	return CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), true)));
      }
      else
	return right;
    case BooleanBinop::bImpl:
      if (value)
	return right;
      else {
	right->destroy ();
	goto True;
      }
    case BooleanBinop::bXor:
      if (value)
	return fold (new class NotExpression (right));
      else
	return right;
    case BooleanBinop::bEquiv:
      if (value)
	return right;
      else
	return fold (new class NotExpression (right));
    }
  }
  else if (right->getKind () == Expression::eConstant) {
    v = static_cast<class Constant*>(right)->getValue ();
    right->destroy ();
    assert (v && v->getKind () == Value::vBool);
    value = static_cast<const class BoolValue*>(v)->getValue ();

    switch (op) {
    case BooleanBinop::bAnd:
      if (value)
	return left;
      else {
	left->destroy ();
	goto False;
      }
    case BooleanBinop::bOr:
      if (value) {
	left->destroy ();
	goto True;
      }
      else
	return left;
    case BooleanBinop::bImpl:
      if (value) {
	left->destroy ();
	goto True;
      }
      else
	return fold (new class NotExpression (left));
    case BooleanBinop::bXor:
      if (value)
	return fold (new class NotExpression (left));
      else
	return left;
    case BooleanBinop::bEquiv:
      if (value)
	return left;
      else
	return fold (new class NotExpression (left));
    }
  }
  else
    return CSE (new class BooleanBinop (op, left, right));

  assert (false);
  return NULL;
}

static class Expression*
newSetExpression (enum SetExpression::Op op,
		  class Expression* left,
		  class Expression* right)
{
  left = ensureSet (left);
  right = ensureSet (right);
  if (!left || !right) {
    left->destroy (), right->destroy ();
    return NULL;
  }

  if (left->getKind () == Expression::eEmptySet) {
    if (right->getKind () == Expression::eEmptySet) {
    Same:
      switch (op) {
      case SetExpression::sIntersection:
      case SetExpression::sMinus:
      case SetExpression::sUnion:
	right->destroy ();
	return left;
      case SetExpression::sSubset:
      case SetExpression::sEquals:
	left->destroy (), right->destroy ();
      True:
	return CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), true)));
      }
      assert (false);
    }
    else {
      switch (op) {
      case SetExpression::sIntersection:
      case SetExpression::sMinus:
	right->destroy ();
	return left;
      case SetExpression::sUnion:
	left->destroy ();
	return right;
      case SetExpression::sSubset:
	right->destroy ();
	goto True;
      case SetExpression::sEquals:
	break;
      }
    }

    if (!left->getType () || !right->getType ()) {
      deerror ("expression has no type");
      left->destroy (), right->destroy ();
      return NULL;
    }
  }
  else if (right->getKind () == Expression::eEmptySet) {
    switch (op) {
    case SetExpression::sIntersection:
      left->destroy ();
      return right;
    case SetExpression::sMinus:
    case SetExpression::sUnion:
      right->destroy ();
      return left;
    case SetExpression::sSubset:
    case SetExpression::sEquals:
      goto Compatible;
    }
  }

  if (left == right)
    goto Same;

  if (Type::isAlwaysCompatible (left->getType (), right->getType ())) {
  Compatible:
    class Expression* expr;
    (expr = CSE (new class SetExpression (op, left, right)))
      ->associateVariables (true);
    return expr;
  }
  else {
    deerror ("expressions have incompatible types");
    left->destroy (), right->destroy ();
    return NULL;
  }
}

static class Expression*
newBufferUnop (enum BufferUnop::Op op, class Expression* expr)
{
  if (!(expr = ensureBuffer (expr)))
    return NULL;

  if (expr->getKind () == Expression::eBufferIndex) {
    class BufferIndex* b = static_cast<class BufferIndex*>(expr);
    class Expression* buffer = b->getBuffer ();
    class Expression* index = b->getIndex ();
    b->destroy ();
    switch (op) {
    case BufferUnop::bFree:
    case BufferUnop::bUsed:
      if (index) {
	dewarn ("ignoring meaningless buffer indexing expression");
	index->destroy (), index = NULL;
      }
      break;
    case BufferUnop::bPeek:
    case BufferUnop::bRemove:
      break;
    }

    return fold (new class BufferUnop (op, buffer, index));
  }

  return fold (new class BufferUnop (op, expr, NULL));
}
