#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <clocale>

#include "version.h"

#include "Expression.h"
#include "Valuation.h"
#include "BoolValue.h"

#include <getopt.h>
#ifdef HAS_GETOPT_LONG
# define getopt(argc,argv,optstring) getopt_long(argc,argv,optstring,opts,0)
static const struct option opts[] = {
  { "include", 1, 0, 'I' },
  { "help", 0, 0, '?' },
  { "verbose", 0, 0, 'v' },
  { "version", 0, 0, 'V' },
  { "warnings", 0, 0, 'W' },
  { "no-warnings", 0, 0, 'w' },
  { "regenerate", 0, 0, 'R' },
  { "no-regenerate", 0, 0, 'r' },
  { "array-limit", 1, 0, 'a' },
  { "quantification-limit", 1, 0, 'q' },
  { "breadth-first-search", 0, 0, 'b' },
  { "depth-first-search", 0, 0, 'd' },
  { "execute", 1, 0, 'e' },
  { "file", 1, 0, 'f' },
  { "width", 1, 0, 'i' },
  { "radix", 1, 0, 'x' },
# ifdef DEBUG
  { "dump-ast", 0, 0, 'S' },
  { "no-dump-ast", 0, 0, 's' },
  { "dump-rg", 0, 0, 'G' },
  { "no-dump-rg", 0, 0, 'g' },
# endif // DEBUG
  { 0, 0, 0, 0 }
};
#endif // HAS_GETOPT_LONG

/** Flag: be verbose */
static bool verbose = false;
/** Flag: generate the reachability graph anew */
static bool regenerate = false;
/** Flag: use breadth-first search (as opposed to depth-first search) */
static bool breadth = false;
#ifdef DEBUG
/** Flag: dump the abstract syntax tree? */
static bool dumpAST = false;
/** Flag: dump the reachability graph? */
static bool dumpRG = false;
# define DEBUGFLAGS "sSgG"
#else // DEBUG
# define DEBUGFLAGS ""
#endif // DEBUG

/** Option string for getopts(3) */
static const char optstring[] = "I:?hvVwWrRa:q:bde:f:i:x:" DEBUGFLAGS;

#include "util.h"
#include "NameList.h"
static class NameList includePath;

#include "Printer.h"
#include "Net.h"
#include "Graph.h"
#include "Property.h"
#include "Product.h"

#ifndef YYDEBUG
# define YYDEBUG 0
#endif

#if YYDEBUG
/** Debugging flag for the Bison-generated parser */
extern int pndebug;
#endif // YYDEBUG

/** The Bison-generated parser
 * @return	zero if exiting due to end-of-file; 1 if parse error occurred
 */
extern int pnparse ();

/** warning flag (true=print warnings) */
extern bool pnwarnings;
/** Maximum index type size of arrays */
extern card_t maxIndex;
/** Maximum range size of quantifications */
extern card_t maxRange;
/** number of errors encountered */
extern unsigned pnerrors;

/** The printer object */
class Printer thePrinter;
/** The Petri Net being parsed */
class Net* net;
/** Name of the Petri Net being parsed */
const char* netname;
/** Name of Petri Nets parsed from the command line */
const char* commandline = "(command line)";
/** Petri Net model to be parsed from the command line */
const char* netstring;

/** Open an input file
 * @param name	name of the input file
 * @return		a descriptor to the file, or NULL on error
 */
FILE* openfile (const char* name);

/** Analyze a net
 * @param argv		command-line arguments
 * @param filename	name of the net to be analyzed
 * @param graphfile	name for the reachability graph files
 * @return		false on a fatal error, true otherwise
 */
static bool analyze (const char*const* argv,
		     const char* filename,
		     const char* graphfile);

/** Flag: has the analysis been interrupted? */
bool interrupted = false;
/** Handle a signal
 * @param num		number of the signal
 */
extern "C" void sig (int num) {
  interrupted = true;
}
#include <csignal>

/** The main function
 * @param argc	argument count
 * @param argv	argument vector
 * @return	0 if successful;
 *		1 on parse error
 */
int main (int argc, char** argv);

int main (int argc, char** argv)
{
  /** Name of the reachability graph file (NULL=use default) */
  const char* graphfile = NULL;
  setlocale (LC_ALL, "");
  signal (SIGINT, sig);

#if YYDEBUG
  {
    char *debug = getenv ("DEBUG");
    if (debug) pndebug = atoi (debug);
  }
#endif // YYDEBUG

  for (;;) {
    int c = getopt (argc, argv, optstring);
    if (c == EOF)
      break;
    switch (c) {
    case ':':
      // fall through
    default:
    case '?':
    case 'h':
      fprintf (stderr, "Usage: %s [options] file...\n", *argv);
      return 0;
    case 'v':
      verbose = true;
      break;
    case 'V':
      fprintf (stderr, "%s version " VERSION "\n", *argv);
      return 0;
    case 'w':
      pnwarnings = false;
      break;
    case 'W':
      pnwarnings = true;
      break;
    case 'r':
      regenerate = false;
      break;
    case 'R':
      regenerate = true;
      break;
    case 'd':
      breadth = false;
      break;
    case 'b':
      breadth = true;
      break;
    case 'a':
      {
	char* end = NULL;
	maxIndex = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in array size limit: `%s'\n", optarg);
	  return 1;
	}
      }
      break;
    case 'q':
      {
	char* end = NULL;
	maxRange = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in quantification limit: `%s'\n", optarg);
	  return 1;
	}
      }
      break;
    case 'i':
      {
	char* end = NULL;
	unsigned width = strtoul (optarg, &end, 0);
	if (!*optarg || *end) {
	  fprintf (stderr, "error in output width limit: `%s'\n", optarg);
	  return 1;
	}
	else
	  thePrinter.setWidth (width);
      }
      break;
    case 'x':
      {
	if (!strcmp (optarg, "oct") || !strcmp (optarg, "octal") ||
	    !strcmp (optarg, "8"))
	  thePrinter.setRadix (Printer::Octal); 
	else if (!strcmp (optarg, "hex") || !strcmp (optarg, "hexadecimal") ||
		 !strcmp (optarg, "16"))
	  thePrinter.setRadix (Printer::Hexadecimal);
	else if (!strcmp (optarg, "dec") || !strcmp (optarg, "decimal") ||
		 !strcmp (optarg, "10"))
	  thePrinter.setRadix (Printer::Decimal);
	else
	  fprintf (stderr, "ignoring unknown radix setting: `%s'\n", optarg);
      }
      break;
    case 'I':
      includePath.append (newString (optarg));
      break;
    case 'e':
      netstring = optarg;
      if (!analyze (argv, commandline, graphfile))
	return 1;
      break;
    case 'f':
      graphfile = optarg;
      break;
#ifdef DEBUG
    case 's':
      dumpAST = false;
      break;
    case 'S':
      dumpAST = true;
      break;
    case 'g':
      dumpRG = false;
      break;
    case 'G':
      dumpRG = true;
      break;
#endif // DEBUG
    }
  }

  if (optind == argc && !netstring)
    fprintf (stderr, "%s: no input files\n", *argv);

  for (int i = optind; i < argc; i++)
    if (!analyze (argv, argv[i], graphfile))
      return 1;

  return 0;
}

static bool
analyze (const char*const* argv, const char* filename, const char* graphfile)
{
  netname = filename;
  net = new class Net;
  pnerrors = 0;
  if (pnparse ()) {
    delete net;
    return false;
  }
  if (pnerrors)
    fprintf (stderr, "%s:%u %s\n", netname, pnerrors,
	     pnerrors == 1 ? "error" : "errors");
  else {
    if (verbose) {
      fprintf (stderr, "%s:%u places and %u transitions\n",
	       netname, net->getNumPlaces (), net->getNumTransitions ());
      fprintf (stderr, "%s:computing the initial marking\n", netname);
    }
    net->computeInitMarking (thePrinter);

#ifdef DEBUG
    if (dumpAST)
      net->prettyPrint (0);
#endif // DEBUG

    class Property* property = NULL;

    if (const class Expression* const f = net->getFormula ()) {
      if (f->isTemporal ())
	property = Property::create (thePrinter, *f);
      else {
	class Valuation v;
	v.setGlobalMarking (net->getInitMarking ());
	if (verbose)
	  fprintf (stderr, "%s:checking the formula in the initial state\n",
		   netname);
	if (class Value* val = f->eval (&v)) {
	  assert (val->getKind () == Value::vBool);
	  if (!static_cast<class BoolValue*>(val)->getValue ()) {
	    fprintf (stderr,
		     "%s:the formula does not hold in the initial state\n",
		     netname);
	    delete val;
	    delete net;
	    return true;
	  }
	  delete val;
	}
	else {
	  fprintf (stderr,
		   "%s:the formula cannot be evaluated in the initial state\n",
		   netname);
	  v.display (thePrinter);
	  thePrinter.finish ();
	  delete net;
	  return true;
	}
      }
    }

    if (!net->getNumPlaces () && !net->getNumTransitions ());
    else if (class GlobalMarking* const g = net->getInitMarking ()) {
      class Graph graph (*net, optind, argv, netname, graphfile);
      if (!graph.openFiles (regenerate)) {
	fprintf (stderr, "%s:could not open or create the graph files\n",
		 netname);
      }
      else {
	if (property) {
	  if (verbose)
	    fprintf (stderr, "%s:performing on-the-fly model checking\n",
		     netname);
	  class Product product (graph, *property);
	  if (!product.analyze (*g))
	    fprintf (stderr, "%s:counterexample found\n", netname);
	}
	else {
	  if (verbose)
	    fprintf (stderr, "%s:generating the reachability graph\n",
		     netname);
	  net->analyze (thePrinter, graph, *g, breadth);
	}
#ifdef DEBUG
	if (dumpRG && !interrupted)
	  graph.prettyPrint (0);
#endif // DEBUG
	if (verbose) {
	  if (interrupted)
	    fprintf (stderr,
		     "partial reachability graph of %u states generated\n",
		     graph.size ());
	  else
	    fprintf (stderr,
		     "complete reachability graph of %u states generated\n",
		     graph.size ());
	}
      }
    }
  }
  delete net;
  return true;
}

FILE*
openfile (const char* name)
{
  FILE* f;
  // first try from the current directory
  if (!(f = fopen (name, "r"))) {
    // next, try from the directory of the original net

    size_t dirlength;
    char* fullname;
    size_t namelength = strlen (name);
    for (dirlength = 0; netname[dirlength]; dirlength++);
    while (dirlength > 0 && netname[dirlength] != '/')
      dirlength--;
    if (dirlength > 0) {
      fullname = new char[namelength + dirlength + 2];
      memcpy (fullname, netname, dirlength + 1);
      memcpy (fullname + dirlength + 1, name, namelength + 1);
      f = fopen (fullname, "r");
      delete[] fullname;
      if (f)
	return f;
    }

    // failing that, try from the include directories
    for (NameList::iterator i = includePath.begin ();
	 i != includePath.end (); i++) {
      const char* dirname = *i;
      dirlength = strlen (dirname);
      fullname = new char[namelength + dirlength + 2];
      memcpy (fullname, dirname, dirlength);
      fullname[dirlength] = '/';
      memcpy (fullname + dirlength + 1, name, namelength + 1);
      f = fopen (fullname, "r");
      delete[] fullname;
      if (f)
	break;
    }
  }

  return f;
}
