/* -*- c++ -*- Parser for the high-level Petri Net language used in Maria */

%{
#include <alloca.h>

#include "util.h"
#include "StringBuffer.h"
#include "VariableStackMap.h"

#include "allExpressions.h"
#include "ExpressionSet.h"
#include "Constraint.h"
#include "Range.h"

#include "Arc.h"
#include "Transition.h"
#include "Place.h"
#include "Net.h"
#include "VariableDefinition.h"
#include "Function.h"

#include "Quantifier.h"

#include "TransitionQualifier.h"

#include "typedefs.h"
#include "allTypes.h"
#include "allValues.h"
#include "Valuation.h"

# include <set>
# include <map>
# include <list>
# include <stack>

# ifndef NDEBUG
#  define ASSERT1(x) x
# else
#  define ASSERT1(x)
# endif

  /** warning flag (true=print warnings) */
  extern bool pnwarnings;
  /** Maximum index type size of arrays */
  card_t maxIndex = CARD_T_MAX;
  /** Maximum range size of quantifications */
  card_t maxRange = CARD_T_MAX;

  /** lexical analyzer function
   * @return		a terminal symbol, and set yylval to its semantic value
   */
  extern int pnlex ();

  /** The Petri Net */
  extern class Net* net;
# define theNet (*net)
  /** Parameters of the function that is currently being processed */
  static class VariableDefinition** currentParameters = NULL;
  /** Arity of the function that is currently being processed */
  static unsigned currentArity = 0;
  /** Function being invoked */
  static std::stack< class Function*, std::list<class Function*> >
    functionStack;
  /** The place that is currently being processed */
  class Place* currentPlace = NULL;
  /** The transition that is currently being processed */
  class Transition* currentTransition = NULL;
  /** The context transition (not being processed) */
  static class Transition* currentContextTransition = NULL;
  /** The type definition that is currently being processed */
  static const class Type* currentType = NULL;
  /** The context type of currentType (StructType or VectorType or NULL) */
  static const class Type* currentContextType = NULL;
  /** Struct or vector component number currently being parsed */
  static card_t currentComponent;
  /** Type context */
  struct TypeContext {
    /** Context type (@see currentContextType), may be NULL */
    const class Type* context;
    /** Component type (@see currentContext) */
    const class Type* component;
    /** Number of current component (@see currentComponent)
     * , meaningless if context==NULL */
    card_t numComponent;
  };
  /** Stack of types, used e.g. when parsing quantifier conditions */
  static std::stack< struct TypeContext, std::list<struct TypeContext> >
    typeStack;
  /** Flag: is an output arc being parsed? */
  static bool isOutputArc = false;
  /** Flag: has a number been successfully parsed? */
  static bool numberParsed = false;

  /** Flag: has an empty list component been parsed? */
  static bool emptyParsed = false;

  /** Flag: are submarking criteria allowed? */
  static bool allowSubmarking = true;
  /** Flag: are cardinality expressions allowed? */
  static bool allowCardinality = false;

  typedef std::pair<const class Type*,card_t> StructLevel;
  /** Stack for parsing structs and vectors */
  static std::stack< StructLevel, std::list<StructLevel> > structLevels;

  /** Convert a StringBuffer to a NUL-terminated string
   * @param sbuf	the string buffer
   * @return		the corresponding NUL-terminated string
   */
  static char* convertString (class StringBuffer* sbuf);

  /** Make a copy of a type if it is not an unnamed user type
   * @param type	the type to be copied
   * @return		type or a copy of it
   */
  static class Type* copyType (class Type* type);

  /** Delete the type definition if it is an unnamed user type
   * @param type	the type to be deleted
   */
  static void deleteType (class Type* type);

  /** Push the type context on structLevels
   * @param type	new type context
   */
  static void pushTypeContext (const class Type* type);
  /** Restore the type context from structLevels */
  static void popTypeContext ();

  /** Get the parameter of the currently parsed function by name
   * @param name	name of the parameter
   * @return		the corresponding VariableDefinition, or NULL
   */
  static const class VariableDefinition* getParameter (char* name);

  /** Add a parameter to the function definition being parsed
   * @param type	type of the parameter
   * @param name	name of the parameter
   */
  static void addParameter (const class Type* type, char* name);

  /** Get a Function by name
   * @param name	the function name
   * @return		a Function object
   */
  static class Function* getFunction (const char* name);

  /** Get a zero-arity Function by name
   * @param name	the function name
   * @return		a Function object
   */
  static class Function* getConstantFunction (const char* name);

  /** Get a VariableDefinition by name
   * @param name	the variable name
   * @return		a VariableDefinition object
   */
  static const class VariableDefinition* getVariable (char* name);

  /** Resolve a name to a constant or variable
   * @param name	the name
   * @return		the resolved expression, or NULL
   */
  static class Expression* resolveName (char* name);

  /** Determine whether a formula is a set-valued expression
   * @param expr	formula to be checked
   * @return		true if the formula is a marking expression
   */
  static bool isSet (const class Expression* expr);

  /** Ensure that the formula is a set-valued expression
   * @param expr	formula to be checked
   * @return		expr, pointer to a Marking object, or NULL
   */
  static class Expression* ensureSet (class Expression* expr);

  /** Ensure that the formula is a marking expression, converting if needed
   * @param expr	formula to be checked
   * @return		pointer to a Marking object, or NULL
   */
  static class Marking* ensureMarking (class Expression* expr);

  /** Ensure that the formula is not a marking expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* noMarking (class Expression* expr);

  /** Ensure that the formula is a basic expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureExpr (class Expression* expr);

  /** Ensure that the formula is a boolean expression
   * @param expr	formula to be checked
   * @param undefined	flag: allow the formula to be Undefined
   * @return		expr or NULL
   */
  static class Expression* ensureBool (class Expression* expr,
				       bool undefined = false);

  /** Ensure that the formula is an integer expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureInt (class Expression* expr);

  /** Ensure that the formula is a buffer expression
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureBuffer (class Expression* expr);

  /** Ensure that the formula is a basic expression of ordered type
   * @param expr	formula to be checked
   * @return		expr or NULL
   */
  static class Expression* ensureOrdered (class Expression* expr);

  /** Add an enumeration constant to an enumerated type, ensuring uniqueness.
   * @param e		the enumeration type
   * @param name	name of the constant
   */
  static void addEnumeration (class EnumType* e, char* name);

  /** Add an enumeration constant to an enumerated type, ensuring uniqueness.
   * @param e		the enumeration type
   * @param name	name of the constant
   * @param value	value of the constant
   */
  static void addEnumeration (class EnumType* e, char* name, int_t value);

  /** Ensure that the formula is a constant literal
   * @param expr	formula to be checked
   * @param type	type of the constant expected
   * @return		the value of expr or NULL
   */
  static class Value* ensureConstant (class Expression* expr,
				      const class Type* type);

  /** Convert an integer to an enumeration value if possible
   * @param type	type of the enumeration value
   * @param value	numerical value to be converted to the enumeration type
   * @return		a constant having the enumeration value or NULL
   */
  static class Expression* enumValue (const class EnumType* type, int_t value);

  /** Ensure that a named component matches a struct being parsed
   * @param name	name of the component
   * @param expr	initialization expression for the component
   * @return		expr or NULL
   */
  static class Expression* namedComponent (char* name, class Expression* expr);

  /** Begin parsing a tester expression
   * @param place	The place the tester is associated with
   */
  static void beginTester (class Place* place);

  /** Finish parsing a tester expression
   * @param kind	kind of the tester
   * @param expr	tester expression
   */
  static void endTester (enum Net::TesterKind kind, class Marking* expr);

  /** Begin parsing a non-determinism expression
   * @param type	type of the expression
   * @param name	name of the non-deterministic variable (optional)
   */
  static void beginAny (class Type*& type, char* name);

  /** End parsing a non-determinisim expression
   * @param type	type of the expression
   * @param name	name of the non-deterministic variable (optional)
   * @param condition	quantification condition (optional)
   * @return		a Variable expression for the variable, or NULL
   */
  static class Variable* endAny (const class Type* type, char* name,
				 class Expression* condition);

  /** Begin parsing a quantifier expression
   * @param name	name of the quantifier variable
   * @param type	type of the quantifier variable
   */
  static void beginQuantifier (char* name, const class Type* type);

  /** Continue parsing a quantifier expression
   *  (after parsing an optional condition)
   * @param type	type of the quantifier variable
   */
  static void continueQuantifier (ASSERT1 (const class Type* type));

  /** End parsing a quantifier expression
   * @param name	name of the quantifier variable
   * @param type	type of the quantifier variable
   * @param condition	quantification condition (optional)
   * @param expr	marking expression
   * @return		expr augmented with the quantifier, or NULL if failed
   */
  static class Marking* endQuantifier (char* name, const class Type* type,
				       class Expression* condition,
				       class Marking* expr);

  /** Determine whether a variable is a quantifier index
   * @param variable	variable to be checked
   * @return		whether the variable is a quantifier index
   */
  static bool isQuantifier (const class VariableDefinition& variable);

  /** Try to make expr compatible with type if expr is a Constant
   * @param expr	expression to be made compatible
   * @param type	type to be made compatible with
   * @param casting	flag: typecasting in progress (suppress a warning)
   * @return		true if expr is not a Constant or
   *			if the conversion succeeded
   */
  static bool makeCompatible (class Expression*& expr,
			      const class Type* type,
			      bool casting = false);

  /** Determine whether an expression is compatible with a type
   * @param expr	expression to be checked
   * @param type	the type
   * @return		true if the expression is compatible with the type
   */
  static bool isCompatible (class Expression*& expr, const class Type* type);

  /** Determine whether an expression is compatible with
   * the struct component being parsed
   * @param expr	expression to be checked
   * @param s		the struct whose current component is considered
   * @return		true if the expression is compatible with the struct
   */
  static bool isCompatible (class Expression*& expr,
			    const class StructExpression* s);

  /** Determine whether two expressions are type compatible
   * @param expr1	first expression
   * @param expr2	second expression
   * @return		true if the expressions are compatible
   */
  static bool isCompatible (class Expression*& expr1,
			    class Expression*& expr2);

  /** Ensure that the expressions have an ordered type and are compatible
   * @param expr1	first expression to be checked
   * @param expr2	second expression to be checked
   * @return		true if the expressions have an ordered type
   */
  static bool isOrdered (class Expression*& expr1,
			 class Expression*& expr2);

  /** Check a valuation
   * @param v		valuation to be checked
   * @return		true if the valuation is OK
   */
  bool check (const class Valuation& v);

  /** Perform constant folding
   * @param expr	expression to be folded
   * @return		expr or a new Constant, or NULL in case of error
   */
  static class Expression* fold (class Expression* expr);

  /** Create a RelopExpression or optimize it
   * @param op		operator
   * @param left	left-hand-side of RelopExpression
   * @param right	right-hand-side of RelopExpression
   * @return		RelopExpression or a Boolean constant
   */
  static class Expression* newRelopExpression (enum RelopExpression::Op op,
					       class Expression* left,
					       class Expression* right);

  /** Create a BooleanBinop or optimize it
   * @param op		operator
   * @param left	left-hand-side of BooleanBinop
   * @param right	right-hand-side of BooleanBinop
   * @return		left, right, Boolean constant, or a BooleanBinop
   */
  static class Expression* newBooleanBinop (enum BooleanBinop::Op op,
					    class Expression* left,
					    class Expression* right);

  /** Create a SetExpression or optimize it
   * @param op		operator
   * @param left	left-hand-side of SetExpression
   * @param right	right-hand-side of SetExpression
   */
  static class Expression* newSetExpression (enum SetExpression::Op op,
					     class Expression* left,
					     class Expression* right);

  /** Create a BufferUnop from a buffer expression
   * @param op		operator
   * @param expr	the buffer expression
   * @return		a BufferUnop object, or NULL
   */
  static class Expression* newBufferUnop (enum BufferUnop::Op op,
					  class Expression* expr);

  /** Variable stack map to be used with variable scoping */
  static class VariableStackMap currentVariables;

  /** Set of subexpressions for common subexpression elimination */
  static class ExpressionSet currentExpressions;

  /** Eliminate common subexpressions
   * @param expr	expression to be searched in currentExpressions
   * @return		expr or an equivalent expression
   */
  inline static class Expression* CSE (class Expression* expr) {
    return expr ? currentExpressions.insert (expr) : NULL;
  }

  /**
   * generic compiler warning reporting function
   * @param s		warning message
   */
  extern void pnwarn (const char* s);

  /**
   * generic printf-like compiler warning reporting function
   * @param fmt		printf format string
   * @param ...		variable arguments
   */
  extern void pnwarnf (const char* fmt, ...)
# ifdef __GNUC__
    __attribute__ ((format (printf, 1, 2)))
# endif /* __GNUC__ */
    ;

  /**
   * generic parser error reporting function
   * @param s		error message
   */
  extern void pnerror (const char* s);

  /**
   * generic printf-like parser error reporting function
   * @param fmt		printf format string
   * @param ...		variable arguments
   */
  extern void pnerrorf (const char* fmt, ...)
# ifdef __GNUC__
    __attribute__ ((format (printf, 1, 2)))
# endif /* __GNUC__ */
    ;
%}

%union{
  int_t i;
  bool flag;
  char_t c;
  char* s;
  class StringBuffer* sbuf;
  class Expression* expr;
  class ExpressionList* exprList;
  class Transition* trans;
  class Place* place;
  class Marking* marking;
  class Type* type;
  class ComponentList* componentList;
  class Value* value;
  class Range* range;
  class Constraint* constraint;
}

%right '?' ':'
%left ALWAYS EXISTS UNTIL RELEASE
%left EQUALS
%left SUBSET
%left MINUS UNION
%left INTERSECT
%left IMPL EQUIV
%left OR
%left XOR
%left AND
%left FINALLY GLOBALLY NEXT
%left '|'
%left '^'
%left '&'
%left NE EQ
%left GE LE '<' '>'
%left ROL ROR
%left '+' '-'
%left '*' '/' '%'
%right '#' IS
%left '~' '!'
%right CARDINALITY
%left ATOM
%left '.' '[' '('

%token <i>
	NUMBER

%token <c>
	CHARACTER

%token <s>
	NAME
%token <sbuf>
	STRING

%token
	TRUE_ FALSE_ UNDEFINED FATAL
	TYPEDEF PLACE TRANS IN OUT GATE
	VERIFY TESTER REJECT_ DEADLOCK LIVELOCK INFINITE
	ENUM ID STRUCT QUEUE STACK RANGE EMPTY

%type <s>
	name
	name_

%type <i>
	number

%type <expr>
	_opt_expr
	_opt__expr
	expr_
	expr
	formula
	marking_list_
	qualifier
	qual_expr
	struct_expr

%type <exprList>
	arg_list
	colon_formula

%type <marking>
	place_init
	marking
	marking_list

%type <place>
	placename

%type <trans>
	transname

%type <type>
	typedefinition
	typereference
	enum_list

%type <componentList>
	comp_list

%type <constraint>
	range_list
	constraint
%type <range>
	range
%type <value>
	value
	value_

%type <flag>
	_list_var_expr/* true if component present */
	_opt_question/* true if present */
	_opt_place/* true if present */
	_opt_trans/* true if present */
	queue_stack/* false=queue, true=stack */

%%

start:
	|
	net
	;

net:	';'
	{ pnwarn ("extraneous semicolon"); }
	|
	netcomponent ';'
	|
	net netcomponent ';'
	;

netcomponent:
	type
	|
	function
	|
	place
	|
	transition
	|
	tester
	|
	verify
	;

type:
	TYPEDEF typedefinition name
	{
	  if ($3 && theNet.getType ($3)) {
	    pnerrorf ("redefined type `%s'", $3); delete[] $3;
	  }
	  else if ($2 && $3)
	    theNet.addType ($2, $3);
	  else
	    deleteType ($2), delete[] $3;
	}
	;

_opt_question:
	{ $$ = false; }
	|
	_opt_question '?'
	{ $$ = true; if ($1) pnerror ("extraneous question mark"); }
	;

queue_stack:
	QUEUE
	{ $$ = false; }
	|
	STACK
	{ $$ = true; }
	|
	queue_stack QUEUE
	{ $$ = $1; pnerror ("ignoring extraneous `queue'"); }
	|
	queue_stack STACK
	{ $$ = $1; pnerror ("ignoring extraneous `stack'"); }
	;

typedefinition:
	ENUM '{' enum_list '}'
	{
	  if (!($$ = $3))
	    $$ = new class StructType (new class ComponentList);
	}
	|
	ID '[' number ']'
	{
	  if (numberParsed && $3 > 0)
	    $$ = new class IdType ($3);
	  else {
	    if (numberParsed)
	      pnerrorf ("unsupported capacity of type: id[%d]", $3);
	    $$ = NULL;
	  }
	}
	|
	STRUCT '{' comp_list '}'
	{ $$ = $3 ? new class StructType ($3) : NULL; emptyParsed = false; }
	|
	UNION '{' comp_list '}'
	{
	  if ($3 && !$3->getSize ()) {
	    pnerror ("missing component list for union");
	    delete $3, $$ = NULL;
	  }
	  else
	    $$ = $3 ? new class UnionType ($3) : NULL;

	  emptyParsed = false;
	}
	|
	typereference
	{ $$ = $1; }
	|
	typedefinition
	{ currentType = $1; }
	constraint
	{
	  if (($$ = $1) && $3) {
	    $$ = copyType ($$);
	    if (class Constraint* c = $$->getConstraint ()) {
	      if (c->isDisjoint ($3)) {
		pnerror ("ignoring constraint that would make the type empty");
		delete $3;
	      }
	      else
		c->restrict ($3, $$);
	    }
	    else
	      $$->setConstraint ($3);
	  }
 	  else
 	    delete $3;
	  currentType = NULL;
	}
	|
	typedefinition '[' _opt_question typedefinition ']'
	{
	  if ($4 && !$4->isOrdered ()) {
	    pnerrorf ("type `%s' cannot be ordered nor used as an index",
		      $4->getName ()
		      ? $4->getName ()
		      : $4->getSyntacticName ());
	    deleteType ($1), deleteType ($4), $$ = NULL;
	  }
	  else if ($1 && $4) {
	    card_t size = $4->getNumValues ();
	    if (!size || size == CARD_T_MAX || size > maxIndex) {
	      pnerrorf ("type `%s' with %u values cannot be used as an index",
			$4->getName ()
			? $4->getName ()
			: $4->getSyntacticName (),
			size);
	      deleteType ($1), deleteType ($4), $$ = NULL;
	    }
	    else {
	      theNet.addType ($1); theNet.addType ($4);
	      $$ = new class VectorType ($1, $4, $3);
	    }
	  }
	  else
	    deleteType ($1), deleteType ($4), $$ = NULL;
	}
	|
	typedefinition '[' queue_stack number ']'
	{
	  if ($1 && numberParsed) {
	    if ($4 < 1) {
	      pnerrorf ("invalid buffer capacity %d<1", $4);
	      deleteType ($1), $$ = NULL;
	    }
	    else {
	      theNet.addType ($1);
	      $$ = new class BufferType ($1, $4, $3);
	    }
	  }
	  else
	    deleteType ($1), $$ = NULL;
	}
	|
	typedefinition '[' NUMBER ']'
	{
	  pnerror ("array dimension must be an ordered type");
	  deleteType ($1), $$ = NULL;
	}
	;

constraint:
	'(' range_list ')'
	{ $$ = $2; }
	;

range_list:
	range_list delim range
	{
 	  if (($$ = $1) && $3)
 	    $$->append ($3);
 	  else
	    delete $3;
	}
	|
	range
	{ if ($1) ($$ = new class Constraint)->append ($1); else $$ = NULL; }
	;

range:
	value_
	{ $$ = $1 ? new class Range ($1, $1) : NULL; }
	|
	range_op value_
	{ $$ = $2 ? new class Range (NULL, $2) : NULL; }
	|
	value range_op value
	{
 	  if ($1 && $3)
 	    $$ = new class Range ($1, $3);
	  else
	    $$ = NULL, delete $1, delete $3;
	}
	|
	value range_op
	{ $$ = $1 ? new class Range ($1, NULL) : NULL; }
	;

range_op:
	RANGE
	|
	':'
	;

value:
	expr
	{ $$ = ensureConstant ($1, currentType); }
	;

value_:
	expr_
	{ $$ = ensureConstant ($1, currentType); }
	;

typereference:
	name_
	{
	  if (!$1) $$ = NULL;
	  else if (($$ = const_cast<class Type*>(theNet.getType ($1))));
	  else if (!strcmp ($1, "bool"))
	    $$ = const_cast<class BoolType*>(Net::getBoolType ());
	  else if (!strcmp ($1, "int"))
	    $$ = const_cast<class IntType*>(Net::getIntType ());
	  else if (!strcmp ($1, "char"))
	    $$ = const_cast<class CharType*>(Net::getCharType ());
	  else
	    pnerrorf ("undefined type name `%s'", $1);
	  delete[] $1;
	}
	;

number:
	expr
	{
	  $$ = 0; numberParsed = false;

	  if ($1 && $1->getType () && ($1 = fold ($1)) &&
	      $1->getKind () == Expression::eConstant) {
	    class Value* v = $1->eval (NULL);
	    assert (v && $1->getType ()->isConstrained (v));
	    switch (v->getKind ()) {
	    case Value::vChar:
	      $$ = static_cast<const class CharValue*>(v)->getValue ();
	      pnwarnf ("converting '%c' to %d", static_cast<char_t>($$), $$);
	      break;
	    case Value::vInt:
	      numberParsed = true;
	      $$ = static_cast<const class IntValue*>(v)->getValue ();
	      break;
	    case Value::vEnum:
	      numberParsed = true;
	      $$ = static_cast<const class EnumValue*>(v)->getValue ();
	      break;
	    default:
	      pnerror ("expression does not evaluate to a number");
	      break;
	    }

	    delete v;
	  }
	  else if ($1)
	    pnerror ("number expected, got expression");

	  $1->destroy ();
	}
	;

name_:
	NAME
	{ $$ = $1; }
	|
	STRING
	{ $$ = convertString ($1); }
	|
	error
	{ $$ = NULL; }
	;

name:
	{ pnerror ("missing name"); $$ = NULL; }
	|
	name_
	{ $$ = $1; }
	;

_opt_eq:
	|
	'='
	;

enum_list:
	{ $$ = NULL; pnerror ("missing enumeration value"); }
	|
	name_
	{
	  if ($1)
	    addEnumeration (static_cast<class EnumType*>($$ =
							 new class EnumType),
			    $1);
	  else
	    $$ = NULL;
	}
	|
	enum_list delim name_
	{
	  if (!($$ = $1) && $3)
	    addEnumeration (static_cast<class EnumType*>($$ =
							 new class EnumType),
			    $3);
	  else if ($$ && $3)
	    addEnumeration (static_cast<class EnumType*>($$), $3);
	}
	|
	name_ _opt_eq number
	{
	  if ($1 && numberParsed)
	    addEnumeration (static_cast<class EnumType*>($$ =
							 new class EnumType),
			    $1, $3);
	  else
	    $$ = NULL, delete[] $1;
	}
	|
	enum_list delim name_ _opt_eq number
	{
	  if (!numberParsed)
	    $$ = $1, delete[] $3;
	  else if (!($$ = $1) && $3)
	    addEnumeration (static_cast<class EnumType*>($$ =
							 new class EnumType),
			    $3, $5);
	  else if ($$ && $3)
	    addEnumeration (static_cast<class EnumType*>($$), $3, $5);
	}
	;

delim:	';' | ',' ;

comp_list:
	typedefinition name
	{
	  if ($1 && $2) {
	    ($$ = new class ComponentList)->addComponent ($2, $1);
	    theNet.addType ($1);
	  }
	  else {
	    deleteType ($1);
	    delete[] $2;
	    $$ = NULL;
	  }

	  emptyParsed = false;
	}
	|
	comp_list delim typedefinition name
	{
	  if (($$ = $1) && $3 && $4) {
	    if (emptyParsed)
	      pnerror ("extraneous delimiter in component list");

	    emptyParsed = false;

	    if ((*$$)[$4]) {
	      pnerrorf ("ignoring redefinition of component `%s'", $4);
	      deleteType ($3);
	      delete[] $4;
	    }
	    else {
	      $$->addComponent ($4, $3);
	      theNet.addType ($3);
	    }
	  }
	  else {
	    delete $$;
	    deleteType ($3);
	    delete[] $4;
	    $$ = NULL;
	  }
	}
	|
	comp_list delim
	{
	  $$ = $1;
	  if (emptyParsed)
	    pnerror ("extraneous delimiter in component list");
	  emptyParsed = true;
	}
	|
	{ $$ = new class ComponentList; emptyParsed = true; }
	;

next_eq:
	NEXT
	|
	'='
	;

function:
	typereference name next_eq
	{ assert (!currentType); currentType = $1; }
	formula
	{
	  assert (currentType == $1); currentType = NULL;
	  if ($1 && $2 && $5) {
	    if ((currentTransition && currentTransition->getFunction ($2)) ||
		(!currentTransition && theNet.getFunction ($2))) {
	      pnerrorf ("ignoring redefinition of function `%s'", $2);
	      delete[] $2, $5->destroy ();
	    }
	    else if (isCompatible ($5, $1)) {
	      class Function* f = new class Function ($2, 0, 0, $5);
	      if (currentTransition) {
		currentTransition->addFunction (*f);
		if (pnwarnings && theNet.getFunction ($2))
		  pnwarnf ("definition of transition function `%s'"
			   " shadows the global definition", $2);
	      }
	      else
		theNet.addFunction (*f);

	      if (pnwarnings && theNet.getType ($2))
		pnwarnf ("function `%s' shadows a type name", $2);
	    }
	    else
	      delete[] $2, $5->destroy ();
	  }
	  else
	    delete[] $2, $5->destroy ();
	}
	|
	typereference name '(' param_list ')'
	{
	  assert (!currentType); currentType = $1;
	  if (emptyParsed && currentArity)
	    pnerror ("extraneous delimiter after formal parameter list");
	  emptyParsed = false;
	}
	formula
	{
	  assert (currentType == $1); currentType = NULL;
	  if ($1 && $2 && $7) {
	    if ((currentTransition && currentTransition->getFunction ($2)) ||
		theNet.getFunction ($2)) {
	      pnerrorf ("ignoring redefinition of function `%s'", $2);
	      delete[] $2, delete[] currentParameters, $7->destroy ();
	    }
	    else if (isCompatible ($7, $1)) {
	      class Function* f =
		new class Function ($2, currentArity, currentParameters, $7);
	      if (currentTransition) {
		currentTransition->addFunction (*f);
		if (pnwarnings && theNet.getFunction ($2))
		  pnwarnf ("definition of transition function `%s'"
			   " shadows the global definition", $2);
	      }
	      else
		theNet.addFunction (*f);

	      if (pnwarnings) {
		if (theNet.getType ($2))
		  pnwarnf ("function `%s' shadows a type name", $2);
#ifdef EXPR_CACHE
		for (unsigned i = 0; i < currentArity; i++)
		  if (!$7->invalidate (currentParameters[i]))
		    pnwarnf ("function `%s' does not use parameter `%s'",
			     $2, currentParameters[i]->getName ());
#endif // EXPR_CACHE
	      }
	    }
	    else
	      delete[] $2, delete[] currentParameters, $7->destroy ();
	  }
	  else
	    delete[] $2, delete[] currentParameters, $7->destroy ();
	  currentArity = 0, currentParameters = 0;
	  emptyParsed = false;
	}
	;

param_list:
	typereference name
	{
	  assert (!currentArity && !currentParameters);
	  addParameter ($1, $2);
	  emptyParsed = false;
	}
	|
	param_list delim typereference name
	{
	  if ($3 && $4 && emptyParsed)
	    pnerror ("extraneous delimiter in parameter list");
	  addParameter ($3, $4);
	  emptyParsed = false;
	}
	|
	param_list delim
	{
	  if (emptyParsed)
	    pnerror ("extraneous delimiter in parameter list");
	  emptyParsed = true;
	}
	|
	{ assert (!currentArity && !currentParameters); emptyParsed = true; }
	;

place:
	_nopt_place name_ typedefinition
	{
	  currentPlace = NULL;
	  if ((currentType = $3))
	    theNet.addType ($3);

	  if ($2 && $3) {
	    if (theNet.getPlace ($2)) {
	      pnerrorf ("ignoring redefinition of place `%s'", $2);
	      delete[] $2;
	    }
	    else
	      currentPlace = theNet.addPlace ($2, $3);
	  }
	  else
	    delete[] $2;
	}
	place_init
	{
	  if (currentPlace && $5)
	    currentPlace->setInitMarking ($5);
	  else
	    $5->destroy ();
	  currentType = NULL, currentPlace = NULL;
	}
	;

placename:
	name
	{
	  if ($1) {
	    if (!($$ = const_cast<class Place*>(theNet.getPlace ($1))))
	      pnerrorf ("undefined place `%s'", $1);
	  }
	  else
	    $$ = NULL;
	  delete[] $1;
	}
	;

place_init:
	{ $$ = NULL; }
	|
	place_init ':' marking_list
	{
	  if (!($$ = $1))
	    $$ = $3;
	  else {
	    pnerror ("multiple init blocks"); $3->destroy ();
	  }
	  currentExpressions.clear ();
	}
	;

transition:
	_nopt_trans name
	{
	  if ($2) {
	    if (theNet.getTransition ($2)) {
	      pnerrorf ("ignoring redefinition of transition `%s'", $2);
	      delete[] $2;
	      currentTransition = NULL;
	    }
	    else
	      currentTransition = theNet.addTransition ($2);
	  }
	  else
	    currentTransition = NULL;

	  currentExpressions.clear ();
	}
	trans
	{
	  currentExpressions.clear ();
	  if (currentTransition) {
#ifdef EXPR_CACHE
	    if (const class QuantifierList* q =
		currentTransition->getOutputVariables ()) {
	      for (QuantifierList::const_iterator i = q->begin ();
		   i != q->end (); i++) {
		const class VariableDefinition* const v = (*i)->getVariable ();

		if (currentTransition->isInput (v))
		  pnerrorf ("output variable `%s' encountered on an input arc",
			    v->getName ());
		if (currentTransition->isGate (v))
		  pnerrorf ("output variable `%s' encountered on a gate",
			    v->getName ());
		if (!currentTransition->isOutput (v))
		  pnerrorf ("output variable `%s' not used on output arcs",
			    v->getName ());
	      }
	    }
#endif // EXPR_CACHE
	    for (Transition::const_iterator i = currentTransition->begin ();
		 i != currentTransition->end (); i++) {
	      const class VariableDefinition* const v = i->second;

	      if (!currentTransition->isOutputVariable (v) &&
		  !currentTransition->isLvalue (v))
		pnerrorf ("variable `%s' cannot be bound from input arcs",
			  v->getName ());
	    }
	    if (!currentTransition->hasInputs () &&
		currentTransition->hasOutputs ())
	      pnwarnf ("transition has output arcs but no inputs");
	  }
	  currentTransition = NULL;
	}
	;

transname:
	_opt_trans name
	{
	  if ($2) {
	    if (!($$ = const_cast<class Transition*>(theNet.getTransition
						     ($2))))
	      pnerrorf ("undefined transition `%s'", $2);
	  }
	  else
	    $$ = NULL;
	  delete[] $2;
	}
	;

trans:
	|
	trans '{' _list_var_expr '}'
	|
	trans IN
	{ isOutputArc = false; }
	trans_places
	|
	trans OUT
	{ isOutputArc = true; }
	trans_places
	|
	trans GATE gate_list
	;

_list_var_expr:
	{ $$ = false; }
	|
	var_expr
	{ $$ = true; }
	|
	_list_var_expr delim var_expr
	{
	  if (!$1)
	    pnerror ("extraneous delimiter before variable definition");
	  $$ = true;
	}
	|
	_list_var_expr delim
	{
	  if (!$1)
	    pnerror ("extraneous delimiter in variable definition list");
	  $$ = false;
	}
	;

var_expr:
	typereference name
	{
	  if ($1 && $2 && currentTransition) {
	    if (currentTransition->getVariable ($2)) {
	      pnerrorf ("redefinition of variable `%s'", $2);
	      delete[] $2;
	    }
	    else {
	      currentTransition->addVariable ($2, $1);
	      if (pnwarnings && getConstantFunction ($2))
		pnwarnf ("variable `%s' shadows a constant", $2);
	      else if (pnwarnings && theNet.getType ($2))
		pnwarnf ("variable `%s' shadows a type name", $2);
	    }
	  }
	  else
	    delete[] $2;
	}
	|
	function
	;

trans_places:
	'{' _list_place_marking '}'
	;

_list_place_marking:
	place_marking
	|
	_list_place_marking ';' place_marking
	;

place_marking:
	|
	_opt_place error
	{}
	|
	_opt_place placename ':'
	{
	  currentType = (currentPlace = $2) ? $2->getType () : NULL;
	  assert (allowSubmarking);
	  allowSubmarking = false;
	}
	marking_list
	{
	  if (!currentPlace || !currentTransition)
	    $5->destroy ();
	  else if ($5) {
	    if (class Arc* arc =
		currentTransition->getArc (isOutputArc, currentPlace)) {
	      pnwarnf (isOutputArc
		       ? "extending arc to place `%s'"
		       : "extending arc from place `%s'",
		       currentPlace->getName ());
	      arc->append ($5);
	    }
	    else
	      currentTransition->addArc (new class Arc
					 (isOutputArc, $5, currentPlace));
	  }
	  currentPlace = NULL, currentType = NULL;
	  assert (!allowSubmarking);
	  allowSubmarking = true;
	}
	;

_opt_place:
	{ $$ = false; }
	|
	_opt_place PLACE
	{ $$ = true; if ($1) pnerror ("ignoring extraneous `place'"); }
	|
	_opt_place TRANS
	{ $$ = $1; pnerror ("expected `place', got `trans'"); }
	;

_nopt_place:
	PLACE
	|
	_nopt_place PLACE
	{ pnerror ("ignoring extraneous `place'"); }
	|
	_nopt_place TRANS
	{ pnerror ("ignoring extraneous `trans'"); }
	;

_opt_trans:
	{ $$ = false; }
	|
	_opt_trans TRANS
	{ $$ = true; if ($1) pnerror ("ignoring extraneous `trans'"); }
	|
	_opt_trans PLACE
	{ $$ = $1; pnerror ("expected `trans', got `place'"); }
	;

_nopt_trans:
	TRANS
	|
	_nopt_trans TRANS
	{ pnerror ("ignoring extraneous `trans'"); }
	|
	_nopt_trans PLACE
	{ pnerror ("ignoring extraneous `place' after `trans'"); }
	;

gate_list:
	gate
	|
	gate_list ',' gate
	;

gate:
	expr
	{
	  if (($1 = ensureBool (ensureExpr ($1))) && currentTransition) {
	    if ($1->getKind () == Expression::eConstant) {
	      const class Value* v =
		static_cast<const class Constant*>($1)->getValue ();
	      assert (v && v->getKind () == Value::vBool);
	      pnwarnf ("gate expression is always %s",
		       static_cast<const class BoolValue*>(v)->getValue ()
		       ? "true" : "false");
	    }
	    currentTransition->addGate ($1);
	  }
	  else
	    $1->destroy ();
	}
	;

tester:
	TESTER _opt_place placename
	{ beginTester ($3); }
	REJECT_ marking_list
	{ endTester (Net::RejectTester, $6); currentExpressions.clear (); }
	|
	TESTER _opt_place placename
	{ beginTester ($3); }
	DEADLOCK marking_list
	{ endTester (Net::DeadlockTester, $6); currentExpressions.clear (); }
	|
	TESTER _opt_place placename
	{ beginTester ($3); }
	LIVELOCK marking_list
	{ endTester (Net::LivelockTester, $6); currentExpressions.clear (); }
	|
	TESTER _opt_place placename
	{ beginTester ($3); }
	INFINITE marking_list
	{ endTester (Net::InfiniteTester, $6); currentExpressions.clear (); }
	;

verify:
	VERIFY
	{
	  assert (!allowCardinality); allowCardinality = true;
	  currentType = Net::getBoolType ();
	}
	formula
	{
	  assert (currentType == Net::getBoolType ());
	  currentType = NULL;

	  if (class Expression* f = ensureBool ($3)) {
	    if (theNet.getFormula ()) {
	      pnerror ("only one model checker formula may be specified");
	      f->destroy ();
	    }
	    else {
	      f->associateVariables (true);
	      if (f->getKind () == Expression::eConstant) {
		const class Value* v =
		  static_cast<class Constant*>(f)->getValue ();
		assert (v->getKind () == Value::vBool);
		pnwarnf ("model checker formula is always %s",
			 static_cast<const class BoolValue*>(v)->getValue ()
			 ? "true" : "false");
	      }
	      theNet.setFormula (f);
	    }
	  }

	  assert (allowCardinality); allowCardinality = false;
	  currentExpressions.clear ();
	}
	;

marking:
	formula
	{
	  class Marking* m = ensureMarking ($1);
	  if (($$ = m)) m->associate ();
	}
	;

marking_list:
	marking
	{ $$ = $1; }
	|
	marking_list ',' marking
	{
	  if ($1 && $3) {
	    $$ = $1->isQuantified ()
	      ? new class Marking ($1->getPlace (), $1)
	      : $1;
	    $$->append ($3);
	  }
	  else
	    $$ = $1 ? $1 : $3;
	}
	;

marking_list_:
	formula
	{
	  if (($$ = $1) && $$->getKind () == Expression::eMarking)
	    static_cast<class Marking*>($$)->associate ();
	}
	|
	marking_list_ ','
	{
	  if (($1 = ensureMarking ($1)) &&
	      $1->getKind () == Expression::eMarking)
	    static_cast<class Marking*>($1)->associate ();
	}
	marking
	{
	  assert (!$1 || $1->getKind () == Expression::eMarking);
	  class Marking* m = static_cast<class Marking*>($1);
	  if (($$ = m) && $4) {
	    if (m->isQuantified ())
	      $$ = m = new class Marking (m->getPlace (), m);
	    m->append ($4);
	  }
	  else
	    $4->destroy ();
	}
	;

_opt_expr:
	{ $$ = NULL; }
	|
	'(' expr_ ')'
	{ $$ = $2; }
	;

_opt__expr:
	{ $$ = NULL; }
	|
	expr
	{ $$ = $1; }
	|
	NAME ':'
	{ $$ = namedComponent ($1, NULL); }
	|
	STRING ':'
	{ $$ = namedComponent (convertString ($1), NULL); }
	|
	NAME ':' expr
	{ $$ = namedComponent ($1, $3); }
	|
	STRING ':' expr
	{ $$ = namedComponent (convertString ($1), $3); }
	;

struct_expr:
	_opt__expr
	{
	  if (currentContextType && !YYRECOVERING ()) {
	    assert (!currentComponent);

	    switch (currentContextType->getKind ()) {
	    case Type::tVector:
	      {
		const class VectorType* type =
		  static_cast<const class VectorType*>(currentContextType);
		class VectorExpression* v = new class VectorExpression (type);
		assert (currentType == type->getItemType ());

		if (!$1) {
		  if (!type->isOptional ()) {
		    pnerror ("missing expression for first vector component");
		    v->destroy (), $$ = NULL;
		  }
		  else
		    $$ = v;
		}
		else if (isCompatible ($1, currentType))
		  $$ = v, v->setComponent (currentComponent, $1);
		else
		  v->destroy (), $1->destroy (), $$ = NULL;
	      }
	      break;
	    case Type::tBuffer:
	      {
		const class BufferType* type =
		  static_cast<const class BufferType*>(currentContextType);
		class BufferExpression* b = new class BufferExpression (type);
		assert (currentType == type->getItemType ());

		if (!$1)
		  $$ = b;
		else if (isCompatible ($1, currentType))
		  $$ = b, b->append ($1);
		else
		  b->destroy (), $1->destroy (), $$ = NULL;
	      }
	      break;
	    case Type::tStruct:
	      {
		if (!$1 &&
		    !static_cast<const class StructType*>(currentContextType)
		    ->getSize ())
		  $$ = CSE (new class Constant
			    (new class StructValue (currentContextType)));
		else {
		  class StructExpression* s =
		    new class StructExpression (currentContextType);
		  assert (currentType == s->getNextType ());

		  if (isCompatible ($1, s))
		    $$ = s, s->append ($1), currentType = s->getNextType ();
		  else
		    s->destroy (), $1->destroy (),
		      $$ = NULL, currentType = NULL;
		}
	      }
	      break;
	    default:
	      assert (false);
	    }
	  }
	  else
	    $1->destroy (), $$ = NULL;

	  currentComponent++;
	}
	|
	struct_expr ',' _opt__expr
	{
	  if (($$ = $1)) {
	    assert (currentContextType != NULL);
	    assert (currentComponent > 0);
	    switch (currentContextType->getKind ()) {
	    case Type::tVector:
	      {
		assert ($$->getKind () == Expression::eVector);
		class VectorExpression* v =
		  static_cast<class VectorExpression*>($$);
		const class VectorType* type =
		  static_cast<const class VectorType*>(currentContextType);
		assert (currentType == type->getItemType ());

		if (currentComponent < type->getSize ()) {
		  if (!$3) {
		    if (!type->isOptional ()) {
		      pnerrorf ("missing expression for vector component %u",
				currentComponent);
		      $$->destroy (), $$ = NULL;
		    }
		  }
		  else if (isCompatible ($3, currentType))
		    v->setComponent (currentComponent, $3);
		  else
		    $$->destroy (), $3->destroy (), $$ = NULL;
		}
		else {
		  if (type->getName ())
		    pnerrorf ("array `%s' only has %u elements",
			      type->getName (), type->getSize ());
		  else
		    pnerrorf ("array only has %u elements", type->getSize ());

		  $$->destroy (), $3->destroy (), $$ = NULL;
		}
	      }
	      break;
	    case Type::tBuffer:
	      {
		assert ($$->getKind () == Expression::eBuffer);
		class BufferExpression* b =
		  static_cast<class BufferExpression*>($$);
		const class BufferType* type =
		  static_cast<const class BufferType*>(currentContextType);
		assert (currentType == type->getItemType ());

		if (b->hasCapacity ()) {
		  if (!$3 || !b->getSize ()) {
		    pnerrorf ("missing expression for buffer component %u",
			      $3 ? currentComponent : 0);
		    $$->destroy (), $3->destroy (), $$ = NULL;
		  }
		  else if (isCompatible ($3, currentType))
		    b->append ($3);
		  else
		    $$->destroy (), $3->destroy (), $$ = NULL;
		}
		else {
		  if (type->getName ())
		    pnerrorf ("%s `%s' only has %u slots",
			      type->isStack () ? "stack" : "queue",
			      type->getName (), type->getSize ());
		  else
		    pnerrorf ("%s only has %u slots",
			      type->isStack () ? "stack" : "queue",
			      type->getSize ());
		  $$->destroy (), $3->destroy (), $$ = NULL;
		}
	      }
	      break;
	    case Type::tStruct:
	      {
		assert ($$->getKind () == Expression::eStruct);
		class StructExpression* s =
		  static_cast<class StructExpression*>($$);
		assert (currentType == s->getNextType ());

		if (isCompatible ($3, s))
		  s->append ($3), currentType = s->getNextType ();
		else
		  $$->destroy (), $3->destroy (),
		    $$ = NULL, currentType = NULL;
	      }
	      break;
	    default:
	      assert (false);
	    }
	  }
	  else
	    $$->destroy (), $3->destroy (), $$ = NULL;

	  currentComponent++;
	}
	;

expr_:
	{ $$ = NULL; pnerror ("missing expression"); }
	|
	expr
	{ $$ = $1; }
	;

expr:
	formula
	{ $$ = ensureExpr ($1); }
	;

formula:
	error
	{ $$ = NULL; }
	|
	name_ NEXT
	{
	  $$ = NULL;
	  if ($1) {
	    if (class Function* f = getFunction ($1)) {
	      if (!($$ = fold (f->expand (NULL))))
		pnerrorf ("could not expand function `%s'", f->getName ());
	    }
	    else
	      pnerrorf ("there is no function `%s'", $1);
	  }
	}
	|
	name_ '('
	{
	  class Function* f = NULL;
	  if ($1 && !(f = getFunction ($1)))
	    pnerrorf ("there is no function `%s'", $1);
	  delete[] $1;
	  functionStack.push (f);
	}
	arg_list ')'
	{
	  assert (!functionStack.empty ());
	  if (class Function* f = functionStack.top ()) {
	    $$ = NULL;
	    if ($4 && !($$ = fold (f->expand ($4))))
	      pnerrorf ("could not expand function `%s'", f->getName ());
	  }
	  else
	    $$ = NULL;
	  $4->destroy ();
	  functionStack.pop ();
	}
	|
	TRUE_
	{
	  $$ = CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), true)));
	}
	|
	FALSE_
	{
	  $$ = CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), false)));
	}
	|
	_nopt_place placename
	{
	  if ($2) {
	    if (!allowSubmarking) {
	      pnerror ("place contents expression is not allowed here");
	      $$ = NULL;
	    }
	    else
	      $$ = CSE (new class PlaceContents ($2));
	  }
	  else
	    $$ = NULL;
	}
	|
	NAME
	{ $$ = resolveName ($1); }
	|
	NUMBER
	{
	  $$ = CSE (new class Constant
		    (new class IntValue (Net::getIntType (), $1)));
	}
	|
	CHARACTER
	{
	  $$ = CSE (new class Constant
		    (new class CharValue (Net::getCharType (), $1)));
	}
	|
	STRING
	{
	  if (currentType && currentType->getKind () == Type::tVector) {
	    const class VectorType* type =
	      static_cast<const class VectorType*>(currentType);
	    if (type->getItemType ()->getKind () == Type::tChar) {
	      const class CharType* charType =
		static_cast<const class CharType*>(type->getItemType ());
	      card_t length = $1->getLength ();
	      if (length == type->getSize ()) {
		class VectorValue* v = new class VectorValue (type);
		const char* const str = $1->getString ();
		for (card_t i = 0; i < length; i++) {
		  class CharValue* c = new class CharValue (charType, str[i]);
		  if (!charType->isConstrained (c)) {
		    pnerrorf ("%s constant 0x%02x out of range for type %s",
			      charType->getSyntacticName (),
			      static_cast<unsigned char>(c->getValue ()),
			      type->getName ()
			      ? type->getName ()
			      : type->getSyntacticName ());
		    delete c;
		    delete v;
		    v = NULL;
		    break;
		  }
		  else
		    (*v)[i] = c;
		}

		$$ = v ? CSE (new class Constant (v)) : NULL;
	      }
	      else {
		pnerrorf ("expecting an array of %u characters, got %u",
			  type->getSize (), length);
		$$ = NULL;
	      }

	      delete $1;
	    }
	    else
	      $$ = resolveName (convertString ($1));
	  }
	  else
	    $$ = resolveName (convertString ($1));
	}
	|
	UNDEFINED
	{ $$ = CSE (new class Undefined (Undefined::fError)); }
	|
	FATAL
	{ $$ = CSE (new class Undefined (Undefined::fFatalError)); }
	|
	'{'
	{
	  structLevels.push (StructLevel
			     (currentContextType, currentComponent));

	  if (!currentType) {
	    pnerror ("cannot determine type of structured expression");
	    currentContextType = NULL, currentComponent = 0;
	  }
	  else {
	    assert (!structLevels.empty () || !currentComponent);
	    currentContextType = currentType;
	    currentComponent = 0;

	    switch (currentType->getKind ()) {
	    case Type::tStruct:
	      currentType =
		static_cast<const class StructType*>(currentContextType)
		->getSize ()
		? (*static_cast<const class StructType*>(currentContextType))
		[currentComponent]
		: NULL;
	      break;
	    case Type::tVector:
	      currentType =
		static_cast<const class VectorType*>(currentContextType)
		->getItemType ();

	      if (pnwarnings) {
		const class Type* indexType =
		  static_cast<const class VectorType*>(currentContextType)
		  ->getIndexType ();
		assert (indexType != NULL);

		if (indexType->getKind () == Type::tInt) {
		  const class Constraint* constraint =
		    indexType->getConstraint ();
		  if (!constraint || constraint->size () <= 1)
		    break;
		}
		if (indexType->getKind () == Type::tBool)
		  break;

		pnwarn ("initializing an array indexed by something else"
			" than a boolean or a simple integer range");
	      }

	      break;
	    case Type::tBuffer:
	      currentType =
		static_cast<const class BufferType*>(currentContextType)
		->getItemType ();

	      break;
	    default:
	      pnerror ("cannot determine type of structured expression");
	      currentType = currentContextType = NULL;
	    }
	  }
	}
	struct_expr
	'}'
	{
	  $$ = $3;
	  if (currentContextType) {
	    if ($$) {
	      switch ($$->getKind ()) {
	      case Expression::eStruct:
		{
		  class StructExpression* s =
		    static_cast<class StructExpression*>($$);
		  if (s->getNextType ()) {
		    if (currentContextType->getName ())
		      pnerrorf ("struct `%s' not fully initialized",
				currentContextType->getName ());
		    else
		      pnerror ("struct not fully initialized");

		    $$->destroy (), $$ = NULL;
		  }
		}
		break;

	      case Expression::eConstant:
	      case Expression::eBuffer:
		break;

	      case Expression::eVector:
		{
		  card_t diff =
		    static_cast<const class VectorType*>(currentContextType)
		    ->getSize () - currentComponent;

		  if (diff) {
		    if (currentContextType->getName ())
		      pnerrorf ("%u %s of array `%s' left uninitialized",
				diff, diff == 1 ? "element" : "elements",
				currentContextType->getName ());
		    else
		      pnerrorf ("%u %s of array left uninitialized",
				diff, diff == 1 ? "element" : "elements");

		    $$->destroy (), $$ = NULL;
		  }
		}
		break;

	      default:
		assert (false);
	      }
	    }

	    StructLevel p = structLevels.top ();
	    currentType = currentContextType;
	    currentContextType = p.first;
	    currentComponent = p.second;
	  }
	  else {
	    assert (!currentType);
	    currentComponent = 0;
	  }

	  structLevels.pop ();

	  if ($$)
	    $$ = fold ($$);
	}
	|
	IS typereference
	{ pushTypeContext ($2); }
	formula %prec ATOM
	{
	  if (($$ = $4) && $$->getKind () == Expression::eMarking &&
	      $$->getType () == currentType);
	  else if (($$ = noMarking ($$)) && currentType) {
	    if (!$$->getType ()) {
	      pnerror ("expression to cast has no type");
	      $$->destroy (), $$ = NULL;
	    }
	    else if (!makeCompatible ($$, currentType, true));
	    else if ($$->getType () == currentType);
	    else if ((Typecast::isCastable (*currentType) &&
		      Typecast::isCastable (*$$->getType ()) ||
		      Type::isCompatible ($$->getType (), currentType)))
	      $$ = fold (new class Typecast (currentType, $$));
	    else {
	      pnerrorf ("cannot cast from `%s' to `%s'",
			$$->getType ()->getName ()
			? $$->getType ()->getName ()
			: $$->getType ()->getSyntacticName (),
			currentType->getName ()
			? currentType->getName ()
			: currentType->getSyntacticName ());
	      $$->destroy (), $$ = NULL;
	    }
	  }
	  popTypeContext ();
	}
	|
	name_ '='
	{
	  StructLevel p (currentContextType, currentComponent);
	  structLevels.push (p);
	  currentContextType = currentType;
	  currentType = NULL, currentComponent = 0;

	  if ($1 && currentContextType &&
	      currentContextType->getKind () == Type::tUnion) {
	    const class UnionType* u =
	      static_cast<const class UnionType*>(currentContextType);

	    if ((currentType = (*u)[$1]))
	      currentComponent = u->getIndex ($1);
	    else {
	      if (u->getName ())
		pnerrorf ("union `%s' does not have component `%s'",
			  u->getName (), $1);
	      else
		pnerrorf ("union does not have component `%s'", $1);
	    }
	  }
	  else if ($1) {
	    if (currentContextType && currentContextType->getName ())
	      pnerrorf ("cannot reference component `%s' of type `%s'",
			$1, currentContextType->getName ());
	    else
	      pnerrorf ("cannot reference component `%s' of non-union type",
			$1);
	  }

	  delete[] $1;
	}
	formula %prec ALWAYS
	{
	  assert (!currentType ||
		  (currentContextType->getKind () == Type::tUnion &&
		   currentType ==
		   (*static_cast<const class UnionType*>(currentContextType))
		   [currentComponent]));

	  if (($4 = ensureExpr ($4)) && currentType) {
	    const class UnionType* u =
	      static_cast<const class UnionType*>(currentContextType);

	    if (isCompatible ($4, currentType))
	      $$ = fold (new class UnionExpression (u, $4, currentComponent));
	    else {
	      if (u->getName ())
		pnerrorf ("type mismatch for union `%s' component `%s'",
			  u->getName (),
			  u->getComponentName (currentComponent));
	      else
		pnerrorf ("type mismatch for union component `%s'",
			  u->getComponentName (currentComponent));
	      $4->destroy (), $$ = NULL;
	    }
	  }
	  else
	    $4->destroy (), $$ = NULL;

	  assert (!structLevels.empty ());
	  StructLevel p = structLevels.top ();
	  currentType = currentContextType;
	  currentContextType = p.first;
	  currentComponent = p.second;
	  structLevels.pop ();

	  if ($$)
	    $$ = CSE ($$);
	}
	|
	formula IS name
	{
	  if (($1 = ensureExpr ($1)) && $3) {
	    if ($1->getType () && $1->getType ()->getKind () == Type::tUnion) {
	      const class UnionType* u =
		static_cast<const class UnionType*>($1->getType ());

	      if ((*u)[$3])
		$$ = fold (new class UnionTypeExpression ($1,
							  u->getIndex ($3)));
	      else {
		if (u->getName ())
		  pnerrorf ("union `%s' does not have component `%s'",
			    u->getName (), $3);
		else
		  pnerrorf ("union does not have component `%s'", $3);

		$1->destroy (), $$ = NULL;
	      }
	    }
	    else {
	      if ($1->getType () && $1->getType ()->getName ())
		pnerrorf ("type `%s' is not a union",
			  $1->getType ()->getName ());
	      else
		pnerror ("type is not a union");

	      $1->destroy (), $$ = NULL;
	    }
	  }
	  else
	    $1->destroy (), $$ = NULL;

	  delete[] $3;

	  if ($$)
	    $$ = CSE ($$);
	}
	|
	ATOM formula
	{ if (($$ = noMarking ($2))) { $$->setAtomic (true); $$ = CSE ($$); } }
	|
	typereference '!'
	{ beginAny ($1, NULL); }
	_opt_expr %prec ':'
	{ $$ = endAny ($1, NULL, $4); }
	|
	typereference name_ '!'
	{ beginAny ($1, $2); }
	_opt_expr %prec ':'
	{ $$ = endAny ($1, $2, $5); }
	|
	formula '<' formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rLess, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula EQ formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isCompatible ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rEqual, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '>' formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rGreater, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula GE formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rGreaterEqual, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula NE formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isCompatible ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rInequal, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula LE formula
	{
	  $1 = ensureExpr ($1); $3 = ensureExpr ($3);
	  if ($1 && $3 && isOrdered ($1, $3))
	    $$ = newRelopExpression (RelopExpression::rLessEqual, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '+' formula
	{
	  if ($1 && $1->getType () &&
	      $1->getType ()->getKind () == Type::tBuffer) {
	    const class BufferType* type =
	      static_cast<const class BufferType*>($1->getType ());
	    if (($1 = ensureExpr ($1)) && ($3 = ensureExpr ($3)) &&
		isCompatible ($3, type->getItemType ())) {
	      if ($1->getKind () == Expression::eBufferIndex) {
		class BufferIndex* b = static_cast<class BufferIndex*>($1);
		$$ = fold (new class BufferWrite (b->getBuffer (), $3,
						  b->getIndex ()));
		b->destroy ();
	      }
	      else
		$$ = fold (new class BufferWrite ($1, $3, NULL));
	    }
	    else
	      $1->destroy (), $3->destroy (), $$ = NULL;
	  }
	  else {
	    $1 = ensureInt ($1); $3 = ensureInt ($3);
	    if ($1 && $3)
	      $$ = fold (new class BinopExpression (BinopExpression::bPlus,
						    $1, $3));
	    else {
	      $1->destroy (); $3->destroy (); $$ = NULL;
	    }
	  }
	}
	|
	formula '-' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bMinus,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '/' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bDiv,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '*' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bMul,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '%' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bMod,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	'~' formula
	{
	  if (($$ = ensureInt ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uNot, $$));
	}
	|
	'#' typereference %prec '!'
	{
	  if ($2) {
	    card_t num = $2->getNumValues ();
	    assert (num > 0);
	    if (num > INT_T_MAX) {
	      pnerrorf ("type `%s' has too many values to count",
			$2->getName ()
			? $2->getName ()
			: $2->getSyntacticName ());
	      $$ = NULL;
	    }
	    else
	      $$ = CSE (new class Constant
			(new class IntValue
			 (Net::getIntType (), int_t (num))));
	  }
	  else
	    $$ = NULL;
	}
	|
	'^' typereference %prec '!'
	{
	  if ($2) {
	    if ($2->isOrdered ())
	      $$ = CSE (new class Constant ($2->getFirstValue ()));
	    else {
	      pnerrorf ("cannot get smallest value of non-ordered type `%s'",
			$2->getName ()
			? $2->getName ()
			: $2->getSyntacticName ());
	      $$ = NULL;
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	'&' typereference %prec '!'
	{
	  if ($2) {
	    if ($2->isOrdered ())
	      $$ = CSE (new class Constant ($2->getLastValue ()));
	    else {
	      pnerrorf ("cannot get largest value of non-ordered type `%s'",
			$2->getName ()
			? $2->getName ()
			: $2->getSyntacticName ());
	      $$ = NULL;
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	'|' formula %prec '!'
	{
	  if (($$ = ensureOrdered ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uPred, $$));
	}
	|
	'+' formula %prec '!'
	{
	  if (($$ = ensureOrdered ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uSucc, $$));
	}
	|
	'-' formula %prec '!'
	{
	  if (($2 = ensureExpr ($2)) && $2->getType () &&
	      $2->getType ()->getKind () == Type::tBuffer)
	    $$ = newBufferUnop (BufferUnop::bRemove, $2);
	  else if (($$ = ensureInt ($2)))
	    $$ = fold (new class UnopExpression (UnopExpression::uNeg, $$));
	}
	|
	'*' formula %prec '!'
	{ $$ = newBufferUnop (BufferUnop::bPeek, $2); }
	|
	'/' formula %prec '!'
	{ $$ = newBufferUnop (BufferUnop::bUsed, $2); }
	|
	'%' formula %prec '!'
	{ $$ = newBufferUnop (BufferUnop::bFree, $2); }
	|
	'?' formula %prec '!'
	{
	  if (($$ = ensureExpr ($2)) &&
	      $$->getKind () == Expression::eVectorIndex)
	    $$ = CSE (static_cast<class VectorIndex*>($$)->toComponent ());
	  else if ($$ && $$->getType () &&
		   $$->getType ()->getKind () == Type::tVector) {
	    const class VectorType* const type =
	      static_cast<const class VectorType*>($$->getType ());
	    if (type->isOptional ())
	      $$ = CSE (new class VectorDefined ($2));
	    else
	      pnerror ("ignoring unary '?':"
		       " vector has no optional components");
	  }
	  else if ($$)
	    pnerror ("ignoring unary `?': no vector context");
	}
	|
	formula ROL formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bRol,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula ROR formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bRor,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '&' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bAnd,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '^' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bXor,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '|' formula
	{
	  $1 = ensureInt ($1); $3 = ensureInt ($3);
	  if ($1 && $3)
	    $$ = fold (new class BinopExpression (BinopExpression::bOr,
						  $1, $3));
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '?' colon_formula
	{
	  $1 = noMarking ($1);
	  if ($1 && $3) {
	    if (!$1->getType ()) {
	      pnerror ("left-hand-side of ? has no type");
	      $1->destroy (); $3->destroy (); $$ = NULL;
	    }
	    else if ($1->getType ()->getNumValues () != $3->size ()) {
	      pnerrorf ("left-hand-side of ? has %u!=%u possible values",
			$1->getType ()->getNumValues (), $3->size ());
	      $1->destroy (); $3->destroy (); $$ = NULL;
	    }
	    else if ($1->getType ()->getNumValues () == 1) {
	      pnwarn ("left-hand-side of ? has only one possible value");
	      $$ = (*$3)[0]->copy (); $1->destroy (); $3->destroy ();
	    }
	    else
	      $$ = CSE (new class IfThenElse ($1, $3));
	  }
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula '.' name
	{
	  $1 = ensureExpr ($1);
	  if ($1 && $3) {
	    if ($1->getType () &&
		$1->getType ()->getKind () == Type::tStruct) {
	      const class StructType *s =
		static_cast<const class StructType*>($1->getType ());
	      if ((*s)[$3])
		$$ = fold (new class StructComponent ($1, s->getIndex ($3)));
	      else {
		if (s->getName ())
		  pnerrorf ("struct `%s' does not have component `%s'",
			    s->getName (), $3);
		else
		  pnerrorf ("struct does not have component `%s'", $3);

		$1->destroy (), $$ = NULL;
	      }
	    }
	    else if ($1->getType () &&
		     $1->getType ()->getKind () == Type::tUnion) {
	      const class UnionType *u =
		static_cast<const class UnionType*>($1->getType ());
	      if ((*u)[$3])
		$$ = fold (new class UnionComponent ($1, u->getIndex ($3)));
	      else {
		if (u->getName ())
		  pnerrorf ("union `%s' does not have component `%s'",
			    u->getName (), $3);
		else
		  pnerrorf ("union does not have component `%s'", $3);

		$1->destroy (), $$ = NULL;
	      }
	    }
	    else {
	      if ($1->getType () && $1->getType ()->getName ())
		pnerrorf ("cannot assign to component `%s' of type `%s'",
			  $3, $1->getType ()->getName ());
	      else
		pnerrorf ("cannot assign to component `%s' of expression", $3);

	      $1->destroy (), $$ = NULL;
	    }
	  }
	  else
	    $1->destroy (), $$ = NULL;

	  delete[] $3;
	}
	|
	formula '['
	{
	  if (($1 = ensureExpr ($1)) && $1->getType ()) {
	    switch ($1->getType ()->getKind ()) {
	    case Type::tVector:
	      {
		const class VectorType* type =
		  static_cast<const class VectorType*>($1->getType ());
		pushTypeContext (type->getIndexType ());
	      }
	      break;
	    case Type::tBuffer:
	      if ($1->getKind () == Expression::eBufferIndex)
		pnerror ("a buffer index has already been specified");
	      break;
	    default:
	      pnerror ("only vector-typed expressions can be indexed");
	      $1->destroy (), $1 = NULL;
	    }
	  }
	  else if ($1) {
	    pnerror ("no type for indexed expression");
	    $1->destroy (), $1 = NULL;
	  }
	}
	expr ']'
	{
	  if ($1) {
	    switch ($1->getType ()->getKind ()) {
	    case Type::tVector:
	      {
		const class VectorType* type =
		  static_cast<const class VectorType*>($1->getType ());
		popTypeContext ();

		if (isCompatible ($4, type->getIndexType ()))
		  $$ = fold (new class VectorIndex ($1, $4));
		else
		  $$ = NULL, $1->destroy (), $4->destroy ();
	      }
	      break;
	    case Type::tBuffer:
	      if ($1->getKind () == Expression::eBufferIndex)
		$4->destroy (), $$ = $1;
	      else
		$$ = CSE (new class BufferIndex ($1, $4));
	      break;
	    default:
	      assert (false);
	    }
	  }
	  else
	    $$ = NULL, $4->destroy ();
	}
	|
	'!' formula
	{
	  if (($$ = ensureBool ($2)))
	    $$ = fold (new class NotExpression ($$));
	}
	|
	formula AND formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3, true);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bAnd, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula OR formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3, true);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bOr, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula IMPL formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3, true);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bImpl, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula XOR formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bXor, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	formula EQUIV formula
	{
	  $1 = ensureBool ($1); $3 = ensureBool ($3);
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bEquiv, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	ALWAYS _no_qualifier formula
	{
	  $3 = ensureBool ($3);
	  $$ = $3
	    ? CSE (new class TemporalQuantifier (TemporalQuantifier::Always,
						 $3))
	    : NULL;
	}
	|
	EXISTS _no_qualifier formula
	{
	  $3 = ensureBool ($3);
	  $$ = $3
	    ? CSE (new class TemporalQuantifier (TemporalQuantifier::Exists,
						 $3))
	    : NULL;
	}
	|
	FINALLY qualifier formula
	{
	  $3 = ensureBool ($3);
	  if ($3)
	    $$ = CSE (new class TemporalUnop (TemporalUnop::Finally, $2, $3));
	  else {
	    $$ = NULL; $2->destroy ();
	  }
	}
	|
	GLOBALLY qualifier formula
	{
	  $3 = ensureBool ($3);
	  if ($3)
	    $$ = CSE (new class TemporalUnop (TemporalUnop::Globally, $2, $3));
	  else {
	    $$ = NULL; $2->destroy ();
	  }
	}
	|
	NEXT qualifier formula
	{
	  $3 = ensureBool ($3);
	  if ($3)
	    $$ = CSE (new class TemporalUnop (TemporalUnop::Next, $2, $3));
	  else {
	    $$ = NULL; $2->destroy ();
	  }
	}
	|
	formula UNTIL qualifier formula
	{
	  $1 = ensureBool ($1); $4 = ensureBool ($4);
	  if ($1 && $4)
	    $$ = CSE (new class TemporalBinop (TemporalBinop::Until,
					       $3, $1, $4));
	  else {
	    $1->destroy (); $3->destroy (); $4->destroy (); $$ = NULL;
	  }
	}
	|
	formula RELEASE qualifier formula
	{
	  $1 = ensureBool ($1); $4 = ensureBool ($4);
	  if ($1 && $4)
	    $$ = CSE (new class TemporalBinop (TemporalBinop::Release,
					       $3, $1, $4));
	  else {
	    $1->destroy (); $3->destroy (); $4->destroy (); $$ = NULL;
	  }
	}
	|
	EMPTY
	{ $$ = CSE (new class EmptySet); }
	|
	formula SUBSET formula
	{ $$ = newSetExpression (SetExpression::sSubset, $1, $3); }
	|
	formula INTERSECT formula
	{ $$ = newSetExpression (SetExpression::sIntersection, $1, $3); }
	|
	formula MINUS formula
	{ $$ = newSetExpression (SetExpression::sMinus, $1, $3); }
	|
	formula UNION formula
	{ $$ = newSetExpression (SetExpression::sUnion, $1, $3); }
	|
	formula EQUALS formula
	{ $$ = newSetExpression (SetExpression::sEquals, $1, $3); }
	|
	CARDINALITY formula
	{
	  if (class Expression* m = ensureSet ($2)) {
	    if (!allowCardinality) {
	      pnerror ("cardinality expressions not allowed here");
	      if (m->getKind () == Expression::eMarking)
		static_cast<class Marking*>(m)->associate (false);
	      m->destroy (), $$ = NULL;
	    }
	    else if (m->getKind () == Expression::eEmptySet) {
	      m->destroy ();
	      pnwarn ("the cardinality of an empty set is always zero");
	      $$ = CSE (new class Constant (new class IntValue
					    (Net::getIntType (), 0)));
	    }
	    else {
	      $$ = CSE (new class CardinalityExpression (m));
	      if (m->getKind () == Expression::eMarking)
		static_cast<class Marking*>(m)->associate ();
	      $$->associateVariables (true);
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	'(' marking_list_ ')'
	{ $$ = $2; }
	|
	formula '#' formula
	{
	  $1 = ensureInt ($1);
	  class Marking* m = ensureMarking ($3);

	  if (($$ = m) && $1) {
	    if ($1->getKind () == Expression::eConstant) {
	      const class Value* v =
		static_cast<class Constant*>($1)->getValue ();
	      assert (v->getKind () == Value::vInt);
	      int_t mult = static_cast<const class IntValue*>(v)->getValue ();
	      if (mult < 1) {
		pnerrorf ("non-positive cardinality %d", mult);
		$1->destroy (), m->associate (false), m->destroy ();
		$$ = m = NULL;
	      }
	      else {
		for (;;) {
		  if (class Expression* c = m->getMultiplicity ()) {
		    class Expression* card =
		      fold (new class BinopExpression (BinopExpression::bMul,
						       $1, c));
		    if (card && card->getKind () != Expression::eConstant)
		      card->associateVariables (true);

		    m->setMultiplicity (card);
		  }
		  else
		    m->setMultiplicity ($1);

		  if ((m = const_cast<class Marking*>(m->getNext ())))
		    $1 = CSE (new class Constant
			      (new class IntValue ($1->getType (), mult)));
		  else
		    break;
		}
	      }
	    }
	    else {
	      if (m->getNext ())
		$$ = m = static_cast<class Marking*>
		  (CSE (new class Marking (m->getPlace (), m)));

	      class Expression* card = $1;

	      if (class Expression* c = m->getMultiplicity ())
		card = fold
		  (new class BinopExpression (BinopExpression::bMul, $1, c));

	      if (card && card->getKind () != Expression::eConstant)
		card->associateVariables (true);

	      m->setMultiplicity (card);
	    }
	  }
	  else
	    $1->destroy ();

	  if (m)
	    m->associate ();
	}
	|
	typereference name_
	{ beginQuantifier ($2, $1); }
	_opt_expr
	{ continueQuantifier (ASSERT1 ($1)); }
	':' formula %prec ':'
	{ $$ = endQuantifier ($2, $1, $4, ensureMarking ($7)); }
	|
	SUBSET name '{' formula '}'
	{
	  if ($4 && $4->getKind () == Expression::eEmptySet) {
	    pnwarn ("any subset of an empty set is an empty set");
	    delete[] $2, $2 = NULL;
	  }
	  else if (($4 = ensureSet ($4)) && $2 && $4->getType ())
	    currentVariables.push (new class VariableDefinition
				   ($2, $4->getType ()));
	  else
	    delete[] $2, $2 = NULL;
	}
	formula %prec CARDINALITY
	{
	  $7 = ensureBool (ensureExpr ($7));
	  if (($$ = $4) && $7 && $2) {
	    if (!allowSubmarking) {
	      pnerror ("subset criteria not allowed here; ignoring");
	      $7->destroy ();
	    }
	    else {
	      ($7 = CSE ($7))->associateVariables (true);
#ifdef EXPR_CACHE
	      if (pnwarnings && !$7->invalidate (currentVariables.top ()))
		pnwarnf ("subset criterion does not depend on `%s'",
			 currentVariables.top ()->getName ());
#endif // EXPR_CACHE
	      if ($7->getKind () == Expression::eConstant) {
		const class Value* v =
		  static_cast<class Constant*>($7)->getValue ();
		assert (v->getKind () == Value::vBool);
		bool result =
		  static_cast<const class BoolValue*>(v)->getValue ();
		$7->destroy ();
		pnwarnf ("subset criterion is always %s",
			 result ? "true" : "false");
		if (result)
		  $$ = $4;
		else {
		  $$->destroy ();
		  $$ = CSE (new class EmptySet);
		}
	      }
	      else
		$$ = CSE (new class Submarking
			  (currentVariables.top (), $4, $7));
	    }
	    if ($4->getKind () == Expression::eMarking)
	      static_cast<class Marking*>($4)->associate ();
	  }
	  else
	    $7->destroy ();
	  if ($2)
	    currentVariables.pop ();
	}
	;

arg_list:
	{
	  if (class Function* f = functionStack.top ()) {
	    if (!f->getArity ())
	      $$ = new class ExpressionList;
	    else {
	      pnerrorf ("function `%s' expects %u!=0 arguments",
			f->getName (), f->getArity ());
	      $$ = NULL;
	    }
	  }
	  else
	    $$ = NULL;
	}
	|
	{
	  const class Type* type = NULL;

	  if (class Function* f = functionStack.top ()) {
	    if (f->getArity ())
	      type = (*f)[0]->getType ();
	    else
	      pnerrorf ("function `%s' has no arguments", f->getName ());
	  }

	  pushTypeContext (type);
	}
	formula
	{
	  if (currentType &&
	      isCompatible (($2 = noMarking ($2)), currentType)) {
	    if (class Function* f = functionStack.top ())
	      ($$ = new class ExpressionList)->append ($2);
	    else {
	      $2->destroy (); $$ = NULL;
	    }
	  }
	  else {
	    $2->destroy (); $$ = NULL;
	  }
	  popTypeContext ();
	}
	|
	arg_list ','
	{
	  const class Type* type = NULL;

	  if (class Function* f = functionStack.top ()) {
	    if ($1) {
	      if (f->getArity () > $1->size ())
		type = (*f)[$1->size ()]->getType ();
	      else
		pnerrorf ("function `%s' takes only %u arguments",
			  f->getName (), f->getArity ());
	    }
	  }

	  pushTypeContext (type);
	}
	formula
	{
	  if (($$ = $1) && currentType &&
	      isCompatible (($4 = noMarking ($4)), currentType)) {
	    if (class Function* f = functionStack.top ())
	      $$->append ($4);
	    else {
	      $$->destroy (); $4->destroy (); $$ = NULL;
	    }
	  }
	  else {
	    $$->destroy (); $4->destroy (); $$ = NULL;
	  }
	  popTypeContext ();
	}
	|
	arg_list ','
	{ if (($$ = $1)) pnerrorf ("extraneous comma in argument list"); }
	;


colon_formula:
	formula %prec ':'
	{
	  if (($1 = noMarking ($1)))
	    ($$ = new class ExpressionList)->prepend ($1);
	  else {
	    $1->destroy (); $$ = NULL;
	  }
	}
	|
	colon_formula ':' formula
	{
	  $3 = noMarking ($3);
	  if (($$ = $1) && $3) {
	    if ($$->getType () && $$->getType () != $3->getType ()) {
	      pnerror ("incompatible types in ?: expression");
	      $$->destroy (); $3->destroy (); $$ = NULL;
	    }
	    else
	      $$->prepend ($3);
	  }
	  else {
	    $$->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	;

_no_qualifier:
	qualifier
	{
	  if ($1) {
	    $1->destroy ();
	    pnerror ("unexpected transition instance qualifier");
	  }
	}
	;

qualifier:
	{ $$ = NULL; }
	|
	qualifier '[' qual_expr ']'
	{
	  if (($$ = $1)) {
	    pnerror ("transition instance qualifier already specified");
	    $3->destroy ();
	  }
	  else
	    $$ = $3;
	}
	;

qual_expr:
	transname ':'
	{ currentContextTransition = $1; }
	formula
	{
	  $4 = ensureBool (ensureExpr ($4));
	  if ($1 && $4)
	    $$ = CSE (new class TransitionQualifier ($1, CSE ($4)));
	  else
	    $4->destroy (), $$ = NULL;
	  assert (currentContextTransition == $1);
	  currentContextTransition = NULL;
	}
	|
	transname
	{ $$ = $1 ? CSE (new class TransitionQualifier ($1, NULL)) : NULL; }
	|
	'(' qual_expr ')'
	{ $$ = $2; }
	|
	'!' qual_expr
	{ if (($$ = $2)) $$ = fold (new class NotExpression ($$)); }
	|
	qual_expr AND qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bAnd, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr OR qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bOr, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr IMPL qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bImpl, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr XOR qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bXor, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	|
	qual_expr EQUIV qual_expr
	{
	  if ($1 && $3)
	    $$ = newBooleanBinop (BooleanBinop::bEquiv, $1, $3);
	  else {
	    $1->destroy (); $3->destroy (); $$ = NULL;
	  }
	}
	;

%%

static char*
convertString (class StringBuffer* sbuf)
{
  assert (sbuf != NULL);
  char* str = newString (sbuf->getString ());
  if (strlen (str) != sbuf->getLength ())
    pnerror ("truncating name containing NUL chars");
  delete sbuf;
  return str;
}

static class Type*
copyType (class Type* type)
{
  if (type->getName () ||
      type == static_cast<const class Type*>(Net::getBoolType ()) ||
      type == static_cast<const class Type*>(Net::getIntType ()) ||
      type == static_cast<const class Type*>(Net::getCharType ()))
    return type->copy ();
  else
    return type;
}

static void
deleteType (class Type* type)
{
  if (type && !type->getName () &&
      type != static_cast<const class Type*>(Net::getBoolType ()) &&
      type != static_cast<const class Type*>(Net::getIntType ()) &&
      type != static_cast<const class Type*>(Net::getCharType ()))
    delete type;
}

static void pushTypeContext (const class Type* type)
{
  StructLevel p (currentContextType, currentComponent);
  structLevels.push (p);
  StructLevel q (currentType, 0);
  structLevels.push (q);

  currentType = type;
  currentContextType = NULL, currentComponent = 0;
}

static void popTypeContext ()
{
  StructLevel p = structLevels.top ();
  structLevels.pop ();
  currentType = p.first;
  assert (!p.second);
  p = structLevels.top ();
  structLevels.pop ();
  currentContextType = p.first;
  currentComponent = p.second;
}

static const class VariableDefinition*
getParameter (char* name)
{
  for (unsigned i = 0; i < currentArity; i++) {
    if (!strcmp (currentParameters[i]->getName (), name))
      return currentParameters[i];
  }

  return NULL;
}

static void
addParameter (const class Type* type, char* name)
{
  if (!type || !name) {
    delete[] name;
    return;
  }

  if (getParameter (name)) {
    pnerrorf ("ignoring redefinition of parameter `%s'", name);
    delete[] name;
    return;
  }

  if (pnwarnings) {
    if (currentTransition && currentTransition->getVariable (name))
      pnwarnf ("parameter `%s' shadows a variable", name);
    else if (getConstantFunction (name))
      pnwarnf ("parameter `%s' shadows a constant", name);
    else if (theNet.getType (name))
      pnwarnf ("parameter `%s' shadows a type name", name);
  }

  class VariableDefinition** v =
    new class VariableDefinition*[currentArity + 1];
  assert (!currentArity || currentParameters);
  assert (currentArity || !currentParameters);

  memcpy (v, currentParameters, currentArity * sizeof *v);
  delete[] currentParameters;
  currentParameters = v;
  currentParameters[currentArity++] =
    new class VariableDefinition (name, type);
}

static class Function*
getFunction (const char* name)
{
  class Function* f = currentTransition
    ? currentTransition->getFunction (name)
    : (currentContextTransition
       ? currentContextTransition->getFunction (name)
       : NULL);
  if (!f)
    f = theNet.getFunction (name);
  return f;
}

static class Function*
getConstantFunction (const char* name)
{
  class Function* f = currentTransition
    ? currentTransition->getFunction (name)
    : (currentContextTransition
       ? currentContextTransition->getFunction (name)
       : NULL);
  if (!f || f->getArity ())
    f = theNet.getFunction (name);
  return f && f->getArity () ? NULL : f;
}

static const class VariableDefinition*
getVariable (char* name)
{
  assert (name != NULL);

  const class VariableDefinition* v = getParameter (name);
  const class Function* f = getConstantFunction (name);

  if (v || (v = currentVariables.find (name)))
    goto found;

  if (!currentTransition) {
    if (currentContextTransition &&
	(v = currentContextTransition->getVariable (name)))
      assert (v->getType () != NULL);
    else if (!f &&
	     !(currentType && currentType->getKind () == Type::tEnum &&
	       static_cast<const class EnumType*>(currentType)
	       ->getValue (name)))
      pnerrorf ("undefined variable `%s'", name);
    delete[] name;
  }
  else if ((v = currentTransition->getVariable (name))) {
  found:
    assert (v->getType () != NULL);
    delete[] name;
  }
  else if (f)
    delete[] name;
  else if (!currentType) {
    pnerrorf ("cannot determine type of variable `%s'", name);
    delete[] name;
  }
  else {
    assert (currentType != NULL);

    if (pnwarnings && theNet.getType (name))
      pnwarnf ("variable `%s' shadows a type name", name);

    if (currentType->getKind () != Type::tEnum ||
	!static_cast<const class EnumType*>(currentType)->getValue (name))
      v = currentTransition->addVariable (name, currentType);
    else
      delete[] name;
  }

  return v;
}

static class Expression*
resolveName (char* name)
{
  class Function* f = getConstantFunction (name);
  const int_t*const i =
    (currentType && currentType->getKind () == Type::tEnum)
    ? static_cast<const class EnumType*>(currentType)->getValue (name)
    : NULL;
  if (const class VariableDefinition* d = getVariable (name))
    return static_cast<class Variable*>(CSE (new class Variable (d)));
  else if (class Expression* expr = f ? f->expand (NULL) : NULL)
    return fold (expr);
  else if (i)
    return enumValue (static_cast<const class EnumType*>(currentType), *i);
  else
    return NULL;
}

bool
isSet (const class Expression* expr)
{
  if (!expr)
    return false;
  switch (expr->getKind ()) {
  case Expression::eMarking:
  case Expression::ePlaceContents:
  case Expression::eSubmarking:
  case Expression::eEmptySet:
    return true;
  case Expression::eSet:
    switch (static_cast<const class SetExpression*>(expr)->getOp ()) {
    case SetExpression::sIntersection:
    case SetExpression::sMinus:
    case SetExpression::sUnion:
      return true;
    case SetExpression::sSubset:
    case SetExpression::sEquals:
      return false;
    }
  default:
    return false;
  }
}

static class Expression*
ensureSet (class Expression* expr)
{
  return isSet (expr) ? expr : ensureMarking (expr);
}

static class Marking*
ensureMarking (class Expression* expr)
{
  if (!expr)
    return NULL;

  assert (expr->getType () || expr->getKind () == Expression::eEmptySet);

  if (expr->getKind () == Expression::eMarking)
    return static_cast<class Marking*>(expr);
  else if (expr->isBasic ()) {
    if (currentPlace) {
      assert (currentType && currentType == currentPlace->getType ());

      if (!expr->getType ()->isAssignable (currentType)) {
	if (currentType->getName ())
	  pnerrorf ("marking expression is incompatible with type `%s'",
		    currentType->getName ());
	else
	  pnerror ("marking expression has incompatible type");

	expr->destroy ();
	return NULL;
      }
    }

    if (!currentType || makeCompatible (expr, currentType)) {
      if (currentType) {
	if (!expr->getType () ||
	    expr->getType ()->isAlwaysAssignable (currentType))
	  expr->setType (currentType);
	else {
	  pnerrorf ("marking expression has too broad type `%s'",
		    expr->getType ()->getName ()
		    ? expr->getType ()->getName ()
		    : expr->getType ()->getSyntacticName ());
	  expr->destroy ();
	  return NULL;
	}
      }
      class Marking* m = new class Marking (currentPlace);
      m->setToken (expr);
      return m;
    }
    else {
      expr->destroy ();
      return NULL;
    }
  }
  else {
    pnerror ("marking expression expected");
    expr->destroy ();
    return NULL;
  }
}

static class Expression*
noMarking (class Expression* expr)
{
  if (expr && expr->getKind () == Expression::eMarking) {
    pnerror ("marking expression not allowed here");
    expr->destroy ();
    return NULL;
  }

  return expr;
}

static class Expression*
ensureExpr (class Expression* expr)
{
  expr = noMarking (expr);

  if (expr && !expr->isBasic ()) {
    pnerror ("basic expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureBool (class Expression* expr, bool undefined)
{
  expr = noMarking (expr);

  if (!expr)
    return NULL;

  if (expr->getKind () == Expression::eUndefined) {
    if (undefined)
      return expr;
    pnerror ("undefined or fatal expression is not meaningful here");
    expr->destroy (); return NULL;
  }

  if (!expr->getType ()) {
    pnerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (expr->getType ()->getKind () != Type::tBool) {
    pnerror ("boolean expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureInt (class Expression* expr)
{
  expr = ensureExpr (expr);

  if (!expr)
    return NULL;

  if (!expr->getType ()) {
    pnerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (expr->getType ()->getKind () != Type::tInt) {
    pnerror ("integer expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureBuffer (class Expression* expr)
{
  expr = ensureExpr (expr);

  if (!expr)
    return NULL;

  if (!expr->getType ()) {
    pnerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (expr->getType ()->getKind () != Type::tBuffer) {
    pnerror ("buffer expression expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static class Expression*
ensureOrdered (class Expression* expr)
{
  expr = ensureExpr (expr);

  if (!expr)
    return NULL;

  if (!expr->getType ()) {
    pnerror ("expression has no type");
    expr->destroy (); return NULL;
  }

  if (!expr->getType ()->isOrdered ()) {
    pnerror ("expression of ordered type expected");
    expr->destroy (); return NULL;
  }

  return expr;
}

static void
addEnumeration (class EnumType* e, char* name)
{
  assert (e && name);
  const int_t* const i = e->getValue (name);
  if (i) {
    pnerrorf ("enumeration constant %s already defined as %d", name, *i);
    delete[] name;
  }
  else
    e->addEnumeration (name);
  if (pnwarnings) {
    if (getConstantFunction (name))
      pnwarnf ("enumeration constant %s shadows a constant", name);
    else if (theNet.getType (name))
      pnwarnf ("enumeration constant %s shadows a type name", name);
  }
}

static void
addEnumeration (class EnumType* e, char* name, int_t value)
{
  assert (e && name);
  const int_t* const i = e->getValue (name);
  if (e->getValue (name)) {
    pnerrorf ("enumeration constant %s defined as %d and %d", name, *i, value);
    delete[] name;
  }
  else if (e->hasValue (value)) {
    pnerrorf ("enumeration already contains the value %d", value);
    delete[] name;
  }
  else
    e->addEnumeration (name, value);
  if (pnwarnings) {
    if (getConstantFunction (name))
      pnwarnf ("enumeration constant %s shadows a constant", name);
    else if (theNet.getType (name))
      pnwarnf ("enumeration constant %s shadows a type name", name);
  }
}

static class Value*
ensureConstant (class Expression* expr, const class Type* type)
{
  if (!(expr = fold (expr)))
    return NULL;

  if (!type) {
    expr->destroy ();
    return NULL;
  }

  if (!type->isOrdered ()) {
    pnerrorf ("cannot define constraint for non-ordered type `%s'",
	      type->getName () ? type->getName () : type->getSyntacticName ());
    expr->destroy ();
    return NULL;
  }

  class Value* v = NULL;
  if (expr->getKind () == Expression::eConstant) {
    v = expr->eval (NULL);
    assert (v && v->getType ());
  }
  else
    pnerror ("constant expected");

  expr->destroy ();
  /** the integer value */
  int_t value;

  switch (v->getType ()->getKind ()) {
  case Type::tInt:
    assert (v->getKind () == Value::vInt);
    if (v->getType () == type)
      return v;
    value = static_cast<class IntValue*>(v)->getValue ();
    break;
  case Type::tBool:
    assert (v->getKind () == Value::vBool);
    if (v->getType () == type)
      return v;
    value = static_cast<class BoolValue*>(v)->getValue () ? 1 : 0;
    break;
  case Type::tChar:
    assert (v->getKind () == Value::vChar);
    if (v->getType () == type)
      return v;
    value = static_cast<class CharValue*>(v)->getValue ();
    break;
  case Type::tEnum:
    assert (v->getKind () == Value::vEnum);
    if (v->getType () == type)
      return v;
    value = static_cast<class EnumValue*>(v)->getValue ();
    break;

  default:
    if (v->getType () == type)
      return v;
    pnerrorf ("cannot define constraint: incompatible type `%s'",
	      v->getType ()->getName ()
	      ? v->getType ()->getName ()
	      : v->getType ()->getSyntacticName ());
    delete v;
    return NULL;
  }

  if (pnwarnings && v->getType ()->getKind () != type->getKind ())
    pnwarnf ("converting constant %d from %s to %s", value,
	     v->getType ()->getName ()
	     ? v->getType ()->getName ()
	     : v->getType ()->getSyntacticName (),
	     type->getName () ? type->getName () : type->getSyntacticName ());

  delete v;
  v = NULL;
  switch (type->getKind ()) {
  case Type::tInt:
    if (type->isConstrained (v = new class IntValue (type, value)))
      return v;
    delete v;
    break;
  case Type::tBool:
    if ((value == 0 || value == 1) &&
	type->isConstrained (v = new class BoolValue (type, value == 1)));
      return v;
    delete v;
    break;
  case Type::tChar:
    if (value >= CHAR_T_MIN && value <= CHAR_T_MAX &&
	type->isConstrained (v = new class CharValue (type, value)))
      return v;
    delete v;
    break;
  case Type::tEnum:
    if (type->isConstrained (v = new class EnumValue (type, value)) ||
	!type->getConstraint ())
      return v;
    delete v;
    break;

  default:
    assert (false);
  }

  pnerrorf ("value %d out of range for type %s", value,
	    type->getName () ? type->getName () : type->getSyntacticName ());
  return NULL;
}

static class Expression*
enumValue (const class EnumType* type, int_t value)
{
  class EnumValue* e = new class EnumValue (type, value);
  if (type->isConstrained (e))
    return CSE (new class Constant (e));
  else {
    if (const char* c = type->getEnumName (value))
      pnerrorf ("enumeration constant %s=%d out of range for %s", c, value,
		type->getName ()
		? type->getName ()
		: type->getSyntacticName ());
    else
      pnerrorf ("enumeration constant %d out of range for %s", value,
		type->getName ()
		? type->getName ()
		: type->getSyntacticName ());
    delete e;
    return NULL;
  }
}

static class Expression*
namedComponent (char* name, class Expression* expr)
{
  if (!name) {
    expr->destroy ();
    return NULL;
  }

  if (!currentContextType || currentContextType->getKind () != Type::tStruct) {
    pnerrorf ("named component `%s' can only be used with struct values",
	      name);
    expr->destroy ();
    delete[] name;
    return NULL;
  }

  const class StructType& type =
    static_cast<const class StructType&>(*currentContextType);

  if (currentComponent < type.getSize ()) {
    const char* comp = type.getComponentName (currentComponent);
    if (strcmp (name, comp)) {
      pnerrorf ("expecting named component `%s' instead of `%s'", comp, name);
      expr->destroy ();
      delete[] name;
      return NULL;
    }
  }
  else {
    pnerrorf ("extraneous named component `%s' at end of struct", name);
    expr->destroy ();
    delete[] name;
    return NULL;
  }

  delete[] name;
  return expr;
}

static void
beginTester (class Place* place)
{
  if (theNet.hasTester ())
    pnerror ("only one tester may be specified");
  else
    currentType = (currentPlace = place) ? place->getType () : NULL;

  assert (allowSubmarking);
  allowSubmarking = false;
}

static void
endTester (enum Net::TesterKind kind, class Marking* expr)
{
  if (currentPlace && expr)
    theNet.setTester (kind, expr);
  else
    expr->destroy ();
  currentPlace = NULL, currentType = NULL;

  assert (!allowSubmarking);
  allowSubmarking = true;
}

static void
beginAny (class Type*& type, char* name)
{
  if (type) {
    if (currentTransition) {
      if (name && currentTransition->getVariable (name)) {
	pnerrorf ("output variable `%s' shadows another name", name);
	type = NULL, delete[] name;
      }
      else if (name)
	currentTransition->addVariable (name, type);
      card_t size = type->getNumValues ();
      if (!size || size == CARD_T_MAX || size > maxRange)
	pnerrorf ("type `%s' has too many (%u) values for non-determinism",
		  type->getName ()
		  ? type->getName ()
		  : type->getSyntacticName (),
		  size);
    }
    else {
      pnerror ("non-determinism is only allowed on output arcs");
      type = NULL, delete[] name;
    }
  }
  else
    delete[] name;

  pushTypeContext (type);
}

static class Variable*
endAny (const class Type* type, char* name, class Expression* condition)
{
  class Variable* var;
  popTypeContext ();

  if (type) {
    if (!condition || (condition = ensureBool (ensureExpr (condition)))) {
      if (condition)
	condition->associateVariables (true);
      var = static_cast<class Variable*>
	(CSE (new class Variable
	      (currentTransition->addOutputVariable (type, name, condition))));
      assert (var != NULL);
#ifdef EXPR_CACHE
      if (condition && !condition->invalidate (var->getVariable ()))
	pnwarnf ("condition does not depend on output variable `%s'",
		 var->getVariable ()->getName ());
#endif // EXPR_CACHE
    }
    else
      var = NULL, condition->destroy ();
  }
  else
    var = NULL, condition->destroy ();
  return var;
}

static void
beginQuantifier (char* name, const class Type* type)
{
  if (name && type) {
    if (pnwarnings) {
      if (currentVariables.find (name))
	pnwarnf ("quantifier `%s' shadows another quantifier", name);
      else if (currentTransition && currentTransition->getVariable (name))
	pnwarnf ("quantifier `%s' shadows a transition variable", name);
      else if (currentContextTransition &&
	       currentContextTransition->getVariable (name))
	pnwarnf ("quantifier `%s' shadows a transition variable", name);
      else if (getConstantFunction (name))
	pnwarnf ("quantifier `%s' shadows a constant", name);
      else if (currentType && currentType->getKind () == Type::tEnum) {
	if (const int_t*const c =
	    static_cast<const class EnumType*>(currentType)->getValue (name))
	  pnwarnf ("quantifier `%s' shadows an enumeration constant %d",
		   name, *c);
      }
      else if (theNet.getType (name))
	pnwarnf ("quantifier `%s' shadows a type name", name);
    }

    card_t size = type->getNumValues ();
    if (!size || size == CARD_T_MAX || size > maxRange)
      pnerrorf ("type `%s' with %u values cannot be used for quantification",
		type->getName ()
		? type->getName ()
		: type->getSyntacticName (),
		size);

    currentVariables.push (new class VariableDefinition (name, type, true));
  }
  assert (!currentContextType);
  struct TypeContext tc = {
    currentContextType,
    currentType,
    currentComponent
  };
  typeStack.push (tc);

  currentContextType = NULL;
  currentComponent = 0;
  currentType = type;
}

static void
continueQuantifier (ASSERT1 (const class Type* type))
{
  assert (!currentContextType);
  assert (currentType == type);
  assert (!typeStack.empty ());
  struct TypeContext tc = typeStack.top ();
  typeStack.pop ();

  currentContextType = tc.context;
  currentType = tc.component;
  currentComponent = tc.numComponent;
}

/** Marking being checked */
static const class Marking* currentMarking;

static class Marking*
endQuantifier (char* name, const class Type* type,
	       class Expression* condition, class Marking* expr)
{
  if (name && type && expr) {
    class VariableDefinition* var = currentVariables.top ();
    currentVariables.pop ();

    if ((condition = ensureBool (ensureExpr (condition)))) {
      if (!currentVariables.empty ()) {
	currentMarking = expr;
	if (!condition->forVariables (isQuantifier))
	  assert (false);
	currentMarking = NULL;
      }
      condition->associateVariables (true);
#ifdef EXPR_CACHE
      if (!condition->invalidate (var))
	pnwarnf ("condition does not depend on quantifier `%s'",
		 var->getName ());
#endif // EXPR_CACHE
    }
      
    expr->addQuantifier (new class Quantifier (var, condition));
  }
  else {
    condition->destroy (), expr->destroy (), expr = NULL;
    if (name && type)
      currentVariables.pop ();
    else
      delete[] name;
  }

  return expr;
}

static bool isQuantifier (const class VariableDefinition& variable)
{
  assert (!!currentMarking);
  if (!variable.isAggregate ())
    variable.associateMarking (*currentMarking);
  return true;
}

static bool
makeCompatible (class Expression*& expr, const class Type* type, bool casting)
{
  assert (expr && expr->getType ());
  assert (type != NULL);

  if (expr->getKind () == Expression::eConstant) {
    const class Value* const cv =
      static_cast<class Constant*>(expr)->getValue ();
    /** the integer value to be converted */
    int_t value;
    switch (cv->getKind ()) {
    case Value::vInt:
      value = static_cast<const class IntValue*>(cv)->getValue ();
      break;
    case Value::vBool:
      value = static_cast<const class BoolValue*>(cv)->getValue () ? 1 : 0;
      break;
    case Value::vChar:
      value = static_cast<const class CharValue*>(cv)->getValue ();
      break;
    case Value::vEnum:
      value = static_cast<const class EnumValue*>(cv)->getValue ();
      break;
    default:
      if (cv->getType ()->isAssignable (type)) {
	if (type->isConstrained (cv)) {
	  expr->setType (type);
	  return true;
	}
	else {
	  pnerrorf ("%s constant out of range for type %s",
		    expr->getType ()->getSyntacticName (),
		    type->getName ()
		    ? type->getName ()
		    : type->getSyntacticName ());
	  return false;
	}
      }
      else {
      incompatible:
	pnerrorf ("%s constant is meaningless for type %s",
		  cv->getType ()->getSyntacticName (),
		  type->getName ()
		  ? type->getName ()
		  : type->getSyntacticName ());
	return false;
      }
    }

    class Value* v = NULL;

    switch (type->getKind ()) {
    case Type::tInt:
      if (!type->isConstrained (v = new class IntValue (type, value)))
	delete v, v = NULL;
      break;
    case Type::tBool:
      if (!(value == 0 || value == 1) ||
	  !type->isConstrained (v = new class BoolValue (type, value == 1)))
	delete v, v = NULL;
      break;
    case Type::tChar:
      if (value < CHAR_T_MIN || value > CHAR_T_MAX ||
	  !type->isConstrained (v = new class CharValue (type, value)))
	delete v, v = NULL;
      break;
    case Type::tEnum:
      if (!type->isConstrained (v = new class EnumValue (type, value)))
	delete v, v = NULL;
      break;

    default:
      goto incompatible;
    }

    if (v) {
      if (!casting && pnwarnings && !expr->getType ()->isAssignable (type))
	pnwarnf ("converting %s constant %d to type %s",
		 expr->getType ()->getSyntacticName (), value,
		 type->getName ()
		 ? type->getName ()
		 : type->getSyntacticName ());
      expr->destroy ();
      expr = CSE (new class Constant (v));
      return true;
    }
    else {
      const char* const name =
	type->getName () ? type->getName () : type->getSyntacticName ();

      switch (cv->getKind ()) {
      case Value::vInt:
	pnerrorf ("%s constant %d out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<const class IntValue*>(cv)->getValue (),
		  name);
	break;
      case Value::vBool:
	pnerrorf ("%s constant %s out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<const class BoolValue*>(cv)->getValue ()
		  ? "true" : "false", name);
	break;
      case Value::vEnum:
	pnerrorf ("%s constant %d out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<const class EnumValue*>(cv)->getValue (),
		  name);
	break;
      case Type::tChar:
	pnerrorf ("%s constant 0x%02x out of range for type %s",
		  cv->getType ()->getSyntacticName (),
		  static_cast<unsigned char>
		  (static_cast<const class CharValue*>(cv)->getValue ()),
		  name);
	break;
      default:
	assert (false);
      }
      return false;
    }
  }
  else
    return true;
}

static bool
isCompatible (class Expression*& expr, const class Type* type)
{
  if (!type) {
    pnerror ("too many components for struct");
    return false;
  }

  if (!expr)
    return false;

  if (!expr->getType ()) {
    pnerror ("expression has no type");
    return false;
  }

  if (!makeCompatible (expr, type))
    return false;

  if (!expr->getType ()->isAlwaysAssignable (type)) {
    const char* reason = Type::isCompatible (expr->getType (), type)
      ? "has too broad type for" : "is incompatible with type";
    pnerrorf ("expression %s `%s'", reason,
	     type->getName ()
	     ? type->getName ()
	     : type->getSyntacticName ());
    return false;
  }

  return true;
}

static bool
isCompatible (class Expression*& expr, const class StructExpression* s)
{
  assert (s != NULL);
  const class Type* type = s->getNextType ();
  const char* structName = s->getType ()->getName ();

  if (!type) {
    if (structName)
      pnerrorf ("too many components for struct `%s'",
		structName);
    else
      pnerror ("too many components for struct");
    return false;
  }

  const char* comp =
    static_cast<const class StructType*>(s->getType ())
    ->getComponentName (s->size ());

  if (!expr) {
    if (structName)
      pnerrorf ("error in expression for struct `%s' component `%s'",
		structName, comp);
    else
      pnerrorf ("error in expression for struct component `%s'", comp);
    return false;
  }

  if (!expr->getType ()) {
    if (structName)
      pnerrorf ("expression for struct `%s' component `%s' has no type",
		structName, comp);
    else
      pnerrorf ("expression for struct component `%s' has no type", comp);
    return false;
  }

  if (!makeCompatible (expr, type))
    return false;

  if (!expr->getType ()->isAlwaysAssignable (type)) {
    const char* reason = Type::isCompatible (expr->getType (), type)
      ? "too broad type" : "type mismatch";

    if (structName)
      pnerrorf ("%s for struct `%s' component `%s'", reason,
		structName, comp);
    else
      pnerrorf ("%s for struct component `%s'", reason, comp);
    return false;
  }

  return true;
}

static bool
isCompatible (class Expression*& expr1, class Expression*& expr2)
{
  if (!expr1 || !expr2)
    return false;

  if (!expr1->getType () || !expr2->getType ()) {
    pnerror ("expression has no type");
    return false;
  }

  if (!makeCompatible (expr1, expr2->getType ()) ||
      !makeCompatible (expr2, expr1->getType ()))
    return false;

  if (!Type::isCompatible (expr1->getType (), expr2->getType ())) {
    pnerror ("subexpressions have incompatible types");
    return false;
  }

  return true;
}

static bool
isOrdered (class Expression*& expr1, class Expression*& expr2)
{
  if (isCompatible (expr1, expr2)) {
    if (expr1->getType ()->isOrdered () && expr2->getType ()->isOrdered ())
      return true;

    pnerror ("invalid comparison for unordered types");
  }

  return false;
}

bool
check (const class Valuation& v)
{
  bool ok = true;

  if (v.isDivisionByZero ())
    ok = false, pnerror ("division by zero");
  if (v.isOverflow ())
    ok = false, pnerror ("overflow");
  if (v.isShiftError ())
    ok = false, pnerror ("shift error");
  if (v.isConstraintViolation ())
    ok = false, pnerror ("constraint violation");
  if (v.isUnionViolation ())
    ok = false, pnerror ("union violation");
  if (v.isUndefinedComponent ())
    ok = false, pnerror ("undefined component");
  if (v.isBufferViolation ())
    ok = false, pnerror ("buffer violation");

  assert (!v.isNegativeCardinality ());
  assert (!v.isUnificationError ());
  assert (!v.isCapacityViolation ());

  return ok;
}

static class Expression*
fold (class Expression* expr)
{
  if (expr && expr->getKind () != Expression::eConstant && expr->isBasic ()) {
    class Valuation v;
    class Value* value = expr->eval (&v);
    if (value) {
      assert (v.isOK ());
      expr->destroy ();
      return CSE (new class Constant (value));
    }
    else {
      if (!check (v)) {
	expr->destroy ();
	return NULL;
      }
      else {
	if (v.isFatalExpression ()) {
	  expr->destroy ();
	  return CSE (new class Undefined (Undefined::fFatalError));
	}
	if (v.isUndefinedExpression ()) {
	  expr->destroy ();
	  return CSE (new class Undefined (Undefined::fError));
	}

	assert (v.isUndefinedValue ());
	return CSE (expr);
      }
    }
  }
  else
    return CSE (expr);
}

static class Expression*
newRelopExpression (enum RelopExpression::Op op,
		    class Expression* left,
		    class Expression* right)
{
  if (left == right) {
    bool result;
    switch (op) {
    case RelopExpression::rLess:
    case RelopExpression::rGreater:
    case RelopExpression::rInequal:
      result = false;
      break;
    case RelopExpression::rEqual:
    case RelopExpression::rGreaterEqual:
    case RelopExpression::rLessEqual:
      result = true;
      break;
    }

    pnwarnf ("result of comparison is always %s", result ? "true" : "false");
    left->destroy ();
    right->destroy ();
    return CSE (new class Constant
		(new class BoolValue (Net::getBoolType (), result)));
  }

  return fold (new class RelopExpression (op, left, right));
}

static class Expression*
newBooleanBinop (enum BooleanBinop::Op op,
		 class Expression* left,
		 class Expression* right)
{
  assert (left && right);
  assert (left->getKind () == Expression::eUndefined ||
	  (left->getType () && left->getType ()->getKind () == Type::tBool));
  assert (right->getKind () == Expression::eUndefined ||
	  (right->getType () && right->getType ()->getKind () == Type::tBool));
  const class Value* v;
  bool value;
  assert (left == CSE (left));
  assert (right == CSE (right));

  if (left->getKind () == Expression::eConstant) {
    v = static_cast<class Constant*>(left)->getValue ();
    left->destroy ();
    assert (v && v->getKind () == Value::vBool);
    value = static_cast<const class BoolValue*>(v)->getValue ();

    switch (op) {
    case BooleanBinop::bAnd:
      if (value)
	return right;
      else {
	right->destroy ();
      False:
	pnwarn ("subexpression is always false");
	return CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), false)));
      }
    case BooleanBinop::bOr:
      if (value) {
	right->destroy ();
      True:
	pnwarn ("subexpression is always true");
	return CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), true)));
      }
      else
	return right;
    case BooleanBinop::bImpl:
      if (value)
	return right;
      else {
	right->destroy ();
	goto True;
      }
    case BooleanBinop::bXor:
      if (value)
	return fold (new class NotExpression (right));
      else
	return right;
    case BooleanBinop::bEquiv:
      if (value)
	return right;
      else
	return fold (new class NotExpression (right));
    }
  }
  else if (right->getKind () == Expression::eConstant) {
    v = static_cast<class Constant*>(right)->getValue ();
    right->destroy ();
    assert (v && v->getKind () == Value::vBool);
    value = static_cast<const class BoolValue*>(v)->getValue ();

    switch (op) {
    case BooleanBinop::bAnd:
      if (value)
	return left;
      else {
	left->destroy ();
	goto False;
      }
    case BooleanBinop::bOr:
      if (value) {
	left->destroy ();
	goto True;
      }
      else
	return left;
    case BooleanBinop::bImpl:
      if (value) {
	left->destroy ();
	goto True;
      }
      else
	return fold (new class NotExpression (left));
    case BooleanBinop::bXor:
      if (value)
	return fold (new class NotExpression (left));
      else
	return left;
    case BooleanBinop::bEquiv:
      if (value)
	return left;
      else
	return fold (new class NotExpression (left));
    }
  }
  else
    return CSE (new class BooleanBinop (op, left, right));

  assert (false);
  return NULL;
}

static class Expression*
newSetExpression (enum SetExpression::Op op,
		  class Expression* left,
		  class Expression* right)
{
  left = ensureSet (left);
  right = ensureSet (right);
  if (!left || !right) {
    left->destroy (), right->destroy ();
    return NULL;
  }

  if (left->getKind () == Expression::eEmptySet) {
    if (right->getKind () == Expression::eEmptySet) {
    Same:
      switch (op) {
      case SetExpression::sIntersection:
      case SetExpression::sMinus:
      case SetExpression::sUnion:
	right->destroy ();
	return left;
      case SetExpression::sSubset:
      case SetExpression::sEquals:
	left->destroy (), right->destroy ();
      True:
	return CSE (new class Constant
		    (new class BoolValue (Net::getBoolType (), true)));
      }
      assert (false);
    }
    else {
      switch (op) {
      case SetExpression::sIntersection:
      case SetExpression::sMinus:
	right->destroy ();
	return left;
      case SetExpression::sUnion:
	left->destroy ();
	return right;
      case SetExpression::sSubset:
	right->destroy ();
	goto True;
      case SetExpression::sEquals:
	break;
      }
    }

    if (!left->getType () || !right->getType ()) {
      pnerror ("expression has no type");
      left->destroy (), right->destroy ();
      return NULL;
    }
  }
  else if (right->getKind () == Expression::eEmptySet) {
    switch (op) {
    case SetExpression::sIntersection:
      left->destroy ();
      return right;
    case SetExpression::sMinus:
    case SetExpression::sUnion:
      right->destroy ();
      return left;
    case SetExpression::sSubset:
    case SetExpression::sEquals:
      goto Compatible;
    }
  }

  if (left == right)
    goto Same;

  if (Type::isAlwaysCompatible (left->getType (), right->getType ())) {
  Compatible:
    class Expression* expr;
    (expr = CSE (new class SetExpression (op, left, right)))
      ->associateVariables (true);
    return expr;
  }
  else {
    pnerror ("expressions have incompatible types");
    left->destroy (), right->destroy ();
    return NULL;
  }
}

static class Expression*
newBufferUnop (enum BufferUnop::Op op, class Expression* expr)
{
  if (!(expr = ensureBuffer (expr)))
    return NULL;

  if (expr->getKind () == Expression::eBufferIndex) {
    class BufferIndex* b = static_cast<class BufferIndex*>(expr);
    class Expression* buffer = b->getBuffer ();
    class Expression* index = b->getIndex ();
    b->destroy ();
    switch (op) {
    case BufferUnop::bFree:
    case BufferUnop::bUsed:
      if (index) {
	pnwarn ("ignoring meaningless buffer indexing expression");
	index->destroy (), index = NULL;
      }
      break;
    case BufferUnop::bPeek:
    case BufferUnop::bRemove:
      break;
    }

    return fold (new class BufferUnop (op, buffer, index));
  }

  return fold (new class BufferUnop (op, expr, NULL));
}
